/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: ulcdshield.c 2416 2019-04-13 22:58:50Z roi $
 */

/* 
 *  UART/LCDץ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "ulcdshield.h"
#include "topamelogo.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))
#define ABS(X)  ((X) > 0 ? (X) : -(X))

#define MAX_SPI_FREQ        (40000000)

/*
 *  Ѥѿ
 */
static const Point testPolygon[7] = {
	{0, 0},
	{127, 47},
	{64, 0},
	{64, 47},
	{127, 0},
	{0, 47},
	{0, 0}
};

LCD_Handler_t  LcdHandle;
LCD_DrawProp_t DrawProp;


/*
 *  GPIOؿ
 */
static void
pinMode(uint8_t no, uint32_t otype, uint32_t pullmode)
{
	Arduino_PortControlBlock *pgcb = getGpioTable(DIGITAL_PIN, no);
	GPIO_Init_t GPIO_Init_Data;

	if(pgcb == NULL)
		return;
	pinClock(no);
	GPIO_Init_Data.mode  = GPIO_MODE_OUTPUT;
	GPIO_Init_Data.pull  = pullmode;
	GPIO_Init_Data.otype = otype;
	GPIO_Init_Data.oxor  = GPIO_OUTPUT_NORMAL;
	GPIO_Init_Data.iof   = GPIO_NOIOF;
	gpio_setup(pgcb->giobase, &GPIO_Init_Data, pgcb->giopin);
}


/*
 *  եåɽƥ
 */
void
graphics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ST7565_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ST7565_WHITE);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5) {
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ST7565_RED);
	}
	for (x=0; x < hlcd->_width; x+=5) {
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ST7565_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7565_GREEN;
	lcd_drawPolygon(&DrawProp, (Point *)testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7565_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 32, 0, (uint8_t *)topamelogo);
	dly_tsk(1000);
}

/*
 *  ᥤ󥿥
 */
void
main_task(intptr_t exinf)
{
	SPI_Init_t spi_initd;
	SPI_Handle_t  *hspi;
	LCD_Handler_t *hlcd;
	ER_UINT	ercd;
	uint32_t sysclock;
	uint8_t sws;
	int i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "## START ##");

	/*
	 *  CS PORT(D10-GPIO2)
	 */
	pinMode(CS_PORT, GPIO_OTYPE_PP, GPIO_PULLUP);
	/*
	 *  RS PORT(D7-GPIO23)
	 */
	pinMode(RS_PORT, GPIO_OTYPE_PP, GPIO_PULLUP);
	/*
	 *  RST PORT(D9-GPIO1)
	 */
	pinMode(RST_PORT, GPIO_OTYPE_PP, GPIO_PULLUP);

	sysclock = get_sysclock(&sws);
	spi_initd.Prescaler     = (sysclock / MAX_SPI_FREQ) - 1;
	spi_initd.Direction     = SPI_DIRECTION_2LINES;
	spi_initd.CLKPhase      = SPI_PHASE_1EDGE;
	spi_initd.CLKPolarity   = SPI_POLARITY_LOW;
	spi_initd.DataSize      = 8;
	spi_initd.SignBit       = SPI_DATA_MSB;
	spi_initd.CSMode        = SPI_CSM_OFF;
	spi_initd.Protocol      = SPI_PROTOCOL_1LINE;
	spi_initd.SCKDelay      = 1;
	spi_initd.CSDelay       = 1;
	spi_initd.MinScInActTime= 1;
	spi_initd.MaxifTime     = 0;
	spi_initd.semid         = SPI1DMA_SEM;
	spi_initd.semlock       = 0;
	syslog_3(LOG_NOTICE, "sysclk(%d) max_freq(%d) prescler(%d) !", sysclock, MAX_SPI_FREQ, spi_initd.Prescaler);

	if((hspi = spi_init(SPI2_PORTID, &spi_initd)) == NULL){
		syslog_0(LOG_ERROR, "spi_start initialize error !");
	}

	cs_set(PORT_HIGH);
	hlcd = &LcdHandle;
	hlcd->hspi              = hspi;
	hlcd->spi_lock          = 0;
	hlcd->cs_pin            = CS_PORT;
	hlcd->rs_pin            = RS_PORT;
	hlcd->rst_pin           = RST_PORT;
	lcd_init(hlcd);

	DrawProp.hlcd      = hlcd;
	DrawProp.BackColor = ST7565_BLACK;
	DrawProp.TextColor = ST7565_WHITE;
	graphics_test(hlcd);

	DrawProp.TextColor = ST7565_BLACK;
	DrawProp.BackColor = ST7565_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ST7565_BLUE);
	DrawProp.TextColor = ST7565_WHITE;
	DrawProp.BackColor = ST7565_BLUE;
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 2, (uint8_t *)"TBP TEST", CENTER_MODE);

	DrawProp.TextColor = ST7565_BLACK;
	DrawProp.BackColor = ST7565_WHITE;
	DrawProp.pFont = &Font8;
	lcd_DisplayStringAt(&DrawProp, 0, 20, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 30, (uint8_t *)"Helle World !", CENTER_MODE);

	DrawProp.TextColor = ST7565_RED;
	for(i = 10 ; i < 14 ; i += 2){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2+10, i);
	}

	syslog_0(LOG_NOTICE, "LCD TEST END");
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}


