/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: test.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  D13USER-SWꤷ硢ˤߤư
 *  WDT˳ߥ⡼ɤWDTԤ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "target_syssvc.h"
#include "kernel_cfg.h"
#include "device.h"
#include "pinmode.h"
#include "test.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

#define SW_PIN   13


static void
pinModex(uint8_t dwPin, uint8_t dwMode){ 
    int gpionumx = gpio_get_gpiohno(getGpioPin(dwPin), true);
	GPIO_Init_t init = {0};
	int gpionum;

	gpionum = gpionumx;
	syslog_4(LOG_NOTICE, "## dwPin(%d) MD_PIN_MAP(%d) gpionum(%d) gpionumx(%d)  ##", dwPin, getGpioPin(dwPin), gpionum, gpionumx);
    if(gpionum >= 0){
        uint8_t function = FUNC_GPIOHS0 + gpionum;
        fpioa_set_function(getGpioPin(dwPin), function);
		switch(dwMode){
		case INPUT:
			init.mode = GPIO_MODE_INPUT;
//			init.mode = GPIO_MODE_IT_RISING;
			init.pull = GPIO_NOPULL;
			break;
		case INPUT_PULLDOWN:
			init.mode = GPIO_MODE_INPUT;
			init.pull = GPIO_PULLDOWN;
			break;
		case INPUT_PULLUP:
			init.mode = GPIO_MODE_INPUT;
			init.pull = GPIO_PULLUP;
			break;
		case OUTPUT:
		default:
			init.mode = GPIO_MODE_OUTPUT;
			init.pull = GPIO_PULLDOWN;
			break;
		}
        gpio_setup(TADR_GPIOHS_BASE, &init, (uint8_t)gpionum);
    }
    return ;
}

/*
 *  WDTߥХå
 */
static void
wdt_int_handler(WDT_Handle_t *hwdt)
{
	syslog_1(LOG_NOTICE, "wdt_int_handler[%x] !", hwdt);
}

/*
 *  USER-SWߥϥɥ
 */
void
sw_int_handler(void)
{
	uint32_t ipr = sil_rew_mem((uint32_t *)(TADR_GPIOHS_BASE+TOFF_GPIOHS_RISE_IP));
	uint32_t ipf = sil_rew_mem((uint32_t *)(TADR_GPIOHS_BASE+TOFF_GPIOHS_FALL_IP));
	syslog_2(LOG_EMERG, "## sw_int_handler r[%08x] f(%08x] ##", ipr, ipf);
	sil_wrw_mem((uint32_t *)(TADR_GPIOHS_BASE+TOFF_GPIOHS_RISE_IP), ipr);
}

/*
 *  ᥤ󥿥
 */
void
main_task(intptr_t exinf)
{
	WDT_Handle_t *hwdt;
	ER_UINT	ercd;
	int     i;
	uint32_t cnt;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	pinModex(SW_PIN, INPUT);
	hwdt = wdt_init(WDOG1_PORTID);
	hwdt->callback = wdt_int_handler;
	cnt = wdt_start(hwdt, WDT_MODE_INTERRUPT, 10000);
	syslog_1(LOG_NOTICE, "## cnt(%d) ##", cnt);
	for(i = 0 ; i < 500000 ; i++){
		syslog_1(LOG_NOTICE, "## sw(%d) ##", digitalRead(SW_PIN));
		dly_tsk(1000);
		if(i == 15){
			wdt_deinit(hwdt);
			syslog_0(LOG_NOTICE, "WDT STOP");
		}
	}
	slp_tsk();

	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}
