#include <kernel.h>
#include <t_syslog.h>
#include <stdio.h>
#include <t_stdlib.h>
#include <stdlib.h>
#if 1	/* ROI DEBUG */
#include "target_syssvc.h"
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#endif	/* ROI DEBUG */
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#define _SYS__STDINT_H
#include <time.h>
#include "kernel_cfg.h"
#include "monitor.h"
#include "device.h"

/*
 *  ǥХޥֹ
 */
static int_t ver_func(int argc, char **argv);
static int_t rck_func(int argc, char **argv);
static int_t rdt_func(int argc, char **argv);
static int_t rtm_func(int argc, char **argv);
static int_t ral_func(int argc, char **argv);
static int_t led_func(int argc, char **argv);
static int_t dip_func(int argc, char **argv);
static int_t cup_func(int argc, char **argv);

/*
 *  ǥХޥɥơ֥
 */
static const COMMAND_INFO device_command_info[] = {
	{"VER",     ver_func},
	{"RCK",		rck_func},
	{"RDT",		rdt_func},
	{"RTM",		rtm_func},
	{"RAL",		ral_func},
	{"LED",		led_func},
	{"DIP",		dip_func},
	{"CUP",		cup_func}
};

#define NUM_DEVICE_CMD   (sizeof(device_command_info)/sizeof(COMMAND_INFO))

static const char device_name[] = "DEVICE";
static const char device_help[] =
"  Device  VER                    display version\n"
"          RCK                    display clock\n"
"          RDT (year #) (month #) (day #) set date\n"
"          RTM (hour #) (min   #) (sec #) set time\n"
"          RAL count              set alarm count(sec)\n"
"          LED (no) (on-1,off-0)  led   control\n"
"          DIP (no) (on-1,off-0)  dipsw control\n"
"          CUP command            cup   control\n";

static const uint16_t led_pattern[3] = {
	LED01, LED02, LED03
};

static const char monthday[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static COMMAND_LINK device_command_link = {
	NULL,
	NUM_DEVICE_CMD,
	device_name,
	NULL,
	device_help,
	&device_command_info[0]
};

static uint8_t cup_command;

static int a2i(char *str)
{
	int num = 0;

	while(*str >= '0' && *str <= '9'){
		num = num * 10 + *str++ - '0';
	}
	return num;
}

/*
 *  ALARM EVENT
 */
static void alarm_event(void)
{
	syslog_0(LOG_NOTICE, "rtc alarm event !");
}

/*
 *  mktimeؿ
 */
time_t mktime(struct tm *ptm)
{
	int day = ptm->tm_year *365;
	int yday, i; 

	day += (ptm->tm_year+2) / 4;
	for(i = 1, yday = 0 ; i < ptm->tm_mon ; i++)
		yday += monthday[i-1];
	if(ptm->tm_mon > 2 && ((ptm->tm_year+2) % 4) == 0)
		yday++;
	yday += ptm->tm_mday - 1;
	day  += yday;
	ptm->tm_wday = (day+4) % 7;
	ptm->tm_yday = yday+1;
	syslog_3(LOG_INFO, "yday[%d] wday[%d] day[%d]", yday+1, ptm->tm_wday, ptm->tm_yday);
	return day * (24*60*60) + ptm->tm_hour * 60*60 + ptm->tm_min * 60 + ptm->tm_sec;
}

/*
 * gmtime_rؿ
 */
struct tm *gmtime_r(const time_t *pt, struct tm *ptm)
{
	int day = *pt / (24*60*60);
	int time, i, yday, mday;

	time = *pt - day * 24 * 60 * 60;
	ptm->tm_wday = (day+4) % 7;
	for(i = 365, ptm->tm_year = 0 ; i < day ; i += 365){
		if(((ptm->tm_year+2) % 4) == 0)
			i++;
		ptm->tm_year++;
	}

	ptm->tm_yday = yday = day - i + 365 + 1;
	for(ptm->tm_mon = 1 ; ptm->tm_mon <= 12 && yday > 0 ; ptm->tm_mon++){
		mday = monthday[ptm->tm_mon-1];
		if(ptm->tm_mon == 2 && ((ptm->tm_year+2) % 4) == 0)
			mday++;
		if(yday < mday){
			ptm->tm_mday = yday;
			break;
		}
		yday -= mday;
	}
	ptm->tm_hour = time / (60*60);
	time -= ptm->tm_hour * 60 * 60;
	ptm->tm_min  = time / 60;
	ptm->tm_sec = time % 60;
	return ptm;
}

/*
 *  DEVICEޥؿ
 */
void device_info_init(intptr_t exinf)
{
	setup_command(&device_command_link);
}

/*
 *  VERSIONɽؿ
 */
static int_t ver_func(int argc, char **argv)
{
	printf("TOPPERS BASE PLATFORM Verion %d.%X.%d\n", 
			(TPLATFORM_PRVER >> 12) &  0x0fU,
			(TPLATFORM_PRVER >> 4) & 0xffU,
				TPLATFORM_PRVER & 0x0fU);
	return 0;
}

/*
 *  Фޥɴؿ
 */
static int_t rck_func(int argc, char **argv)
{
	struct tm2 timedate;

	rtc_get_time(&timedate);
	mktime((struct tm *)&timedate);
	printf("        %04d/%02d/%02d\n", timedate.tm_year+1970, timedate.tm_mon, timedate.tm_mday);
	printf("y[%d][%d] %02d:%02d:%02d\n", timedate.tm_yday, timedate.tm_wday, timedate.tm_hour, timedate.tm_min, timedate.tm_sec);
	return 0;
}

/*
 *  ꥳޥɴؿ
 */
static int_t rdt_func(int argc, char **argv)
{
	struct tm timedate;
	int arg1 = 2019;

	rtc_get_time((struct tm2 *)&timedate);
	if(argc > 1)
		arg1 = a2i(argv[1]);
	if(argc > 2)
		timedate.tm_mon = a2i(argv[2]);
	else
		timedate.tm_mon = 1;
	if(argc > 3)
		timedate.tm_mday = a2i(argv[3]);
	else
		timedate.tm_mday = 1;
	printf("%04d/%02d/%02d\n", arg1, timedate.tm_mon, timedate.tm_mday);
	timedate.tm_year = arg1-1970;
	timedate.tm_isdst = 0;
	mktime((struct tm *)&timedate);
	rtc_set_time((struct tm2 *)&timedate);
	return 0;
}

/*
 *  ꥳޥɴؿ
 */
static int_t rtm_func(int argc, char **argv)
{
	struct tm2 timedate;

	rtc_get_time(&timedate);
	if(argc > 1)
		timedate.tm_hour = a2i(argv[1]);
	else
		timedate.tm_hour = 0;
	if(argc > 2)
		timedate.tm_min = a2i(argv[2]);
	else
		timedate.tm_min = 0;
	if(argc > 3)
		timedate.tm_sec = a2i(argv[3]);
	else
		timedate.tm_sec = 0;
	printf("%02d:%02d:%02d\n", timedate.tm_hour, timedate.tm_min, timedate.tm_sec);
	timedate.tm_isdst = 0;
	mktime((struct tm *)&timedate);
	rtc_set_time(&timedate);
	return 0;
}

/*
 *  顼ꥳޥɴؿ
 */
static int_t ral_func(int argc, char **argv)
{
	struct tm2 timedate;
	uint32_t count = 20;
	ER ercd;

	if(argc > 1)
		count = atoi(argv[1]);
	rtc_get_time(&timedate);
	timedate.tm_sec += count;
	while(timedate.tm_sec >= 60){
		timedate.tm_sec -= 60;
		timedate.tm_min++;
	}
	while(timedate.tm_min >= 60){
		timedate.tm_min -= 60;
		timedate.tm_hour++;
	}
	ercd = rtc_setalarm(&timedate, alarm_event);
	printf("alarm result(%d) !\n", ercd);
	return 1;
}

/*
 *  LEDꥳޥɴؿ
 */
static int_t led_func(int argc, char **argv)
{
	int_t    arg1=0, arg2;
	uint16_t reg;

	if(argc < 3)
		return -1;
	arg1 = atoi(argv[1]);
	arg2 = atoi(argv[2]);
	if(arg1 >= 1 && arg1 <= 3){
		reg = led_in();
		if(arg2 != 0)
			reg |= led_pattern[arg1-1];
		else
			reg &= ~led_pattern[arg1-1];
		led_out(reg);
	}
	return 0;
}

/*
 *  DIPSWꥳޥɴؿ
 */
static int_t dip_func(int argc, char **argv)
{
	int_t   arg1=0, arg2=0, i;

	if(argc > 1)
		arg1 = a2i(argv[1]);
	if(argc > 2)
		arg2 = a2i(argv[2]);
	if(arg1 >= 1 && arg1 <= 8){
		if(arg2 != 0)
			dipsw_value |= (0x1<<(arg1-1));
		else
			dipsw_value &= ~(0x1<<(arg1-1));
	}
	printf("dipsw\n1   2   3   4   5   6   7   8\n");
	for(i = 0 ; i < 8 ; i++){
		if(dipsw_value & (1<<i))
			printf("ON  ");
		else
			printf("OFF ");
	}
	printf("\n");
	return 0;
}

/*
 *  åץ顼󥿥ޥɴؿ
 */
static int_t cup_func(int argc, char **argv)
{
	if(argc > 1)
		cup_command = *argv[1];
	return 0;
}

/*
 * CUP顼󥳥ޥ
 */
int8_t
get_cup_command(void)
{
	uint8_t cmd = cup_command;
	cup_command = 0;
	return cmd;
}


