/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: demo.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  LONGAN NANOǥץ
 */

#include <kernel.h>
#include <string.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "spi.h"
#include "spi_driver.h"
#include "sipeed_st7735.h"
#ifdef SDEV_SENSE_ONETIME
#include "storagedevice.h"
#include "sddiskio.h"
#endif
#include "demo.h"
#include "topamelogo.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

/*
 * ST7735Sɽ
 */

#define USE_HORIZONTAL  2

extern uint8_t aTxBuffer[];

/*
 *  Ѥѿ
 */
static const Point testPolygon[7] = {
	{0, 0},
	{ST7735_TFTWIDTH-1, ST7735_TFTHEIGHT-1},
	{ST7735_TFTWIDTH/2, 0},
	{ST7735_TFTWIDTH/2, ST7735_TFTHEIGHT-1},
	{ST7735_TFTWIDTH-1, 0},
	{0, ST7735_TFTHEIGHT-1},
	{0, 0}
};

static const char banner[] = "TOPPERS BASE PLATFORM";

LCD_Handler_t LcdHandle;
LCD_DrawProp_t DrawProp;

#ifndef SDEV_SENSE_ONETIME
uint32_t sdbuffer[514/4];
#endif


static void
set_value(uint8_t *buf, int value)
{
    buf[1] = (value % 10) + '0';
    buf[0] = (value / 10) + '0';
}

static void
set_time(uint8_t *buf, struct tm2 *ptim)
{
	buf[8] = 0;
    set_value(&buf[6], ptim->tm_sec);
    buf[5] = ':';
    set_value(&buf[3], ptim->tm_min);
    buf[2] = ':';
    set_value(&buf[0], ptim->tm_hour);
}

void
digitalWrite(uint16_t no, int sw)
{
	uint32_t base;

	if(no == RST_PIN)
		base = RST_GPIO;
	else if(no == DC_PIN)
		base = DC_GPIO;
	else if(no == CS2_PIN)
		base = CS2_GPIO;
	else
		base = CS_GPIO;

	if(no >= 3)
		return;
	if(sw == 0)
		sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_BC), (1<<no));
	else
		sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_BOP), (1<<no));
}


/*
 *  եåɽƥ
 */
void
grapics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ST7735_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ST7735_GREEN);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5){
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ST7735_RED);
	}
	for (x=0; x < hlcd->_width; x+=5){
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ST7735_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7735_GREEN;
	lcd_drawPolygon(&DrawProp, (Point *)testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7735_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 50, 4, (uint8_t *)topamelogo);
	dly_tsk(1000);
}


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
#ifdef SDEV_SENSE_ONETIME
	StorageDevice_t *psdev;
#endif
	GPIO_Init_t GPIO_Init_Data;
	SPI_Init_t spi_initd;
	SPI_Handle_t   *hspi;
	SDCARD_Handler_t *hsd = NULL;
	LCD_Handler_t  *hlcd;
#ifndef SDEV_SENSE_ONETIME
	SDCARD_CardInfo_t CardInfo;
#endif
	struct tm2 tim;
	uint8_t buf[12];
	ER_UINT	ercd;
	uint32_t tmp;
	int i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	/*
	 *  GPIO PORTBͭ
	 */
	tmp  = sil_rew_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN));
	tmp |= RCU_APB2EN_PBEN;
	sil_wrw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN), tmp);

	/*
	 *  CS PIN
	 */
	GPIO_Init_Data.mode = GPIO_MODE_OUTPUT;		/* GPIO */
	GPIO_Init_Data.pull = GPIO_NOPULL;
	GPIO_Init_Data.otype = GPIO_OTYPE_PP;		/* GPIOץ륢å׽ */
	GPIO_Init_Data.speed = GPIO_OSPEED_50MHZ;	/* GPIOԡ */
	gpio_setup(CS_GPIO, &GPIO_Init_Data, CS_PIN);

	/*
	 *  C2 PIN
	 */
	gpio_setup(CS2_GPIO, &GPIO_Init_Data, CS2_PIN);

	/*
	 *  RST PIN
	 */
	gpio_setup(RST_GPIO, &GPIO_Init_Data, RST_PIN);
	digitalWrite(RST_PIN, 0);
	sil_wrw_mem((uint32_t *)(RST_GPIO+TOFF_GPIO_BC), (1<<RST_PIN));

	/*
	 *  DC PIN
	 */
	gpio_setup(DC_GPIO, &GPIO_Init_Data, DC_PIN);
	digitalWrite(DC_PIN, PORT_LOW);

	/*
	 *  LCD SPI 
	 */
    spi_initd.Prescaler         = SPI_BAUDRATEPRESCALER_8;
	spi_initd.Direction         = SPI_DIRECTION_1LINE;
	spi_initd.CLKPhase          = SPI_PHASE_2EDGE;
	spi_initd.CLKPolarity       = SPI_POLARITY_HIGH;
	spi_initd.CRC               = SPI_CRC_DISABLE;
	spi_initd.DataSize          = SPI_DATASIZE_8BIT;
	spi_initd.SignBit           = SPI_DATA_MSB;
	spi_initd.NSS               = SPI_NSS_SOFT;
	spi_initd.NSSPMode          = SPI_NSS_PULSE_DISABLE;
	spi_initd.TIMode            = SPI_TIMODE_DISABLE;
	spi_initd.TxDMAChannel      = -1;
//	spi_initd.TxDMAChannel      = DMALTX_CH;
	spi_initd.RxDMAChannel      = -1;
	spi_initd.Mode  = SPI_MODE_MASTER;
	spi_initd.semid = SPI1TRN_SEM;
	spi_initd.semlock = 0;

	if((hspi = spi_init(SPIL_PORTID, &spi_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_NOTICE, "spi(lcd) init(int-1) error");
		slp_tsk();
	}

	hlcd = &LcdHandle;
	hlcd->hspi = hspi;
	hlcd->cs_pin  = CS_PIN;
	hlcd->rs_pin  = DC_PIN;
	hlcd->rst_pin = RST_PIN;
	lcd_init(&LcdHandle, USE_HORIZONTAL);
	syslog_4(LOG_NOTICE, "h(%d) w(%d) col(%d) low(%d)", hlcd->_width, hlcd->_height, hlcd->colstart, hlcd->rowstart);

	DrawProp.hlcd      = hlcd;
	DrawProp.BackColor = ST7735_BLACK;
	DrawProp.TextColor = ST7735_WHITE;
	lcd_fillScreen(&DrawProp);

	/*
	 *  LCD SD-CARD 
	 */
    spi_initd.Prescaler         = SPI_BAUDRATEPRESCALER_64;
	spi_initd.Direction         = SPI_DIRECTION_2LINES;
	spi_initd.TxDMAChannel      = -1;
	spi_initd.RxDMAChannel      = -1;
	spi_initd.Mode  = SPI_MODE_MASTER;
	spi_initd.semid = SPI2TRN_SEM;
	spi_initd.semlock = 0;
	if((hspi = spi_init(SPIS_PORTID, &spi_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_ERROR, "spi(card) init(int-1) error");
		slp_tsk();
	}
	sdcard_setspi2(SPISDCARD_PORTID, hspi, CS2_PIN);

	for(i = 0 ; i < 10 ; i++)
		aTxBuffer[i] = 0xFF;
	cs2_set(PORT_LOW);
	if((ercd = spi_transmit(hspi, (uint8_t*)aTxBuffer, 10)) != E_OK){
		/* Transfer error in transmission process */
		syslog_1(LOG_ERROR, "sd-card init error(%d) !", ercd);
	}
#if SPI_WAIT_TIME == 0
	spi_wait(hspi, 50);
#endif
	dly_tsk(100);
#ifndef SDEV_SENSE_ONETIME
	if(sdcard_init(SPISDCARD_PORTID)){
		hsd = sdcard_open(SPISDCARD_PORTID);
		syslog_1(LOG_NOTICE, "## card(%d) ##", hsd->cardtype);
		SVC_PERROR(sdcard_checkCID(hsd));
		SVC_PERROR(sdcard_sendCSD(hsd));
		SVC_PERROR(sdcard_getcardinfo(hsd, &CardInfo));
		SVC_PERROR(sdcard_configuration(hsd));
	}
	else
		syslog_0(LOG_NOTICE, "SD-CARD INITIAL ERROR !");
#else
	SDMSence_task(0);
	psdev = SDMGetStorageDevice(SDCARD_DEVNO);
	hsd = (SDCARD_Handler_t *)psdev->_sdev_local[1];
	if(hsd == NULL)
		syslog_0(LOG_ERROR, "sd-card init no card or error !");
	else
		syslog_1(LOG_NOTICE, "sd-card type(%d) !", hsd->cardtype);
#endif
#ifndef SDEV_SENSE_ONETIME
	if(hsd != NULL && hsd->hspi != NULL){
		uint32_t j;
		uint8_t *p;
		i = 0;
		ercd = sdcard_blockread(hsd, sdbuffer, i*512, 512, 1);
		syslog_3(LOG_NOTICE, "## ercd(%d) i(%d) sdbuffer[%08x] ##", ercd, i, sdbuffer);
		dly_tsk(300);
		p = (unsigned char *)sdbuffer;
		printf("\nsec(%d) ", i);
		for(j = 0 ; j < 512 ; j++){
			if((j % 16) == 0)
				printf("\n%03x ", j);
			printf("%02x ", p[j]);
		}
		printf("\n");
	}
#endif

	grapics_test(hlcd);
	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ST7735_BLUE);
	DrawProp.TextColor = ST7735_WHITE;
	DrawProp.BackColor = ST7735_BLUE;
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 6, (uint8_t *)banner, CENTER_MODE);

	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;
	DrawProp.pFont = &Font8;
	lcd_DisplayStringAt(&DrawProp, 0, 30, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 44, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font16;
	lcd_DisplayStringAt(&DrawProp, 0, 56, (uint8_t *)"Helle World !", CENTER_MODE);

	DrawProp.TextColor = ST7735_RED;
	for(i = 20 ; i < 30 ; i++){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2, i);
		if(i == 24)
			i += 5;
	}
	dly_tsk(5000);

	DrawProp.TextColor = ST7735_BLACK;
	DrawProp.BackColor = ST7735_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/2, ST7735_BLUE);
	DrawProp.TextColor = ST7735_WHITE;
	DrawProp.BackColor = ST7735_BLUE;
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 6, (uint8_t *)banner, CENTER_MODE);
	lcd_DisplayStringAt(&DrawProp, 0, 24, (uint8_t *)"RTC TIMER", CENTER_MODE);
	DrawProp.TextColor = ST7735_BLUE;
	DrawProp.BackColor = ST7735_WHITE;
	DrawProp.pFont = &Font16;
	for(i = 0 ; i < 50000 ; i++){
		rtc_get_time(&tim);
		set_time(buf, &tim);
		lcd_DisplayStringAt(&DrawProp, 0, 50, buf, CENTER_MODE);
		dly_tsk(400);
	}

	syslog(LOG_NOTICE, "Sample program ends.");
	slp_tsk();
//	SVC_PERROR(ext_ker());
}
