/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: lcdshield.c 2416 2019-04-15 22:39:15Z roi $
 */

/* 
 *  adafruit2.8"LCDץ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "spi.h"
#include "spi_driver.h"
#include "adafruit_ILI9341.h"
#include "adafruitts_stmp610.h"
#ifdef SDEV_SENSE_ONETIME
#include "storagedevice.h"
#include "sddiskio.h"
#endif
#include "lcdshield.h"
#include "topamelogo.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))
#define ABS(X)  ((X) > 0 ? (X) : -(X))

#define NUM_POINTS   8

/*
 *  Ѥѿ
 */
static const Point testPolygon[7] = {
	{0, 0},
	{239, 319},
	{120, 0},
	{120, 319},
	{239, 0},
	{0, 319},
	{0, 0}
};

LCD_Handler_t  LcdHandle;
LCD_DrawProp_t DrawProp;
TouchScreen_Handle_t TsHandle;

/*
 *  SPI-SDCARDåѥΰ
 */
extern uint8_t aTxBuffer[BUFFERSIZE];

#ifndef SDEV_SENSE_ONETIME
uint32_t sdbuffer[514/4];
#endif

/* Buffer used for transmission */


/*
 *  SW1
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

/*
 *  եåɽƥ
 */
void
grapics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ILI9341_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ILI9341_GREEN);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5) {
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ILI9341_RED);
	}
	for (x=0; x < hlcd->_width; x+=5) {
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ILI9341_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ILI9341_GREEN;
	lcd_drawPolygon(&DrawProp, (Point *)testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ILI9341_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 33, 48, (uint8_t *)topamelogo);
	dly_tsk(1000);
}

/*
 *  ᥤ󥿥
 */
void
main_task(intptr_t exinf)
{
#ifdef SDEV_SENSE_ONETIME
	StorageDevice_t *psdev;
#endif
	SPI_Setup_t     SPI_Setup;
	SDCARD_Handler_t *hsd = NULL;
	SPI_Handle_t  *hspi;
	LCD_Handler_t   *hlcd;
#ifndef SDEV_SENSE_ONETIME
	SDCARD_CardInfo_t CardInfo;
#endif
	TouchScreen_Handle_t *hts;
	TouchScreen_Point_t  touch[NUM_POINTS];
	int16_t x, y;
	ER_UINT	ercd;
	int sens = 1000;
	int i;


	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "## START ##");
	SPI_Setup.port    = SPI1_PORTID;
	SPI_Setup.baud    = SPI_BAUDRATEPRESCALER_256;
	SPI_Setup.spi_sem = SPI1DMA_SEM;
	SPI_Setup.cs_pin  = CS_PORT;
	SPI_Setup.rs_pin  = DC_PORT;
	SPI_Setup.rst_pin = 32;
	SPI_Setup.cs2_pin = CS2_PORT;
	SPI_Setup.otype   = GPIO_OTYPE_PP;
	SPI_Setup.pullmode= GPIO_PULLUP;
	hspi = spi_start(&SPI_Setup);
	sdcard_setspi(SPISDCARD_PORTID, hspi);

	for(i = 0 ; i < 10 ; i++)
		aTxBuffer[i] = 0xFF;
	cs_set(PORT_HIGH);
	cs2_set(PORT_HIGH);
	if(spi_transmit(hspi, (uint8_t*)aTxBuffer, 10) != E_OK){
		/* Transfer error in transmission process */
		syslog_0(LOG_NOTICE, "## call Error_Handler(1) ##");
	}
#if SPI_WAIT_TIME == 0
	spi_wait(hspi, 50);
#endif
	dly_tsk(100);

#ifndef SDEV_SENSE_ONETIME
	if(sdcard_init(SPISDCARD_PORTID)){
		hsd = sdcard_open(SPISDCARD_PORTID);
		syslog_1(LOG_NOTICE, "## card(%d) ##", hsd->cardtype);
		SVC_PERROR(sdcard_checkCID(hsd));
		SVC_PERROR(sdcard_sendCSD(hsd));
		SVC_PERROR(sdcard_getcardinfo(hsd, &CardInfo));
		SVC_PERROR(sdcard_configuration(hsd));
	}
	else
		syslog_0(LOG_NOTICE, "SD-CARD INITIAL ERROR !");
#else
	SDMSence_task(0);
	psdev = SDMGetStorageDevice(SDCARD_DEVNO);
	hsd = (SDCARD_Handler_t *)psdev->_sdev_local[1];
#endif
	if(hsd != NULL)
		hsd->spi_lock = SPI1LOCK_SEM;
	spi_end(hspi);
	dly_tsk(100);
#if 0
	SPI_Setup.baud = SPI_BAUDRATEPRESCALER_16;
#else
	SPI_Setup.baud = SPI_BAUDRATEPRESCALER_64;
#endif
	hspi = spi_start(&SPI_Setup);
#ifndef SDEV_SENSE_ONETIME
	if(hsd != NULL && hsd->hspi != NULL){
		unsigned char *p;
		int j, bpb;
		int SecPerCls, Rsv, NumFat, Fsize, RootDirC, RootSec;
		for(i = 0 ; i < 1 ; i++){
			ercd = sdcard_blockread(hsd, sdbuffer, i*512, 512, 1);
			syslog_3(LOG_NOTICE, "## ercd(%d) i(%d) sdbuffer[%08x] ##", ercd, i, sdbuffer);
			dly_tsk(300);
			p = (unsigned char *)sdbuffer;
			if(i == 0){
				if(p[0x36] == 'F' && p[0x37] == 'A' && p[0x38] == 'T')
					bpb = 0;
				else{
					bpb = p[0x1c6];
					bpb |= p[0x1c7]<<8;
					bpb |= p[0x1c8]<<16;
					bpb |= p[0x1c9]<<24;
				}
			}
			printf("\nsec(%d) ", i);
			for(j = 0 ; j < 512 ; j++){
				if((j % 16) == 0)
					printf("\n%03x ", j);
				printf("%02x ", p[j]);
			}
			printf("\n");
		}
		syslog_1(LOG_NOTICE, "## bpb(%08x) ##", bpb);
		if(bpb != 0){
			ercd = sdcard_blockread(hsd, sdbuffer, bpb*512, 512, 1);
			syslog_3(LOG_NOTICE, "## ercd(%d) i(%d) sdbuffer[%08x] ##", ercd, bpb, sdbuffer);
			dly_tsk(300);
			printf("\nsec(%d) ", bpb);
			for(j = 0 ; j < 512 ; j++){
				if((j % 16) == 0)
					printf("\n%03x ", j);
				printf("%02x ", p[j]);
			}
			printf("\n");
		}
		p = (unsigned char *)sdbuffer;
		SecPerCls = p[13];
		Rsv = p[14] + (p[15] << 8);
		NumFat = p[16];
		RootDirC = p[17] + (p[18]<<8);
		Fsize = p[22] + (p[23] << 8);
		syslog_4(LOG_NOTICE, "## SecPerCls(%d) Rsv(%d) NumFat(%d) RootDirC(%d) ##", SecPerCls, Rsv, NumFat, RootDirC);
		RootSec = bpb + Rsv + (NumFat * Fsize);
		syslog_2(LOG_NOTICE, "## Fsize(%d) RootSec(%d) ##", Fsize, RootSec);

		for(i = 0 ; i < 2 ; i++){
			ercd = sdcard_blockread(hsd, sdbuffer, (i+RootSec)*512, 512, 1);
			syslog_3(LOG_NOTICE, "## ercd(%d) i(%d) sdbuffer[%08x] ##", ercd, (i+RootSec), sdbuffer);
			dly_tsk(300);
			p = (unsigned char *)sdbuffer;
			printf("\nsec(%d) ", (i+RootSec));
			for(j = 0 ; j < 512 ; j++){
				if((j % 16) == 0)
					printf("\n%03x ", j);
				printf("%02x ", p[j]);
			}
			printf("\n");
		}
	}
#endif
	(void)(hsd);

	hlcd = &LcdHandle;
	hlcd->hspi = hspi;
	hlcd->spi_lock= SPI1LOCK_SEM;
	hlcd->cs_pin  = CS_PORT;
	hlcd->dc_pin  = DC_PORT;
	hlcd->rst_pin = 32;
	DrawProp.hlcd = hlcd;
	lcd_init(hlcd, 0);
	syslog_2(LOG_NOTICE, "LCD width(%d) height(%d)", hlcd->_width, hlcd->_height);
	DrawProp.BackColor = ILI9341_BLACK;
	DrawProp.TextColor = ILI9341_WHITE;
	grapics_test(hlcd);
#if 1	/* LOCAL DEFINE */
	DrawProp.TextColor = ILI9341_BLACK;
	DrawProp.BackColor = ILI9341_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ILI9341_BLUE);
	DrawProp.TextColor = ILI9341_WHITE;
	DrawProp.BackColor = ILI9341_BLUE;
	DrawProp.pFont = &Font20;
	lcd_DisplayStringAt(&DrawProp, 0, 20, (uint8_t *)"TOPPERS", CENTER_MODE);
	lcd_DisplayStringAt(&DrawProp, 0, 40, (uint8_t *)"BASE PLATFORM(ST)", CENTER_MODE);

	DrawProp.TextColor = ILI9341_BLACK;
	DrawProp.BackColor = ILI9341_WHITE;
	DrawProp.pFont = &Font8;
	lcd_DisplayStringAt(&DrawProp, 0, 120, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 140, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font16;
	lcd_DisplayStringAt(&DrawProp, 0, 160, (uint8_t *)"Helle World !", CENTER_MODE);

	DrawProp.TextColor = ILI9341_RED;
	for(i = 20 ; i < 40 ; i++){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2+30, i);
		if(i == 25)
			i += 10;
	}
#endif	/* LOCAL DEFINE */
	syslog_0(LOG_NOTICE, "LCD TEST END");

	hts = &TsHandle;
	hts->hspi = hspi;
	hts->cs_pin = CS3_PORT;
	hts->otype  = GPIO_OTYPE_PP;
	hts->pullmode = GPIO_PULLUP;
	hts->spi_lock = SPI1LOCK_SEM;
	ercd = touchscreen_init(hts, 0);
	if(ercd != E_OK){
		syslog_1(LOG_NOTICE, "touch screen initial error(%d)", ercd);
		slp_tsk();
	}
	while(sens > 0){
		if(touchscreen_touched(hts) && touchscreen_buffersize(hts) > 0){
			ercd = touchscreen_getpoint(&TsHandle, touch, NUM_POINTS);
			if(ercd > 0){
				syslog_3(LOG_NOTICE, "sence x(%d) y(%d) z(%d)", touch[0].x, touch[0].y, touch[0].z);
				if(touch[0].x <= hts->min_x)
					x = 0;
				else
					x = ((int)touch[0].x * LcdHandle._width) / (hts->max_x - hts->min_x);
				if(touch[0].y <= hts->min_y)
					y = 0;
				else
					y = ((int)touch[0].y * LcdHandle._height) / (hts->max_y - hts->min_y);
				syslog_2(LOG_NOTICE, "lcd x(%d) y(%d)", x, y);
				sens--;
				DrawProp.TextColor = ILI9341_GREEN;
				lcd_DrawCircle(&DrawProp, x, y, 10);
				dly_tsk(500);
			}
		}
		else{
			if((ercd = touchscreen_buffersize(hts)) > 0)
				touchscreen_getpoint(hts, touch, NUM_POINTS);
		}
		dly_tsk(50);
	}
#ifdef SDEV_SENSE_ONETIME
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
#endif
	spi_end(hspi);

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

