/**
  ******************************************************************************
  * @file    BSP/Src/audio_loopback.c
  * @author  MCD Application Team
  * @version V1.1.0
  * @date    23-September-2016  
  * @brief   This example code shows how to use the audio feature in the
  *          stm32f769i_discovery driver
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; COPYRIGHT(c) 2016 STMicroelectronics</center></h2>
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */

/* Includes ------------------------------------------------------------------*/
#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "sample1.h"
#include "string.h"

/** @addtogroup STM32F7xx_HAL_Examples
  * @{
  */

/** @addtogroup BSP
  * @{
  */

/* Private typedef -----------------------------------------------------------*/
typedef enum
{
  BUFFER_OFFSET_NONE = 0,
  BUFFER_OFFSET_HALF = 1,
  BUFFER_OFFSET_FULL = 2,
}BUFFER_StateTypeDef;

#define CACHE_LINE_SIZE    32

#ifdef AUDIO_IN_OUT
#define RECORD_BUFFER_SIZE  4096
#else
#define RECORD_BUFFER_SIZE  8192
#endif

/* Private define ------------------------------------------------------------*/
/* Private macro -------------------------------------------------------------*/
/* Private variables ---------------------------------------------------------*/
uint32_t  audio_rec_buffer_state;
//uint8_t ChannelNumber = 0;
/* Buffer containing the PCM samples coming from the microphone */
int16_t RecordBuffer[RECORD_BUFFER_SIZE] __attribute__ ((aligned (CACHE_LINE_SIZE)));

#ifdef AUDIO_IN_OUT
/* Buffer used to stream the recorded PCM samples towards the audio codec. */
int16_t PlaybackBuffer[RECORD_BUFFER_SIZE] __attribute__ ((aligned (CACHE_LINE_SIZE)));
#endif

/* Private function prototypes -----------------------------------------------*/
static void AudioLoopback_SetHint(void);
static void CopyBuffer(int16_t *pbuffer1, int16_t *pbuffer2, uint16_t BufferSize);
/* Private functions ---------------------------------------------------------*/

/**
  * @brief  Audio Play demo
  * @param  None
  * @retval None
  */
void AudioLoopback_demo (void)
{
  uint32_t audio_loop_back_init = RESET ; 
  
  AudioLoopback_SetHint();
  BSP_LCD_SetFont(&Font12);

  /* Initialize Audio Recorder with 4 channels to be used */
#ifdef AUDIO_IN_OUT
  if (BSP_AUDIO_IN_OUT_Init(INPUT_DEVICE_DIGITAL_MICROPHONE_2, OUTPUT_DEVICE_BOTH, BSP_AUDIO_FREQUENCY_16K, DEFAULT_AUDIO_IN_BIT_RESOLUTION, DEFAULT_AUDIO_IN_CHANNEL_NBR) == AUDIO_OK)
#else
  if (BSP_AUDIO_IN_InitEx(INPUT_DEVICE_DIGITAL_MICROPHONE_2, BSP_AUDIO_FREQUENCY_16K, DEFAULT_AUDIO_IN_BIT_RESOLUTION, DEFAULT_AUDIO_IN_CHANNEL_NBR) == AUDIO_OK)
#endif
  {
    BSP_LCD_SetBackColor(LCD_COLOR_WHITE);
    BSP_LCD_SetTextColor(LCD_COLOR_GREEN);
    BSP_LCD_DisplayStringAt(0, BSP_LCD_GetYSize() - 95, (uint8_t *)"  AUDIO RECORD INIT OK  ", CENTER_MODE);
  }
  else
  {
    BSP_LCD_SetBackColor(LCD_COLOR_WHITE);
    BSP_LCD_SetTextColor(LCD_COLOR_RED);
    BSP_LCD_DisplayStringAt(0, BSP_LCD_GetYSize() - 95, (uint8_t *)"  AUDIO RECORD INIT FAIL", CENTER_MODE);
    BSP_LCD_DisplayStringAt(0, BSP_LCD_GetYSize() - 80, (uint8_t *)" Try to reset board ", CENTER_MODE);
  }

  /* Display the state on the screen */
  BSP_LCD_SetBackColor(LCD_COLOR_WHITE);
  BSP_LCD_SetTextColor(LCD_COLOR_BLUE);
  BSP_LCD_DisplayStringAt(0, BSP_LCD_GetYSize() - 20, (uint8_t *)"Microphones sound streamed to headphones", CENTER_MODE);

  /* Start Recording */
  BSP_AUDIO_IN_Record((uint16_t*)&RecordBuffer[0], RECORD_BUFFER_SIZE);

 // ChannelNumber = BSP_AUDIO_IN_GetChannelNumber();
  audio_rec_buffer_state = BUFFER_OFFSET_NONE; 
  
  BSP_LCD_SetBackColor(LCD_COLOR_WHITE);  
  BSP_LCD_SetTextColor(LCD_COLOR_BLUE);
  BSP_LCD_DisplayStringAt(0, 130, (uint8_t*)"RECORD/PLAYING ONGOING", CENTER_MODE);
  BSP_LCD_DisplayStringAt(0, 145, (uint8_t*)"at Sample rate 44 Khz", CENTER_MODE);  
  BSP_LCD_DisplayStringAt(0, 175, (uint8_t *)"Press User push-button", CENTER_MODE);
  BSP_LCD_DisplayStringAt(0, 190, (uint8_t *)"to stop recording", CENTER_MODE);

  while (1)
  {
    /* 1st or 2nd half of the record buffer ready for being copied 
    to the playbakc buffer */
    if(audio_rec_buffer_state != BUFFER_OFFSET_NONE)
    {
      /* Copy half of the record buffer to the playback buffer */
      if(audio_rec_buffer_state == BUFFER_OFFSET_HALF)
      {
#ifdef AUDIO_IN_OUT
        CopyBuffer(&PlaybackBuffer[0], &RecordBuffer[0], RECORD_BUFFER_SIZE/2);
#endif
        
        if (audio_loop_back_init == RESET)
        {
#ifdef AUDIO_IN_OUT
          /* Initialize the audio device*/ 
//          if (BSP_AUDIO_OUT_Init(OUTPUT_DEVICE_BOTH, 
//                                 70, 
//                                 BSP_AUDIO_FREQUENCY_8K) != AUDIO_OK)
//          {
//            syslog_0(LOG_ERROR, "## BSP_AUDIO_OUT_Init ERROR ##");
//            slp_tsk();
//          }
          /* Play the recorded buffer */
          BSP_AUDIO_OUT_Play((uint16_t *) &PlaybackBuffer[0], 2*RECORD_BUFFER_SIZE);
#endif
          
          /* Audio device is initialized only once */
          audio_loop_back_init = SET; 
        }
      }
#ifdef AUDIO_IN_OUT
      else /* if(audio_rec_buffer_state == BUFFER_OFFSET_FULL)*/
      {
        CopyBuffer(&PlaybackBuffer[RECORD_BUFFER_SIZE/2], &RecordBuffer[RECORD_BUFFER_SIZE/2], RECORD_BUFFER_SIZE/2);
        
      }
#endif
      
      /* Wait for next data */
      audio_rec_buffer_state = BUFFER_OFFSET_NONE;  
    }
    
    if (CheckForUserInput() > 0)
    {
      /* Stop Player before close Test */
#ifdef AUDIO_IN_OUT
      BSP_AUDIO_OUT_Stop(CODEC_PDWN_HW);
      BSP_AUDIO_OUT_DeInit();
#endif
      BSP_AUDIO_IN_Stop(CODEC_PDWN_SW);
      BSP_AUDIO_IN_DeInit(); 
#ifdef AUDIO_IN_OUT
      return;
#else
      break;
#endif
    }
  }

#ifndef AUDIO_IN_OUT
  if (BSP_AUDIO_OUT_Init(OUTPUT_DEVICE_BOTH, 
                         70, 
                         BSP_AUDIO_FREQUENCY_8K) != AUDIO_OK)
  {
     syslog_0(LOG_ERROR, "## BSP_AUDIO_OUT_Init ERROR ##");
     slp_tsk();
  }
  BSP_AUDIO_OUT_SetAudioFrameSlot(CODEC_AUDIOFRAME_SLOT_02);
  /* Play the recorded buffer */
  BSP_AUDIO_OUT_Play((uint16_t *) &RecordBuffer[0], 2*RECORD_BUFFER_SIZE);

  while (CheckForUserInput() == 0){
    dly_tsk(10);
  }
  /* Stop Player before close Test */
  BSP_AUDIO_OUT_Stop(CODEC_PDWN_HW);
  BSP_AUDIO_OUT_DeInit();
  return;
#endif
}

/**
  * @brief  Copy content of pbuffer2 to pbuffer1
  * @param  None
  * @retval None
  */
static void CopyBuffer(int16_t *pbuffer1, int16_t *pbuffer2, uint16_t BufferSize)
{
  uint32_t i = 0;
  for(i = 0; i < BufferSize; i++)
  {
    pbuffer1[i] = pbuffer2[i];
  }
  flushinvalidatedcache_by_addr((uint8_t *)pbuffer1, BufferSize*2);
}

/**
  * @brief  Display Audio Record demo hint
  * @param  None
  * @retval None
  */
static void AudioLoopback_SetHint(void)
{
  /* Clear the LCD */
  BSP_LCD_Clear(LCD_COLOR_WHITE);

  /* Set Audio Demo description */
  BSP_LCD_SetTextColor(LCD_COLOR_BLUE);
  BSP_LCD_FillRect(0, 0, BSP_LCD_GetXSize(), 90);
  BSP_LCD_SetTextColor(LCD_COLOR_WHITE);
  BSP_LCD_SetBackColor(LCD_COLOR_BLUE);
  BSP_LCD_SetFont(&Font16);
  BSP_LCD_DisplayStringAt(0, 0, (uint8_t *)"AUDIO LOOPBACK EXAMPLE", CENTER_MODE);
  BSP_LCD_SetFont(&Font8);
  BSP_LCD_DisplayStringAt(0, 30, (uint8_t *)"Press User button for next menu", CENTER_MODE);

  /* Set the LCD Text Color */
  BSP_LCD_SetTextColor(LCD_COLOR_BLUE);
  BSP_LCD_DrawRect(10, 100, BSP_LCD_GetXSize() - 20, BSP_LCD_GetYSize() - 80);
  BSP_LCD_DrawRect(11, 101, BSP_LCD_GetXSize() - 22, BSP_LCD_GetYSize() - 82);

}

/**
  * @brief Manages the DMA Transfer complete interrupt.
  * @param None
  * @retval None
  */
void BSP_AUDIO_IN_TransferComplete_CallBack(void *handle)
{
#if 1	/* ROI DEBUG */
 static int count = 0;
 if((count % 20) == 0){
 //  int i;
   syslog_1(LOG_NOTICE, "## BSP_AUDIO_IN_TransferComplete_CallBack(%d) ##", count);
   syslog_4(LOG_NOTICE, "## [%04x][%04x][%04x][%04x] ##", RecordBuffer[RECORD_BUFFER_SIZE/2], RecordBuffer[RECORD_BUFFER_SIZE/2+1], RecordBuffer[RECORD_BUFFER_SIZE/2+2], RecordBuffer[RECORD_BUFFER_SIZE/2+3]);
//   for(i = RECORD_BUFFER_SIZE/4 ; i < RECORD_BUFFER_SIZE ; i += 2){
//     RecordBuffer[i] = 0x003f;
//     RecordBuffer[i+1] = 0x0014;
//   }
 }
 count++;
#endif	/* ROI DEBUG */
  audio_rec_buffer_state = BUFFER_OFFSET_FULL;
}

/**
  * @brief  Manages the DMA Half Transfer complete interrupt.
  * @param  None
  * @retval None
  */
void BSP_AUDIO_IN_HalfTransfer_CallBack(void *handle)
{
#if 1	/* ROI DEBUG */
 static int count = 0;
 if((count % 100) == 0){
//   int i;
   syslog_1(LOG_NOTICE, "## BSP_AUDIO_IN_HalfTransfer_CallBack(%d) ##", count);
   syslog_4(LOG_NOTICE, "## [%04x][%04x][%04x][%04x] ##", RecordBuffer[0], RecordBuffer[1], RecordBuffer[2], RecordBuffer[3]);
//   for(i = 0 ; i < RECORD_BUFFER_SIZE/2 ; i++)
//     RecordBuffer[i] = 0xd1d1;
 }
 count++;
#endif	/* ROI DEBUG */
  audio_rec_buffer_state = BUFFER_OFFSET_HALF;
}

/**
  * @brief  Audio IN Error callback function.
  * @param  None
  * @retval None
  */
void BSP_AUDIO_IN_Error_CallBack(void)
{
  /* This function is called when an Interrupt due to transfer error on or peripheral
     error occurs. */
  /* Display message on the LCD screen */
  syslog_0(LOG_ERROR, "## BSP_AUDIO_IN_Error_CallBack ##");
  BSP_LCD_SetBackColor(LCD_COLOR_RED);
  BSP_LCD_DisplayStringAt(0, LINE(14), (uint8_t *)"       DMA  ERROR     ", CENTER_MODE);

#if 0	/* ROI DEBUG */
  /* Stop the program with an infinite loop */
  while (BSP_PB_GetState(BUTTON_WAKEUP) != RESET)
  {
    return;
  }
#endif	/* ROI DEBUG */
  /* could also generate a system reset to recover from the error */
  /* .... */
}

/**
  * @}
  */

/**
  * @}
  */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
