/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: bletest.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  SPIƥȤ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"
#include "spi.h"
#include "bletest.h"

#include <utility/lib_aci.h>
#include <utility/aci_setup.h>
#include "BLECentral.h"
#include "BLECharacteristic.h"
#include "BLEPeripheral.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define BLE_REQ   10
#define BLE_RDY   2
#define BLE_RST   7

static unsigned char _charvalue[BLE_ATTRIBUTE_MAX_VALUE_LENGTH+1];
#if APPLE_NO == 2
static unsigned char _charvalue2[BLE_ATTRIBUTE_MAX_VALUE_LENGTH+1];
#else
static unsigned char _descvalue[] = "value";
#endif

// create one or more services
static BLEService service;

// create one or more characteristics
static BLECharacteristic characteristic;
#if APPLE_NO == 2
static BLECharacteristic characteristic2;
static char rbuffer[512];
#endif

#if APPLE_NO != 2
// create one or more descriptors (optional)
static BLEDescriptor descriptor = {
  {{"2901", {0x01, 0x29, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, 2},
   BLETypeDescriptor},
  5,
  _descvalue,
  0,
};
#endif


// create peripheral instance, see pinouts above
// BLEPeripheral blePeripheral = BLEPeripheral(BLE_REQ, BLE_RDY, BLE_RST);
BLEPeripheral blePeripheral;

static uint8_t buffer[64];
static uint8_t key_req;

/*
 *  SW1
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
	key_req++;
}

void __ble_assert(const char *file, uint16_t line)
{
	syslog_2(LOG_ERROR, "assert:%s %d!", file, line);
}


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	SPI_Init_t spi_initd;
	SPI_Handle_t *hspi;
	ER_UINT	ercd;
	BLEPeripheral *pblep;
	BLECentral    *pcentral;
	int count = 0;
	int wcount = 0;
	bool_t connect = false;
#if APPLE_NO == 2
	T_SERIAL_RPOR k_rpor;
	int  rlen = 0;
	bool_t sflag = false;
#endif

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "BLE SEND/RECV START");
	spi_initd.Prescaler         = SPI_BAUDRATEPRESCALER_32 /*SPI_CLOCK_DIV8*/;
	spi_initd.Direction         = SPI_DIRECTION_2LINES;
	spi_initd.CLKPhase          = SPI_PHASE_1EDGE;
	spi_initd.CLKPolarity       = SPI_POLARITY_LOW;
	spi_initd.CRC               = SPI_CRC_DISABLE;
	spi_initd.CRCPolynomial     = 7;
	spi_initd.DataSize          = SPI_DATASIZE_8BIT;
	spi_initd.SignBit           = SPI_DATA_LSB;
	spi_initd.NSS               = SPI_NSS_SOFT;
	spi_initd.TIMode            = SPI_TIMODE_DISABLE;
	spi_initd.Mode = SPI_MODE_MASTER;
	spi_initd.semid = SPI1DMA_SEM;
	spi_initd.semlock = SPI1LOCK_SEM;
	if((hspi = spi_init(SPI1_PORTID, &spi_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_NOTICE, "## call Error_Handler(1) ##");
		slp_tsk();
	}

	pblep = &blePeripheral;
	BLEPeripheral_setup(pblep, BLE_REQ, BLE_RDY, BLE_RST);
	exti_func[10] = m_aci_isr;
#if APPLE_NO == 2
	BLEService_setup(&service, "713d0000503e4c75ba943148f18d941e");
	BLECharacteristic_setup(&characteristic, "713d0002503e4c75ba943148f18d941e", BLENotify, _charvalue, sizeof(_charvalue));
	BLECharacteristic_setup(&characteristic2, "713d0003503e4c75ba943148f18d941e", BLEWriteWithoutResponse, _charvalue2, sizeof(_charvalue2));
	BLEPeripheral_setLocalName(pblep, "ble-Micro");
#elif APPLE_NO == 1
	BLEService_setup(&service, "19b10000e8f2537e4f6cd104768a1214");
	BLECharacteristic_setup(&characteristic, "19b10001e8f2537e4f6cd104768a1214", (BLERead | BLEWrite), _charvalue, sizeof(_charvalue));
	BLEPeripheral_setLocalName(pblep, "LED");
#else
	BLEService_setup(&service, "fff0");
	BLECharacteristic_setup(&characteristic, "fff1", (BLERead | BLEWrite), _charvalue, sizeof(_charvalue));
	BLEPeripheral_setLocalName(pblep, "local-name");
#endif

	BLEPeripheral_setAdvertisedServiceUuid(pblep, service._uuid._str);

	pblep->_nRF8001._aciState.aci_pins.spihandle         = hspi;
	pblep->_nRF8001._aciState.aci_pins.interface_is_interrupt = true;
	pblep->_nRF8001._aciState.aci_pins.event_taskid      = ACI_TASK;

	pblep->_nRF8001.nRF8001Connected = BLEPeripheral_nRF8001Connected;
	pblep->_nRF8001.nRF8001Disconnected = BLEPeripheral_nRF8001Disconnected;
	pblep->_nRF8001.nRF8001CharacteristicValueChanged = BLEPeripheral_nRF8001CharacteristicValueChanged;
	pblep->_nRF8001.nRF8001CharacteristicSubscribedChanged = BLEPeripheral_nRF8001CharacteristicSubscribedChanged;
	pblep->_nRF8001.nRF8001AddressReceived = BLEPeripheral_nRF8001AddressReceived;
	pblep->_nRF8001.nRF8001TemperatureReceived = BLEPeripheral_nRF8001TemperatureReceived;
	pblep->_nRF8001.nRF8001BatteryLevelReceived = BLEPeripheral_nRF8001BatteryLevelReceived;


	// add attributes (services, characteristics, descriptors) to peripheral
    BLEPeripheral_addAttribute(pblep, &service);
    BLEPeripheral_addAttribute(pblep, (BLEAttribute*)&characteristic);
#if APPLE_NO == 2
    BLEPeripheral_addAttribute(pblep, (BLEAttribute*)&characteristic2);
#else
    BLEPeripheral_addAttribute(pblep, (BLEAttribute*)&descriptor);
#endif

	// set initial value
	memset(_charvalue, 0, sizeof(_charvalue));

	// begin initialization
	BLEPeripheral_begin(pblep);
	syslog_2(LOG_NOTICE, "## addr[%08x][%08x] ##\n", &service, &characteristic);

	dly_tsk(100);
	BLEPeripheral_version(pblep);

    while(count < 50000){
		pcentral = BLEPeripheral_central(pblep);
		char value[16];
		uint8_t *p;

		while(BLECentral_connected(pcentral)){
			if(!connect){
        	    // central connected to peripheral
				p = pcentral->_address;
				syslog_3(LOG_NOTICE, "## central Connected to periphral1 %02x:%02x:%02x ##", p[0], p[1], p[2]);
				syslog_3(LOG_NOTICE, "## central Connected to periphral2 %02x:%02x:%02x ##", p[3], p[4], p[5]);
				connect = true;
			}
			// central still connected to peripheral
			if(BLECharacteristic_written(&characteristic)){
#if APPLE_NO == 2
				unsigned char len = characteristic2._valueLength;
				const char *val = (const char *)characteristic2._value;
				syslog_1(LOG_NOTICE, "didCharacteristicWritten, Length:%d", len);
				serial_wri_dat(TASK_PORTID, val, len);
				serial_wri_dat(TASK_PORTID, "\r\n", 2);
			}
			serial_ref_por(TASK_PORTID, &k_rpor);
			if(k_rpor.reacnt > 0){
				int i;
				for(i = 0 ; i < k_rpor.reacnt ; i++, rlen++){
					serial_rea_dat(TASK_PORTID, &rbuffer[rlen], 1);
					if(rbuffer[rlen] == '\r' || rbuffer[rlen] == '\n')
						sflag = true;
				}
			}
			if(sflag && rlen > 0){
				syslog_1(LOG_NOTICE, "write(%d)", rlen);
				BLECharacteristic_setValue1(&characteristic, (const unsigned char *)rbuffer, rlen);
				sflag = false;
				rlen =0;
			}
#else
				// central wrote new value to characteristic
				memcpy(buffer, BLECharacteristic_getvalue(&characteristic), BLECharacteristic_getvalueLength(&characteristic));
				buffer[BLECharacteristic_getvalueLength(&characteristic)] = 0;
				syslog_2(LOG_NOTICE, "value[%s] notified(%d)", buffer, BLECharacteristic_notified(&characteristic));
				dly_tsk(100);
			}
                  // set value on characteristic
#if APPLE_NO == 0
			if(key_req != 0){
				sprintf(value, "%d", wcount++);
				BLECharacteristic_setValue(&characteristic, pcentral, (uint8_t *)value, strlen(value));
				key_req = 0;
			}
#endif
#endif
			dly_tsk(5);
		}
		if(connect){
			/* central disconnected */
			syslog_0(LOG_NOTICE, "## central Disconnected ##");
			connect = false;
		}

		dly_tsk(100);
		if((count % 10) == 0)
			syslog_1(LOG_NOTICE, "## wait count(%d) ##", count);
		count++;
	}

	syslog(LOG_NOTICE, "Sample program ends.");
	slp_tsk();
//	SVC_PERROR(ext_ker());
}
