/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: usbdtest.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  USB DEVICEƥȤ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <target_syssvc.h>
#include "device.h"

#include "usb_device.h"
#include "usbdtest.h"

#include "tusbd_base.h"
#include "tusbd_device.h" 



/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

USBD_HandleTypeDef USBD_Device;
uint8_t ukey;

extern void MakeUsbDescriptor(USBD_HandleTypeDef *);
extern void setup_usbcontrol(void *func);

void sw_handle(void)
{
	ukey = 1;
#ifdef USE_BCD
	SVC_PERROR(iact_tsk(BCD_TASK));
#endif
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

#define CURSOR_STEP     5

uint8_t usb_mode = 0;

#if USB_DEVICE_ID == 0

uint8_t HID_Buffer[4];

/*
 *  HID MOUSE DEVICE CLASS REQUEST
 *  parameter1:  pbuf  ǡХåե
 */
static void GetPointerData(uint8_t *pbuf)
{
  static int8_t cnt = 0;
  int8_t  x = 0, y = 0 ;
  
  if(cnt++ > 0)
  {
    x = CURSOR_STEP;
  }
  else
  {
    x = -CURSOR_STEP;
  }
  pbuf[0]  = 0;
  pbuf[1] += x;
  pbuf[2] += y;
  pbuf[3]  = 0;
}
#else

#define APP_TX_DATA_SIZE   2048

#define FLG_CDC_SET_LINE_CODING         0x0001
#define FLG_CDC_GET_LINE_CODING         0x0002
#define FLG_CDC_SET_CONTROL_LINE_STATE  0x0004
#define FLG_CDC_SEND_BREAK              0x0008
#define FLG_CDC_RECEIVE_DATA           	0x0010
#define FLG_CDC_SEND_FLUSH              0x0100

static uint8_t    txBuffer[APP_TX_DATA_SIZE] __attribute__ ((aligned (32)));/* Received Data over UART (CDC interface) are stored in this buffer */
static uint8_t    *pTransRxBuffer;
static uint16_t   TransRxLength;
static uint16_t   TransRxCount;


Line_Coding_t LineCoding =
{
	115200, /* baud rate*/
	0x00,   /* stop bits-1*/
	0x00,   /* parity - none*/
	0x08    /* nb. of bits 8*/
};

/*
 *  CDC DEVICE CLASS REQUEST
 *  parameter1:  hcdc   CDCϥɥ
 *  parameter2:  cmd    ׵ᥳޥ
 *  parameter3:  length ǡĹ
 */
static void CdcDeviceRequest(TUSBD_CDC_Handle_t *hcdc, uint8_t cmd, uint32_t length)
{
	switch (cmd){
	case CDC_SEND_ENCAPSULATED_COMMAND:
		syslog_2(LOG_NOTICE, "## CdcDeviceRequest CDC_SEND_ENCAPSULATED_COMMAND hcdc[%08x] length(%d) ##", hcdc, length);
		break;

	case CDC_GET_ENCAPSULATED_RESPONSE:
		syslog_2(LOG_NOTICE, "## CdcDeviceRequest CDC_GET_ENCAPSULATED_RESPONSE hcdc[%08x] length(%d) ##", hcdc, length);
		break;

	case CDC_SET_COMM_FEATURE:
		syslog_2(LOG_NOTICE, "## CdcDeviceRequest CDC_SET_COMM_FEATURE hcdc[%08x] length(%d) ##", hcdc, length);
		break;

	case CDC_GET_COMM_FEATURE:
		syslog_2(LOG_NOTICE, "## CDC_GET_COMM_FEATURE hcdc[%08x] length(%d) ##", hcdc, length);
		break;

	case CDC_CLEAR_COMM_FEATURE:
		syslog_2(LOG_NOTICE, "## CDC_CLEAR_COMM_FEATURE hcdc[%08x] length(%d) ##", hcdc, length);
		break;

	case CDC_SET_LINE_CODING:
		memcpy(&LineCoding, hcdc->cmddata, sizeof(Line_Coding_t));
		iset_flg(USBCDC_FLG, FLG_CDC_SET_LINE_CODING);
		break;

	case CDC_GET_LINE_CODING:
		memcpy(hcdc->cmddata, &LineCoding, sizeof(Line_Coding_t));
		break;

	case CDC_SET_CONTROL_LINE_STATE:
		iset_flg(USBCDC_FLG, FLG_CDC_SET_CONTROL_LINE_STATE);
		break;

	case CDC_SEND_BREAK:
		iset_flg(USBCDC_FLG, FLG_CDC_SEND_BREAK);
		break;

	case CDC_RECEIVED:
		pTransRxBuffer = hcdc->rxdata;
		TransRxLength = length;
		TransRxCount  = 0;
		iset_flg(USBCDC_FLG, FLG_CDC_RECEIVE_DATA);
		break;

	default:
		syslog_3(LOG_NOTICE, "## CdcDeviceRequest cmd[%02x] hcdc[%08x] length(%d) ##", cmd, hcdc, length);
		break;
	}
}
#endif


static void
usbd_bcd_callback(USB_DEV_Handle_t *husb, int8_t mes)
{
	syslog_2(LOG_NOTICE, "## usbd_bcd_callback [%08x] (%d) ##", husb, mes);
}

static TUSBD_ERCODE
usbd_control(int mode)
{
	USB_DEV_Init_t USB_Data_Init;
	USB_DEV_Handle_t *husb;
	TUSBD_ERCODE   ercd = TUSBD_E_OK;

	if(mode == 1){
		/*
		 *  USB DEVICE
		 */
		memset(&USB_Data_Init, 0, sizeof(USB_DEV_Init_t));
		USB_Data_Init.dev_endpoints = 8;
		USB_Data_Init.phy_itface = USB_PHY_EMBEDDED;
		USB_Data_Init.speed = USB_SPEED_FULL;
#if 0	/* ROI DEBUG */
		USB_Data_Init.low_power_enable = 1;
#else	/* ROI DEBUG */
		USB_Data_Init.low_power_enable = 0;
#endif	/* ROI DEBUG */
#ifdef USE_BCD
		USB_Data_Init.battery_charging_enable = 1;
#endif
		husb = usbd_init(USB1_PORTID, &USB_Data_Init);
		if(husb == NULL)
			return TUSBD_E_ERR;
		husb->devbcdcallback = usbd_bcd_callback;

		/*
		 *  USB DEVICEߥɥ륦
		 */
		USBD_Device.pSysData = husb;
		ercd = tusbdInit(&USBD_Device, USB_DEVICE_ID);
		if(ercd != TUSBD_E_OK)
			return ercd;
		MakeUsbDescriptor(&USBD_Device);

#if USB_DEVICE_ID != 0
		USBD_Device.pUsrData = CdcDeviceRequest;
#endif

		/*
		 *  USB DEVICE
		 */
		ercd = tusbdStart(&USBD_Device);
	}
	else{
		ercd = tusbdStop(&USBD_Device);
		usbd_deinit(USBD_Device.pSysData);
	}
	usb_mode = mode;
	return ercd;
}

/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	ER_UINT	ercd;
	TUSBD_ERCODE result;
	FLGPTN crcflag;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	dly_tsk(300);
	/*
	 *  USB DEVICE
	 */
	result = usbd_control(1);
	syslog_1(LOG_NOTICE, "usbd_control result(%d) !", result);
#if 0	/* ROI DEBUG */
	setup_usbcontrol(usbd_control);
#endif	/* ROI DEBUG */

	while (1){
		ercd = twai_flg(USBCDC_FLG, 0xffff, TWF_ORW, &crcflag, 1000);
		if(usb_mode == 1){
#if USB_DEVICE_ID == 0
			GetPointerData(HID_Buffer);
			if(USBD_HID_SendReport(&USBD_Device, HID_Buffer, 4) == TUSBD_E_OK)
				memset(&HID_Buffer, 0, sizeof(HID_Buffer));
#else
			if(ercd == E_OK){
				if((crcflag & FLG_CDC_RECEIVE_DATA) != 0 && TransRxLength > TransRxCount){
					pTransRxBuffer[TransRxLength] = 0;
					syslog_2(LOG_NOTICE, "## revlen(%d)[%s] ##", TransRxLength, pTransRxBuffer);
					memcpy(txBuffer, pTransRxBuffer, TransRxLength);
					tusbdCdcStartTransmit(&USBD_Device, txBuffer, TransRxLength);
					TransRxLength = 0;
					tusbdCdcSetReceivePacket(&USBD_Device);
				}
			}
#endif
			syslog_1(LOG_NOTICE, "## dev_state(%d) ##", USBD_Device.dev_state);
		}
	}

#if 1	/* ROI DEBUG */
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
#endif	/* ROI DEBUG */

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

