/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: i2ctest.c 2416 2020-09-29 10:09:46Z roi $
 */

/* 
 *  I2CƥȤ
 *  SCL/SDA10KΥץꥢå񹳤NACKȯ뤿ᡢ
 *  ľ³Ȥʤޤ
 *  EEPROMƱͤ³ơEEPROM_ADDRꤹEEPROM
 *  Ǥޤ
 *   _________________________              __________________ 
 *  |          |I2C5          |            |  ADT7410-DIP |
 *  |          |              |            |                  |
 *  |          |          SCL |____________| SCL              |
 *  |          |              |            |                  |
 *  |          |              |            |                  |
 *  |          |          SDA |____________| SDA              |
 *  |          |              |            |                  |
 *  |          |              |            |                  |
 *  |          |           GND|____________|GND               |
 *  |          |              |            |                  |
 *  |                     3.3V|____________|3.3V              |
 *  |_STM32_Board 1___________|            |__________________|
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <string.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <target_syssvc.h>
#include "device.h"
#include "i2ctest.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define RXBUFFERSIZE       8

/*
 *  I2C TIMING Register define when I2C clock source is APB(100MHz:SYSCLK/4)
 *  This example use TIMING to 0x00901954 to reach 400 kHz speed (Rise time = 100 ns, Fall time = 10 ns)
 */
#define I2C_TIMING      0x00701837

#define LCD_ADDR           (0x3E<<1)
//#define EEPROM_ADDR        (0x50<<1)
#define ADT7410_ADDR       (0x48<<1)

#define ADT7410_TEMP_H	0x00
#define ADT7410_TEMP_L	0x01
#define ADT7410_STATUS	0x02

/* Buffer used for reception */
uint8_t aRxBuffer[RXBUFFERSIZE];
uint8_t aRxBuffer2[RXBUFFERSIZE];


/*
 *  桼åߴؿ
 */
void sw_handle(intptr_t exinf)
{
	syslog_1(LOG_NOTICE, "## sw int(%d) ##", (int)exinf);
}

/*
 *  I2C SEND CALLBACK FUNCTION
 */
static void
I2C_TxCpltCallback(I2C_Handle_t *hi2c)
{
	syslog_1(LOG_INFO, "## I2C_TxCpltCallback(%08x) ##", hi2c);
}

/*
 *  I2C RECEIVE CALLBACK
 */
static void
I2C_RxCpltCallback(I2C_Handle_t *hi2c)
{
	syslog_1(LOG_INFO, "## I2C_RxCpltCallback(%08x) ##", hi2c);
}

/*
 *  I2C ERROR CALLBACK
 */
static void
I2C_ErrorCallback(I2C_Handle_t *hi2c)
{
	syslog_2(LOG_ERROR, "## I2C_ErrorCallback(%08x)[%08x] ##", hi2c, hi2c->ErrorCode);
}


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	I2C_Init_t i2c_initd;
	I2C_Handle_t *hi2c;
	ER_UINT	ercd;
	uint32_t temp;
	uint32_t count = 0;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_NOTICE), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "I2C SEND/RECV START");

    i2c_initd.Timing          = I2C_TIMING;
	i2c_initd.AddressingMode  = I2C_ADDRESSINGMODE_7BIT;
	i2c_initd.DualAddressMode = I2C_DUALADDRESS_DISABLE;
	i2c_initd.GeneralCallMode = I2C_GENERALCALL_DISABLE;
	i2c_initd.NoStretchMode   = I2C_NOSTRETCH_DISABLE;
	i2c_initd.OwnAddress1     = 0;
	i2c_initd.OwnAddress2     = 0;
	i2c_initd.AnalogFilter    = I2C_ANALOGFILTER_ENABLE;
	i2c_initd.DigitalFilter   = 0;
	i2c_initd.semid           = I2CTRS_SEM;
	i2c_initd.semlock         = I2CLOC_SEM;

	if((hi2c = i2c_init(I2C_PORTID, &i2c_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_ERROR, "## I2C ERROR(1) ##");
	}
	i2c_setfastmodeplus(I2C_PORTID, true);
	syslog_0(LOG_NOTICE, "LCD INITAILIZE !");

	hi2c->writecallback = I2C_TxCpltCallback;
	hi2c->readcallback  = I2C_RxCpltCallback;
	hi2c->errorcallback = I2C_ErrorCallback;

#ifdef EEPROM_ADDR
	syslog_1(LOG_NOTICE, "READ EEPROM [%08x] !", aRxBuffer2);
	if(i2c_memread(hi2c, (uint16_t)EEPROM_ADDR, 0, I2C_MEMADD_SIZE_16BIT, (uint8_t*)aRxBuffer2, RXBUFFERSIZE)!= E_OK){
	    /* I2C RECEIVE ERROR */
		syslog_1(LOG_ERROR, "## I2C ERROR(12)[%08x] ##", hi2c->ErrorCode);
		goto stop_task;
	}
#endif

	syslog_0(LOG_NOTICE, "TEMP SENSER !");
	do{
		/* STATUS READ */
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_STATUS, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(13)[%08x] ##", hi2c->ErrorCode);
			goto stop_task;
		}
		temp = aRxBuffer[0];
	}while((temp & 0x80) != 0);

	while(count < 1000){
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_H, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(14)[%08x] ##", hi2c->ErrorCode);
			goto stop_task;
		}
		temp = aRxBuffer[0] << 8;
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_L, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(15)[%08x] ##", hi2c->ErrorCode);
			goto stop_task;
		}
		temp += aRxBuffer[0];
		temp /= 13;

		if((count % 10) == 0)
			syslog_2(LOG_NOTICE, "count(%d) temp(%d)", count, temp);
		dly_tsk(1000);
		count++;
	}

stop_task:
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}
