/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: lcdtest.c 2416 2020-05-11 10:25:10Z roi $
 */

/* 
 *  IWHI 1.54" LCDץ
 *  SPI̿楿ॢȤȯ롣
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "spi.h"
#include "adafruit_st7789.h"
#include "lcdtest.h"
#include "topamelogo.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))
#define ABS(X)  ((X) > 0 ? (X) : -(X))

#if LCDTYPE == 1
#define LCD_WIDTH    135
#define LCD_HEIGHT   240
#else
#define LCD_WIDTH    240
#define LCD_HEIGHT   240
#endif

/*
 *  Ѥѿ
 */
static const Point testPolygon[7] = {
	{0, 0},
	{LCD_WIDTH-1, LCD_HEIGHT-1},
	{LCD_WIDTH/2, 0},
	{LCD_WIDTH/2, LCD_HEIGHT-1},
	{LCD_WIDTH-1, 0},
	{0, LCD_HEIGHT-1},
	{0, 0}
};

LCD_Handler_t  LcdHandle;
LCD_DrawProp_t DrawProp;


/*
 *  SW1
 */
void sw_handle(intptr_t exinf)
{
	syslog_1(LOG_NOTICE, "## sw int(%d) ##", (int)exinf);
}


/*
 *  GPIOؿ
 */
static void
pinMode(uint8_t no, uint32_t mode)
{
	Arduino_PortControlBlock *pgcb = getGpioTable(DIGITAL_PIN, no);
	GPIO_Init_t GPIO_Init_Data;

	if(pgcb == NULL)
		return;
	pinClock(no);
	GPIO_Init_Data.mode  = mode;
	GPIO_Init_Data.pull  = GPIO_PULLUP;
	GPIO_Init_Data.otype = GPIO_OTYPE_PP;
	GPIO_Init_Data.speed = GPIO_SPEED_FAST;
	gpio_setup(pgcb->giobase, &GPIO_Init_Data, pgcb->giopin);
}

/*
 *  եåɽƥ
 */
void
grapics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ST7789_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ST7789_GREEN);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5){
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ST7789_RED);
	}
	for (x=0; x < hlcd->_width; x+=5){
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ST7789_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7789_GREEN;
	lcd_drawPolygon(&DrawProp, (Point *)testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ST7789_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 33, 48, (uint8_t *)topamelogo);
	dly_tsk(1000);
}

/*
 *  ᥤ󥿥
 */
void
main_task(intptr_t exinf)
{
	SPI_Init_t spi_initd;
	LCD_Handler_t   *hlcd;
	SPI_Handle_t  *hspi;
	ER_UINT	ercd;
	int i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "## START ##");
	/*
	 *  CS PORT(D10-PB6)
	 */
	pinMode(CS_PORT, GPIO_MODE_OUTPUT);
	/*
	 *  RS PORT(D8-PA9)
	 */
	pinMode(RS_PORT, GPIO_MODE_OUTPUT);
	/*
	 *  RST PORT(D9-PC7)
	 */
	pinMode(RST_PORT, GPIO_MODE_OUTPUT);
	/*
	 *  CS2 PORT(D4-PB5)
	 */
	pinMode(CS2_PORT, GPIO_MODE_OUTPUT);

	spi_initd.Prescaler     = SPI_BAUDRATEPRESCALER_128;
	spi_initd.Direction         = SPI_DIRECTION_2LINES;
	spi_initd.CLKPhase          = SPI_PHASE_1EDGE;
	spi_initd.CLKPolarity       = SPI_POLARITY_HIGH;
	spi_initd.CRC               = SPI_CRC_DISABLE;
	spi_initd.CRCPolynomial     = 7;
	spi_initd.CRCLength         = SPI_CRC_LENGTH_DATASIZE;
	spi_initd.DataSize          = SPI_DATASIZE_8BIT;
	spi_initd.SignBit           = SPI_DATA_MSB;
	spi_initd.NSS               = SPI_NSS_SOFT;
	spi_initd.NSSPMode          = SPI_NSS_PULSE_DISABLE;
	spi_initd.NSSPolarity       = SPI_NSS_POLARITY_HIGH;
	spi_initd.TIMode            = SPI_TIMODE_DISABLE;
	spi_initd.TxCRCInitializationPattern = SPI_CRC_INITIALIZATION_ALL_ONE_PATTERN;
	spi_initd.RxCRCInitializationPattern = SPI_CRC_INITIALIZATION_ALL_ONE_PATTERN;
	spi_initd.MasterSSIdleness  = SPI_MASTER_SS_IDLENESS_00CYCLE;
	spi_initd.MasterInterDataIdleness = SPI_MASTER_INTERDATA_IDLENESS_00CYCLE;
	spi_initd.MasterKeepIOState = SPI_MASTER_KEEP_IO_STATE_ENABLE;
	spi_initd.IOSwap            = SPI_IO_SWAP_DISABLE;
	spi_initd.GpioClkPull       = -1;
	spi_initd.GpioDataPull      = -1;
	spi_initd.TxDMAStream       = TXDMA_SID;
	spi_initd.RxDMAStream       = RXDMA_SID;
	spi_initd.Mode = SPI_MODE_MASTER;
	spi_initd.semid = SPI1DMA_SEM;
	spi_initd.semlock = 0;

	if((hspi = spi_init(SPI_PORTID, &spi_initd)) == NULL){
		syslog_0(LOG_ERROR, "spi_start initialize error !");
		slp_tsk();
	}

	hlcd = &LcdHandle;
	hlcd->hspi = hspi;
	hlcd->rotation = 0;
	hlcd->spi_lock = SPI1LOCK_SEM;
	hlcd->cs_pin   = CS_PORT;
	hlcd->rs_pin   = RS_PORT;
	hlcd->rst_pin  = RST_PORT;
	hlcd->cs2_pin  = CS2_PORT;
	DrawProp.hlcd  = hlcd;
	lcd_init(hlcd, LCD_WIDTH, LCD_HEIGHT);
	DrawProp.BackColor = ST7789_BLACK;
	DrawProp.TextColor = ST7789_WHITE;
	grapics_test(hlcd);
#if 1	/* LOCAL DEFINE */
	DrawProp.TextColor = ST7789_BLACK;
	DrawProp.BackColor = ST7789_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ST7789_BLUE);
	DrawProp.TextColor = ST7789_WHITE;
	DrawProp.BackColor = ST7789_BLUE;
	DrawProp.pFont = &Font24;
	lcd_DisplayStringAt(&DrawProp, 0, 30, (uint8_t *)"TBP TEST", CENTER_MODE);

	DrawProp.TextColor = ST7789_BLACK;
	DrawProp.BackColor = ST7789_WHITE;
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 100, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font16;
	lcd_DisplayStringAt(&DrawProp, 0, 130, (uint8_t *)"Helle World !", CENTER_MODE);
	DrawProp.pFont = &Font20;
	lcd_DisplayStringAt(&DrawProp, 0, 160, (uint8_t *)"Helle World !", CENTER_MODE);

	DrawProp.TextColor = ST7789_RED;
	for(i = 20 ; i < 40 ; i++){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2+30, i);
		if(i == 25)
			i += 10;
	}
#endif	/* LOCAL DEFINE */
	slp_tsk();
	spi_deinit(hspi);

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

