/* USER CODE BEGIN Header */
/**
 ******************************************************************************
 * @file    p2p_server_app.c
 * @author  MCD Application Team
 * @brief   peer to peer Server Application
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; Copyright (c) 2019 STMicroelectronics.
 * All rights reserved.</center></h2>
 *
 * This software component is licensed by ST under Ultimate Liberty license
 * SLA0044, the "License"; You may not use this file except in compliance with
 * the License. You may obtain a copy of the License at:
 *                             www.st.com/SLA0044
 *
 ******************************************************************************
 */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"
#include "app_common.h"
#include "dbg_trace.h"
#include "ble.h"
#include "p2p_server_app.h"
#include "stm32_seq.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */

/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */
 typedef struct{
    uint8_t             Device_Led_Selection;
    uint8_t             Led1;
 }P2P_LedCharValue_t;

 typedef struct{
    uint8_t             Device_Button_Selection;
    uint8_t             ButtonStatus;
 }P2P_ButtonCharValue_t;

#if 1	/* ROI MODIFIED */
 typedef struct{
    uint8_t             Device_Temp_Selection;
    uint8_t             TempStatus[3];
 }P2P_TempCharValue_t;
#endif	/* ROI MODIFIED */

typedef struct
{
  uint8_t               Notification_Status; /* used to chek if P2P Server is enabled to Notify */
  P2P_LedCharValue_t    LedControl;
  P2P_ButtonCharValue_t ButtonControl;
#if 1	/* ROI MODIFIED */
  P2P_TempCharValue_t   TempControl;
#endif	/* ROI MODIFIED */
  uint16_t              ConnectionHandle;
} P2P_Server_App_Context_t;
/* USER CODE END PTD */

/* Private defines ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

/* USER CODE END PD */

/* Private macros -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
/* USER CODE BEGIN PV */
/**
 * START of Section BLE_APP_CONTEXT
 */

PLACE_IN_SECTION("BLE_APP_CONTEXT") static P2P_Server_App_Context_t P2P_Server_App_Context;

/**
 * END of Section BLE_APP_CONTEXT
 */
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
/* USER CODE BEGIN PFP */
static void P2PS_Send_Notification(void);
#if 1	/* ROI MODIFIED */
static void P2PS_Send2_Data(void);
#endif	/* ROI MODIFIED */
static void P2PS_APP_LED_BUTTON_context_Init(void);
/* USER CODE END PFP */

/* Functions Definition ------------------------------------------------------*/
void P2PS_STM_App_Notification(P2PS_STM_App_Notification_evt_t *pNotification)
{
/* USER CODE BEGIN P2PS_STM_App_Notification_1 */

#if 1	/* ROI DEBUG */
  syslog_5(LOG_NOTICE, "## P2PS_STM_App_Notification P2P_Evt_Opcode(%d) data0[%02x] data1[%02x] connect[%04x] serv[%02x] ##",
		pNotification->P2P_Evt_Opcode,
		pNotification->DataTransfered.pPayload[0],
		pNotification->DataTransfered.pPayload[1],
		pNotification->ConnectionHandle,
		pNotification->ServiceInstance);
#endif	/* ROI DEBUG */
/* USER CODE END P2PS_STM_App_Notification_1 */
  switch(pNotification->P2P_Evt_Opcode)
  {
/* USER CODE BEGIN P2PS_STM_App_Notification_P2P_Evt_Opcode */
#if(BLE_CFG_OTA_REBOOT_CHAR != 0)
    case P2PS_STM_BOOT_REQUEST_EVT:
      syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER : BOOT REQUESTED\n");

      *(uint32_t*)SRAM1_BASE = *(uint32_t*)pNotification->DataTransfered.pPayload;
      NVIC_SystemReset();
      break;
#endif
/* USER CODE END P2PS_STM_App_Notification_P2P_Evt_Opcode */

    case P2PS_STM__NOTIFY_ENABLED_EVT:
/* USER CODE BEGIN P2PS_STM__NOTIFY_ENABLED_EVT */
      P2P_Server_App_Context.Notification_Status = 1;
      syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER : NOTIFICATION ENABLED\n"); 
/* USER CODE END P2PS_STM__NOTIFY_ENABLED_EVT */
      break;

    case P2PS_STM_NOTIFY_DISABLED_EVT:
/* USER CODE BEGIN P2PS_STM_NOTIFY_DISABLED_EVT */
      P2P_Server_App_Context.Notification_Status = 0;
      syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER : NOTIFICATION DISABLED\n");
/* USER CODE END P2PS_STM_NOTIFY_DISABLED_EVT */
      break;

    case P2PS_STM_WRITE_EVT:
/* USER CODE BEGIN P2PS_STM_WRITE_EVT */
      if(pNotification->DataTransfered.pPayload[0] == 0x00){ /* ALL Deviceselected - may be necessary as LB Routeur informs all connection */
        if(pNotification->DataTransfered.pPayload[1] == 0x01)
        {
          BSP_LED_On(LED_BLUE);
          syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER  : LED1 ON"); 
          P2P_Server_App_Context.LedControl.Led1=0x01; /* LED1 ON */
        }
        if(pNotification->DataTransfered.pPayload[1] == 0x00)
        {
          BSP_LED_Off(LED_BLUE);
          syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER  : LED1 OFF"); 
          P2P_Server_App_Context.LedControl.Led1=0x00; /* LED1 OFF */
        }
      }
				/* end device 1 selected - may be necessary as LB Routeur informs all connection */
      if(pNotification->DataTransfered.pPayload[0] == 0x01 || pNotification->DataTransfered.pPayload[0] == '1'){
        if(pNotification->DataTransfered.pPayload[1] == 0x01 || pNotification->DataTransfered.pPayload[1] == '1')
        {
          BSP_LED_On(LED_BLUE);
          syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER 1 : LED1 ON"); 
          P2P_Server_App_Context.LedControl.Led1=0x01; /* LED1 ON */
        }
        if(pNotification->DataTransfered.pPayload[1] == 0x00 || pNotification->DataTransfered.pPayload[1] == '0')
        {
          BSP_LED_Off(LED_BLUE);
          syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER 1 : LED1 OFF"); 
          P2P_Server_App_Context.LedControl.Led1=0x00; /* LED1 OFF */
        }
      }
/* USER CODE END P2PS_STM_WRITE_EVT */
      break;

    default:
/* USER CODE BEGIN P2PS_STM_App_Notification_default */
      
/* USER CODE END P2PS_STM_App_Notification_default */
      break;
  }
/* USER CODE BEGIN P2PS_STM_App_Notification_2 */

/* USER CODE END P2PS_STM_App_Notification_2 */
  return;
}

void P2PS_APP_Notification(P2PS_APP_ConnHandle_Not_evt_t *pNotification)
{
#if 1	/* ROI DEBUG */
  syslog_1(LOG_NOTICE, "## P2PS_APP_Notification P2P_Evt_Opcode(%d) ##", pNotification->P2P_Evt_Opcode);
#endif	/* ROI DEBUG */
/* USER CODE BEGIN P2PS_APP_Notification_1 */

/* USER CODE END P2PS_APP_Notification_1 */
  switch(pNotification->P2P_Evt_Opcode)
  {
/* USER CODE BEGIN P2PS_APP_Notification_P2P_Evt_Opcode */

/* USER CODE END P2PS_APP_Notification_P2P_Evt_Opcode */
  case PEER_CONN_HANDLE_EVT :
/* USER CODE BEGIN PEER_CONN_HANDLE_EVT */
          
/* USER CODE END PEER_CONN_HANDLE_EVT */
    break;

    case PEER_DISCON_HANDLE_EVT :
/* USER CODE BEGIN PEER_DISCON_HANDLE_EVT */
       P2PS_APP_LED_BUTTON_context_Init();       
/* USER CODE END PEER_DISCON_HANDLE_EVT */
    break;

    default:
/* USER CODE BEGIN P2PS_APP_Notification_default */

/* USER CODE END P2PS_APP_Notification_default */
      break;
  }
/* USER CODE BEGIN P2PS_APP_Notification_2 */

/* USER CODE END P2PS_APP_Notification_2 */
  return;
}

void P2PS_APP_Init(void)
{
/* USER CODE BEGIN P2PS_APP_Init */
  UTIL_SEQ_RegTask( 1<< CFG_TASK_SW1_BUTTON_PUSHED_ID, UTIL_SEQ_RFU, P2PS_Send_Notification );

#if 1	/* ROI MODIFIED */
  UTIL_SEQ_RegTask( 1<< CFG_TASK_CONN_UPDATE_REG_ID, UTIL_SEQ_RFU, P2PS_Send2_Data );
#endif	/* ROI MODIFIED */

  /**
   * Initialize LedButton Service
   */
  P2P_Server_App_Context.Notification_Status=0; 
  P2PS_APP_LED_BUTTON_context_Init();
/* USER CODE END P2PS_APP_Init */
  return;
}

/* USER CODE BEGIN FD */
void P2PS_APP_LED_BUTTON_context_Init(void){

  BSP_LED_Off(LED_BLUE);

  P2P_Server_App_Context.LedControl.Device_Led_Selection=0x01; /* Device1 */
  P2P_Server_App_Context.LedControl.Led1=0x00; /* led OFF */
  P2P_Server_App_Context.ButtonControl.Device_Button_Selection=0x01;/* Device1 */
  P2P_Server_App_Context.ButtonControl.ButtonStatus=0x00;
#if 1	/* ROI MODIFIED */
  P2P_Server_App_Context.TempControl.Device_Temp_Selection=0x01;
  P2P_Server_App_Context.TempControl.TempStatus[0] = 0x00;
  P2P_Server_App_Context.TempControl.TempStatus[1] = 0x00;
  P2P_Server_App_Context.TempControl.TempStatus[2] = 0x00;
#endif	/* ROI MODIFIED */
}

void P2PS_APP_SW1_Button_Action(void)
{
  UTIL_SEQ_SetTask( 1<<CFG_TASK_SW1_BUTTON_PUSHED_ID, CFG_SCH_PRIO_0);

  return;
}

#if 1	/* ROI MODIFIED */
void P2PS_APP_Temp_Action(void)
{
  UTIL_SEQ_SetTask( 1<<CFG_TASK_CONN_UPDATE_REG_ID, CFG_SCH_PRIO_0);

  return;
}
#endif	/* ROI MODIFIED */

/* USER CODE END FD */
#if 1	/* ROI DEBUG */
extern uint16_t temp;
#endif	/* ROI DEBUG */
/*************************************************************
 *
 * LOCAL FUNCTIONS
 *
 *************************************************************/
/* USER CODE BEGIN FD_LOCAL_FUNCTIONS*/
void P2PS_Send_Notification(void)
{
 
  if(P2P_Server_App_Context.ButtonControl.ButtonStatus == 0x00){
    P2P_Server_App_Context.ButtonControl.ButtonStatus=0x01;
  } else {
    P2P_Server_App_Context.ButtonControl.ButtonStatus=0x00;
  }
  
   if(P2P_Server_App_Context.Notification_Status){ 
    syslog_0(LOG_NOTICE, "-- P2P APPLICATION SERVER  : INFORM CLIENT BUTTON 1 PUSHED");
    P2PS_STM_App_Update_Char(P2P_NOTIFY_CHAR_UUID, (uint8_t *)&P2P_Server_App_Context.ButtonControl);
   } else {
    syslog_0(LOG_ERROR, "-- P2P APPLICATION SERVER(SW) : CAN'T INFORM CLIENT -  NOTIFICATION DISABLED"); 
   }

  return;
}

#if 1	/* ROI MODIFIED */
void P2PS_Send2_Data(void)
{
  P2P_Server_App_Context.TempControl.TempStatus[0] = temp >> 8;
  P2P_Server_App_Context.TempControl.TempStatus[1] = (uint8_t)temp;

  syslog_2(LOG_NOTICE, "-- P2P APPLICATION SERVER  : SEND TEMP(%d)[%04x]", temp, temp);
  P2PS_STM_App_Update_Char(0xFE43, (uint8_t *)&P2P_Server_App_Context.TempControl);

  return;
}
#endif	/* ROI MODIFIED */

/* USER CODE END FD_LOCAL_FUNCTIONS*/

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
