/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: ble_demo.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  BLE DEMO¹ԥƥȤ
 *
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "i2c.h"
#include "aqm0802_st7032.h"
#include "ble_demo.h"

#if 1	/* ROI DEBUG */
#include "app_entry.h"
#include "app_common.h"
#include "stm32_lpm.h"
#include "stm32_seq.h"
#include "dbg_trace.h"
#include "hw_conf.h"
#endif	/* ROI DEBUG */

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

#if 1	/* FOR I2C */
/* I2C TIMING Register define when I2C clock source is SYSCLK */
/* I2C TIMING is calculated in case of the I2C Clock source is the SYSCLK = 64 MHz */
/* This example use TIMING to 107F0727 to reach 400KHz speed (Rise time = 1250ns, Fall time = 50ns) */
//#if MSI_RANGE == 0
#define I2C_TIMING      0x007F0727
//#else
//#define I2C_TIMING      0x107F0727
//#endif

/* Size of Reception buffer */
#define RXBUFFERSIZE    8

/*
 *  졼֥ɥ쥹
 */
#define LCD_ADDR        (0x3E<<1)
#define EEPROM_ADDR     (0x50<<1)
#define	ADT7410_ADDR    (0x48<<1)

#define ADT7410_TEMP_H	0x00
#define ADT7410_TEMP_L	0x01
#define ADT7410_STATUS	0x02

#define LCD_locate(h, x, y) aqm0802_set_command((h), (0x80 + ((y) << 6) + (x)))
#define LCD_puts(h, b)      aqm0802_set_data((h), (b), strlen((const char *)(b)))


static I2C_Handle_t *hi2c;

uint32_t temp, old_temp;

/* Buffer used for reception */
uint8_t aRxBuffer[RXBUFFERSIZE];

uint8_t lcd_buf[2][12];

#endif	/* FOR I2C */

#if 1	/* ROI DEBUG */
void sw_handle(void)
{
	syslog_0(LOG_NOTICE, "## sw1 int ##");
	HAL_GPIO_EXTI_Callback(PSW1);
}

void sw_handle2(void)
{
	syslog_0(LOG_NOTICE, "## sw2 int ##");
	HAL_GPIO_EXTI_Callback(PSW2);
}

void Error_Handler(void)
{
	syslog_0(LOG_ERROR, "Error_Handler !");
	slp_tsk();
}
#endif	/* ROI DEBUG */


void BSP_LED_On(uint32_t led)
{
	uint32_t cled = led_in();
	cled |= led;
	led_out(cled);
}

void BSP_LED_Off(uint32_t led)
{
	uint32_t cled = led_in();
	cled &= ~led;
	led_out(cled);
}

int HAL_Init(void)
{
//  __HAL_FLASH_PREFETCH_BUFFER_ENABLE();
	sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_PRFTEN);
//  __HAL_RCC_HSEM_CLK_ENABLE();
	hsem_init(0);
	return 0;
}

/* USER CODE BEGIN 4 */

void PeriphClock_Config(void)
{
#ifdef USB_DEVICE_ID
	LL_HSEM_1StepLock(HSEM, 5);
#endif

	/*
	 * Select wakeup source of BLE RF
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR), RCC_CSR_RFWKPSEL, RCC_RFWKP_CLKSOURCE_LSE);
}

static void Init_Exti( void )
{
	/**< Disable all wakeup interrupt on CPU1  except IPCC(36), HSEM(38) */
//	LL_EXTI_DisableIT_0_31(~0);
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), 0xFFF00000);
	exti_func[0] = sw_handle2;
//	LL_EXTI_DisableIT_32_63( (~0) & (~(LL_EXTI_LINE_36 | LL_EXTI_LINE_38)) );
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR2), ~(EXTI_LINE_IPCC | EXTI_LINE_HSEM));
}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
	/* GPIO Ports Clock Enable(AB) */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOAEN);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOBEN);
}

/**
  * @brief RF Initialization Function
  * @param None
  * @retval None
  */
static void MX_RF_Init(void)
{

  /* USER CODE BEGIN RF_Init 0 */

  /* USER CODE END RF_Init 0 */

  /* USER CODE BEGIN RF_Init 1 */

  /* USER CODE END RF_Init 1 */
  /* USER CODE BEGIN RF_Init 2 */

  /* USER CODE END RF_Init 2 */

}

#if 1	/* FOR I2C */
static void
set_value(uint8_t *buf, int value)
{
    buf[1] = (value % 10) + '0';
    buf[0] = (value / 10) + '0';
}

static void
set_time(uint8_t *buf, long time)
{
	struct tm2 timedate;

	rtc_get_time(&timedate);
    set_value(&buf[6], timedate.tm_sec);
    buf[5] = ':';
    set_value(&buf[3], timedate.tm_min);
    buf[2] = ':';
    set_value(&buf[0], timedate.tm_hour);
}
#endif	/* FOR I2C */


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
#if 1	/* FOR I2C */
	I2C_Init_t i2c_initd;
#endif	/* FOR I2C */
	ER ercd;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));


	/* Reset of all peripherals, Initializes the Flash interface and the Systick. */
	HAL_Init();

	/* USER CODE BEGIN SysInit */
	PeriphClock_Config();
	Init_Exti(); /**< Configure the system Power Mode */
	/* USER CODE END SysInit */

	/* Initialize all configured peripherals */
	MX_GPIO_Init();
//	MX_DMA_Init();
	MX_RF_Init();
//	MX_RTC_Init();
	/* USER CODE BEGIN 2 */

#if 1	/* FOR I2C */
	/*
	 *  I2CݡȤϥѥ륹å˽
	 */
	i2c_initd.Timing          = I2C_TIMING;
	i2c_initd.AddressingMode  = I2C_ADDRESSINGMODE_7BIT;
	i2c_initd.DualAddressMode = I2C_DUALADDRESS_DISABLE;
	i2c_initd.GeneralCallMode = I2C_GENERALCALL_DISABLE;
	i2c_initd.NoStretchMode   = I2C_NOSTRETCH_DISABLE;
	i2c_initd.OwnAddress1     = 0;
	i2c_initd.OwnAddress2     = 0;
	i2c_initd.OwnAddress2Masks = I2C_OA2_NOMASK;
	i2c_initd.semid           = I2CTRS_SEM;
	i2c_initd.semlock         = I2CLOC_SEM;

	if((hi2c = i2c_init(I2C_PORTID, &i2c_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_ERROR, "## I2C ERROR(1) ##");
	}
	syslog_0(LOG_NOTICE, "LCD INITAILIZE !");

//	hi2c->writecallback = I2C_TxCpltCallback;
//	hi2c->readcallback  = I2C_RxCpltCallback;
//	hi2c->errorcallback = I2C_ErrorCallback;

	memset(&lcd_buf[0][0], ' ', 8);
	memset(&lcd_buf[1][0], ' ', 8);
	act_tsk(SENSE_TASK);
#endif	/* FOR I2C */

	/* USER CODE END 2 */
	APPE_Init();
	syslog_0(LOG_NOTICE, "## APPE_Init done ##");

	/* Infinite loop */
	/* USER CODE BEGIN WHILE */
	while(1){
		UTIL_SEQ_Run( UTIL_SEQ_DEFAULT );
		/* USER CODE END WHILE */

	/* USER CODE BEGIN 3 */
	}
	/* USER CODE END 3 */

	syslog_0(LOG_NOTICE, "### STOP ###");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}

void
sense_task(intptr_t exinf)
{
	CLCD_Handler_t  CLcdHandle;
	SYSTIM  tim;
	ER      ercd;
	uint32_t count = 0;

	CLcdHandle.hi2c  = hi2c;
	CLcdHandle.saddr = LCD_ADDR;
	if((ercd = aqm0802_init(&CLcdHandle)) != E_OK){
		syslog_2(LOG_ERROR, "## LCD INIT ERROR(%d)[%08x] ##", ercd, hi2c->ErrorCode);
		return;
	}

	get_tim(&tim);
	set_time(&lcd_buf[0][0], tim);
	LCD_locate(&CLcdHandle, 0, 0);
	LCD_puts(&CLcdHandle, &lcd_buf[0][0]);

	syslog_0(LOG_NOTICE, "TEMP SENSER !");
	do{
		/* STATUS READ */
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_STATUS, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(13)[%08x] ##", hi2c->ErrorCode);
			return;
		}
	}while((aRxBuffer[0] & 0x80) != 0);
	temp = old_temp = 0;

	while(count < 100000){
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_H, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(14)[%08x] ##", hi2c->ErrorCode);
			return;
		}
		temp = aRxBuffer[0] << 8;
		if(i2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_L, I2C_MEMADD_SIZE_8BIT, (uint8_t*)aRxBuffer, 1)!= E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(15)[%08x] ##", hi2c->ErrorCode);
			return;
		}
		temp += aRxBuffer[0];
		temp /= 13;
		if(old_temp != temp){
			APP_BLE_Temp_Action();
			old_temp = temp;
		}

		get_tim(&tim);
		set_time(&lcd_buf[0][0], tim);
		LCD_locate(&CLcdHandle, 0, 0);
		LCD_puts(&CLcdHandle, &lcd_buf[0][0]);
		set_value(&lcd_buf[1][0], temp/10);
		lcd_buf[1][2] = '.';
		lcd_buf[1][3] = (temp % 10) + '0';
		LCD_locate(&CLcdHandle, 0, 1);
		LCD_puts(&CLcdHandle, &lcd_buf[1][0]);
		if((count % 10) == 0)
			syslog_2(LOG_NOTICE, "count(%d) temp(%d)", count, temp);
		dly_tsk(1000);
		count++;
	}
	syslog_0(LOG_NOTICE, "SENSE END");
}

