/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: p2pserver.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  BLE p2pserver¹ԥƥȤ
 *
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "p2pserver.h"

#if 1	/* ROI DEBUG */
#include "app_entry.h"
#include "app_common.h"
#include "stm32_lpm.h"
#include "stm32_seq.h"
#include "dbg_trace.h"
#include "hw_conf.h"
#endif	/* ROI DEBUG */

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#if 1	/* ROI DEBUG */
void sw_handle(void)
{
	syslog_0(LOG_NOTICE, "## sw1 int ##");
	HAL_GPIO_EXTI_Callback(PSW1);
}

void sw_handle2(void)
{
	syslog_0(LOG_NOTICE, "## sw2 int ##");
	HAL_GPIO_EXTI_Callback(PSW2);
}

void Error_Handler(void)
{
	syslog_0(LOG_ERROR, "Error_Handler !");
	slp_tsk();
}
#endif	/* ROI DEBUG */


void BSP_LED_On(uint32_t led)
{
	uint32_t cled = led_in();
	cled |= led;
	led_out(cled);
}

void BSP_LED_Off(uint32_t led)
{
	uint32_t cled = led_in();
	cled &= ~led;
	led_out(cled);
}

int HAL_Init(void)
{
//  __HAL_FLASH_PREFETCH_BUFFER_ENABLE();
	sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_PRFTEN);
//  __HAL_RCC_HSEM_CLK_ENABLE();
	hsem_init(0);
	return 0;
}

/* USER CODE BEGIN 4 */

void PeriphClock_Config(void)
{
#ifdef USB_DEVICE_ID
	LL_HSEM_1StepLock(HSEM, 5);
#endif

	/*
	 * Select wakeup source of BLE RF
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR), RCC_CSR_RFWKPSEL, RCC_RFWKP_CLKSOURCE_LSE);
}

static void Init_Exti( void )
{
	/**< Disable all wakeup interrupt on CPU1  except IPCC(36), HSEM(38) */
//	LL_EXTI_DisableIT_0_31(~0);
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), 0xFFF00000);
	exti_func[0] = sw_handle2;
//	LL_EXTI_DisableIT_32_63( (~0) & (~(LL_EXTI_LINE_36 | LL_EXTI_LINE_38)) );
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR2), ~(EXTI_LINE_IPCC | EXTI_LINE_HSEM));
}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
	/* GPIO Ports Clock Enable(AB) */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOAEN);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOBEN);
}

/**
  * @brief RF Initialization Function
  * @param None
  * @retval None
  */
static void MX_RF_Init(void)
{

  /* USER CODE BEGIN RF_Init 0 */

  /* USER CODE END RF_Init 0 */

  /* USER CODE BEGIN RF_Init 1 */

  /* USER CODE END RF_Init 1 */
  /* USER CODE BEGIN RF_Init 2 */

  /* USER CODE END RF_Init 2 */

}

/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	ER ercd;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));


	/* Reset of all peripherals, Initializes the Flash interface and the Systick. */
	HAL_Init();

	/* USER CODE BEGIN SysInit */
	PeriphClock_Config();
	Init_Exti(); /**< Configure the system Power Mode */
	/* USER CODE END SysInit */

	/* Initialize all configured peripherals */
	MX_GPIO_Init();
//	MX_DMA_Init();
	MX_RF_Init();
//	MX_RTC_Init();
	/* USER CODE BEGIN 2 */

	/* USER CODE END 2 */
	APPE_Init();
	syslog_0(LOG_NOTICE, "## APPE_Init done ##");

	/* Infinite loop */
	/* USER CODE BEGIN WHILE */
	while(1){
		UTIL_SEQ_Run( UTIL_SEQ_DEFAULT );
		/* USER CODE END WHILE */

	/* USER CODE BEGIN 3 */
	}
	/* USER CODE END 3 */

	syslog_0(LOG_NOTICE, "### STOP ###");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}
