/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: ble_remote.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  BLE REMOTE CONTROL¹ԥƥȤ
 *
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "i2c.h"
#include "adc.h"
#include "pwm.h"
#include "ble_remote.h"
#include "protocol.h"

#if 1	/* ROI DEBUG */
#include "app_entry.h"
#include "app_common.h"
#include "stm32_lpm.h"
#include "stm32_seq.h"
#include "dbg_trace.h"
#include "hw_conf.h"
#endif	/* ROI DEBUG */


/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define I2C_TIMING      0x007F0727

/* USER CODE BEGIN Private defines */
/* Compute the prescaler value to have TIM2 counter clock equal to 100000 Hz */

#define PRESCALER_VALUE     (uint32_t)(((SystemFrequency) / 100000) - 1)

/* Initialize TIMx peripheral as follows:
   + Prescaler = (SystemCoreClock / 1000000) - 1
   + Period = (41 - 1)
   + ClockDivision = 0
   + Counter direction = Up
*/
#define PERIOD_VALUE        (uint32_t)(101 - 1)  /* Period Value  */

#define set_stby(pin, val)  sil_wrw_mem((uint32_t *)(TADR_GPIOA_BASE+TOFF_GPIO_BSRR), (1 << (pin))<<(((val)^1)*16))


volatile char    *pcommand = NULL;
volatile uint8_t command_length;
volatile uint8_t command_dev;

static TIM_Handle_t htim1, htim2;
static ADC_Handle_t *hadc;
static TIM_Handle_t *pPWM[NUM_MOTORS] = {
    &htim2,
    &htim1
};
static const uint8_t stby_pin[NUM_MOTORS] = {
    TC1_STBY_PIN,
    TC2_STBY_PIN
};

static EEPROM_t eeprom;
static volatile uint32_t event_flags;
static volatile uint32_t adc_value;
static volatile int8_t  uss_result;
static volatile uint8_t uss_dev;
static volatile uint8_t t_stby[NUM_MOTORS];
static volatile bool_t use_eeprom;
static uint8_t  sys_error = 0;
static uint32_t sys_time = 0;
static uint32_t motor_event_time = 0;
static uint32_t uss_event_time = 0;
static uint32_t uss_value;
static uint32_t uss_time1 = 0;
static uint32_t uss_time2 = 0;
static uint8_t  uss_eventno = 0;

static char     command[COM_MAX_DATA_LEN];
static char     response[COM_MAX_DATA_LEN];


#if 1	/* ROI DEBUG */
void sw_handle(void)
{
	syslog_0(LOG_NOTICE, "## sw1 int ##");
	HAL_GPIO_EXTI_Callback(PSW1);
}

void sw_handle2(void)
{
	syslog_0(LOG_NOTICE, "## sw2 int ##");
	HAL_GPIO_EXTI_Callback(PSW2);
}

void Error_Handler(void)
{
	syslog_0(LOG_ERROR, "Error_Handler !");
	slp_tsk();
}
#endif	/* ROI DEBUG */

/*
 *  ϥɥ
 */
void
cyclic_handler(intptr_t exinf)
{
	if(motor_event_time > 0){
		if(--motor_event_time == 0)
			event_flags |= EVENT_MOTOR;
	}
	if(uss_event_time > 0){
		if(--uss_event_time == 0)
			event_flags |= EVENT_USS;
	}
	sys_time++;
}

/*
 *  EEPROMΰΥݥ󥿼Фr.
 *  return: EEPROM dataΰؤΥݥ.
 */
EEPROM_t *
get_eeprom(void)
{
	return &eeprom;
}

/*
 *  ƥ֤μФ
 *  return: System time(sec).
 */
uint32_t
get_systime(void)
{
	return sys_time;
}

/*
 *  ƥࡦ顼
 *  param1:  id Error ID.
 */
void
set_syserror(uint8_t id)
{
	if(sys_error < id)
		sys_error = id;
}

/*
 *  ߤΥƥࡦ顼ͼФ
 *  return: val  System Error code.
 */
uint8_t
get_syserror(void)
{
	return sys_error;
}

/*
 *  ߤΥХåƥŰͤФ.
 *  return: val  Battery level value.
 */
uint32_t
get_batterylevel(void)
{
	uint32_t timeout = 10;
	uint32_t v = 0;
	ER ercd;

	if(hadc != NULL){
		adc_value = 0;
		ercd = adc_start_int(hadc);
		if(ercd == E_OK){
			while(adc_value == 0 && timeout != 0){
				twai_sem(ADC_SEM, 10);
				timeout--;
			}
			v = (adc_value * 330) / 4096;
			adc_end_int(hadc);
		}
	}
	return v;
}

/*
 *  ⡼ǥ塼ƥؿ
 *  parameter1   id     Motor id(pwm_id).
 *  parameter2   value1 Motor in1 duty.
 *  parameter3   value2 Motor in2 duty.
 */
void
motor_duty_set(uint8_t id, int32_t value1, int32_t value2)
{
	TIM_OC_Init_t sConfigOC = {0};

    syslog_3(LOG_NOTICE, "## motor_duty_set(%d, %d, %d) ##", (int)id, (int)value1, (int)value2);
	sConfigOC.OCMode = TIM_OCMODE_PWM1;
	sConfigOC.OCPolarity = TIM_OCPOLARITY_HIGH;
	sConfigOC.OCFastMode = TIM_OCFAST_DISABLE;

	pwm_stop(pPWM[id], TIM_CHANNEL1_ID);
	sConfigOC.Pulse  = value1;
	if(pwm_start(pPWM[id], &sConfigOC, TIM_CHANNEL1_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
	pwm_stop(pPWM[id], TIM_CHANNEL2_ID);
	sConfigOC.Pulse  = value2;
	if(pwm_start(pPWM[id], &sConfigOC, TIM_CHANNEL2_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
}

/*
 *  ⡼Хؿ
 *  parameter1   id     Motor id(pwm_id).
 *  parameter2   value  Motor stby mode 1:ready, 0:stby.
 */
void
motor_stby_set(uint8_t id, uint8_t mode)
{
#if NUM_STBYS == 1
    uint32_t sval = 0;
    uint32_t no;

    t_stby[id] = mode;
    for(no = 0 ; no < NUM_MOTORS ; no++)
    {
        if(t_stby[no] != 0)
            sval = 1;
    }
    set_stby(stby_pin[MOTOR1_ID], sval);
#else
    if(t_stby[id] != mode)
    {
        t_stby[id] = mode;
        set_stby(stby_pin[id], t_stby[id]);
    }
#endif
}

/*
 *  Ķȥ󥵡ؿ
 *  parameter1   interval ultla sonic sensor scan interval time.
 *  parameter2   time     sensor timeout value(ms).
 */
void
uss_event_request(uint32_t interval, uint32_t time)
{
    if(interval <= time)
        uss_time1 = 0;
	else
    {
        uss_dev     = command_dev;
        uss_eventno = 0;
        uss_value = -1;
        uss_time1 = interval - time;
        uss_time2 = time;
        event_flags |= EVENT_USS;
    }
}

/*
 *  EEPROMߴؿ
 *  parameter1   addr  Start address to write.
 *  parameter2   pdata Pointer to data to send.
 *  parameter3   size  Byte count of data to send.
 *
 *  return 0 or reason of error.
 */
static uint32_t
eeprom_block_write(uint16_t addr, uint8_t const *pdata, size_t len)
{
    uint32_t i;
    uint32_t err_code = 0;

    while(len > 0)
    {
        if(((addr % EEPROM_WRITE_MAX_BYTES) + len) > EEPROM_WRITE_MAX_BYTES)
            i = EEPROM_WRITE_MAX_BYTES - (addr % EEPROM_WRITE_MAX_BYTES);
        else
            i = len;
        err_code = eeprom_write(addr, pdata, i);
        if(err_code != 0)
        {
            set_syserror(BLECOM_ERROR_EEPROM);
            break;
        }
        addr  += i;
        pdata += i;
        len   -= i;
        dly_tsk(10);
    }
    return err_code;
}

/*
 *  LED ONؿ
 */
void
BSP_LED_On(uint32_t led)
{
	uint32_t cled = led_in();
	cled |= led;
	led_out(cled);
}

/*
 *  LED OFFؿ
 */
void
BSP_LED_Off(uint32_t led)
{
	uint32_t cled = led_in();
	cled &= ~led;
	led_out(cled);
}

/*
 *  ץĶؿ
 */
int
HAL_Init(void)
{
	sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_PRFTEN);
	hsem_init(0);
	return 0;
}

/* USER CODE BEGIN 4 */

void
PeriphClock_Config(void)
{
#ifdef USB_DEVICE_ID
	LL_HSEM_1StepLock(HSEM, 5);
#endif

	/*
	 * Select wakeup source of BLE RF
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR), RCC_CSR_RFWKPSEL, RCC_RFWKP_CLKSOURCE_LSE);
}

/*
 *  Receive Callback Function
 */
static void
receive_data_func(void * prxq, uint8_t *pdata, uint16_t len)
{
	memcpy(command, pdata, len);
	pcommand = (char *)command;
	command_length = len;
	command_dev    = DEVICE_BLE;
}

/*
 *  ADCžλХåؿ
 */
static void
adc_convcallback(ADC_Handle_t* hadc)
{
	adc_value = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_DR));
	isig_sem(ADC_SEM);
}

/*
 *  ĥ߽ؿ
 */
static void Init_Exti( void )
{
	/**< Disable all wakeup interrupt on CPU1  except IPCC(36), HSEM(38) */
//	LL_EXTI_DisableIT_0_31(~0);
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), 0xFFF00000);
	exti_func[0] = sw_handle2;
//	LL_EXTI_DisableIT_32_63( (~0) & (~(LL_EXTI_LINE_36 | LL_EXTI_LINE_38)) );
	sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR2), ~(EXTI_LINE_IPCC | EXTI_LINE_HSEM));
}

/*
 *  PWMؿ
 */
static void
MX_PWM_Init( void )
{
	ER ercd;

	htim1.Init.Prescaler = PRESCALER_VALUE;
	htim1.Init.CounterMode = TIM_COUNTERMODE_UP;
	htim1.Init.Period = PERIOD_VALUE;
	htim1.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
	htim1.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
	ercd = pwm_init(TIM1_PORTID, &htim1);
	if(ercd != E_OK)
		set_syserror(BLECOM_ERROR_PWM);

	htim2.Init.Prescaler = PRESCALER_VALUE;
	htim2.Init.CounterMode = TIM_COUNTERMODE_UP;
	htim2.Init.Period = PERIOD_VALUE;
	htim2.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
	htim2.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
	ercd = pwm_init(TIM2_PORTID, &htim2);
	if(ercd != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
}

/*
 *  GPIOؿ
 */
static void
MX_GPIO_Init(void)
{
	GPIO_Init_t GPIO_Init_Data = {0};
	uint32_t i;

	/* GPIO Ports Clock Enable(AB) */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOAEN);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), RCC_AHB2ENR_GPIOBEN);

	GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
	GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
	GPIO_Init_Data.pull      = GPIO_PULLUP;
	GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
	for(i = 0 ; i < NUM_MOTORS ; i++){
		gpio_setup(TADR_GPIOA_BASE, &GPIO_Init_Data, stby_pin[i]);
		set_stby(stby_pin[i], t_stby[i]);
	}

	pwm_gpio_channel(&htim1, TIM_CHANNEL1_ID, TADR_GPIOA_BASE, TC2_IN1_PIN);
	pwm_gpio_channel(&htim1, TIM_CHANNEL2_ID, TADR_GPIOA_BASE, TC2_IN2_PIN);
	pwm_gpio_channel(&htim2, TIM_CHANNEL1_ID, TADR_GPIOA_BASE, TC1_IN1_PIN);
	pwm_gpio_channel(&htim2, TIM_CHANNEL2_ID, TADR_GPIOA_BASE, TC1_IN2_PIN);
}

/*
 *  ADCؿ
 */
static void
MX_ADC_Init(void)
{
	ADC_Init_t     aInit;
	ADC_ChannelConf_t sConfig;
	ER ercd;

	/*
	 *  ADC
	 */
	aInit.ClockPrescaler = ADC_CLOCK_SYNC_PCLK_DIV2;
	aInit.Resolution = ADC_RESOLUTION_12B;
	aInit.ScanConvMode = ADC_SCANMODE_DISABLE;
	aInit.ContinuousConvMode = ADC_CONTINUOUS_DISABLE;
	aInit.DiscontinuousConvMode = ADC_DISCONTINUOUS_DISABLE;
	aInit.ExternalTrigConvEdge = ADC_EXTERNALTRIGCONVEDGE_NONE;
	aInit.ExternalTrigConv = ADC_SOFTWARE_START;
	aInit.DataAlign = ADC_DATAALIGN_RIGHT;
	aInit.DMAContinuousRequests = ADC_DMACONTINUOUS_ENABLE;
	aInit.NumConversion = 1;
	aInit.NumDiscConversion = 0;
	aInit.EOCSelection = ADC_EOC_SEQ_DISABLE;

	aInit.LowPowerAutoWait = ADC_LOWAUTOWAIT_DISABLE;
	aInit.Overrun = ADC_OVR_DATA_OVERWRITTEN;
	aInit.OversamplingMode = ADC_OVR_SAMPLING_DISABLE;
	aInit.OversamplingRatio = ADC_OVERSAMPLING_RATIO_2;
	aInit.OversamplingRightBitShift = ADC_RIGHTBITSHIFT_NONE;
	aInit.OversamplingTriggeredMode = ADC_TRIGGEREDMODE_SINGLE_TRIGGER;
	aInit.DMAChannel        = -1;
	if((hadc = adc_init(ADC1_PORTID, &aInit)) == NULL){
		/* ADC顼 */
		syslog_0(LOG_ERROR, "## adc_init ERROR ##");
		slp_tsk();
	}
	hadc->xfercallback = adc_convcallback;

	/*
	 *  ADCͥ
	 */
	sConfig.Channel = ADC_CHANNEL;
	sConfig.Rank = 1;
	sConfig.SamplingTime = ADC_SAMPLETIME_24CYCLES_5;
	sConfig.SingleDiff   = ADC_SINGLE_ENDED;            /* Single-ended input channel */
	sConfig.OffsetNumber = ADC_OFFSET_NONE;             /* No offset subtraction */ 
	sConfig.Offset = 0;                                 /* Parameter discarded because offset correction is disabled */
	sConfig.GpioBase = ADC_CHANNEL_GPIO_PORT;
	sConfig.GpioPin  = ADC_CHANNEL_PINNO;

	if((ercd = adc_setupchannel(hadc, &sConfig)) != E_OK){
		/* ADCͥ顼 */
		syslog_1(LOG_ERROR, "## adc_setupchannel ERROR(%d) ##", ercd);
		slp_tsk();
	}
}

/*
 *  RFؿ̤
 */
static void
MX_RF_Init(void)
{

  /* USER CODE BEGIN RF_Init 0 */

  /* USER CODE END RF_Init 0 */

  /* USER CODE BEGIN RF_Init 1 */

  /* USER CODE END RF_Init 1 */
  /* USER CODE BEGIN RF_Init 2 */

  /* USER CODE END RF_Init 2 */

}


/*
 *  ᥤ󥿥
 */
void
main_task(intptr_t exinf)
{
	I2C_Init_t i2c_initd;
	I2C_Handle_t *hi2c;
	TIM_OC_Init_t sConfigOC = {0};
    uint16_t len;
	ER ercd;
	uint32_t err_code;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));


	/* Reset of all peripherals, Initializes the Flash interface and the Systick. */
	HAL_Init();

	/* USER CODE BEGIN SysInit */
	PeriphClock_Config();
	Init_Exti(); /**< Configure the system Power Mode */
	/* USER CODE END SysInit */

	/* Initialize all configured peripherals */
    event_flags = 0;
    pcommand = NULL;
    command_length = 0;
	t_stby[MOTOR1_ID]  = 0;
	t_stby[MOTOR2_ID]  = 0;

	MX_PWM_Init();
	MX_GPIO_Init();
	MX_ADC_Init();
	MX_RF_Init();
	/* USER CODE BEGIN 2 */

	/*
	 *  I2CݡȤϥѥ륹å˽
	 */
	i2c_initd.Timing          = I2C_TIMING;
	i2c_initd.AddressingMode  = I2C_ADDRESSINGMODE_7BIT;
	i2c_initd.DualAddressMode = I2C_DUALADDRESS_DISABLE;
	i2c_initd.GeneralCallMode = I2C_GENERALCALL_DISABLE;
	i2c_initd.NoStretchMode   = I2C_NOSTRETCH_DISABLE;
	i2c_initd.OwnAddress1     = 0;
	i2c_initd.OwnAddress2     = 0;
	i2c_initd.OwnAddress2Masks = I2C_OA2_NOMASK;
	i2c_initd.semid           = I2CTRS_SEM;
	i2c_initd.semlock         = I2CLOC_SEM;

	if((hi2c = i2c_init(I2C_PORTID, &i2c_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_ERROR, "## I2C ERROR(1) ##");
	}
	set_hi2c(hi2c);
	ercd = sta_cyc(CYCHDR1);
	if(ercd != E_OK)
		set_syserror(BLECOM_ERROR_TIME);

    //Intialize eeprom
    err_code = eeprom_read(EEPROM_INFO_ADDRESS, (uint8_t *)&eeprom, sizeof(EEPROM_t));
    if(err_code == 0)
        use_eeprom = true;
    else
    {
        use_eeprom = false;
        set_syserror(BLECOM_ERROR_EEPROM);
    }
    if(eeprom.magic[0] != 0xAA || eeprom.magic[1] != 0x55){
        eeprom.magic[0] = 0xAA;
        eeprom.magic[1] = 0x55;
        strcpy(eeprom.devicename, "BLE_Remote");
        eeprom.t_mselect  = 0;
        eeprom.t_m_pulse[MOTOR1_ID] = 0;
        eeprom.t_m_pulse[MOTOR2_ID] = 1;
        eeprom.t_m_bias[MOTOR1_ID]  = DEFAULT_BIAS;
        eeprom.t_m_bias[MOTOR2_ID]  = DEFAULT_BIAS;
        eeprom.t_m_time             = DEFAULT_MOTOR_TIME;
        eeprom.u_s_time             = DEFAULT_MIN_USS_TIME;
        eeprom.u_s_interval         = DEFAULT_MIN_USS_INT;
        eeprom.u_s_report_des       = DEAFULT_MIN_USS_REPORT;
        if(use_eeprom){
            err_code = eeprom_block_write(EEPROM_INFO_ADDRESS, (uint8_t *)&eeprom, sizeof(EEPROM_t));
            if(err_code != 0)
                syslog_1(LOG_ERROR, "eeprom info write err_code(%d) !", (int)err_code);
        }
    }

	sConfigOC.OCMode = TIM_OCMODE_PWM1;
	sConfigOC.Pulse  = 0;
	sConfigOC.OCPolarity = TIM_OCPOLARITY_HIGH;
	sConfigOC.OCFastMode = TIM_OCFAST_DISABLE;

	if(pwm_start(&htim2, &sConfigOC, TIM_CHANNEL1_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
	if(pwm_start(&htim2, &sConfigOC, TIM_CHANNEL2_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
	if(pwm_start(&htim1, &sConfigOC, TIM_CHANNEL1_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
	if(pwm_start(&htim1, &sConfigOC, TIM_CHANNEL2_ID) != E_OK)
		set_syserror(BLECOM_ERROR_PWM);
    motor_init();
    uss_time1 = 0;

	/* USER CODE END 2 */
	APP_DeviceName(eeprom.devicename);
	NUS_APP_SetParam(NULL, receive_data_func);
	APPE_Init();
	syslog_0(LOG_NOTICE, "## APPE_Init done ##");

	/* Infinite loop */
	while(1){
		UTIL_SEQ_Run( UTIL_SEQ_DEFAULT );

		if(command_length > 0){
			int      res;

			len = command_length;
			res = ble_dispatch(response, (char *)pcommand, &len);
			switch(res){
			case EVENT_SEND_RESPONSE:
				if(command_dev == DEVICE_UART){
					response[len] = 0;
					printf("%s\n", response);
				}
				else if(NUS_APP_GetConnect() != 0){
					if(len > COM_MAX_DATA_LEN)
						len = COM_MAX_DATA_LEN;
					err_code = NUS_APP_SetTxData((uint8_t *)response, len);
					if(err_code == 0)
						set_syserror(BLECOM_ERROR_SYS);
				}
				break;
			case EVENT_MOTOR_EVENT:
				event_flags |= EVENT_MOTOR;
				break;
			case EVENT_WRITE_FLASH:
				err_code = eeprom_block_write(EEPROM_INFO_ADDRESS, (uint8_t *)&eeprom, sizeof(EEPROM_t));
				if(err_code != 0)
					syslog_1(LOG_ERROR, "eeprom info write err_code(%d) !", (int)err_code);
				break;
			default:
				break;
			}
			command_dev    = DEVICE_BLE;
			command_length = 0;
		}
		if((event_flags & EVENT_MOTOR) != 0){
			event_flags &= ~EVENT_MOTOR;
			motor_event_time = motor_event();
		}
		if((event_flags & EVENT_USS) != 0){
            event_flags &= ~EVENT_USS;
			if(uss_eventno == 0){
				uss_result  = uss_start();
				if(uss_result >= 0){
					uss_eventno = 1;
					uss_event_time = uss_time2;
				}
				else
					set_syserror(BLECOM_ERROR_USS);
			}
            else{
				uint32_t result = uss_get_result(uss_result);
				int32_t  difference, min, max;
				if(result > 0)
					result = ((((float)result - 160.0) / 2) * 0.315);

				uss_eventno = 0;
				uss_event_time = uss_time1;
				difference = result - uss_value;
				max = eeprom.u_s_report_des;
				min = -eeprom.u_s_report_des;
				if((difference >= max || difference <= min) && uss_result != 0){
					uss_value = result;
					uss_make_event(response, uss_result, result, &len);
					if(uss_dev == DEVICE_UART){
						response[len] = 0;
                        printf("%s\n", response);
					}
					else if(NUS_APP_GetConnect() != 0){
						if(len > COM_MAX_DATA_LEN)
							len = COM_MAX_DATA_LEN;
						err_code = NUS_APP_SetTxData((uint8_t *)response, len);
						if(err_code == 0)
							set_syserror(BLECOM_ERROR_SYS);
					}
					uss_result = 0;
				}
			}
		}
	}

	syslog_0(LOG_NOTICE, "### STOP ###");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}

