#include <kernel.h>
#include <t_syslog.h>
#include <stdio.h>
#include <t_stdlib.h>
#include <stdlib.h>
#include <string.h>
#include "kernel_impl.h"
#include "time_event.h"
#include "target_timer.h"
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "monitor.h"
#include "device.h"
#include "ble_remote.h"
#include "protocol.h"

/*
 *  ǥХޥֹ
 */
static int_t clk_func(int argc, char **argv);
static int_t ver_func(int argc, char **argv);
static int_t led_func(int argc, char **argv);
static int_t dip_func(int argc, char **argv);
static int_t cup_func(int argc, char **argv);
static int_t rmt_func(int argc, char **argv);

/*
 *  ǥХޥɥơ֥
 */
static const COMMAND_INFO device_command_info[] = {
	{"CLK",     clk_func},
	{"VER",     ver_func},
	{"LED",		led_func},
	{"DIP",		dip_func},
	{"CUP",		cup_func},
	{"RMT",		rmt_func}
};

#define NUM_DEVICE_CMD   (sizeof(device_command_info)/sizeof(COMMAND_INFO))

static const char device_name[] = "DEVICE";
static const char device_help[] =
"  Device  CLK                    return sysclk\n"
"          VER                    display version\n"
"          LED (no) (on-1,off-0)  led   control\n"
"          DIP (no) (on-1,off-0)  dipsw control\n"
"          CUP command            cup   control\n"
"          RMT command            ble   remote\n";

static const uint16_t led_pattern[4] = {
	LED01, LED02, LED03, LED04
};

static const char *sysclock_source[4] = {
	"MSI",
	"HSI",
	"HSE",
	"PLL"
};

static COMMAND_LINK device_command_link = {
	NULL,
	NUM_DEVICE_CMD,
	device_name,
	NULL,
	device_help,
	&device_command_info[0]
};

static uint8_t cup_command;

static int a2i(char *str)
{
	int num = 0;

	while(*str >= '0' && *str <= '9'){
		num = num * 10 + *str++ - '0';
	}
	return num;
}

/*
 *  DEVICEޥؿ
 */
void device_info_init(intptr_t exinf)
{
	setup_command(&device_command_link);
}

/*
 *  SYSCLK֤ؿ
 */
static int_t clk_func(int argc, char **argv)
{
	ER ercd = E_OK;
	uint32_t sysclock;
	uint8_t sws;
	uint_t  arg1;

	if(argc > 1){
		arg1 = a2i(argv[1]);
		serial_cls_por(SIO_PORTID);
		dly_tsk(100);
		ercd = sysclock_change(arg1);
		target_timer_initialize(0);
		serial_opn_por(SIO_PORTID);
		serial_ctl_por(SIO_PORTID, (IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV));
		printf("restart(%d)\n", ercd);
	}

	sysclock = get_sysclock(&sws);
	printf("sysclk[%s] = %d\n", sysclock_source[sws], sysclock);
	return 0;
}

/*
 *  VERSIONɽؿ
 */
static int_t ver_func(int argc, char **argv)
{
	printf("TOPPERS BASE PLATFORM Verion %d.%X.%d\n", 
			(TPLATFORM_PRVER >> 12) &  0x0fU,
			(TPLATFORM_PRVER >> 4) & 0xffU,
				TPLATFORM_PRVER & 0x0fU);
	return 0;
}

/*
 *  LEDꥳޥɴؿ
 */
static int_t led_func(int argc, char **argv)
{
	int_t    arg1=0, arg2;
	uint16_t reg;

	if(argc < 3)
		return -1;
	arg1 = a2i(argv[1]);
	arg2 = a2i(argv[2]);
	if(arg1 >= 1 && arg1 <= 4){
		reg = led_in();
		if(arg2 != 0)
			reg |= led_pattern[arg1-1];
		else
			reg &= ~led_pattern[arg1-1];
		led_out(reg);
	}
	return 0;
}

/*
 *  DIPSWꥳޥɴؿ
 */
static int_t dip_func(int argc, char **argv)
{
	int_t   arg1=0, arg2=0, i;

	if(argc > 1)
		arg1 = a2i(argv[1]);
	if(argc > 2)
		arg2 = a2i(argv[2]);
	if(arg1 >= 1 && arg1 <= 8){
		if(arg2 != 0)
			dipsw_value |= (0x1<<(arg1-1));
		else
			dipsw_value &= ~(0x1<<(arg1-1));
	}
	printf("dipsw\n1   2   3   4   5   6   7   8\n");
	for(i = 0 ; i < 8 ; i++){
		if(dipsw_value & (1<<i))
			printf("ON  ");
		else
			printf("OFF ");
	}
	printf("\n");
	return 0;
}

/*
 *  åץ顼󥿥ޥɴؿ
 */
static int_t cup_func(int argc, char **argv)
{
	if(argc > 1)
		cup_command = *argv[1];
	return 0;
}

/*
 *  BLE REMOTEޥɴؿ
 */
static int_t rmt_func(int argc, char **argv)
{
	static uint8_t data_array[COM_MAX_DATA_LEN];
	int timeout = 10;
	int len;
	if(argc > 1){
		len = strlen(argv[1]);
		strcpy((char *)data_array, argv[1]);
		printf("[%s]\n", data_array);
		pcommand = (char *)data_array;
		command_dev    = DEVICE_UART;
		command_length = len;
		while(command_dev == DEVICE_UART && timeout > 0){
			dly_tsk(2);
			--timeout;
		}
	}
	return 0;
}

/*
 * CUP顼󥳥ޥ
 */
int8_t
get_cup_command(void)
{
	uint8_t cmd = cup_command;
	cup_command = 0;
	return cmd;
}


