/**
  ******************************************************************************
  * @file    stm32wbxx_hal.h
  * @author  MCD Application Team
  * @brief   This file contains all the functions prototypes for the HAL
  *          module driver.
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; Copyright (c) 2019 STMicroelectronics. 
  * All rights reserved.</center></h2>
  *
  * This software component is licensed by ST under BSD 3-Clause license,
  * the "License"; You may not use this file except in compliance with the 
  * License. You may obtain a copy of the License at:
  *                        opensource.org/licenses/BSD-3-Clause
  *
  ******************************************************************************
  */

/* Define to prevent recursive inclusion -------------------------------------*/
#ifndef STM32WBxx_HAL_H
#define STM32WBxx_HAL_H

#ifdef __cplusplus
 extern "C" {
#endif

/* Includes ------------------------------------------------------------------*/
#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <kernel_impl.h>
#include <target_syssvc.h>
#include "stm32wbxx_hal_def.h"
#include "device.h"


/*  CMSIS CM4 definitions */
#define __CM4_CMSIS_VERSION_MAIN  (0x04U)                                      /*!< [31:16] CMSIS HAL main version */
#define __CM4_CMSIS_VERSION_SUB   (0x1EU)                                      /*!< [15:0]  CMSIS HAL sub version */
#define __CM4_CMSIS_VERSION       ((__CM4_CMSIS_VERSION_MAIN << 16U) | \
                                    __CM4_CMSIS_VERSION_SUB           )        /*!< CMSIS HAL version number */

#define __CORTEX_M                (0x04U)                                      /*!< Cortex-M Core */

#define __disable_irq   set_primask
#define __get_PRIMASK   read_primask

Inline void
__set_PRIMASK(uint32_t val){
	Asm("msr primask, %0" : : "r"(val) : "memory");
}

/**
  \brief   Count leading zeros
  \details Counts the number of leading zeros of a data value.
  \param [in]  value  Value to count the leading zeros
  \return             number of leading zeros in value
 */
#define __CLZ             __builtin_clz

#define __WEAK            __weak

#define HAL_StatusTypeDef ER
#define HAL_OK            E_OK

#define LED_BLUE   LED01
#define LED_GREEN  LED02
#define LED_RED    LED03

#define BLUE_LED   LED01
#define GREEN_LED  LED02
#define RED_LED    LED03

#define BUTTON_SW1_PIN   PSW1
#define BUTTON_SW2_PIN   PSW2
#define BUTTON_SW3_PIN   PSW3

/* stm32wbxx_hal_flash.h */
#define FLASH_SIZE                      (((uint32_t)(*((uint16_t *)FLASHSIZE_BASE)) & (0x07FFUL)) << 10U)
#define FLASH_END_ADDR                  (FLASH_BASE + FLASH_SIZE - 1U)

#define FLASH_BANK_SIZE                 FLASH_SIZE   /*!< FLASH Bank Size */
#define FLASH_PAGE_SIZE                 0x00001000U  /*!< FLASH Page Size, 4 KBytes */

/** @defgroup FLASH_FLAGS FLASH Flags Definition
  * @{
  */
#define FLASH_FLAG_EOP                  FLASH_SR_EOP      /*!< FLASH End of operation flag */
#define FLASH_FLAG_WRPERR               FLASH_SR_WRPERR   /*!< FLASH Write protection error flag */
#define FLASH_FLAG_OPTVERR              FLASH_SR_OPTVERR  /*!< FLASH Option validity error flag  */


#define FLASH_OBProgramInitTypeDef FLASH_WRPAREA_t
#define FLASH_EraseInitTypeDef  FLASH_Erase_t
#define __HAL_FLASH_CLEAR_FLAG(f)  	                sil_wrw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR), (f))
#define LL_FLASH_IsActiveFlag_OperationSuspended()  ((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR)) & FLASH_SR_PESD) != 0)
#define HAL_FLASHEx_OBGetConfig flash_getwrparea
#define HAL_FLASH_Unlock        flash_unlock
#define HAL_FLASH_Lock          flash_lock
#define HAL_FLASHEx_Erase       flash_erase
#define HAL_FLASH_Program       flash_program

extern void BSP_LED_On(uint32_t led);
extern void BSP_LED_Off(uint32_t led);

extern void HAL_GPIO_EXTI_Callback( uint16_t GPIO_Pin );

/**
  * @}
  */

#ifdef __cplusplus
}
#endif

#endif /* STM32WBxx_HAL_H */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
