/* USER CODE BEGIN Header */
/**
 ******************************************************************************
 * @file    p2p_server_app.c
 * @author  MCD Application Team
 * @brief   peer to peer Server Application
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; Copyright (c) 2019 STMicroelectronics.
 * All rights reserved.</center></h2>
 *
 * This software component is licensed by ST under Ultimate Liberty license
 * SLA0044, the "License"; You may not use this file except in compliance with
 * the License. You may obtain a copy of the License at:
 *                             www.st.com/SLA0044
 *
 ******************************************************************************
 */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"
#include "app_common.h"
#include "kernel_cfg.h"
#include "dbg_trace.h"
#include "ble.h"
#include "nus_stm.h"
#include "nus_server_app.h"
#include "stm32_seq.h"


/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */

/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */
/* USER CODE END PTD */

/* Private defines ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

typedef struct
{
  uint16_t              ConnectionHandle;
  void                  *RxQueue;
  void                  (*rxcallback)(void *, uint8_t *, uint16_t);
  uint8_t               TxBuffer[BLE_NUS_MAX_DATA_LEN];
  uint16_t              TxLen;
  uint8_t               Notification_Status; /* used to chek if P2P Server is enabled to Notify */
  uint8_t               Connect;
} NUS_Server_App_Context_t;


/* USER CODE END PD */

/* Private macros -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
/* USER CODE BEGIN PV */
/**
 * START of Section BLE_APP_CONTEXT
 */

PLACE_IN_SECTION("BLE_APP_CONTEXT") static NUS_Server_App_Context_t NUS_Server_App_Context;

/**
 * END of Section BLE_APP_CONTEXT
 */
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
/* USER CODE BEGIN PFP */
static void NUS_Send_Notification(void);
static void NUS_APP_SERIAL_context_Init(void);
/* USER CODE END PFP */

/* Functions Definition ------------------------------------------------------*/
void Nus_STM_App_Notification(NUS_STM_App_Notification_evt_t *pNotification)
{
/* USER CODE BEGIN Nus_STM_App_Notification_1 */

#if 1	/* ROI DEBUG */
  syslog_4(LOG_NOTICE, "## Nus_STM_App_Notification NUS_Evt_Opcode(%d) length(%d) connect[%04x] serv[%02x] ##",
		pNotification->P2P_Evt_Opcode,
		pNotification->DataTransfered.Length,
		pNotification->ConnectionHandle,
		pNotification->ServiceInstance);
#endif	/* ROI DEBUG */
/* USER CODE END Nus_STM_App_Notification_1 */
  switch(pNotification->P2P_Evt_Opcode)
  {
/* USER CODE BEGIN Nus_STM_App_Notification_P2P_Evt_Opcode */
#if(BLE_CFG_OTA_REBOOT_CHAR != 0)
    case NUS_STM_BOOT_REQUEST_EVT:
      syslog_0(LOG_NOTICE, "-- NUS APPLICATION SERVER : BOOT REQUESTED\n");

      *(uint32_t*)SRAM1_BASE = *(uint32_t*)pNotification->DataTransfered.pPayload;
      NVIC_SystemReset();
      break;
#endif
/* USER CODE END Nus_STM_App_Notification_P2P_Evt_Opcode */

    case NUS_STM__NOTIFY_ENABLED_EVT:
/* USER CODE BEGIN NUS_STM__NOTIFY_ENABLED_EVT */
      NUS_Server_App_Context.Notification_Status = 1;
      UTIL_SEQ_SetTask( 1<<CFG_TASK_NUS_SEND_DATA_ID, CFG_SCH_PRIO_0);
      syslog_0(LOG_NOTICE, "-- NUS APPLICATION SERVER : NOTIFICATION ENABLED\n");
/* USER CODE END NUS_STM__NOTIFY_ENABLED_EVT */
      break;

    case NUS_STM_NOTIFY_DISABLED_EVT:
/* USER CODE BEGIN NUS_STM_NOTIFY_DISABLED_EVT */
      NUS_Server_App_Context.Notification_Status = 0;
      syslog_0(LOG_NOTICE, "-- NUS APPLICATION SERVER : NOTIFICATION DISABLED\n");
/* USER CODE END NUS_STM_NOTIFY_DISABLED_EVT */
      break;

    case NUS_STM_WRITE_EVT:
/* USER CODE BEGIN NUS_STM_WRITE_EVT */
				/* ALL Deviceselected - may be necessary as LB Routeur informs all connection */
      if(NUS_Server_App_Context.rxcallback != NULL)
        NUS_Server_App_Context.rxcallback(NUS_Server_App_Context.RxQueue,
             pNotification->DataTransfered.pPayload, pNotification->DataTransfered.Length);
/* USER CODE END NUS_STM_WRITE_EVT */
      break;

    default:
/* USER CODE BEGIN Nus_STM_App_Notification_default */
      
/* USER CODE END Nus_STM_App_Notification_default */
      break;
  }
/* USER CODE BEGIN Nus_STM_App_Notification_2 */

/* USER CODE END Nus_STM_App_Notification_2 */
  return;
}

void NUS_APP_Notification(P2PS_APP_ConnHandle_Not_evt_t *pNotification)
{
#if 1	/* ROI DEBUG */
  syslog_1(LOG_NOTICE, "## NUS_APP_Notification P2P_Evt_Opcode(%d) ##", pNotification->P2P_Evt_Opcode);
#endif	/* ROI DEBUG */
/* USER CODE BEGIN NUS_APP_Notification_1 */

/* USER CODE END NUS_APP_Notification_1 */
  switch(pNotification->P2P_Evt_Opcode)
  {
/* USER CODE BEGIN NUS_APP_Notification_P2P_Evt_Opcode */

/* USER CODE END NUS_APP_Notification_P2P_Evt_Opcode */
  case PEER_CONN_HANDLE_EVT :
/* USER CODE BEGIN PEER_CONN_HANDLE_EVT */
    NUS_Server_App_Context.Connect = 0x01;

/* USER CODE END PEER_CONN_HANDLE_EVT */
    break;

  case PEER_DISCON_HANDLE_EVT :
/* USER CODE BEGIN PEER_DISCON_HANDLE_EVT */
    NUS_APP_SERIAL_context_Init();
/* USER CODE END PEER_DISCON_HANDLE_EVT */
    break;

  default:
/* USER CODE BEGIN NUS_APP_Notification_default */

/* USER CODE END NUS_APP_Notification_default */
    break;
  }
/* USER CODE BEGIN NUS_APP_Notification_2 */

/* USER CODE END NUS_APP_Notification_2 */
  return;
}

/*
 *  Main intarface functions
 */
void NUS_APP_SetParam(void *prxq, void *func)
{
  NUS_Server_App_Context.RxQueue    = prxq;
  NUS_Server_App_Context.rxcallback = func;
}

uint8_t NUS_APP_GetConnect(void)
{
  return NUS_Server_App_Context.Connect;
}

uint16_t NUS_APP_SetTxData(uint8_t *pdata, uint16_t len)
{
  if(len == 0)
    return 0;
  memcpy(NUS_Server_App_Context.TxBuffer, pdata, len);
  NUS_Server_App_Context.TxLen = len;
  UTIL_SEQ_SetTask( 1<<CFG_TASK_NUS_SEND_DATA_ID, CFG_SCH_PRIO_0);
  return len;
}


void NUS_APP_Init(void)
{
/* USER CODE BEGIN NUS_APP_Init */
  UTIL_SEQ_RegTask( 1<< CFG_TASK_NUS_SEND_DATA_ID, UTIL_SEQ_RFU, NUS_Send_Notification );

  /**
   * Initialize LedButton Service
   */
  NUS_Server_App_Context.Notification_Status=0; 
  NUS_APP_SERIAL_context_Init();
/* USER CODE END NUS_APP_Init */
  return;
}

/* USER CODE BEGIN FD */
void NUS_APP_SERIAL_context_Init(void){

  BSP_LED_Off(LED_BLUE);

  NUS_Server_App_Context.TxBuffer[0]  = 0x00;/* Device1 */
  NUS_Server_App_Context.TxBuffer[1]  = 0x00;
  NUS_Server_App_Context.Connect      = 0x00;
}


/*************************************************************
 *
 * LOCAL FUNCTIONS
 *
 *************************************************************/
/* USER CODE BEGIN FD_LOCAL_FUNCTIONS*/
void NUS_Send_Notification(void)
{
  uint16_t no  = 0;

  if(NUS_Server_App_Context.Notification_Status){
    no = NUS_Server_App_Context.TxLen;
    if(no > 0){
      Nus_STM_App_Update_Char(P2P_NOTIFY_CHAR_UUID, (uint8_t *)&NUS_Server_App_Context.TxBuffer, no);
      NUS_Server_App_Context.TxLen = 0;
      syslog_1(LOG_NOTICE, "-- NUS : INFORM CLIENT SEND(%d)", no);
    }
  }
  else
    syslog_0(LOG_ERROR, "-- NUS : CAN'T INFORM CLIENT -  NOTIFICATION DISABLED");

  return;
}

/* USER CODE END FD_LOCAL_FUNCTIONS*/

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
