/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: spitest.c 2416 2012-09-07 08:06:20Z ertl-hiro $
 */

/* 
 *  SPIƥȤ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "spi.h"
#include "spitest.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


/*
 *  SW1
 */
void sw_int(int arg)
{
	syslog_1(LOG_NOTICE, "## sw_int(%d) ##", arg);
}

#include <target_syssvc.h>



/* Size of buffer */
#define BUFFERSIZE                       (COUNTOF(aTxBuffer) - 1)

/* Exported macro ------------------------------------------------------------*/
#define COUNTOF(__BUFFER__)   (sizeof(__BUFFER__) / sizeof(*(__BUFFER__)))


/* Exported functions ------------------------------------------------------- */

/* Private variables ---------------------------------------------------------*/
/* SPI handler declaration */

/* Buffer used for transmission */
uint8_t aTxBuffer[] = "SPI123456789012345678901";

/* Buffer used for reception */
uint8_t aRxBuffer[BUFFERSIZE];


/*
 *  ᥤ󥿥
 */
void main_task(intptr_t exinf)
{
	SPI_Init_t spi_initd;
	SPI_Handle_t *hspi;
	DMA_Handle_t *hdmarx, *hdmatx;
	ER_UINT	ercd;
	int i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  ꥢݡȤν
	 *
	 *  ƥƱꥢݡȤȤʤɡꥢ
	 *  ݡȤץѤߤξˤϤE_OBJ顼ˤʤ뤬پ
	 *  ʤ
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "SPI SEND/RECV START");
	/*##-1- Configure the SPI peripheral #######################################*/
	/* Set the SPI parameters */
	spi_initd.Prescaler         = SPI_BAUDRATEPRESCALER_32;
//    spi_initd.Prescaler         = SPI_BAUDRATEPRESCALER_64;
	spi_initd.Direction         = SPI_DIRECTION_2LINES;
	spi_initd.CLKPhase          = SPI_PHASE_1EDGE;
	spi_initd.CLKPolarity       = SPI_POLARITY_HIGH;
	spi_initd.CRC               = SPI_CRC_DISABLE;
	spi_initd.CRCPolynomial     = 7;
	spi_initd.CRCLength         = SPI_CRC_LENGTH_8BIT;
	spi_initd.DataSize          = SPI_DATASIZE_8BIT;
	spi_initd.SignBit           = SPI_DATA_MSB;
	spi_initd.NSS               = SPI_NSS_SOFT;
	spi_initd.NSSPMode          = SPI_NSS_PULSE_DISABLE;
	spi_initd.TIMode            = SPI_TIMODE_DISABLE;
	spi_initd.TxDMAChannel      = TXDMA_CID;
	spi_initd.RxDMAChannel      = RXDMA_CID;
	spi_initd.Mode = SPI_MODE_MASTER;
	spi_initd.semid = SPI1DMA_SEM;
	spi_initd.semlock = SPI1LOCK_SEM;

	if((hspi = spi_init(SPI1_PORTID, &spi_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_NOTICE, "## call Error_Handler(1) ##");
	}

	/*##-2- Start the Full Duplex Communication process ########################*/  
  /* While the SPI in TransmitReceive process, user can transmit data through 
     "aTxBuffer" buffer & receive data through "aRxBuffer" */
	if((ercd = spi_transrecv(hspi, (uint8_t*)aTxBuffer, (uint8_t *)aRxBuffer, BUFFERSIZE)) != E_OK){
		/* Transfer error in transmission process */
		syslog_1(LOG_NOTICE, "## call Error_Handler(2)(%d) ##", ercd);
	}
	if((ercd = spi_wait(hspi, 100)) != E_OK){
		syslog_0(LOG_NOTICE, "## call Error_Handler(3) ##");
	}
	dly_tsk(500);
	syslog_1(LOG_NOTICE, "## ercd(%d) ##", ercd);
	for(i = 0 ; i < BUFFERSIZE ; i++){
		syslog_3(LOG_NOTICE, "## (%d) T[%02x] R[%02x] ##", i, aTxBuffer[i], aRxBuffer[i]);
	}

	dly_tsk(1000);
	if(spi_transrecv(hspi, (uint8_t*)aTxBuffer, (uint8_t *)aRxBuffer, BUFFERSIZE) != E_OK){
		/* Transfer error in transmission process */
		syslog_0(LOG_NOTICE, "## call Error_Handler(4) ##");
	}
	if((ercd = spi_wait(hspi, 100)) != E_OK){
		syslog_0(LOG_NOTICE, "## call Error_Handler(5) ##");
	}
	dly_tsk(500);
	syslog_1(LOG_NOTICE, "## ercd(%d) ##", ercd);
	for(i = 0 ; i < BUFFERSIZE ; i++){
		syslog_3(LOG_NOTICE, "## (%d) T[%02x] R[%02x] ##", i, aTxBuffer[i], aRxBuffer[i]);
	}
	hdmarx = hspi->hdmarx;
	hdmatx = hspi->hdmatx;

	syslog_2(LOG_NOTICE, "## INST RX[%08x] TX[%08x] ##", hdmarx->cbase, hdmatx->cbase);
	syslog_2(LOG_NOTICE, "## CCR  RX[%08x] TX[%08x] ##", sil_rew_mem((uint32_t *)(hdmarx->cbase+TOFF_DMACH_CCR)), sil_rew_mem((uint32_t *)(hdmatx->cbase+TOFF_DMACH_CCR)));
	syslog_2(LOG_NOTICE, "## CMAR RX[%08x] TX[%08x] ##", sil_rew_mem((uint32_t *)(hdmarx->cbase+TOFF_DMACH_CMAR)), sil_rew_mem((uint32_t *)(hdmatx->cbase+TOFF_DMACH_CMAR)));
	syslog_2(LOG_NOTICE, "## NDTR RX[%08x] TX[%08x] ##", sil_rew_mem((uint32_t *)(hdmarx->cbase+TOFF_DMACH_CNDTR)), sil_rew_mem((uint32_t *)(hdmatx->cbase+TOFF_DMACH_CNDTR)));
	if(spi_deinit(hspi) != E_OK){
		syslog_0(LOG_NOTICE, "## call Error_Handler(6) ##");
	}
	syslog_0(LOG_NOTICE, "SPI SEND/RECV END");
	spi_deinit(hspi);
	dly_tsk(500);

	syslog_0(LOG_NOTICE, "SPI SEND START");
	spi_initd.Direction         = SPI_DIRECTION_1LINE;
	if((hspi = spi_init(SPI1_PORTID, &spi_initd)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_NOTICE, "## call Error_Handler(7) ##");
	}

	/*##-2- Start the Full Duplex Communication process ########################*/  
  /* While the SPI in TransmitReceive process, user can transmit data through 
     "aTxBuffer" buffer & receive data through "aRxBuffer" */
	if(spi_transmit(hspi, (uint8_t*)aTxBuffer, BUFFERSIZE) != E_OK){
		/* Transfer error in transmission process */
		syslog_0(LOG_NOTICE, "## call Error_Handler(8) ##");
	}
	if((ercd = spi_wait(hspi, 100)) != E_OK){
		syslog_0(LOG_NOTICE, "## call Error_Handler(9) ##");
	}
	dly_tsk(500);
	syslog_1(LOG_NOTICE, "## ercd(%d) ##", ercd);
	syslog_0(LOG_NOTICE, "SPI SEND END");
	spi_deinit(hspi);

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}
