/*
 *  TOPPERS BASE PLATFORM MIDDLEWARE
 * 
 *  Copyright (C) 2017-2019 by TOPPERS PROJECT
 *                             Educational Working Group.
 * 
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: rom_file.c,v 1.1 2019/12/03 18:00:05 roi Exp $
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "device.h"
#include <fcntl.h>
#include "storagedevice.h"
#include "rom_file.h"


static int           num_files = 0;
static int           dirreadcount;
static fs_descrip_t  fs_descrip_table[MAX_DESCRIP];

static void  *romf_opendir(const char *pathname);
static int   romf_closedir(void *dir);
static int   romf_readdir(void *dir, void *pdirent);
static int   romf_open(int devno, const char *pathname, int flags);
static int   romf_close(int fd);
static int   romf_fstat(int fd, struct stat *buf);
static off_t romf_lseek(int fd, off_t offset, int whence, int *res);
static long  romf_read(int fd, void *buf, long count, int *res);

static const StorageDeviceFileFunc_t romfSDeviceFunc = {
	romf_opendir,
	romf_closedir,
	romf_readdir,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	romf_open,
	romf_close,
	romf_fstat,
	romf_lseek,
	romf_read,
	0,
	0
};


/*
 *  FLASH-ROMեؿν
 */
ER
rom_file_init(void)
{
	StorageDevice_t *psdev;
	int result;

	num_files = 0;
	memset(fs_descrip_table, 0, sizeof(fs_descrip_table));
	result = SDMSetupDevice(DEFAULT_DEVNO, &psdev);
	if(result == E_OK){
		psdev->pdevff           = (StorageDeviceFileFunc_t *)&romfSDeviceFunc;
		psdev->_sdev_attribute |= SDEV_EMPLOY;
		return MAX_DESCRIP;
	}
	else
		return 0;
}

/*
 *  ROMե
 */
int
create_rom_file(const char *pathname, uintptr_t start, uint32_t size)
{
	fs_descrip_t  *fdp;
	int           cd = -1;
	int           i;

	dis_dsp();			/* disable dispatch */
	for(i = 0 ; i < MAX_DESCRIP ; i++){
		fdp = &fs_descrip_table[i];
		if(fdp->create == 0){
			fdp->create   = 1;
			fdp->inuse    = 0;
			fdp->fp       = 0;
			fdp->addr     = start;
			fdp->size     = size;
			fdp->filename = pathname;
			cd = i;
			num_files++;
			break;
		}
	}
	ena_dsp();			/* enable dispatch */
	return cd;
}

/*
 *  FLASH-ROMեǥ쥯ȥꥪץ
 */
static void *
romf_opendir(const char *pathname)
{
	dirreadcount = 0;
	return (void *)&num_files;
}

/*
 *  ROMեǥ쥯ȥꥯ
 */
static int
romf_closedir(void *dir)
{
	return 0;
}

/*
 *  ROMեǥ쥯ȥɤ߹
 */
static int
romf_readdir(void *dir, void *pdirent)
{
	fs_descrip_t   *pf;
	struct dirent2 *pd = pdirent;
	unsigned int   i;

	if(dirreadcount >= num_files)
		return 0;
	else{
		pf = &fs_descrip_table[dirreadcount];
		pd->d_fsize = pf->size;
		pd->d_date  = 0;
		pd->d_time  = 0;
		pd->d_type  = AM_RDO;
		for(i = 0 ; i < 255 && pf->filename[i] != 0 ; i++){
			pd->d_name[i] = pf->filename[i];
		}
		pd->d_name[i] = 0;
		dirreadcount++;
		return dirreadcount;
	}
}

/*
 *  ROMե륪ץ
 */
static int
romf_open(int devno, const char *pathname, int flags)
{
	fs_descrip_t  *fdp = 0;
	int           fd;

	SDMGetDeviceNo(&pathname);
	if(*pathname == '/')
		pathname++;
	dis_dsp();			/* disable dispatch */
	for(fd = 0 ; fd < num_files ; fd++){
		fdp = &fs_descrip_table[fd];
		if(fdp->create && strcmp(pathname, fdp->filename) == 0){
			break;
		}
	}
	if(fd < num_files && (flags & 0x0f) == O_RDONLY && fdp->inuse == 0){
		fdp->inuse = 1;
		fdp->fp = 0;
		ena_dsp();			/* enable dispatch */
		return fd;
	}
	else{
		ena_dsp();			/* enable dispatch */
		return -1;
	}
}

/*
 *  ROMե륯
 */
static int
romf_close(int fd)
{
	fs_descrip_t  *fdp;

	if(fd < 0 || num_files < fd)
		return -1;
	fdp = &fs_descrip_table[fd];
	if(fdp->inuse == 1){
		fdp->inuse = 0;
		return 0;
	}
	else
		return -1;
}

/*
 *  ROMե륹ơȼ
 */
static int
romf_fstat(int fd, struct stat *buf)
{
	fs_descrip_t  *fdp;

	if(fd < 0 || num_files < fd)
		return -1;
	fdp = &fs_descrip_table[fd];
	if(fdp->create == 1){
		memset(buf, 0, sizeof(struct stat));
		buf->st_size = fdp->size;
		return 0;
	}
	else
		return -1;
}

/*
 *  ROMե륷
 */
static off_t
romf_lseek(int fd, off_t offset, int whence, int *res)
{
	fs_descrip_t  *fdp;

	*res = -1;
	if(fd < 0 || num_files < fd)
		return -1;
	fdp = &fs_descrip_table[fd];
	if(fdp->inuse == 1){
		switch (whence) {
		case SEEK_SET:
			fdp->fp = offset;
			break;
		case SEEK_CUR:
			fdp->fp += offset;
			break;
		case SEEK_END:
			fdp->fp = fdp->size + offset;
			break;
		}
		offset = fdp->fp;
		*res = 0;
		return offset;
	}
	else
		return -1;
}

/*
 *  ROMեɤߤ
 */
static long
romf_read(int fd, void *buf, long count, int *res)
{
	fs_descrip_t*fdp;
	int         readsize;

	*res = -1;
	if(fd < 0 || num_files < fd)
		return -1;
	fdp = &fs_descrip_table[fd];
	if(fdp->inuse == 1){
		if(fdp->fp < fdp->size){
			if((fdp->fp + count) <= fdp->size)
				readsize = count;	/* all counts can be read */
			else
				readsize = fdp->size - fdp->fp;	/* a part of counts can be read */
			memcpy(buf, ((uint8_t *)fdp->addr + fdp->fp), readsize);
			buf += readsize;
			fdp->fp += readsize;
		}
		else
			readsize = 0;	/* no data can be read */
		*res = 0;
		return readsize;
	}
	else
		return -1;
}

