/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: sipeed_st3375.c 2416 2020-10-21 14:21:09Z roi $
 */

/* 
 *  SIPEED ST7735 0.96"LCDץ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "device.h"
#include "pinmode.h"
#include "sipeed_st7735.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))
#define ABS(X)  ((X) > 0 ? (X) : -(X))


// Rather than a bazillion writecommand() and writedata() calls, screen
// initialization commands and arguments are organized in these tables
// stored in PROGMEM.  The table may look bulky, but that's mostly the
// formatting -- storage-wise this is hundreds of bytes more compact
// than the equivalent code.  Companion function follows.
#define DELAY 0x80
static const uint8_t
  Scmd[] = {                  // Initialization commands for 7735S screens
   15,                        // 18 commands in list:
    ST7735_SLPOUT ,   DELAY,  //  1: Out of sleep mode, no args, w/delay
      100,                    //     255 = 500 ms delay
    ST7735_INVON,     DELAY,  //  2: Display inversion mode
      1,                      //     1 ms delay
    ST7735_FRMCTR1, 3      ,  //  3: Frame rate control1, 3 args:
      0x05,                   //     RTNA
      0x3A,                   //     FPA
      0x3A,                   //     BPA
    ST7735_FRMCTR2, 3      ,  //  4: Frame rate control2, 3 args:
                              //     Set the frame frequency of the Idle mode
                              //     Frame rate=fosc/((RTNB x 2 + 40) x (LINE + FPB + BPB +2))
                              //     fosc = 850kHz
      0x05,                   //     RTNB
      0x3A,                   //     FPB
      0x3A,                   //     BPB
    ST7735_FRMCTR3, 6      ,  //  5: Frame rate control2, 6 args:
      0x05,                   //     
      0x3A,                   //     
      0x3A,                   //     
      0x05,                   //     
      0x3A,                   //     
      0x3A,                   //     
    ST7735_INVCTR , 1      ,  //  6: Display inversion control, 1 arg:
      0x03,                   //     Line inversion
    ST7735_PWCTR1 , 3      ,  //  7: Power control1, 2 args + no delay:
      0x62,                   //     
      0x02,                   //     
      0x04,                   //     
    ST7735_PWCTR2 , 1      ,  //  8: Power control2, 1 arg, no delay:
      0xC0,                   //     
    ST7735_PWCTR3 , 2      ,  //  9: Power control3, 2 args, no delay:
      0x0D,                   //     
      0x00,                   //     
    ST7735_PWCTR4 , 2      ,  // 10: Power control4, 2 args, no delay:
      0x8D,                   //     
      0x6A,                   //     
    ST7735_PWCTR5 , 2      ,  // 11: Power control5, 2 args, no delay:
      0x8D,                   //     
      0xEE,                   //     
    ST7735_VMCTR1 , 1+DELAY,  // 12: VCOM control, 1 args, delay:
      0x0E,                   //     
      10,                     //     10 ms delay
    ST7735_GMCTRP1,16      ,  // 13: Magical unicorn dust, 16 args, no delay:
      0x10, 0x0E, 0x02, 0x03, //     (seriously though, not sure what
      0x0E, 0x07, 0x02, 0x07, //      these config values represent)
      0x0A, 0x12, 0x27, 0x37,
      0x00, 0x0D, 0x0E, 0x10,
    ST7735_GMCTRN1,16+DELAY,  // 14: Sparkles and rainbows, 16 args + delay:
      0x10, 0x0E, 0x03, 0x03, //     (ditto)
      0x0F, 0x06, 0x02, 0x08,
      0x0A, 0x13, 0x26, 0x36,
      0x00, 0x0D, 0x0E, 0x10,
      10,                     //     10 ms delay
    ST7735_COLMOD , 1+DELAY,  // 15: Set color mode, 1 arg + delay:
      0x05,                   //     16-bit color
      1};                     //     1 ms delay


/* Buffer used for transmission */
static uint8_t aTxBuffer[4];

#define PORT_HIGH           1
#define PORT_LOW            0

#define cs_set(sw)  digitalWrite(hlcd->cs_pin, sw)
#define rs_set(sw)  digitalWrite(hlcd->rs_pin, sw)
#define rst_set(sw) digitalWrite(hlcd->rst_pin, sw)
#define cs2_set(sw) digitalWrite(hlcd->cs2_pin, sw)


/*
 *  LCDؤΥޥؿ
 */
ER
lcd_writecommand(LCD_Handler_t *hlcd, uint8_t c)
{
	ER ercd = E_OK;

	if(hlcd->spi_lock != 0){
		if((ercd = wai_sem(hlcd->spi_lock)) != E_OK)
			return ercd;
	}
	rs_set(PORT_LOW);
	cs_set(PORT_LOW);

	aTxBuffer[0] = c;
#if SPI_WAIT_TIME != 0
	ercd = spi_transmit(hlcd->hspi, (uint8_t*)aTxBuffer, 1);
#else
	if((ercd = spi_transmit(hlcd->hspi, (uint8_t*)aTxBuffer, 1)) == E_OK){
		ercd = spi_wait(hlcd->hspi, 100);
	}
#endif

	cs_set(PORT_HIGH);
	if(hlcd->spi_lock != 0)
		sig_sem(hlcd->spi_lock);
	return ercd;
}

/*
 *  LCDؤΥǡؿ
 */
ER
lcd_writedata(LCD_Handler_t *hlcd, uint8_t *p, uint32_t len)
{
	ER ercd = E_OK;

	if(hlcd->spi_lock != 0){
		if((ercd = wai_sem(hlcd->spi_lock)) != E_OK)
			return ercd;
	}
	rs_set(PORT_HIGH);
	cs_set(PORT_LOW);
	sil_dly_nse(100);

#if SPI_WAIT_TIME != 0
	ercd = spi_transmit(hlcd->hspi, (uint8_t*)p, len);
#else
	if((ercd = spi_transmit(hlcd->hspi, (uint8_t*)p, len)) == E_OK){
		ercd = spi_wait(hlcd->hspi, 100);
	}
#endif

	sil_dly_nse(100);
	cs_set(PORT_HIGH);
	if(hlcd->spi_lock != 0)
		sig_sem(hlcd->spi_lock);
	return ercd;
}

/*
 *  LCDؤΣХȥǡؿ
 */
ER
lcd_writedata2(LCD_Handler_t *hlcd, uint16_t c, int cnt)
{
	ER  ercd = E_OK;
	int i;

	if(cnt == 0)
		return ercd;
	if(hlcd->spi_lock != 0){
		if((ercd = wai_sem(hlcd->spi_lock)) != E_OK)
			return ercd;
	}
	rs_set(PORT_HIGH);
	cs_set(PORT_LOW);
	sil_dly_nse(100);
	aTxBuffer[0] = c>>8;
	aTxBuffer[1] = (uint8_t)c;
	for(i = 0 ; i < cnt ; i++){
		if((ercd = spi_transmit(hlcd->hspi, (uint8_t*)aTxBuffer, 2)) != E_OK){
			break;
		}
#if SPI_WAIT_TIME == 0
		if((ercd = spi_wait(hlcd->hspi, 100)) != E_OK){
			break;
		}
#endif
	}
	sil_dly_nse(100);
	cs_set(PORT_HIGH);
	if(hlcd->spi_lock != 0)
		sig_sem(hlcd->spi_lock);
	return ercd;
}

/*
 *  LCDؤRGBǡؿ
 */
ER
lcd_writedata3(LCD_Handler_t *hlcd, uint8_t *pbmp, uint16_t width, uint16_t height)
{
	ER  ercd = E_OK;
	int index, i;

	if(width == 0 || height == 0)
		return ercd;
	if(hlcd->spi_lock != 0){
		if((ercd = wai_sem(hlcd->spi_lock)) != E_OK)
			return ercd;
	}
	rs_set(PORT_HIGH);
	cs_set(PORT_LOW);
	sil_dly_nse(100);
	for(index=0; index < height; index++){
		uint8_t *p = pbmp;
		for (i = 0; i < width; i++){
			aTxBuffer[0]  = (*p++) & 0xf8;
			aTxBuffer[0] |= *p >> 5;
			aTxBuffer[1]  = (*p++ << 3) & 0xE0;
			aTxBuffer[1] |= (*p++ >> 3) & 0x1F;
			if((ercd = spi_transmit(hlcd->hspi, (uint8_t*)aTxBuffer, 2)) != E_OK){
				break;
			}
#if SPI_WAIT_TIME == 0
			if((ercd = spi_wait(hlcd->hspi, 100)) != E_OK){
				break;
			}
#endif
		}
		pbmp -= width*3;
	}
	sil_dly_nse(100);
	cs_set(PORT_HIGH);
	if(hlcd->spi_lock != 0)
		sig_sem(hlcd->spi_lock);
	return ercd;
}

/*
 *  ST7735SؤΥޥɥꥹ
 */
static void
Sipeed_ST7735_commandList(LCD_Handler_t *hlcd, const uint8_t *addr)
{
	uint8_t  numCommands, numArgs;
	uint8_t  data[4];
	uint16_t ms;
	ER       ercd;

	numCommands = *addr++;   // Number of commands to follow
	while(numCommands--) {                 // For each command...
		ercd = lcd_writecommand(hlcd, *addr);
		if(ercd != E_OK){
			if(ercd == E_OBJ && (hlcd->hspi->ErrorCode & SPI_ERROR_TIMEOUT) != 0)
				SVC_PERROR(lcd_writecommand(hlcd, *addr)); //   Read, issue command
			else
				syslog_1(LOG_ERROR, "Sipeed_ST7735_commandList SPI Error(%d) !", ercd);
		}
		addr++;
		numArgs  = *addr++;    //   Number of args to follow
		ms       = numArgs & DELAY;          //   If hibit set, delay follows args
		numArgs &= ~DELAY;                   //   Mask out delay bit
		while(numArgs--) {                   //   For each argument...
			data[0] = *addr++;
			SVC_PERROR(lcd_writedata(hlcd, data, 1));  //     Read, issue argument
		}

		if(ms) {
			ms = *addr++; // Read post-command delay time (ms)
			if(ms == 255) ms = 500;     // If 255, delay for 500 ms
			dly_tsk(ms);
		}
	}
}


/*
 *  ɽɥ
 */
void
lcd_setAddrWindow(LCD_Handler_t *hlcd, uint16_t x0, uint16_t y0, uint16_t x1, uint16_t y1)
{
	uint8_t  data[4];

	SVC_PERROR(lcd_writecommand(hlcd, ST7735_CASET)); // Column addr set
	data[0] = (uint8_t)((x0 + hlcd->colstart) >> 8);
	data[1] = (uint8_t)(x0 + hlcd->colstart);
	data[2] = (uint8_t)((x1 + hlcd->colstart) >> 8);
	data[3] = (uint8_t)(x1 + hlcd->colstart);
 	SVC_PERROR(lcd_writedata(hlcd, data, 4));

	SVC_PERROR(lcd_writecommand(hlcd, ST7735_RASET)); // Row addr set
	data[0] = (uint8_t)((y0 + hlcd->rowstart) >> 8);
	data[1] = (uint8_t)(y0 + hlcd->rowstart);
	data[2] = (uint8_t)((y1 + hlcd->rowstart) >> 8);
	data[3] = (uint8_t)(y1 + hlcd->rowstart);
	SVC_PERROR(lcd_writedata(hlcd, data, 4));

	SVC_PERROR(lcd_writecommand(hlcd, ST7735_RAMWR)); // write to RAM
}

/*
 *  ST7735S screensνؿ
 */
void
lcd_init(LCD_Handler_t *hlcd, uint8_t rotation)
{
	uint8_t buf[4];

	hlcd->rotate   = rotation;

	rs_set(PORT_LOW);

	/*
	 *  OLED CLEAR RESET
	 */
	rst_set(PORT_LOW);
	dly_tsk(200);
	/*
	 *  OLED SET RESET
	 */
	rst_set(PORT_HIGH);
	dly_tsk(20);
	/*
	 *  NO BLK CONTROL
	 */
//	OLED_BLK_Set();

	Sipeed_ST7735_commandList(hlcd, Scmd);
	SVC_PERROR(lcd_writecommand(hlcd, ST7735_MADCTL));
	if(rotation == 0){
		hlcd->_width  = ST7735_TFTHEIGHT;
		hlcd->_height  = ST7735_TFTWIDTH;
		hlcd->colstart = 26;
		hlcd->rowstart = 1;
		buf[0] = 0x08;
	}
	else if(rotation == 1){
		hlcd->_width = ST7735_TFTHEIGHT;
		hlcd->_height = ST7735_TFTWIDTH;
		hlcd->colstart = 26;
		hlcd->rowstart = 1;
		buf[0] = 0xC8;
	}
	else if(rotation == 2){
		hlcd->_width = ST7735_TFTWIDTH;
		hlcd->_height = ST7735_TFTHEIGHT;
		hlcd->colstart = 1;
		hlcd->rowstart = 26;
		buf[0] = 0x78;
	}
	else{
		hlcd->_width = ST7735_TFTWIDTH;
		hlcd->_height = ST7735_TFTHEIGHT;
		hlcd->colstart = 1;
		hlcd->rowstart = 26;
		buf[0] = 0xA8;
	}
	SVC_PERROR(lcd_writedata(hlcd, buf, 1));
	SVC_PERROR(lcd_writecommand(hlcd, ST7735_DISPON));
}


/*
 *  RECTANGLEΥե
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x: start X position
 *  param3  y: start Y position
 *  param4  w: width
 *  param5  h: height
 *  param6  color: color value
 */
void
lcd_fillRect(LCD_Handler_t *hlcd, int16_t x, int16_t y, int16_t w, int16_t h, color_t color)
{
	// rudimentary clipping (drawChar w/big text requires this)
	if((x >= hlcd->_width) || (y >= hlcd->_height)) return;
	if((x + w - 1) >= hlcd->_width)  w = hlcd->_width  - x;
	if((y + h - 1) >= hlcd->_height) h = hlcd->_height - y;

	lcd_setAddrWindow(hlcd, x, y, x+w-1, y+h-1);
	SVC_PERROR(lcd_writedata2(hlcd, color, h*w));
}

void
lcd_pushColor(LCD_Handler_t *hlcd, color_t color)
{
	SVC_PERROR(lcd_writedata2(hlcd, color, 1));
}

/*
 *  PIXEL
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x: X position
 *  param3  y: Y position
 *  param4  color: color value
 */
void
lcd_drawPixel(LCD_Handler_t *hlcd, int16_t x, int16_t y, color_t color)
{
	if((x < 0) ||(x >= hlcd->_width) || (y < 0) || (y >= hlcd->_height)) return;

	lcd_setAddrWindow(hlcd, x, y, x, y);
	SVC_PERROR(lcd_writedata2(hlcd, color, 1));
}

/*
 *  ľLINEι®
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x: start X position
 *  param3  y: start Y position
 *  param4  h: height
 *  param5  color: color value
 */
void
lcd_drawFastVLine(LCD_Handler_t *hlcd, int16_t x, int16_t y, int16_t h, color_t color)
{
	// Rudimentary clipping
	if((x >= hlcd->_width) || (y >= hlcd->_height)) return;
	if((y+h-1) >= hlcd->_height) h = hlcd->_height-y;
	lcd_setAddrWindow(hlcd, x, y, x, y+h-1);
	SVC_PERROR(lcd_writedata2(hlcd, color, h));
}

/*
 *  ʿLINEι®
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x: start X position
 *  param3  y: start Y position
 *  param4  w: width
 *  param5  color: color value
 */
void lcd_drawFastHLine(LCD_Handler_t *hlcd, int16_t x, int16_t y, int16_t w, color_t color)
{
	// Rudimentary clipping
	if((x >= hlcd->_width) || (y >= hlcd->_height)) return;
	if((x+w-1) >= hlcd->_width)  w = hlcd->_width-x;
	lcd_setAddrWindow(hlcd, x, y, x+w-1, y);
	SVC_PERROR(lcd_writedata2(hlcd, color, w));
}

/*
 *  DRAW IMAGE LINE
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x: X position
 *  param3  y: Y position
 *  param4  w: width
 *  param5  pcolor: color value
 */
void
lcd_drawImageHLine(LCD_Handler_t *hlcd, int16_t x, int16_t y, uint16_t w, color_t *pcolor)
{
	uint32_t i;

	if((x >= hlcd->_width) || (y >= hlcd->_height)) return;
	if((x+w-1) >= hlcd->_width)  w = hlcd->_width-x;

	lcd_setAddrWindow(hlcd, x, y, x+w-1, y);
	for(i = 0 ; i < w ; i++, pcolor++)
		SVC_PERROR(lcd_writedata2(hlcd, *pcolor, 1));
}

/*
 *  BITMAP
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  x0: Bmp X position in the LCD
 *  param3  y0: Bmp Y position in the LCD
 *  param4  pbmp: Pointer to Bmp picture address in the internal Flash
 */
void
lcd_drawBitmap(LCD_Handler_t *hlcd, uint16_t x0, uint16_t y0, uint8_t *pbmp)
{
	uint32_t index = 0, width = 0, height = 0, bit_pixel = 0;
	uint32_t input_color_mode = 0;

	/* Get bitmap data address offset */
	index = *(uint16_t *) (pbmp + 10);
	index |= (*(uint16_t *) (pbmp + 12)) << 16;

	/* Read bitmap width */
	width = *(uint16_t *) (pbmp + 18);
	width |= (*(uint16_t *) (pbmp + 20)) << 16;

	/* Read bitmap height */
	height = *(uint16_t *) (pbmp + 22);
	height |= (*(uint16_t *) (pbmp + 24)) << 16;

	/* Read bit/pixel */
	bit_pixel = *(uint16_t *) (pbmp + 28);

	/* Get the layer pixel format */
	if ((bit_pixel/8) == 4){
		input_color_mode = CM_ARGB8888;
	}
	else if ((bit_pixel/8) == 2){
		input_color_mode = CM_RGB565;
	}
	else{
		input_color_mode = CM_RGB888;
	}

	/* Bypass the bitmap header */
	pbmp += (index + (width * (height - 1) * (bit_pixel/8)));  
	syslog_4(LOG_INFO, "input_color_mode(%d) width(%d) height(%d) bit_pixel(%d)", input_color_mode, width, height, bit_pixel);

	lcd_setAddrWindow(hlcd, x0, y0, x0+width-1, y0+height-1);
	lcd_writedata3(hlcd, pbmp, width, height);
}

/*
 *  INVERT DISPLAY
 *  param1  hlcd: Pointer to LCD Handler
 *  param2  i: invert value
 */
void
lcd_invertDisplay(LCD_Handler_t *hlcd, bool_t i)
{
  SVC_PERROR(lcd_writecommand(hlcd, i ? ST7735_INVOFF : ST7735_INVON));
}


/*
 *  ꡼ե
 *  param1  pDrawProp: Pointer to Draw Prop
 */
void
lcd_fillScreen(LCD_DrawProp_t *pDrawProp)
{
	lcd_fillRect(pDrawProp->hlcd, 0, 0,  pDrawProp->hlcd->_width, pDrawProp->hlcd->_height, pDrawProp->BackColor);
}

/*
 *  RECTANGLE
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x: left X position
 *  param3  y: top Y position
 *  param4  w: width
 *  param5  h: height
 */
void
lcd_drawRect(LCD_DrawProp_t *pDrawProp, int16_t x, int16_t y, int16_t w, int16_t h)
{
	LCD_Handler_t *hlcd = pDrawProp->hlcd;
	color_t color;

	// rudimentary clipping (drawChar w/big text requires this)
	if((x >= hlcd->_width) || (y >= hlcd->_height)) return;
	if((x + w - 1) >= hlcd->_width)  w = hlcd->_width  - x;
	if((y + h - 1) >= hlcd->_height) h = hlcd->_height - y;

	color = pDrawProp->TextColor;
	lcd_drawFastVLine(hlcd, x, y, h, color);
	lcd_drawFastHLine(hlcd, x, y+h-1, w, color);
	lcd_drawFastVLine(hlcd, x+w-1, y, h, color);
	lcd_drawFastHLine(hlcd, x, y, w, color);
}

/*
 *  
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x0: X position
 *  param3  y0: Y position
 *  param4  Radius: Circle radius
 */
void
lcd_DrawCircle(LCD_DrawProp_t *pDrawProp, uint16_t x0, uint16_t y0, uint16_t Radius)
{
	LCD_Handler_t *hlcd = pDrawProp->hlcd;
 	int32_t   decision;    /* Decision Variable */ 
	uint32_t  current_x;   /* Current X Value */
	uint32_t  current_y;   /* Current Y Value */

	decision = 3 - (Radius << 1);
	current_x = 0;
	current_y = Radius;

	while(current_x <= current_y){
		lcd_drawPixel(hlcd, (x0 + current_x), (y0 - current_y), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 - current_x), (y0 - current_y), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 + current_y), (y0 - current_x), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 - current_y), (y0 - current_x), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 + current_x), (y0 + current_y), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 - current_x), (y0 + current_y), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 + current_y), (y0 + current_x), pDrawProp->TextColor);
		lcd_drawPixel(hlcd, (x0 - current_y), (y0 + current_x), pDrawProp->TextColor);

		if (decision < 0){
			decision += (current_x << 2) + 6;
		}
		else{
			decision += ((current_x - current_y) << 2) + 10;
			current_y--;
		}
		current_x++;
	}
}

/*
 *  
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x1: Point 1 X position
 *  param3  y1: Point 1 Y position
 *  param4  x2: Point 2 X position
 *  param5  y2: Point 2 Y position
 */
void
lcd_drawLine(LCD_DrawProp_t *pDrawProp, uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
{
	LCD_Handler_t *hlcd = pDrawProp->hlcd;
	int16_t deltax = 0, deltay = 0, x = 0, y = 0, xinc1 = 0, xinc2 = 0, 
	yinc1 = 0, yinc2 = 0, den = 0, num = 0, num_add = 0, num_pixels = 0, 
	curpixel = 0;

	deltax = ABS(x2 - x1);		/* The difference between the x's */
	deltay = ABS(y2 - y1);		/* The difference between the y's */
	x = x1;						/* Start x off at the first pixel */
	y = y1;						/* Start y off at the first pixel */

	if(x2 >= x1){				/* The x-values are increasing */
		xinc1 = 1;
		xinc2 = 1;
	}
	else{						/* The x-values are decreasing */
		xinc1 = -1;
		xinc2 = -1;
	}

	if(y2 >= y1){				/* The y-values are increasing */
		yinc1 = 1;
		yinc2 = 1;
	}
	else{						/* The y-values are decreasing */
		yinc1 = -1;
		yinc2 = -1;
	}

	if(deltax >= deltay){		/* There is at least one x-value for every y-value */
		xinc1 = 0;				/* Don't change the x when numerator >= denominator */
		yinc2 = 0;				/* Don't change the y for every iteration */
		den = deltax;
		num = deltax / 2;
		num_add = deltay;
		num_pixels = deltax;	/* There are more x-values than y-values */
	}
	else{						/* There is at least one y-value for every x-value */
		xinc2 = 0;				/* Don't change the x for every iteration */
		yinc1 = 0;				/* Don't change the y when numerator >= denominator */
		den = deltay;
		num = deltay / 2;
		num_add = deltax;
		num_pixels = deltay;	/* There are more y-values than x-values */
	}

	for (curpixel = 0; curpixel <= num_pixels; curpixel++){
		lcd_drawPixel(hlcd, x, y, pDrawProp->TextColor);	/* Draw the current pixel */
		num += num_add;			/* Increase the numerator by the top of the fraction */
		if(num >= den){			/* Check if numerator >= denominator */
			num -= den;			/* Calculate the new numerator value */
			x += xinc1;			/* Change the x as appropriate */
			y += yinc1;			/* Change the y as appropriate */
		}
		x += xinc2;				/* Change the x as appropriate */
		y += yinc2;				/* Change the y as appropriate */
	}
}

/*
 *  PLOY-LINE
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  Points: Pointer to the points array
 *  param3  PointCount: Number of points
 */
void
lcd_drawPolygon(LCD_DrawProp_t *pDrawProp, pPoint Points, uint16_t PointCount)
{
	int16_t x = 0, y = 0;

	if(PointCount < 2){
		return;
	}

	lcd_drawLine(pDrawProp, Points->X, Points->Y, (Points+PointCount-1)->X, (Points+PointCount-1)->Y);
	while(--PointCount){
		x = Points->X;
		y = Points->Y;
		Points++;
		lcd_drawLine(pDrawProp, x, y, Points->X, Points->Y);
	}
}

