/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2003-2016 by Ryosuke Takeuchi
 *                     GJ Business Division RICOH COMPANY,LTD. JAPAN
 *  Copyright (C) 2017-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭ԤϡFree Software Foundation ˤäƸɽƤ 
 *  GNU General Public License  Version 2 ˵ҤƤ狼
 *  (1)(4)ξ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥѲǽʥХʥꥳɡʥ֥륪
 *      ȥե饤֥ʤɡˤηѤˤϡ
 *      ȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭ɽ
 *      Ѿ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥԲǽʥХʥꥳɤηޤϵ
 *      ߹ѤˤϡΤ줫ξȡ
 *    (a) ѤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) Ѥη֤̤ˡˤäơ嵭Ԥ𤹤
 *        ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭Ԥդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥϡ
 *  ܥեȥ˴ؤơŬѲǽޤơʤݾڤԤ
 *  ʤޤܥեȥѤˤľŪޤϴŪ
 *  ʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: riscv64.c,v 1.1 2019/10/12 12:24:33 roi Exp $
 */

/* 
 *  TOPPERS/ASPѥ˥CPU¸ץࡥ
 *
 */

#include <itron.h>
#include <sil.h>
#include <stdio.h>
#include "kernel_impl.h"
#include "task.h"
#include "task_expansion.h"
#include "monitor.h"

extern void dispatch_r(void);

#define set_align(a, b)	((ulong_t)a & ~(b-1))

/*
 * 쥸ι¤
 */
typedef struct t_reg{
	unsigned long ra;
	unsigned long gp;
	unsigned long tp;
	unsigned long s0;
	unsigned long s1;
	unsigned long s2;
	unsigned long s3;
	unsigned long s4;
	unsigned long s5;
}T_REG;

/*
 * Υޥåԥ¤
 */

typedef struct t_memdef{
	uintptr_t mstart;
	uint32_t  mend;
	uint8_t   mtype;
	uint8_t   mstate;
}T_MEMDEF;

/*
 * RISC-V/64Υޥåԥ
 */

static T_MEMDEF const memdefine[] = {
	{0x02000000L, 0x02FFFFFFL,         PORT_AREA,   MREAD_WRITE},
	{0x0C000000L, 0x0CFFFFFFL,         PORT_AREA,   MREAD_WRITE},
	{0x38000000L, 0x3807FFFFL,         PORT_AREA,   MREAD_WRITE},
	{0x40000000L, 0x413FFFFFL,         PORT_AREA,   MREAD_WRITE},
	{0x50000000L, 0x543FFFFFL,         PORT_AREA,   MREAD_WRITE},
#if defined(TOPPERS_K210)
	{0x80000000L, 0x807FFFFFL,         MEMORY_AREA, MREAD_WRITE},
	{0x88000000L, 0x881FFFFFL,         MEMORY_AREA, MREAD_ONLY },
#else
#error "No support board type in RISC-V/64 groups."
#endif
};

static T_REG  sreg;

/******************************************************************************
 * ϡɥݡ°ȴؿ
 ******************************************************************************/
/*
 *  ɥ쥹ΰ°Ф
 *  mode=0:ΰη
 *  mode=1:ɤ߼񤭹°
 */

char
getMemoryType(ulong_t address, int_t mode)
{
	int_t count = sizeof(memdefine) / sizeof(T_MEMDEF);
	int_t i;

	for(i = 0 ; i < count ; i++){
		if(address >= memdefine[i].mstart && address <= memdefine[i].mend){
			if(mode == 0)
				return memdefine[i].mtype;
			else
				return memdefine[i].mstate;
		}
	}
	return NONE_AREA;
}

/*
 *  ɥ쥹饢饤Υɥ쥹Ф
 */

ulong_t
MonAlignAddress(ulong_t address)
{
	return address;
}

/******************************************************************************
 * ꥢѴؿ
 ******************************************************************************/
/*
 *  ΰФɤ߽Фؿ
 *  ΰΥåԤ顼ʤ饼֤
 */
int_t
MemoryRead(ulong_t address, intptr_t p, int_t type)
{
	int_t len;

	switch(getMemoryType(address, 0)){
	case PORT_AREA:
		if(type == 2){
			len = 2;
			*((UH *)p) = sil_reh_mem((VP)address);
		}
		else if(type == 4){
			len = 4;
			*((UW *)p) = sil_rew_mem((VP)address);
		}
		else{
			len = 1;
			*((UB *)p) = sil_reb_mem((VP)address);
		}
		break;
	case MEMORY_AREA:
		if(type == 2){
			len = 2;
			*((UH *)p) = *((UH *)set_align(address, len));
		}
		else if(type == 4){
			len = 4;
			*((UW *)p) = *((UW *)set_align(address, len));
		}
		else{
			len = 1;
			*((UB *)p) = *((UB *)address);
		}
		break;
	default:
		len = 0;
		break;
	}
	return len;
}

/*
 *  ΰФ񤭹ߴؿ
 *  ΰΥåԤ顼ʤ饼֤
 */
int_t
MemoryWrite(ulong_t address, intptr_t p, int_t type)
{
	int_t len;

	switch(getMemoryType(address, 0)){
	case PORT_AREA:
		if(type == 2){
			len = 2;
			address = set_align(address, len);
			sil_wrh_mem((VP)address, *((UH *)p));
		}
		else if(type == 4){
			len = 4;
			address = set_align(address, len);
			sil_wrw_mem((VP)address, *((UW *)p));
		}
		else{
			len = 1;
			sil_wrb_mem((VP)address, *((UB *)p));
		}
		break;
	case MEMORY_AREA:
		if(getMemoryType(address, 1) == MREAD_ONLY){
			len = 0;
		}
		else if(type == 2){
			len = 2;
			*((UH *)address) = *((UH *)set_align(p, len));
		}
		else if(type == 4){
			len = 4;
			*((UW *)address) = *((UW *)set_align(p, len));
		}
		else{
			len = 1;
			*((UB *)address) = *((UB *)p);
		}
		break;
	default:
		len = 0;
		break;
	}
	return len;
}

/******************************************************************************
 * ˥Ѵؿ
 ******************************************************************************/
/*
 *  쥸Ƥɽ
 */
void
display_registers(ID tskid)
{
	ER     ercd;
	T_RTST rtst;

	ercd = ref_tst(tskid, &rtst);
	if(ercd == E_OK){
		if(rtst.tskpc == (FP)dispatch_r){
			sreg = *((T_REG *)rtst.tsksp);
			printf(" RA  =%lx SP =%lx\n", (long)sreg.ra, (long)((long)rtst.tsksp+sizeof(T_REG)));
			printf(" GP  =%lx TP =%lx S0 =%lx S1 =%lx\n", sreg.gp, sreg.tp, sreg.s0, sreg.s1);
			printf(" S2  =%lx S3 =%lx S4 =%lx S5 =%lx\n", sreg.s2, sreg.s3, sreg.s4, sreg.s5);
			printf("     %lx    %04x\n", (long)sreg.ra, *((UH*)(sreg.ra & ~1)));
			return;
		}
		else if(rtst.tskstat == TTS_DMT){
			printf("  wait in activate_r() !!\n");
			return;
		}
	}
	printf("  wait in dispatch() !!\n");
}

