/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2019-04-13 19:53:59Z roi $
 */
/*
 *  FE310ѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

static const char monthday[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static void (*alarm_callback)(void);	/* RTC Alam callback */
uint_t dipsw_value;


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_LEVEL         0x00040000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define LEVEL_LOW               0x00150000
#define LEVEL_HIGH              0x00250000


/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	/*
	 *  GPIO⡼
	 */
	if((init->mode & GPIO_MODE) == GPIO_MODE_INPUT){
		sil_andw_mem((uint32_t *)(base+TOFF_GPIO_OUTPUT_EN), 1 << pin);
		sil_andw_mem((uint32_t *)(base+TOFF_GPIO_PULLUP_EN), 1 << pin);
		sil_orw_mem((uint32_t *)(base+TOFF_GPIO_INPUT_EN), 1 << pin);
	}
	else{
		sil_andw_mem((uint32_t *)(base+TOFF_GPIO_INPUT_EN), 1 << pin);
		if(init->pull == GPIO_PULLUP)
			sil_orw_mem((uint32_t *)(base+TOFF_GPIO_PULLUP_EN), 1 << pin);
		else
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_PULLUP_EN), 1 << pin);
		if((init->mode & GPIO_MODE) == GPIO_MODE_OUTPUT){
			if(init->oxor == GPIO_OUTPUT_XOR)
				sil_orw_mem((uint32_t *)(base+TOFF_GPIO_OUTPUT_XOR), 1 << pin);
			else
				sil_andw_mem((uint32_t *)(base+TOFF_GPIO_OUTPUT_XOR), 1 << pin);
		}
		else
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_OUTPUT_XOR), 1 << pin);
		sil_orw_mem((uint32_t *)(base+TOFF_GPIO_OUTPUT_EN), 1 << pin);
	}

	/*
	 *  EXTI⡼
	 */
	if((init->mode & GPIO_MODE_IT) != 0){
		if((init->mode & GPIO_MODE_LEVEL) != 0){
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_HIGH_IE), 1 << pin);
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_LOW_IE), 1 << pin);
			if((init->mode & LEVEL_LOW) != 0)
				sil_orw_mem((uint32_t *)(base+TOFF_GPIO_LOW_IE), 1 << pin);
			else
				sil_orw_mem((uint32_t *)(base+TOFF_GPIO_HIGH_IE), 1 << pin);
		}
		else{
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_RISE_IE), 1 << pin);
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_FALL_IE), 1 << pin);
			if((init->mode & RISING_EDGE) != 0)
				sil_orw_mem((uint32_t *)(base+TOFF_GPIO_RISE_IE), 1 << pin);
			if((init->mode & FALLING_EDGE) != 0)
				sil_orw_mem((uint32_t *)(base+TOFF_GPIO_FALL_IE), 1 << pin);
		}
	}
	/* IOF1 */
	if(init->iof != 0){
		if(init->iof & 2)
			sil_orw_mem((uint32_t *)(base+TOFF_GPIO_IOF_SEL), 1 << pin);
		else
			sil_andw_mem((uint32_t *)(base+TOFF_GPIO_IOF_SEL), 1 << pin);
		sil_orw_mem((uint32_t *)(base+TOFF_GPIO_IOF_EN), 1 << pin);
	}
	else
		sil_andw_mem((uint32_t *)(base+TOFF_GPIO_IOF_EN), 1 << pin);
}


/*
 *  ƥ९å֤
 *  return SYSCLK(Hz)
 */
uint32_t
get_sysclock(uint8_t *sws)
{
	if(sws != NULL)
		*sws = 0;
	return SystemFrequency;
}


/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  wdogtime: WDOGȯ(sec)
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, uint16_t wdogtime)
{
	uint32_t base = TADR_AON_BASE;

	if(wdogtime == 0 || portid != WDOG1_PORTID)
		return E_PAR;

	sil_wrw_mem((uint32_t *)(base+TOFF_AON_WDOGKEY), AON_WDOGKEY_VALUE);
	sil_wrw_mem((uint32_t *)(base+TOFF_AON_WDOGCMP), wdogtime);
	/*
	 *  ե졼󥻥å
	 */
	sil_wrw_mem((uint32_t *)(base+TOFF_AON_WDOGKEY), AON_WDOGKEY_VALUE);
	sil_orw_mem((uint32_t *)(base+TOFF_AON_WDOGCFG),
		 (AON_WDOGCFG_RSTEN | AON_WDOGCFG_ENALWAYS | AON_WDOGCFG_ZEROCMP | AON_WDOGCFG_SCALE));
	sil_wrw_mem((uint32_t *)(base+TOFF_AON_WDOGKEY), AON_WDOGKEY_VALUE);
	sil_wrw_mem((uint32_t *)(base+TOFF_AON_WDOGFEED), AON_WDOGFEED_VALUE);

	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_WDOGKEY), AON_WDOGKEY_VALUE);
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_WDOGCFG),
		 (AON_WDOGCFG_ZEROCMP | AON_WDOGCFG_SCALE));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_WDOGKEY), AON_WDOGKEY_VALUE);
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_WDOGFEED), AON_WDOGFEED_VALUE);
	return E_OK;
}


/*
 *  tm2timesecѴ
 */
static uint32_t
mktimesec(struct tm2 *ptm)
{
	int day = ptm->tm_year * 365;
	int yday, i;

	day += ptm->tm_year / 4;
	for(i = 1, yday = 0 ; i < ptm->tm_mon ; i++)
		yday += monthday[i-1];
	if(ptm->tm_mon > 2 && ((ptm->tm_year+2) % 4) == 0)
		yday++;
	yday += ptm->tm_mday - 1;
	day  += yday;
	ptm->tm_wday = (day+4) % 7;
	ptm->tm_yday = yday+1;
	return day * (24*60*60) + ptm->tm_hour * 60*60 + ptm->tm_min * 60 + ptm->tm_sec;
}

/*
 *  timesectm2ؤѴ
 */
static void
mktime2(struct tm2 *ptm, uint32_t timesec)
{
	int day = timesec / (24*60*60);
	int time, i, yday, mday;

	time = timesec - day * 24 * 60 * 60;
	ptm->tm_wday = (day+4) % 7;
	for(i = 365, ptm->tm_year = 0 ; i < day ; i += 365){
		if(((ptm->tm_year+2) % 4) == 0)
			i++;
		ptm->tm_year++;
	}

	ptm->tm_yday = yday = day - i + 365 + 1;
	for(ptm->tm_mon = 1 ; ptm->tm_mon <= 12 && yday > 0 ; ptm->tm_mon++){
		mday = monthday[ptm->tm_mon-1];
		if(ptm->tm_mon == 2 && ((ptm->tm_year+2) % 4) == 0)
			mday++;
		if(yday < mday){
			ptm->tm_mday = yday;
			break;
		}
		yday -= mday;
	}
	ptm->tm_hour = time / (60*60);
	time -= ptm->tm_hour * 60 * 60;
	ptm->tm_min  = time / 60;
	ptm->tm_sec = time % 60;
}

/*
 *  RTCѽ
 */
void
rtc_init(intptr_t exinf)
{
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCMP), 0xFFFFFFFF);
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCFG), (AON_RTCCFG_ENALWAYS | AON_RTCCFG_SCALE));
}

/*
 *  RTCλؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_set_time(struct tm2 *pt)
{
	uint64_t rtccount = 0;

	if(pt == NULL)
		return E_PAR;

	SVC_PERROR(wai_sem(RTCSEM));

	rtccount = (uint64_t)mktimesec(pt) * RTC_FREQ;
	sil_andw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCFG), AON_RTCCFG_ENALWAYS);
	sil_dly_nse(100*1000);
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCHI), (uint32_t)(rtccount >> 32));
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCLO), (uint32_t)rtccount);
	sil_orw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCFG), AON_RTCCFG_ENALWAYS);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTCλФؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_get_time(struct tm2 *pt)
{
	uint64_t rtccount = 0;

	if(pt == NULL)
		return E_PAR;

	SVC_PERROR(wai_sem(RTCSEM));

	rtccount  = sil_rew_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCLO));
	rtccount |= (uint64_t)sil_rew_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCHI)) << 32;
	rtccount /= RTC_FREQ;
	mktime2(pt, (uint32_t)rtccount);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC顼
 */
ER
rtc_setalarm(struct tm2 *ptm, void *func)
{
	uint32_t comptime;

	if(ptm == NULL)
		return E_PAR;
	SVC_PERROR(wai_sem(RTCSEM));

	alarm_callback = (void(*)(void))func;
	comptime = mktimesec(ptm);
	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCMP), comptime);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC ALRAMϥɥ
 */
void
rtc_handler(void)
{
	uint32_t temp;

	sil_wrw_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCMP), 0xFFFFFFFF);
	temp = sil_rew_mem((uint32_t *)(TADR_AON_BASE+TOFF_AON_RTCCMP));
	sil_dly_nse(1000);
	if(alarm_callback != NULL)
		alarm_callback();
	(void)(temp);
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
	uint32_t	pinxor;
	uint32_t    piniof;
	uint32_t	offpwm;
} gio_conftype;

static const gio_conftype led_confdata[] = {
	{ TADR_GPIO_BASE, PINPOSITION22, GPIO_NOPULL, GPIO_OUTPUT_XOR,    GPIO_IOF1, TOFF_PWM_CMP3},/* R */
	{ TADR_GPIO_BASE, PINPOSITION19, GPIO_NOPULL, GPIO_OUTPUT_NORMAL, GPIO_IOF1, TOFF_PWM_CMP1},/* G */
	{ TADR_GPIO_BASE, PINPOSITION21, GPIO_NOPULL, GPIO_OUTPUT_NORMAL, GPIO_IOF1, TOFF_PWM_CMP2}	/* B */
};
#define NUM_LED         (sizeof(led_confdata)/sizeof(gio_conftype))
#define LED_PWM_BASE    TADR_PWM1_BASE

/*
 *  PWMͤμФ
 */
static uint16_t
get_pwm_value(uint8_t no, uint32_t pwmval)
{
	uint32_t value;

	switch(no){
	case 0:
		value = (0xFFFF - pwmval) >> 4;
		break;
	case 1:
		value = pwmval >> 4;
		break;
	case 2:
		value = pwmval >> 5;
		break;
	default:
		value = 0;
	}
	return value;
}

/*
 *  PWMѴ
 */
static uint32_t
make_pwm_value(uint8_t no, uint16_t value)
{
	uint32_t result;

	switch(no){
	case 0:
		result = 0xFFFF - (value << 4);
		break;
	case 1:
		result = value << 4;
		break;
	case 2:
		result = value << 5;
		break;
	default:
		result = 0;
	}
	return result;
}

/*
 *  LED³ݡȽ
 */
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t pwmbase = LED_PWM_BASE;
	uint32_t i;

	sil_wrw_mem((uint32_t *)(pwmbase+TOFF_PWM_CFG), 0);
	/*
	 *  ϤΥХ
	 */
	sil_wrw_mem((uint32_t *)(pwmbase+TOFF_PWM_CFG), (PWM_CFG_ENALWAYS | PWM_CFG_CMP2CENTER));
	sil_wrw_mem((uint32_t *)(pwmbase+TOFF_PWM_COUNT), 0);

	/*
	 *  ȿ244Hz, CMP0˥
	 */
	sil_wrw_mem((uint32_t *)(pwmbase+TOFF_PWM_CMP0), 0);

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.oxor      = pled->pinxor;
		/* IOF#1 */
		GPIO_Init_Data.iof       = pled->piniof;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pwmbase+pled->offpwm), make_pwm_value(i, 0));
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in_color(uint8_t no)
{
	const gio_conftype *pled = &led_confdata[no];

	if(no >= NUM_LED)
		return 0;
	return get_pwm_value(no, sil_rew_mem((uint32_t *)(LED_PWM_BASE+pled->offpwm)));
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out_color(uint8_t no, uint16_t led_data)
{
	const gio_conftype *pled = &led_confdata[no];

	if(no >= NUM_LED)
		return;
	if(led_data > 255)
		led_data = 255;
	sil_wrw_mem((uint32_t *)(LED_PWM_BASE+pled->offpwm), make_pwm_value(no, led_data));
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

