/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: i2c.h 59 2020-11-18 10:15:39Z roi $
 */

/*
 *  FE310 I2Cデバイスドライバの外部宣言
 */

#ifndef _I2C_H_
#define _I2C_H_

#ifdef __cplusplus
 extern "C" {
#endif

/*
 *  I2Cポート定義
 */
#define I2C1_PORTID             1
#define NUM_I2CPORT             1

/*
 *  I2Cメモリアドレスモード
 */
#define I2C_MEMADD_SIZE_NONE    0x00000000
#define I2C_MEMADD_SIZE_8BIT    0x00000001
#define I2C_MEMADD_SIZE_16BIT   0x00000002

#ifndef TOPPERS_MACRO_ONLY

/*
 *  I2Cハードウェア設定構造体
 */

typedef struct _I2C_PortControlBlock{
	uint32_t              base;
	uint16_t              sclpin;
	uint16_t              sdapin;
	uint32_t              gpiopull;
	uint32_t              gpioiof1;
} I2C_PortControlBlock;

/*
 *  I2Cコンフュギュレーション構造体定義
 */
typedef struct
{
	uint32_t              ClockSpeed;		/* I2C クロックスピード */
	int                   semlock;			/* I2C ロックセマフォ値 */
}I2C_Init_t;

/*
 *  I2Cエラー定義
 */
#define I2C_ERROR_NONE    0x00000000		/* No error              */

/*
 *  拡張UARTハンドラ定義
 */
typedef struct i2c_struct I2C_Handle_t;
struct i2c_struct {
	uint32_t              base;				/* I2C register base address */
	I2C_Init_t            Init;				/* I2C初期設定パラメータ */
	uint32_t	          portid;			/* I2C ポートID */
	uint32_t              clkCount;			/* I2C clockカウンタ */
	uint16_t              sclpin;			/* I2C SCL ピン番号 */
	uint16_t              sdapin;			/* I2C SDA ピン番号 */
	uint32_t              gpiopull;			/* I2C GPIO PULLMODE */
	uint32_t              gpioiof1;			/* I2C GPIO IOF#1選択 */
	uint32_t              sdaoe;
	uint32_t              ErrorCode;
};

#endif /* TOPPERS_MACRO_ONLY */

/*
 *  拡張I2C関数のプロトタイプ宣言
 */
#ifndef TOPPERS_MACRO_ONLY

extern I2C_Handle_t *i2c_init(ID portid, I2C_Init_t *ii2c);
extern ER i2c_deinit(I2C_Handle_t *hi2c);
extern ER i2c_memwrite(I2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, const void * data, uint32_t Size);
extern ER i2c_memread(I2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, void * data, uint32_t Size);
extern ER i2c_ready(I2C_Handle_t *hi2c, uint16_t DevAddress, uint32_t Trials, uint32_t Timeout);


#endif /* TOPPERS_MACRO_ONLY */

#ifdef __cplusplus
}
#endif

#endif	/* _I2C_H_ */

