/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: spi.c 698 2019-04-22 17:53:22Z roi $
 */
/*
 * 
 *  SPIɥ饤дؿ
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "spi.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))


/*
 *  SPIOݡID֥åФΥޥ
 */
#define INDEX_SPI(spiid)        ((uint_t)((spiid) - 1))

#define SPI_TXFIFO_SIZE    8
#define SPI_TIMEOUT_VALUE  (10*1000)

static const SPI_PortControlBlock spi_pcb[NUM_SPIPORT] = {
  {	TADR_SPI0_BASE, TADR_GPIO_BASE, 0x00000000, 0x00000000, 0x00000000 },

  {	TADR_SPI1_BASE, TADR_GPIO_BASE,
	(IOF_SPI1_SCK | IOF_SPI1_MISO | IOF_SPI1_MOSI),
	(IOF_SPI1_SS0 | IOF_SPI1_SS1  | IOF_SPI1_SS2 | IOF_SPI1_SS3),
	(IOF_SPI1_DQ2 | IOF_SPI1_DQ3) },

  {	TADR_SPI2_BASE, TADR_GPIO_BASE,
	(IOF_SPI2_SCK | IOF_SPI2_MISO | IOF_SPI2_MOSI),
	(IOF_SPI2_SS0),
	(IOF_SPI2_DQ2 | IOF_SPI2_DQ3) },
};

static SPI_Handle_t SpiHandle[NUM_SPIPORT];



/*
 *  SPIžλԤ
 */
ER
spi_inwait(SPI_Handle_t *hspi, uint32_t timeout)
{
	ER ercd = E_OK;
	int tick = timeout;

	while((hspi->status & SPI_STATUS_BUSY) != 0 && tick > 0){
		if(hspi->Init.semid != 0){
	 		ercd = twai_sem(hspi->Init.semid, 50);
		}
		else
			dly_tsk(1);
		tick--;
	}
	if(hspi->ErrorCode != 0)
		ercd = E_OBJ;
	else if(tick == 0){
		hspi->ErrorCode = SPI_ERROR_TIMEOUT;
		ercd = E_TMOUT;
	}
	return ercd;
}

/*
 *  SPI
 *  parameter1  port: SPIݡֹ
 *  parameter2  spii: SPI깽¤ΤؤΥݥ
 *  return SPIϥɥؤΥݥ󥿡NULLǥ顼
 */
SPI_Handle_t *
spi_init(ID port, SPI_Init_t *spii)
{
	SPI_Handle_t *hspi;
	const SPI_PortControlBlock *spcb;
	uint32_t no;
	volatile uint32_t tmpreg;

	if(port < SPI2_PORTID || port > NUM_SPIPORT)
		return NULL;
	if(spii == NULL)
		return NULL;

	no = INDEX_SPI(port);
	hspi = &SpiHandle[no];
	if(hspi->status != SPI_STATUS_RESET)
		return NULL;
	spcb = &spi_pcb[no];
	memcpy(&hspi->Init, spii, sizeof(SPI_Init_t));
	hspi->base = spcb->base;
	hspi->port = port;

	/*
	 *  SPI GPIOԥ
	 */
	tmpreg = spcb->gpio1pin;
	if(spii->CSMode != SPI_CSMODE_OFF)
		tmpreg |= spcb->gpio2pin;
	if(spii->Protocol == SPI_PROTOCOL_4LINES)
		tmpreg |= spcb->gpio3pin;
	sil_andw_mem((uint32_t *)(spcb->giobase+TOFF_GPIO_INPUT_EN), tmpreg);
	sil_orw_mem((uint32_t *)(spcb->giobase+TOFF_GPIO_OUTPUT_VAL), tmpreg);
	sil_orw_mem((uint32_t *)(spcb->giobase+TOFF_GPIO_OUTPUT_EN), tmpreg);
	sil_andw_mem((uint32_t *)(spcb->giobase+TOFF_GPIO_IOF_SEL), tmpreg);
	sil_orw_mem((uint32_t *)(spcb->giobase+TOFF_GPIO_IOF_EN), tmpreg);

	/*
	 *  ץꥹ/å
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_SCKDIV), spii->Prescaler);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_SCKMODE), (spii->CLKPolarity | spii->CLKPhase));

	/*
	 *  CS
	 */
	if(spii->CSMode == SPI_CSMODE_OFF)
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSID), 0);
	else
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSID), spcb->gpio2pin);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSDEF), 0xffffffff);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSMODE), spii->CSMode);
	if(spii->CSMode == SPI_CSM_AUTO){
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_DCSSCK), ((spii->CSDelay << 16) | (spii->SCKDelay)));
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_DINTERCS), ((spii->MaxifTime << 16) | (spii->MinScInActTime)));
	}

	/*
	 *  եޥå
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_FMT), ((spii->DataSize << 16) | spii->SignBit | spii->Protocol));

	hspi->ErrorCode = SPI_ERROR_NONE;
	hspi->status = SPI_STATUS_READY;
	return hspi;
}

/*
 *  SPIλ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  return ER
 */
ER
spi_deinit(SPI_Handle_t *hspi)
{
	if(hspi == NULL)
		return E_PAR;

	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), 0);

	hspi->ErrorCode = SPI_ERROR_NONE;
	hspi->status = SPI_STATUS_RESET;
	return E_OK;
}

/*
 *  SPI⥸塼Υꥻå
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  return ER
 */
ER
spi_reset(SPI_Handle_t *hspi)
{
	SPI_Init_t *spii;
	const SPI_PortControlBlock *spcb;

	if(hspi == NULL)
		return E_PAR;

	spii = &hspi->Init;
	spcb = &spi_pcb[INDEX_SPI(hspi->port)];
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), 0);

	/*
	 *  ץꥹ/å
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_SCKDIV), spii->Prescaler);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_SCKMODE), (spii->CLKPolarity | spii->CLKPhase));

	/*
	 *  CS
	 */
	if(spii->CSMode == SPI_CSMODE_OFF)
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSID), 0);
	else
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSID), spcb->gpio2pin);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSDEF), 0xffffffff);
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CSMODE), spii->CSMode);
	if(spii->CSMode == SPI_CSM_AUTO){
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_DCSSCK), ((spii->CSDelay << 16) | (spii->SCKDelay)));
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_DINTERCS), ((spii->MaxifTime << 16) | (spii->MinScInActTime)));
	}

	/*
	 *  եޥå
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_FMT), ((spii->DataSize << 16) | spii->SignBit | spii->Protocol));

	hspi->ErrorCode = SPI_ERROR_NONE;
	hspi->status = SPI_STATUS_RESET;
	return E_OK;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ
 *  parameter3  length: 
 *  return ER
 */
ER
spi_transmit(SPI_Handle_t *hspi, uint8_t *pdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || pdata == NULL || length == 0)
		return E_PAR;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

	/* Configure communication */
	hspi->xmode       = SPI_XMODE_TX;
	hspi->status      = SPI_STATUS_BUSY;
	hspi->ErrorCode   = SPI_ERROR_NONE;

	hspi->pTxBuffPtr  = pdata;
	hspi->TxXferSize  = length;
	hspi->TxXferCount = 0;
	hspi->pRxBuffPtr  = NULL;
	hspi->RxXferSize  = length;
	hspi->RxXferCount = 0;

	/*
	 *  TXFIFOʬǡ񤭹
	 */
	while(hspi->TxXferSize > hspi->TxXferCount && hspi->TxXferCount < SPI_TXFIFO_SIZE){
		if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO)) & SPI_TXFIFO_FULL) != 0)
			break;
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO), *hspi->pTxBuffPtr);
		hspi->pTxBuffPtr++;
		hspi->TxXferCount++;
	}
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), (SPI_IP_TXWM | SPI_IP_RXWM));

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ
 *  parameter3  length: 
 *  return ER
 */
ER
spi_receive(SPI_Handle_t *hspi, uint8_t *pdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || pdata == NULL || length == 0)
		return E_PAR;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

    /*
	 *  ž
	 */
	hspi->xmode       = SPI_XMODE_RX;
    hspi->status      = SPI_STATUS_BUSY;
    hspi->ErrorCode   = SPI_ERROR_NONE;
    hspi->pRxBuffPtr  = pdata;
    hspi->RxXferSize  = length;
    hspi->RxXferCount = 0;
	hspi->pTxBuffPtr  = NULL;
    hspi->TxXferSize  = length;
    hspi->TxXferCount = 0;

	/*
	 *  TXFIFOʬǡ񤭹
	 */
	while(hspi->TxXferSize > hspi->TxXferCount && hspi->TxXferCount < SPI_TXFIFO_SIZE){
		if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO)) & SPI_TXFIFO_FULL) != 0)
			break;
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO), 0xFF);
		hspi->TxXferCount++;
	}
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), (SPI_IP_TXWM | SPI_IP_RXWM));

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  ptxdata: ХåեؤΥݥ
 *  parameter3  prxdata: ХåեؤΥݥ
 *  parameter4  length: 
 *  return ER
 */
ER
spi_transrecv(SPI_Handle_t *hspi, uint8_t *ptxdata, uint8_t *prxdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || prxdata == NULL || length == 0)
		return E_PAR;
	if(hspi->Init.Direction != SPI_DIRECTION_2LINES && hspi->Init.Protocol != SPI_PROTOCOL_1LINE)
		return E_PAR;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

	hspi->xmode = SPI_XMODE_TXRX;
	hspi->status = SPI_STATUS_BUSY;

	/*
	 *  ž
	 */
	hspi->ErrorCode   = SPI_ERROR_NONE;
	hspi->pTxBuffPtr  = (uint8_t*)ptxdata;
	hspi->TxXferSize  = length;
	hspi->TxXferCount = 0;
	hspi->pRxBuffPtr  = (uint8_t*)prxdata;
	hspi->RxXferSize  = length;
	hspi->RxXferCount = 0;

	/*
	 *  TXFIFOʬǡ񤭹
	 */
	while(hspi->TxXferSize > hspi->TxXferCount && hspi->TxXferCount < SPI_TXFIFO_SIZE){
		if((sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO)) & SPI_TXFIFO_FULL) != 0)
			break;
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_TXFIFO), *hspi->pTxBuffPtr);
		hspi->pTxBuffPtr++;
		hspi->TxXferCount++;
	}
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), (SPI_IP_TXWM | SPI_IP_RXWM));

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPIžλԤ
 */
ER
spi_wait(SPI_Handle_t *hspi, uint32_t timeout)
{
	ER ercd = E_OK;

#if SPI_WAIT_TIME == 0
	if(hspi == NULL)
		return E_PAR;
	ercd = spi_inwait(hspi, timeout);
	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPIߥӥ롼
 */
void
spi_handler(SPI_Handle_t *hspi)
{
	uint32_t base = hspi->base;
	volatile uint32_t ip = sil_rew_mem((uint32_t *)(base+TOFF_SPI_IP));
	volatile uint32_t tmp;

	if((ip & SPI_IP_RXWM) != 0){
		while(hspi->RxXferSize > hspi->RxXferCount){
			tmp = sil_rew_mem((uint32_t *)(base+TOFF_SPI_RXFIFO));
			if((tmp & SPI_RXFIFO_EMPTY) != 0)
				break;
			if(hspi->pRxBuffPtr != NULL)
				*hspi->pRxBuffPtr++ = (uint8_t)tmp;
			hspi->RxXferCount++;
		}
		if(hspi->RxXferSize <= hspi->RxXferCount)
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), SPI_IP_RXWM);
	}
	while((hspi->TxXferCount - hspi->RxXferCount) < SPI_TXFIFO_SIZE  && hspi->TxXferSize > hspi->TxXferCount){
		if((sil_rew_mem((uint32_t *)(base+TOFF_SPI_TXFIFO)) & SPI_TXFIFO_FULL) != 0)
			break;
		if(hspi->pTxBuffPtr != NULL)
			tmp = (uint32_t)*hspi->pTxBuffPtr++;
		else
			tmp = 0xff;
		sil_wrw_mem((uint32_t *)(base+TOFF_SPI_TXFIFO), tmp);
		hspi->TxXferCount++;
	}
	if(hspi->TxXferSize <= hspi->TxXferCount)
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_IE), SPI_IP_TXWM);

	if(hspi->status == SPI_STATUS_BUSY && hspi->RxXferSize <= hspi->RxXferCount
		&& hspi->TxXferSize <= hspi->TxXferCount){
		hspi->status = SPI_STATUS_READY;
		if(hspi->xfercallback != NULL)
			hspi->xfercallback(hspi);
		if(hspi->Init.semid != 0)
			isig_sem(hspi->Init.semid);
	}
}


/*
 *  SPIߥӥ롼
 */
void spi_isr(intptr_t exinf)
{
  spi_handler(&SpiHandle[INDEX_SPI((uint32_t)exinf)]);
}

