/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: adc.c 698 2020-10-29 09:15:37Z roi $
 */
/*
 * 
 *  GD32V103 ADCɥ饤дؿ
 *
 */
#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include <target_syssvc.h>
#include "device.h"
#include "adc.h"

/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  I2CݡID֥åФΥޥ
 */
#define INDEX_ADC(adcid)        ((uint_t)((adcid) - 1))

/*
 *  ADC¹Ԥ(sec)
 */
#ifndef ADC_STAB_DELAY_US
#define ADC_STAB_DELAY_US              3
#endif

/*
 *  ADCݡơ֥
 */
static const ADC_PortControlBlock adc_pcb[NUM_ADCPORT] = {
  {	TADR_ADC0_BASE,
	(TADR_RCU_BASE+TOFF_RCU_APB2EN),  RCU_APB2EN_ADC0EN,
	(TADR_RCU_BASE+TOFF_RCU_APB2RST), RCU_APB2RST_ADC0RST,
	DMA1CH0_ID },
  {	TADR_ADC1_BASE,
	(TADR_RCU_BASE+TOFF_RCU_APB2EN),  RCU_APB2EN_ADC1EN,
	(TADR_RCU_BASE+TOFF_RCU_APB2RST), RCU_APB2RST_ADC1RST,
	-1 }
};

static ADC_Handle_t AdcHandle[NUM_ADCPORT];

/*
 *  ADC DMAžλХåؿ
 */
static void
adc_dmatrans_func(DMA_Handle_t *hdma)
{
	ADC_Handle_t* hadc = ( ADC_Handle_t* )((DMA_Handle_t* )hdma)->localdata;

	if(hadc->Init.ContinuousConvMode == ADC_CONTINUOUS_DISABLE)
		hadc->status = ADC_STATUS_EOC;
	else
		hadc->status = ADC_STATUS_BUSY_EOC;
	if(hadc->xfercallback != NULL)
		hadc->xfercallback(hadc);
}

/*
 *  ADC DMAϡžХåؿ
 */
static void
adc_dmahalftrans_func(DMA_Handle_t *hdma)
{
	ADC_Handle_t* hadc = ( ADC_Handle_t* )((DMA_Handle_t* )hdma)->localdata;

	if(hadc->xferhalfcallback != NULL)
		hadc->xferhalfcallback(hadc);
}

/*
 *  ADC DMA顼Хåؿ
 */
static void
adc_dmaerror_func(DMA_Handle_t *hdma)
{
	ADC_Handle_t* hadc = ( ADC_Handle_t* )((DMA_Handle_t* )hdma)->localdata;

	hadc->status = ADC_STATUS_READY;
	hadc->ErrorCode |= ADC_ERROR_DMA;
	if(hadc->errorcallback != NULL)
		hadc->errorcallback(hadc);
}

/*
 *  ADC
 *  parameter1  port: ADCݡֹ
 *  parameter2  pini: ADC깽¤ΤؤΥݥ
 *  return ADCϥɥؤΥݥ󥿡NULLǥ顼
 */
ADC_Handle_t *
adc_init(ID portid, ADC_Init_t* pini)
{
	static DMA_Handle_t  hdma_adc;
	DMA_Handle_t  *hdma;
	ADC_Handle_t *hadc;
	const ADC_PortControlBlock *apcb;
	uint32_t reg;

	/*
	 *  ѥ᡼å
	 */
	if(portid < 1 || portid > NUM_ADCPORT)
		return NULL;

	apcb = &adc_pcb[INDEX_ADC(portid)];
	if(pini->DMAChannel >= 0 && pini->DMAChannel != apcb->dmachannel)
		return NULL;

	hadc = &AdcHandle[INDEX_ADC(portid)];
	memcpy(&hadc->Init, pini, sizeof(ADC_Init_t));

	hadc->base = apcb->base;
	hadc->apcb = (ADC_PortControlBlock *)apcb;

	/*
	 *  ADCå
	 */
	sil_orw_mem((uint32_t *)(apcb->adcclockbase), apcb->adcclockbit);
	sil_rew_mem((uint32_t *)(apcb->adcclockbase));

	/*
	 *  DMAбСDMAԤ
	 */
	hdma = &hdma_adc;
	if((hdma->cbase = dma_channel_base(pini->DMAChannel)) != 0){
		hdma->Init.Direction = DMA_PERIPH_TO_MEMORY;
		hdma->Init.PeriphInc = DMA_PINC_DISABLE;
		hdma->Init.MemInc = DMA_MINC_ENABLE;
		hdma->Init.PeriphDataAlignment = DMA_PDATAALIGN_WORD;
		hdma->Init.MemDataAlignment = DMA_MDATAALIGN_WORD;
		hdma->Init.Mode = DMA_NORMAL;
		hdma->Init.Priority = DMA_PRIORITY_HIGH;
		dma_init(hdma);

		hadc->hdmarx = hdma;
		hdma->localdata = hadc;
	}
	else{
		hadc->hdmarx = NULL;
	}

	/*
	 *  ADCΥץꥹ
	 */
	reg  = pini->ClockPrescaler & RCU_CFG0_ADCPSC;
	sil_modw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_CFG0), RCU_CFG0_ADCPSC, reg);
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_CFG0), RCU_CFG0_ADCPSC_2);

	/*
	 *  ADC
	 */
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), (ADC_CTL0_EOCIE | ADC_CTL0_EOICIE));
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_SM, pini->ScanConvMode);
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_DAL, pini->DataAlign);
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_OVSCR), ADC_OVSCR_DRES, pini->Resolution);

	/*
	 *  ADCȥꥬ
	 */
	if(hadc->Init.ExternalTrigConv != ADC_SOFTWARE_START){
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), \
			(ADC_CTL1_ETERC | ADC_CTL1_ETSIC), hadc->Init.ExternalTrigConv);
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), \
			(ADC_CTL1_ETERC | ADC_CTL1_ETEIC), hadc->Init.ExternalTrigConvEdge);
	}
	else{
		sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), (ADC_CTL1_ETERC | ADC_CTL1_ETEIC));
	}

	/*
	 *  ADC³⡼
	 */
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_CTN, pini->ContinuousConvMode);
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_DISRC, pini->DiscontinuousConvMode);
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_DISNUM, pini->NumDiscConversion << 13);

	/*
	 *  С
	 */
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_RSQ0), ADC_RSQ0_RL, (pini->NumConversion - 1) << 20);

	/*
	 *  ADCͭ
	 */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_ADCON);

	/*
	 *  ADC֤
	 */
	hadc->ErrorCode = ADC_ERROR_NONE;
	hadc->status = ADC_STATUS_READY;
	return hadc;
}

/*
 *  ADCλ
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  return ER
 */
ER
adc_deinit(ADC_Handle_t* hadc)
{
	const ADC_PortControlBlock *apcb;

	if(hadc == NULL)
		return E_PAR;
	if((apcb = (const ADC_PortControlBlock *)hadc->apcb) == NULL)
		return E_OBJ;

	/*
	 *  ADC̵
	 */
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_ADCON);

	/*
	 *  ADCꥻå()
	 */
	sil_orw_mem((uint32_t *)(apcb->adcresetbase), apcb->adcresetbit);
	sil_andw_mem((uint32_t *)(apcb->adcresetbase), apcb->adcresetbit);

	/*
	 *  DMA
	 */
	if(hadc->hdmarx != NULL)
		dma_deinit(hadc->hdmarx);

	/*
	 *  ADCξֽ
	 */
	hadc->ErrorCode = ADC_ERROR_NONE;
	hadc->status = ADC_STATUS_RESET;
	return E_OK;
}

/*
 *  ADC-ߥ⡼ɳϽ
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  return ER
 */
ER
adc_start_int(ADC_Handle_t* hadc)
{
	if(hadc == NULL)
		return E_PAR;

	/*
	 *  ¹Ծ֤˰ܹ
	 */
	hadc->status = ADC_STATUS_BUSY;
	hadc->ErrorCode = ADC_ERROR_NONE;
	hadc->NumCurrentConversionRank = hadc->Init.NumConversion;

	/*
	 *  ȥꥬǤʤХեȥ
	 */
	if((sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1)) & (ADC_CTL1_ETERC | ADC_CTL1_ETEIC)) == 0){
		sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_SWRCST);
	}

	/*
	 *  ADCͭ
	 */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_ADCON);
	sil_dly_nse(ADC_STAB_DELAY_US*1000);

	/*
	 *  ž
	 */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_EOCIE);
	return E_OK;
}

/*
 *  ADC-ߥ⡼߽
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  return ER
 */
ER
adc_end_int(ADC_Handle_t* hadc)
{
	if(hadc == NULL)
		return E_PAR;

	/*
	 *  ž
	 */
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), (ADC_CTL0_EOCIE | ADC_CTL0_EOICIE));

	hadc->status = ADC_STATUS_READY;
	return E_OK;
}

/*
 *  ADC-DMA⡼ɳϽ
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  parameter2  pdata: ΰΥݥ
 *  parameter3  length: ǡĹ
 *  return ER
 */
ER
adc_start_dma(ADC_Handle_t* hadc, uint32_t* pdata, uint32_t length)
{
	if(hadc == NULL || hadc->hdmarx == NULL)
		return E_PAR;

	/*
	 *  ADC DMA
	 */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_DMA);

	/*
	 *  DMAХåؿ
	 */
	hadc->hdmarx->xfercallback = adc_dmatrans_func;
	hadc->hdmarx->xferhalfcallback = adc_dmahalftrans_func;
	hadc->hdmarx->errorcallback = adc_dmaerror_func;

	/*
	 *  ¹Ծ֤˰ܹ
	 */
	hadc->status = ADC_STATUS_BUSY;
	hadc->ErrorCode = ADC_ERROR_NONE;
	hadc->NumCurrentConversionRank = hadc->Init.NumConversion;

	/*
	 *  DMA
	 */
	dma_start(hadc->hdmarx, hadc->base+TOFF_ADC_RDATA, (uint32_t)pdata, length);

	/*
	 *  ȥꥬ꤬ʤСեȥС󥹥Ԥ
	 */
	if((sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1)) & (ADC_CTL1_ETERC | ADC_CTL1_ETEIC)) == 0){
		sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_SWRCST);
	}

	/*
	 *  ADCͭ
	 */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_ADCON);
	sil_dly_nse(ADC_STAB_DELAY_US*1000);

	return E_OK;
}

/*
 *  ADC-DMA⡼߽
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  return ER
 */
ER
adc_end_dma(ADC_Handle_t* hadc)
{
	if(hadc == NULL || hadc->hdmarx == NULL)
		return E_PAR;

	/*
	 *  ADCDMA⡼ɤ򥪥
	 */
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1), ADC_CTL1_DMA);

	/*
	 *  DMA
	 */
	if(hadc->hdmarx != NULL)
		dma_end(hadc->hdmarx);

	hadc->status = ADC_STATUS_READY;
	return E_OK;
}

/*
 *  ADC ߥϥɥ¹Դؿ
 */
void
adc_handler(ADC_Handle_t* hadc)
{
	uint32_t sr  = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT));
	uint32_t ct0 = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0));
	uint32_t ct1 = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL1));

	/*
	 *  쥮顼쥸νλȽ
	 */
	if((sr & ADC_STAT_EOC) != 0 && (ct0 & ADC_CTL0_EOCIE) != 0){
		if((ct1 & ADC_CTL1_CTN) == 0)		/* ơѹ */
			hadc->status = ADC_STATUS_EOC;
		else
			hadc->status = ADC_STATUS_BUSY_EOC;

		if((hadc->Init.ContinuousConvMode == ADC_CONTINUOUS_DISABLE) && ((ct1 & ADC_CTL1_ETERC) == 0)){
			if (hadc->NumCurrentConversionRank != 0){
				hadc->NumCurrentConversionRank--;
			}

			/*
			 *  RANKλȽ
			 */
			if(hadc->NumCurrentConversionRank == 0){
				/*
				 *  žȥ顼ߤ
				 */
				sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_EOCIE);
			}
		}
		/*
		 * 쥮顼žγߥꥢ
		 */
		sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT), ADC_STAT_EOC);

		/*
		 *  λХåؿɤ߽Ф
		 */
		if(hadc->xfercallback != NULL)
			hadc->xfercallback(hadc);
	}

	/*
	 *  󥸥ƥåȥͥνλȽ
	 */
	sr = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT));
	if((sr & ADC_STAT_EOIC) != 0 && (ct0 & ADC_CTL0_EOICIE) != 0){
		if((ct1 & ADC_CTL1_CTN) == 0)		/* ơѹ */
			hadc->status = ADC_STATUS_EOC;
		else
			hadc->status = ADC_STATUS_BUSY_EOC;

		if(((hadc->Init.ContinuousConvMode == ADC_CONTINUOUS_DISABLE) || (ct0 & ADC_CTL0_ICA) == 0) && (ct1 & ADC_CTL1_ETEIC) == 0){
			/*
			 *  󥸥ƥåȥͥž
			 */
			sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_EOICIE);
		}

		/*
		 *  󥸥ƥåȥͥߥꥢ
		 */
		sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT), ADC_STAT_EOIC);
	}

	/*
	 *  ʥåɥåȽ
	 */
	sr = sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT));
	if((sr & ADC_STAT_WDE) != 0 && (ct0 & ADC_CTL0_WDEIE) != 0){
		/*
		 *  åɥåȯ֤ѹ
		 */
		hadc->status = ADC_STATUS_AWD;

		/*
		 *  åɥåߥꥢ
		 */
		sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_STAT), ADC_STAT_WDE);

		/*
		 *  ɥΰ襢ȤΥХåؿɤ߽Ф
		 */
		if(hadc->outofwincallback != NULL)
			hadc->outofwincallback(hadc);
	}
}

/*
 *  ADCߥϥɥ
 */
void
adc_int_handler(void)
{
	uint32_t i;
	ADC_Handle_t* hadc = &AdcHandle[0];

	for(i = 0 ; i < NUM_ADCPORT ; i++, hadc++){
		if(hadc->status > ADC_STATUS_READY)
			adc_handler(hadc);
	}
}


/*
 *  쥮顼쥸μФ
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  return 쥮顼쥸
 */
uint32_t
adc_getvalue(ADC_Handle_t* hadc)
{
	return sil_rew_mem((uint32_t *)(hadc->base+TOFF_ADC_RDATA));
}

/*
 *  ADCͥ
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  parameter2  sConfig: ͥ깽¤ΤؤΥݥ
 *  return ER
 */
ER
adc_setupchannel(ADC_Handle_t* hadc, ADC_ChannelConf_t* sConfig)
{
	GPIO_Init_t GPIO_InitStruct;
	uint32_t shift, clkpin;

	if(hadc == NULL || sConfig == NULL)
		return E_PAR;

	clkpin = (sConfig->GpioBase - TADR_GPIOA_BASE) / 0x400;
	if(clkpin >= 12 || sConfig->GpioPin >= 16)
		return E_PAR;

	/*
	 *  ͥͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN), 1<<clkpin);
	sil_rew_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN));

	GPIO_InitStruct.mode = GPIO_MODE_ANALOG;
	GPIO_InitStruct.pull = GPIO_NOPULL;
	gpio_setup(sConfig->GpioBase, &GPIO_InitStruct, sConfig->GpioPin);

	/*
	 *  ץ󥰻
	 */
	if (sConfig->Channel > ADC_CHANNEL_9){	/* ͥ1018 */
		shift = 3 * (sConfig->Channel - 10);
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_SAMPT0), \
				(ADC_SAMPTX_SPTN << shift), (sConfig->SamplingTime << shift));
	}
	else{ 	/* ͥ09 */
		shift = 3 * sConfig->Channel;
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_SAMPT1), \
				(ADC_SAMPTX_SPTN << shift), (sConfig->SamplingTime << shift));
	}

	/*
	 *  
	 */
	if (sConfig->Rank < 7){	/* 16 */
		shift = 5 * (sConfig->Rank - 1);
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_RSQ2), \
				(ADC_RSQX_RSQN << shift), (sConfig->Channel << shift));
	}
	else if (sConfig->Rank < 13){	/* 712 */
		shift = 5 * (sConfig->Rank - 7);
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_RSQ1), \
				(ADC_RSQX_RSQN << shift), (sConfig->Channel << shift));
	}
	else{	/* 1316 */
		shift = 5 * (sConfig->Rank - 13);
		sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_RSQ0), \
				(ADC_RSQX_RSQN << shift), (sConfig->Channel << shift));
	}

	return E_OK;
}

/*
 *  ADCʥåɥå
 *  parameter1  hadc: ADCϥɥؤΥݥ
 *  parameter2  AnalogWDGConfig: ʥåɥå깽¤ΤؤΥݥ
 *  return ER
 */
ER
adc_setupwatchdog(ADC_Handle_t* hadc, ADC_AnalogWDGConf_t* AnalogWDGConfig)
{
	if(hadc == NULL || AnalogWDGConfig == NULL)
		return E_PAR;

	/*
	 *  
	 */
	sil_modw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_WDEIE, AnalogWDGConfig->ITMode);

	/*
	 *  ⡼
	 */
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), (ADC_CTL0_WDSC | ADC_CTL0_RWDEN | ADC_CTL0_IWDEN));

	/* Set the analog watchdog enable mode */
	sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), AnalogWDGConfig->WatchdogMode);
	sil_wrw_mem((uint32_t *)(hadc->base+TOFF_ADC_WDHT), AnalogWDGConfig->HighThreshold);
  
	/*
	 *  HIGH/LOWΥåۡ
	 */
	sil_wrw_mem((uint32_t *)(hadc->base+TOFF_ADC_WDLT), AnalogWDGConfig->LowThreshold);
	sil_andw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), ADC_CTL0_WDCHSEL);

	/*
	 *  бͥ
	 */
    sil_orw_mem((uint32_t *)(hadc->base+TOFF_ADC_CTL0), AnalogWDGConfig->Channel);
	return E_OK;
}


