/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2020-10-22 16:43:59Z roi $
 */
/*
 *  GD32VF103ѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

static const char monthday[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static void (*alarm_callback)(void);	/* RTC Alam callback */
void (*sw_func)(int arg);
void (*exti_func[NUM_EXTI_FUNC])(intptr_t) = { 0 };
uint_t dipsw_value;


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE_MASK          0x0000000F
#define GPIO_MODE_OUTPUT_AF     0x00000008

#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVENT         0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000

#define AFIO_EXTI_SOURCE_MASK   0x03U	/* AFIO exti source selection mask */  
#define AFIO_EXTI_SOURCE_FIELDS 0x04U	/* select AFIO exti source registers */

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	TADR_GPIOE_BASE				/* index 4 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t cbase, abase, index, ishift;
	uint8_t  mode  = init->mode & 0x0FU;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	if(pin < 8)
		cbase = base+TOFF_GPIO_CTL0;
	else
		cbase = base+TOFF_GPIO_CTL1;
	abase  = TADR_AFIO_BASE + TOFF_AFIO_EXTISS0 + (pin & ~AFIO_EXTI_SOURCE_MASK);
	ishift = (pin & AFIO_EXTI_SOURCE_MASK) * AFIO_EXTI_SOURCE_FIELDS;

	/*
	 *  GPIO⡼
	 */
	if((init->mode & GPIO_MODE_OUTPUT) != 0){
		mode = (init->mode & GPIO_MODE_OUTPUT_AF) | \
			(init->otype & GPIO_OTYPE_OD) | (init->speed & 3);
	}
	else if((init->mode & GPIO_MODE_INPUT) != 0){
		if(init->pull == GPIO_NOPULL)
			mode = GPIO_OTYPE_OD;
		else{
			mode = GPIO_OTYPE_PP;
			if(init->pull == GPIO_PULLUP)
				sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_BOP), (1U << pin));
			else
				sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_BC), (1U << pin));
		}
	}
	sil_modw_mem((uint32_t *)(cbase), (GPIO_MODE_MASK << (4U * (pin & 7))),
			 (mode << (4U * (pin & 7))));

	/*
	 *  EXTI⡼
	 */
	if((init->mode & (GPIO_MODE_IT | GPIO_MODE_EVENT)) != 0){
		sil_modw_mem((uint32_t *)(abase), (0xF << ishift), (index << ishift));
		sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTEN), (1 << pin));
		sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTEN), (1 << pin));
		if((init->mode & RISING_EDGE) != 0)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTEN), (1 << pin));
		if((init->mode & FALLING_EDGE) != 0)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTEN), (1 << pin));
		if((init->mode & GPIO_MODE_IT) != 0)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_INTEN), (1 << pin));
		else
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EVEN), (1 << pin));
	}
	else{	/* ̵ */
		if(((sil_rew_mem((uint32_t *)(abase)) >> ishift) & 0xF) == index){
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_INTEN), (1 << pin));
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EVEN), (1 << pin));
		}
	}
}


/*
 *  DMACؿ
 */
#define NUM_CHDMA       12
#define NUM_CHANNEL     7

#define DMA_TRS_TIMEOUT 2000	/* 2 */

#define DMA_CHXCTL_MASK (DMA_CHXCTL_PWIDTH | DMA_CHXCTL_MWIDTH | DMA_CHXCTL_PRIO | \
                         DMA_CHXCTL_PNAGA  | DMA_CHXCTL_MNAGA  | DMA_CHXCTL_DIR  | \
                         DMA_CHXCTL_CMEN)

#define DMA_CHINTF_RST  (DMA_INTF_GIF | DMA_INTF_FTFIF | \
                         DMA_INTF_HTFIF | DMA_INTF_ERRIF)

#define DMAmux_Channel(h)   (TADR_DMAMUX1_BASE+((h)->chid2)*4)
#define DMAmux_RreqGEN(h)   (TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GEN+((h)->genreqno)*4)

static const uint32_t dma_base[NUM_CHDMA] = {
	TADR_DMA0_CH0_BASE,
	TADR_DMA0_CH1_BASE,
	TADR_DMA0_CH2_BASE,
	TADR_DMA0_CH3_BASE,
	TADR_DMA0_CH4_BASE,
	TADR_DMA0_CH5_BASE,
	TADR_DMA0_CH6_BASE,
	TADR_DMA1_CH0_BASE,
	TADR_DMA1_CH1_BASE,
	TADR_DMA1_CH2_BASE,
	TADR_DMA1_CH3_BASE,
	TADR_DMA1_CH4_BASE
};

static DMA_Handle_t *pDmaHandler[NUM_CHDMA];

/*
 *  CHANNEL DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0, req = 0, i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid  = i % NUM_CHANNEL;
	hdma->chid2 = i;
	pDmaHandler[i] = hdma;
	hdma->base = hdma->cbase & ~0x3FF;

	/* DMA-CHXCTL쥸 */
    tmp  = sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL));
    tmp &= ~DMA_CHXCTL_MASK;
	req  = (hdma->Init.PeriphDataAlignment | hdma->Init.MemDataAlignment | hdma->Init.Priority | \
			hdma->Init.PeriphInc | hdma->Init.MemInc | hdma->Init.Direction | hdma->Init.Mode);
	req &= DMA_CHXCTL_MASK;
	/* DMA⡼ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), (tmp | req));

	/* Memory to Memory⡼ */
	req  = hdma->Init.Direction & DMA_CHXCTL_M2M;
	sil_modw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), DMA_CHXCTL_M2M, req);

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid = i % NUM_CHANNEL;
	pDmaHandler[i] = NULL;

	/* DMA֥͡ */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), DMA_CHXCTL_CHEN);

	/* ƥե饰򥯥ꥢ */
	sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMA_INTC), DMA_CHINTF_RST);

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHAANEL DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), DMA_CHXCTL_CHEN);
	hdma->status = DMA_STATUS_BUSY;

	/* ǡĹ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCNT), DataLength);

	if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
		/* ꤫ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXPADDR), DstAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXMADDR), SrcAddress);
	}
	else{
		/* ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXPADDR), SrcAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXMADDR), DstAddress);
	}

	/* ߥ֥͡ */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), (DMA_CHXCTL_FTFIE | DMA_CHXCTL_HTFIE | DMA_CHXCTL_ERRIE));

	/* DMA */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), DMA_CHXCTL_CHEN);

	return E_OK;
}

/*
 *  CHANNEL DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	if(hdma == NULL)
		return E_PAR;
	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMA_CHXCTL), DMA_CHXCTL_CHEN);
	return E_OK;
}

/*
 *  CHANNEL DMAΥ١ɥ쥹
 */
uintptr_t
dma_channel_base(int32_t chid)
{
	if(chid < 0 || chid >= NUM_CHDMA)
		return 0;
	if(chid < NUM_CHANNEL)
		sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN), RCU_AHBEN_DMA0EN);
	else
		sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN), RCU_AHBEN_DMA1EN);
	sil_rew_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN));
	return (uintptr_t)dma_base[chid];
}

/*
 *  CHANNEL DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	uint32_t index = hdma->chid * 4;
	uint32_t intf  = sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_INTF));

	/*
	 *  ž顼
	 */
	if((intf & (DMA_INTF_ERRIF << index)) != 0){
		/* ž顼ե饰򥯥ꥢ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMA_INTC), (DMA_INTC_ERRIFC << index));

		/* ž顼֤ */
		hdma->ErrorCode |= DMA_ERROR_TE;
		hdma->status = DMA_STATUS_READY_ERROR;
		/* 顼Хå */
		if(hdma->errorcallback != NULL)
			hdma->errorcallback(hdma);
	}

	/*
	 *  ϡžλ
	 */
	if((intf & (DMA_INTF_HTFIF << index)) != 0){

		/* ϡžλե饰򥯥ꥢ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMA_INTC), (DMA_INTC_FTFIFC << index));
		/* ϡžơѹ */
		hdma->status = DMA_STATUS_READY_HMEM0;
		/* ϡžλХå */
		if(hdma->xferhalfcallback != NULL)
			hdma->xferhalfcallback(hdma);
	}

	/*
	 *  žλ
	 */
	if((intf & (DMA_INTF_FTFIF << index)) != 0){

		/* žλե饰򥯥ꥢ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMA_INTC), (DMA_INTC_FTFIFC << index));
			/* ơλ */
		hdma->ErrorCode = DMA_ERROR_NONE;
		hdma->status = DMA_STATUS_READY_MEM0;
		/* žλХå */
		if(hdma->xfercallback != NULL)
			hdma->xfercallback(hdma);
	}

	/*
	 *  Х߽
	 */
	if((intf & (DMA_INTF_FTFIF << index)) != 0){

		/* Хߤ򥯥ꥢ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMA_INTC), (DMA_INTC_GIFC << index));
	}
}

/*
 *  CHANNEL DMA ߥӥ롼
 */
void
channel_dma_isr(intptr_t exinf)
{
	dma_inthandler(pDmaHandler[(uint32_t)exinf]);
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint16_t mcount, mcounter;
static uint8_t  wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	uint32_t base = TADR_WWDGT_BASE;

	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;

	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB1EN), RCU_APB1EN_WWDGTEN);

	hini->Window    &= WWDGT_CFG_WIN;
	hini->Counter   &= WWDGT_CTL_CNT;
	hini->PreScaler &= WWDGT_CFG_PSC;
	wdogcount        = hini->Counter;
	mcount           = hini->MCounter;
	mcounter         = 0;

	/*
	 *  ե졼󥻥å
	 */
	sil_modw_mem((uint32_t *)(base+TOFF_WWDGT_CTL), WWDGT_CTL_CNT, hini->Counter);
	sil_modw_mem((uint32_t *)(base+TOFF_WWDGT_CFG), (WWDGT_CFG_WIN | WWDGT_CFG_PSC), \
		(hini->Window | hini->PreScaler));
	if(hini->MCounter > 0)
		sil_orw_mem((uint32_t *)(base+TOFF_WWDGT_CFG), WWDGT_CFG_EWIE);
	sil_orw_mem((uint32_t *)(base+TOFF_WWDGT_CTL), WWDGT_CTL_WDGTEN);

	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID)
		return E_PAR;
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB1RST), RCU_APB1RST_WWDGTRST);
	sil_andw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB1RST), RCU_APB1RST_WWDGTRST);
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID)
		return E_PAR;

	sil_modw_mem((uint32_t *)(TADR_WWDGT_BASE+TOFF_WWDGT_CTL), WWDGT_CTL_CNT, wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDGT_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDGT_CFG)) & WWDGT_CFG_EWIE) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDGT_STAT)) & WWDGT_STAT_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDGT_STAT), WWDGT_STAT_EWIF);
		if(mcounter++ < mcount)
			sil_modw_mem((uint32_t *)(base+TOFF_WWDGT_CTL), WWDGT_CTL_CNT, wdogcount);
	}
}


/*
 *  RTCɥ饤(ưˤrtc_initɬ)
 */
#define RTC_SOURCE_NONE     0x00000000
#define RTC_SOURCE_LXTAL    RCU_BDCTL_RTCSRC_0
#define RTC_SOURCE_IRC40K   RCU_BDCTL_RTCSRC_1
#define RTC_SOURCE_HXTAL128 (RCU_BDCTL_RTCSRC_1 | RCU_BDCTL_RTCSRC_0);
/*
 *  tm2timesecѴ
 */
static uint32_t
mktimesec(struct tm2 *ptm)
{
	int day = ptm->tm_year * 365;
	int yday, i;

	day += ptm->tm_year / 4;
	for(i = 1, yday = 0 ; i < ptm->tm_mon ; i++)
		yday += monthday[i-1];
	if(ptm->tm_mon > 2 && ((ptm->tm_year+2) % 4) == 0)
		yday++;
	yday += ptm->tm_mday - 1;
	day  += yday;
	ptm->tm_wday = (day+4) % 7;
	ptm->tm_yday = yday+1;
	return day * (24*60*60) + ptm->tm_hour * 60*60 + ptm->tm_min * 60 + ptm->tm_sec;
}

/*
 *  timesectm2ؤѴ
 */
static void
mktime2(struct tm2 *ptm, uint32_t timesec)
{
	int day = timesec / (24*60*60);
	int time, i, yday, mday;

	time = timesec - day * 24 * 60 * 60;
	ptm->tm_wday = (day+4) % 7;
	for(i = 365, ptm->tm_year = 0 ; i < day ; i += 365){
		if(((ptm->tm_year+2) % 4) == 0)
			i++;
		ptm->tm_year++;
	}

	ptm->tm_yday = yday = day - i + 365 + 1;
	for(ptm->tm_mon = 1 ; ptm->tm_mon <= 12 && yday > 0 ; ptm->tm_mon++){
		mday = monthday[ptm->tm_mon-1];
		if(ptm->tm_mon == 2 && ((ptm->tm_year+2) % 4) == 0)
			mday++;
		if(yday < mday){
			ptm->tm_mday = yday;
			break;
		}
		yday -= mday;
	}
	ptm->tm_hour = time / (60*60);
	time -= ptm->tm_hour * 60 * 60;
	ptm->tm_min  = time / 60;
	ptm->tm_sec = time % 60;
}

/*
 *  RTCѽ
 */
void
rtc_init(intptr_t exinf)
{
	/*
	 *  RTCꥻå
	 */
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB1EN), (RCU_APB1EN_BKPIEN | RCU_APB1EN_PMUEN));
	sil_orw_mem((uint32_t *)(TADR_PMU_BASE+TOFF_PMU_CTL), PMU_CTL_BKPWEN);
	sil_dly_nse(10*1000);

	/*
	 *  LXTALͭ, RTCͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_BDCTL), RCU_BDCTL_LXTALEN);
	sil_modw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_BDCTL), RCU_BDCTL_RTCSRC, RTC_SOURCE_LXTAL);
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_BDCTL), RCU_BDCTL_RTCEN);

	/*
	 *  ץꥹ
	 */
	sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_PSCH), (LXTAL_VALUE >> 16));
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_PSCL), (LXTAL_VALUE & RTC_PSCL_PSC));
	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);
}

/*
 *  RTCλؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_set_time(struct tm2 *pt)
{
	uint32_t rtccount = 0;

	if(pt == NULL)
		return E_PAR;

	SVC_PERROR(wai_sem(RTCSEM));

	rtccount = (uint64_t)mktimesec(pt);
	sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CNTH), (rtccount >> 16));
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CNTL), (rtccount & RTC_CNTL_CNT));
	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTCλФؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_get_time(struct tm2 *pt)
{
	uint32_t rtccount = 0;

	if(pt == NULL)
		return E_PAR;

	SVC_PERROR(wai_sem(RTCSEM));

	rtccount  = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CNTH)) << 16;
	rtccount |= sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CNTL));
	mktime2(pt, (uint32_t)rtccount);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC顼
 */
ER
rtc_setalarm(struct tm2 *ptm, void *func)
{
	uint32_t comptime;

	if(ptm == NULL)
		return E_PAR;
	SVC_PERROR(wai_sem(RTCSEM));

	alarm_callback = (void(*)(void))func;
	comptime = mktimesec(ptm);
	sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMH), (comptime >> 16));
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRML), (comptime & RTC_ALRML_ALRM));
	sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_INTEN), RTC_INTEN_ALRMIE);
	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), RTC_CTL_CMF);

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC ALRAMϥɥ
 */
void
rtc_handler(void)
{
	uint32_t intr = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_INTEN));
	uint32_t ctlr = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL));

	ctlr &= RTC_CTL_SCIF | RTC_CTL_ALRMIF | RTC_CTL_OVIF;
	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CTL), ctlr);
	if((intr & RTC_INTEN_ALRMIE) != 0 && (ctlr & RTC_CTL_ALRMIF) != 0){
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_INTEN), RTC_INTEN_ALRMIE);
		if(alarm_callback != NULL)
			alarm_callback();
	}
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    clkbit;
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
	{ RCU_APB2EN_PCEN, TADR_GPIOC_BASE, PINPOSITION13, GPIO_OTYPE_PP},
	{ RCU_APB2EN_PAEN, TADR_GPIOA_BASE, PINPOSITION1,  GPIO_OTYPE_PP},
	{ RCU_APB2EN_PAEN, TADR_GPIOA_BASE, PINPOSITION2,  GPIO_OTYPE_PP}
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))


/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN), pled->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = pled->pinpp;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_OSPEED_50MHZ;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BOP), (1 << pled->pinpos));
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_OCTL)) & (1<<(pled->pinpos))) == 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		/* 񤭹ߥǡƽ񤭹 */
		if((led_data & (1<<i)) != 0)
			sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BC), (1 << pled->pinpos));
		else
			sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BOP), (1 << pled->pinpos));
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


/*
 *  USERSWݡ
 *
 *  USE_USERSWƤPA3桼åݡȤȤ롥
 */

static const gio_conftype sw_confdata[] = {
	{ RCU_APB2EN_PAEN, TADR_GPIOA_BASE, PINPOSITION3, GPIO_NOPULL}
};
#define NUM_SW    (sizeof(sw_confdata)/sizeof(gio_conftype))

static void
sw_int(intptr_t exinf)
{
	if(sw_func != NULL)
		sw_func((int)exinf);
}

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	const gio_conftype *psw = &sw_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_SW ; psw++, i++){
		/* ϥݡȤ */
		/* Enable the BUTTON Clock */
		sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_APB2EN), psw->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode  = GPIO_MODE_IT_FALLING;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull  = psw->pinpp;
		gpio_setup(psw->giobase, &GPIO_Init_Data, psw->pinpos);
		exti_func[psw->pinpos] = sw_int;
	}
}

#ifdef USE_USERSW
/*
 *  PUSHåХåؿ
 */
void setup_sw_func(intptr_t exinf)
{
	sw_func = (void (*)(int))exinf;
}
#endif	/* USE_USESW */

/*
 *  EXTI3ߥϥɥ
 */
void
exti3_handler(void)
{
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PD));
	if(exti_func[3] != NULL)
		exti_func[3]((intptr_t)3);
    sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PD), istatus);
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

