/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: i2c.c 699 2020-10-27 17:18:36Z roi $
 */
/*
 *
 *  GD32VF103 I2Cɥ饤дؿ
 *
 */
#include "kernel.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "i2c.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  I2CݡID֥åФΥޥ
 */
#define INDEX_I2C(i2cid)        ((uint_t)((i2cid) - 1))


#define I2C_TIMEOUT_FLAG            35			/* 35 ms */
#define I2C_TIMEOUT_BUSY_FLAG       10000		/* 10 s  */

#define I2C_7BIT_ADD_WRITE(_ADDR_)  ((uint8_t)((_ADDR_) & (~I2C_SADDR0_ADDRESS0)))
#define I2C_7BIT_ADD_READ(_ADDR_)   ((uint8_t)((_ADDR_) | I2C_SADDR0_ADDRESS0))

#define I2C_10BIT_ADDRESS(_ADDR_)   ((uint8_t)(((_ADDR_) & 0x00FF)))
#define I2C_10BIT_HEADWRITE(_ADDR_) ((uint8_t)(((((_ADDR_) & 0x0300) >> 7) | 0xF0)))
#define I2C_10BIT_HEADREAD(_ADDR_)  ((uint8_t)(((((_ADDR_) & 0x0300) >> 7) | 0xF1)))

#define I2C_MEM_ADD_MSB(_ADDR_)     ((uint8_t)(((_ADDR_) & 0xFF00) >> 8))
#define I2C_MEM_ADD_LSB(_ADDR_)     ((uint8_t)((_ADDR_) & 0x00FF))

/*
 *  I2Cå硢Ǿ
 */
#define I2CCLK_MAX                  ((uint32_t)0x00000048U)
#define I2CCLK_MIN                  ((uint32_t)0x00000002U)

/*
 *  I2C쥸ޥ
 */
#define I2C_CTL_MASK                (I2C_CTL0_GCEN | I2C_CTL0_SS)
#define I2C_ADDRESS_MASK            (I2C_SADDR0_ADDRESS0 | I2C_SADDR0_ADDRESS | I2C_SADDR0_ADDRESS_H)


/*
 *  I2Cݡơ֥
 */
static const I2C_PortControlBlock i2c_pcb[NUM_I2CPORT] = {
  {	TADR_I2C0_BASE,
	(TADR_RCU_BASE+TOFF_RCU_APB2EN),  RCU_APB2EN_PBEN, RCU_APB2EN_AFEN,
	(TADR_RCU_BASE+TOFF_RCU_APB1EN),  RCU_APB1EN_I2C0EN,
	(TADR_RCU_BASE+TOFF_RCU_APB1RST), RCU_APB1RST_I2C0RST,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	6,  7 },

  {	TADR_I2C1_BASE,
	(TADR_RCU_BASE+TOFF_RCU_APB2EN),  RCU_APB2EN_PBEN, RCU_APB2EN_AFEN,
	(TADR_RCU_BASE+TOFF_RCU_APB1EN),  RCU_APB1EN_I2C1EN,
	(TADR_RCU_BASE+TOFF_RCU_APB1RST), RCU_APB1RST_I2C1RST,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	10, 11 }
};


/*
 * I2C ϥɥ
 */
static I2C_Handle_t I2cHandle[NUM_I2CPORT];

/*
 *  STAT0쥸åԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Flag:    Ԥե饰
 *  parameter3 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_sr0flagsetwait(I2C_Handle_t *hi2c, uint32_t Flag, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0)) & Flag) == 0){
		if(ctime == 0 || tick > ctime){
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  BUS READYԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_busreadywait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

    while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT1)) & I2C_STAT1_I2CBSY) != 0){
		if(ctime == 0 || tick > ctime){
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 * ɥ쥹ϥե饰Ԥ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Flag:    Ԥե饰
 *  parameter3 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_addrflagwait(I2C_Handle_t *hi2c, uint32_t Flag, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0)) & Flag) == 0){
		/*
		 *  ΥåեȽ
		 */
		if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0)) & I2C_STAT0_AERR) != 0){
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), ~I2C_STAT0_AERR);
			hi2c->ErrorCode = I2C_ERROR_AF;
			hi2c->status= I2C_STATUS_READY;
			return E_SYS;
		}
		if(ctime == 0 || tick > ctime){
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
		sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  ɥ쥹ե饰ꥢ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 */
static void
i2c_clear_addr(I2C_Handle_t *hi2c)
{
	volatile uint32_t tmp;

	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT1));
	(void)(tmp);
}

/*
 *  I2CžԤ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 Timeout:ॢȻ(ms)
 */
static ER
i2c_transwait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;

	while(hi2c->status != I2C_STATUS_READY){
		if(tick >= Timeout)
			return E_TMOUT;
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		tick++;
	}
	return E_OK;
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterreadaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	/*
	 *  Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);

	/*
	 *  Ԥ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_START);
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_SBSEND, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  ǥХɥ쥹Ԥ
	 */
	if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_7BIT){
		if(MemAddSize == 0)	/* ꥢɥ쥹ꤷʤ硢 */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_7BIT_ADD_READ(DevAddress));
		else
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_7BIT_ADD_WRITE(DevAddress));
	}
	else{	/* ꥢɥ쥹ʤ */
		/*
		 *  10ӥåȥإåԤ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_10BIT_HEADWRITE(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_STAT0_ADD10SEND , Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  10ӥåȥɥ쥹Ԥ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_10BIT_ADDRESS(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_STAT0_ADDSEND, Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  ɥ쥹ե饰ꥢ
		 */
		i2c_clear_addr(hi2c);

 		/*
		 *  ꥹԤ
		 */
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_START);
		if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_SBSEND, Timeout) != E_OK){
			return E_TMOUT;
	    }

	   	/*
		 *  졼֥ɥ쥹⡼ɤ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_10BIT_HEADREAD(DevAddress));
	}

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	if((ercd = i2c_addrflagwait(hi2c, I2C_STAT0_ADDSEND, Timeout)) != E_OK){
		return ercd;
	}

	/*
	 *  ꥢɥ쥹ʤʤｪλ
	 */
	if(MemAddSize == 0)
		return E_OK;

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_TBE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥢɥ쥹Ԥ
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_LSB(MemAddress));
	}
	else{	/* 16ӥåȥ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_MSB(MemAddress));
		if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_TBE, Timeout) != E_OK){
			return E_TMOUT;
		}
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_LSB(MemAddress));
	}

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_TBE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥹԤ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_START);
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_SBSEND, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ǥХɥ쥹⡼ɤ(7ӥåȤΤ)
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_7BIT_ADD_READ(DevAddress));

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	return i2c_addrflagwait(hi2c, I2C_STAT0_ADDSEND, Timeout);
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterwriteaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	/*
	 *  Ԥ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_START);
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_SBSEND, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  ǥХɥ쥹Ԥ
	 */
	if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_7BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_7BIT_ADD_WRITE(DevAddress));
	}
	else{
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_10BIT_HEADWRITE(DevAddress));
		if((ercd = i2c_addrflagwait(hi2c, I2C_STAT0_ADD10SEND, Timeout)) != E_OK){
			return ercd;
		}

		/*
		 *  10ӥåȥɥ쥹Ԥ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_10BIT_ADDRESS(DevAddress));
	}

	/*
	 *  ɥ쥹ե饰Ԥ
	 */
	if((ercd = i2c_addrflagwait(hi2c, I2C_STAT0_ADDSEND, Timeout)) != E_OK){
		return ercd;
	}

	/*
	 *  ɥ쥹ե饰ꥢ
	 */
	i2c_clear_addr(hi2c);

	/*
	 *  ꥢɥ쥹ʤʤｪλ
	 */
	if(MemAddSize == 0)
		return E_OK;

	/*
	 *  TXEե饰ޤԤ
	 */
	if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_TBE, Timeout) != E_OK){
		return E_TMOUT;
	}

	/*
	 *  ꥢɥ쥹Ԥ
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_LSB(MemAddress));
	}
	else{	/* 16ӥåȥ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_MSB(MemAddress));
		if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_TBE, Timeout) != E_OK){
			return E_TMOUT;
		}
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_MEM_ADD_LSB(MemAddress));
	}
	return E_OK;
}


/*
 *  I2CǥХν
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
I2C_Handle_t *
i2c_init(ID port, I2C_Init_t *ii2c)
{
	RCU_ClocksTypeDef RCU_Clock;
	GPIO_Init_t GPIO_Init_Data;
	const I2C_PortControlBlock *ipcb;
	I2C_Handle_t *hi2c;
    uint32_t clkc, freq, risetime, reg;
	uint32_t no;

	if(port < I2C1_PORTID || port > NUM_I2CPORT)
		return NULL;
	if(ii2c == NULL || ii2c->ClockSpeed < 100000U)
		return NULL;

	no = INDEX_I2C(port);
	ipcb = &i2c_pcb[no];
	hi2c = &I2cHandle[no];
	if(hi2c->status != I2C_STATUS_RESET)
		return NULL;
	memcpy(&hi2c->Init, ii2c, sizeof(I2C_Init_t));
	hi2c->base  = ipcb->base;
	hi2c->i2cid = port;

	/*
	 *  SCL/SDAԥ/I2Cå
	 */
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit1);
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit2);
	sil_orw_mem((uint32_t *)ipcb->i2cclockbase, ipcb->i2cclockbit);

	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	GPIO_Init_Data.otype     = GPIO_OTYPE_OD;
	GPIO_Init_Data.speed     = GPIO_OSPEED_50MHZ;
	gpio_setup(ipcb->giobase1, &GPIO_Init_Data, ipcb->sclpin);
	gpio_setup(ipcb->giobase2, &GPIO_Init_Data, ipcb->sdapin);

	hi2c->status = I2C_STATUS_BUSY;

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_I2CEN);

	/*
	 *  I2CΥå
	 */
	RCC_GetClocksFreq(&RCU_Clock);
	freq = (uint32_t) (RCU_Clock.APB1_Frequency / 1000000U);
	if(freq >= I2CCLK_MAX)
		freq = I2CCLK_MAX;
	sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), I2C_CTL1_I2CCLK, freq);

	if(400000U < ii2c->ClockSpeed){
		/*
		 *  SCL饤 1000ns(standard mode)
		 */
		risetime = (uint32_t) ((RCU_Clock.APB1_Frequency / 1000000U) + 1U);
		if(risetime >= I2CCLK_MAX)
			risetime = I2CCLK_MAX;
		else if(risetime <= I2CCLK_MIN)
			risetime = I2CCLK_MIN;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_RT), risetime);
        clkc = (uint32_t) (RCU_Clock.APB1_Frequency / (ii2c->ClockSpeed * 2U));
		if(clkc < 0x04U)	/* CLKCκǾͤ4 */
			clkc = 0x04U;
		clkc &= I2C_CKCFG_CLKC;
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CKCFG), I2C_CKCFG_CLKC, clkc);
	}
	else{
		/*
		 *  SCL饤  300ns(fast mode)
		 */
		risetime = ((freq * 300U) / 1000U) + 1U;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_RT), risetime);
		if(ii2c->DutyCycle == I2C_DUTYCYCLE_2){	/* I2C duty cycle is 2 */
			clkc = (uint32_t) (RCU_Clock.APB1_Frequency / (ii2c->ClockSpeed * 3U));
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CKCFG), I2C_CKCFG_DTCY);
        }
		else{	/* I2C duty cycle is 16/9 */
			clkc = (uint32_t) (RCU_Clock.APB1_Frequency / (ii2c->ClockSpeed * 25U));
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CKCFG), I2C_CKCFG_DTCY);
        }
        if((clkc & I2C_CKCFG_CLKC) == 0)	/* CLKCκǾͤ1 */
            clkc = 0x01U;
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CKCFG), I2C_CKCFG_FAST);
		sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CKCFG), I2C_CKCFG_CLKC, clkc);
    }

	/*
	 *  I2C CTL0
	 */
	reg  = ii2c->GeneralCallMode | ii2c->NoStretchMode /*| ii2c->AckEnable | ii2c->AckPosition*/;
	reg &= I2C_CTL_MASK;
	sil_modw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL_MASK, reg);

	/*
	 *  I2CǥХɥ쥹
	 */
	reg = (ii2c->AddressingMode | ii2c->OwnAddress1) & (I2C_SADDR0_ADDFORMAT | I2C_ADDRESS_MASK);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SADDR0), reg);

	/*
	 *  I2CǥХɥ쥹
	 */
	reg = (ii2c->DualAddressMode | ii2c->OwnAddress2) & (I2C_SADDR1_DUADEN | I2C_SADDR1_ADDRESS2);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SADDR1), reg);

	/*
	 *  I2Cư
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_I2CEN);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_READY;
	return hi2c;
}

/*
 *  I2CǥХ̵
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
ER
i2c_deinit(I2C_Handle_t *hi2c)
{
	const I2C_PortControlBlock *ipcb;

	if(hi2c == NULL){
		return E_PAR;
	}

	hi2c->status = I2C_STATUS_BUSY;
	ipcb = &i2c_pcb[INDEX_I2C(hi2c->i2cid)];

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_I2CEN);

	/*
	 *  I2Cåꥻå
	 */
	sil_orw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);
	sil_andw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_RESET;
	return E_OK;
}

/*
 *  I2C졼ּ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slaveread(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

    /*
	 *  ɥ쥹Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2C졼
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slavewrite(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);

    hi2c->status = I2C_STATUS_BUSY_TX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

    /*
	 *  ɥ쥹Υå
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ꡼
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ɤ߽ФХåեؤΥݥ
 *  parameter6 Size:       ɤ߽Ф
 *  return                 ｪλʤE_OK
 */
ER
i2c_memread(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;

	/* 졼֥ɥ쥹ɤ߽Фɥ쥹 */
	if((ercd = i2c_masterreadaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
        return ercd;
	}

	if(hi2c->XferCount == 1){
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);
		i2c_clear_addr(hi2c);
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
	}
	else if(hi2c->XferCount == 2){
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);
		i2c_clear_addr(hi2c);
	}
	else{
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);
		i2c_clear_addr(hi2c);
	}

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
	ercd = i2c_transwait(hi2c, 100);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ饤
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ߥХåեؤΥݥ
 *  parameter6 Size:       ߥ
 *  return                 ｪλʤE_OK
 */
ER
i2c_memwrite(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return  E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return  E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	/*
	 *  Acknowledge/PEC Positionػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);

	hi2c->status = I2C_STATUS_BUSY_TX;
	hi2c->ErrorCode = I2C_ERROR_NONE;

	hi2c->pBuffPtr = pData;
	hi2c->XferSize = Size;
	hi2c->XferCount = Size;

	/* 졼֥ɥ쥹Ƚߥɥ쥹 */
	if((ercd = i2c_masterwriteaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return ercd;
	}

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_ERRIE | I2C_CTL1_EVIE | I2C_CTL1_BUFIE));
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǥå
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 Trials:     ȥ饤
 *  parameter4 Timeout:    ॢ
 *  return                 ǥʤE_OK
 */
ER
i2c_ready(I2C_Handle_t *hi2c, uint16_t DevAddress, uint32_t Trials, uint32_t Timeout)
{
	volatile uint32_t I2C_Trials = 0;
	uint32_t tmp1 = 0, tmp2 = 0;
	uint32_t tick = 0;

	if(hi2c->status == I2C_STATUS_READY){
		if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT1)) & I2C_STAT1_I2CBSY) != 0)
			return E_OBJ;

		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			wai_sem(hi2c->Init.semlock);

		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);
		hi2c->status    = I2C_STATUS_BUSY;
		hi2c->ErrorCode = I2C_ERROR_NONE;

		do{
			/* START */
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_START);

			/* START BITԤ */
			if(i2c_sr0flagsetwait(hi2c, I2C_STAT0_SBSEND, Timeout) != E_OK){
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_TMOUT;
			}

			/*
			 *  ɥ쥹ե饰ꥢ
			 */
			i2c_clear_addr(hi2c);

			/* 졼֥ɥ쥹 */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), I2C_7BIT_ADD_WRITE(DevAddress));

			/* ɥ쥹ΥåԤ */
			tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0)) & (I2C_STAT0_ADDSEND | I2C_STAT0_AERR);
			tmp2 = hi2c->status;
			while(tmp1 == 0 && tmp2 != I2C_STATUS_TIMEOUT){
				if(tick > Timeout){
					hi2c->status = I2C_STATUS_TIMEOUT;
				}
				tmp1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0)) & (I2C_STAT0_ADDSEND | I2C_STAT0_AERR);
				tmp2 = hi2c->status;
				dly_tsk(1);
				tick++;
			}
			hi2c->status = I2C_STATUS_READY;

			/* ɥ쥹ե饰å */
			if((tmp1 & I2C_STAT0_ADDSEND) != 0){
				/* ɥ쥹ե饰ꥢ */
				i2c_clear_addr(hi2c);

				/* ȥå */
				sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);


				/* ǥԤ */
				if(i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
					/*
					 *  å
					 */
					if(hi2c->Init.semlock != 0)
						sig_sem(hi2c->Init.semlock);
					return E_TMOUT;
				}

				hi2c->status = I2C_STATUS_READY;
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_OK;
			}
			else{	/* Υå */
				/* ɥ쥹ե饰ꥻå */
				i2c_clear_addr(hi2c);

				/* ȥå */
				sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);

				/* ǥԤ */
				if(i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
					/*
					 *  å
					 */
					if(hi2c->Init.semlock != 0)
						sig_sem(hi2c->Init.semlock);
					return E_TMOUT;
				}
			}
		}while(I2C_Trials++ < Trials);

		hi2c->status = I2C_STATUS_READY;
		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
	    return E_TMOUT;
	}
	else
		return E_OBJ;
}

/*
 *  I2CХ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 */
static void i2c_transmit_byte(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), (uint32_t)(*hi2c->pBuffPtr++));
		hi2c->XferCount--;
	}
}

/*
 *  I2CХȼ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 */
static void i2c_receive_byte(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;
	}
}

/*
 *  I2Cޥ(BTC)
 */
static void
i2c_mastertrans_BTC(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount != 0){	/* ǡ */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), (uint32_t)(*hi2c->pBuffPtr++));
		hi2c->XferCount--;
	}
	else{	/* λ */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
	}
}

/*
 *  I2Cޥ(TBE)
 */
static void
i2C_MasterTransmit_TBE(I2C_Handle_t *hi2c)
{
	/*
	 *  Х
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA), (uint32_t)(*hi2c->pBuffPtr++));
	hi2c->XferCount--;

	if(hi2c->XferCount == 0){	/*  */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), I2C_CTL1_BUFIE);
	}
}

/*
 *  I2Cޥ(RBNE)
 */
static void
i2c_masterreceiv_RBNE(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount > 3){	/* Ĥ4ХȰʾ */
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;
	}
	else if(hi2c->XferCount == 2 || hi2c->XferCount == 3){	/* Ĥ2ޤ3Х */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), I2C_CTL1_BUFIE);
	}
	else{	/* λ */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));

		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
}

/*
 *  I2Cޥ(BTC)
 */
static void
i2c_masterreceiv_BTC(I2C_Handle_t *hi2c)
{
	if(hi2c->XferCount == 3){		/* Ĥꥫ3Х */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;
	}
	else if(hi2c->XferCount == 2){	/* Ĥꥫ2Х */
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_STOP);
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;

		/*
		 *  
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_ERRIE));

		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
	else{
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_DATA));
		hi2c->XferCount--;
	}
}

/*
 *  I2C EVߥϥɥ
 */
void
i2c_ev_handler(I2C_Handle_t *hi2c)
{
	uint32_t sr0 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));
	uint32_t sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT1));
	uint32_t ct1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1));
	volatile uint32_t tmp;

	/*
	 *  ޥ⡼
	 */
	if((sr1 & I2C_STAT1_MASTER) != 0){
		if((sr1 & I2C_STAT1_TR) != 0){	/* ⡼ */
			if(((sr0 & I2C_STAT0_TBE) != 0) && ((ct1 & I2C_CTL1_BUFIE) != 0) && ((sr0 & I2C_STAT0_BTC) == 0)){
				i2C_MasterTransmit_TBE(hi2c);
			}
			else if(((sr0 & I2C_STAT0_BTC) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){
				i2c_mastertrans_BTC(hi2c);
			}
		}
		else{							/* ⡼ */
			if(((sr0 & I2C_STAT0_RBNE) != 0) && ((ct1 & I2C_CTL1_BUFIE) != 0) && ((sr0 & I2C_STAT0_BTC) == 0)){
				i2c_masterreceiv_RBNE(hi2c);
			}
 			else if(((sr0 & I2C_STAT0_BTC) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){
				i2c_masterreceiv_BTC(hi2c);
			}
		}
		if(hi2c->status == I2C_STATUS_READY && hi2c->Init.semid != 0)
			isig_sem(hi2c->Init.semid);
	}

	/*
	 *  졼֥⡼
	 */
	else{
		if(((sr0 & I2C_STAT0_ADDSEND) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){	/* ADDRե饰 */
			i2c_clear_addr(hi2c);
		}
		else if(((sr0 & I2C_STAT0_STPDET) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){	/* STOPե饰 */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
			/*
			 * STOPե饰ꥢ
			 */
			tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));
			sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_PECTRANS);
			(void)(tmp);

		    /*
			 *  ɥ쥹Υåػ
			 */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);

			hi2c->status = I2C_STATUS_READY;
			if(hi2c->readcallback != NULL)
				hi2c->readcallback(hi2c);
		}
		else if((sr1 & I2C_STAT1_TR) != 0){	/* 졼 */
			if(((sr0 & I2C_STAT0_TBE) != 0) && ((ct1 & I2C_CTL1_BUFIE) != 0) && ((sr0 & I2C_STAT0_BTC) == 0)){
				i2c_transmit_byte(hi2c);
			}
			else if(((sr0 & I2C_STAT0_BTC) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){
				i2c_transmit_byte(hi2c);
			}
		}
		else{								/* 졼ּ */
			if(((sr0 & I2C_STAT0_RBNE) != 0) && ((ct1 & I2C_CTL1_BUFIE) != 0) && ((sr0 & I2C_STAT0_BTC) == 0)){
				i2c_receive_byte(hi2c);
			}
			else if(((sr0 & I2C_STAT0_BTC) != 0) && ((ct1 & I2C_CTL1_EVIE) != 0)){
				i2c_receive_byte(hi2c);
			}
		}
	}
}

/*
 *  I2C ERߥϥɥ
 */
void
i2c_er_handler(I2C_Handle_t *hi2c)
{
	uint32_t sr0 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));
	uint32_t ct1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL1));

	/*
	 *  I2C Х顼
	 */
	if(((sr0 & I2C_STAT0_BERR) != 0) && ((ct1 & I2C_CTL1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_BERR;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), I2C_STAT0_BERR);
	}
	sr0 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));

	/*
	 *  I2C ӥơ󡦥顼
	 */
	if(((sr0 & I2C_STAT0_LOSTARB) != 0) && ((ct1 & I2C_CTL1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_ARLO;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), I2C_STAT0_LOSTARB);
	}
	sr0 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));

	/*
	 *  I2C Υåե롦顼
	 */
	if(((sr0 & I2C_STAT0_AERR) != 0) && ((ct1 & I2C_CTL1_ERRIE) != 0)){
		uint32_t sr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT1));
		if(((sr1 & I2C_STAT1_MASTER) == 0) && (hi2c->XferCount == 0) && (hi2c->status == I2C_STATUS_BUSY_TX)){
			/* Disable EVT, BUF and ERR interrupt */
			sil_andw_mem((uint32_t *)(hi2c->base+ TOFF_I2C_CTL1), (I2C_CTL1_EVIE | I2C_CTL1_BUFIE | I2C_CTL1_ERRIE));
			/* Clear AF flag */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), I2C_STAT0_AERR);
			/* Disable Acknowledge */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_ACKEN);
			hi2c->status = I2C_STATUS_READY;
			if(hi2c->writecallback != NULL)
				hi2c->writecallback(hi2c);
		}
		else{
			hi2c->ErrorCode |= I2C_ERROR_AF;
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), I2C_STAT0_AERR);
		}
	}
	sr0 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0));

	/*
	 *  I2C С/󡦥顼
	 */
	if(((sr0 & I2C_STAT0_OUERR) != 0) && ((ct1 & I2C_CTL1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_OVR;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_STAT0), I2C_STAT0_OUERR);
	}

	if(hi2c->ErrorCode != I2C_ERROR_NONE){
		hi2c->status = I2C_STATUS_READY;

		/*
		 *  Acknowledge/PEC Positionػ
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CTL0), I2C_CTL0_POAP);

		if(hi2c->errorcallback != NULL)
			hi2c->errorcallback(hi2c);
	}
}


/*
 *  I2C EVߥӥ롼
 */
void
i2c_ev_isr(intptr_t exinf)
{
	i2c_ev_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

/*
 *  I2C ERߥӥ롼
 */
void
i2c_er_isr(intptr_t exinf)
{
	i2c_er_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

