/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: usb_otg.c 2418 2020-11-16 15:58:59Z roi $
 */

/* 
 *  GD32VF103:USB_OTGɥ饤Ф
 */

#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <target_syssvc.h>
#include "device.h"
#include "usb_otg.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define get_usbmode(a)          (sil_rew_mem((uint32_t *)(USBOTG_BASE((a))+TOFF_USBO_GINTF)) & 1)
#define get_int_status(a, b)    (sil_rew_mem((uint32_t *)(USBOTG_BASE((a))+TOFF_USBO_GINTF)) & (b))
#define hc_int_status(a, b, c)  (sil_rew_mem((uint32_t *)(USBHC_BASE((a), (b))+TOFF_USBHC_HCHINTF)) & (c))
#define get_haint_status(a)     (sil_rew_mem((uint32_t *)(USBH_BASE((a))+TOFF_USBH_HACHINT)) & 0xFFFF)
#define clear_host_int(a, b)    sil_wrw_mem((uint32_t *)(USBOTG_BASE((a))+TOFF_USBO_GINTF), (b))
#define clear_dev_int(a, b)		sil_wrw_mem((uint32_t *)(USBOTG_BASE((a))+TOFF_USBO_GINTF), (b))
#define clear_hc_int(a, b, c)   sil_wrw_mem((uint32_t *)(USBHC_BASE((a), (b))+TOFF_USBHC_HCHINTF), (c))
#define clear_outep_int(a, b, c) sil_wrw_mem((uint32_t *)(USBOEP_BASE((a), (b))+TOFF_USBEO_DOEPINTF), (c))
#define clear_inep_int(a, b, c) sil_wrw_mem((uint32_t *)(USBIEP_BASE((a), (b))+TOFF_USBEI_DIEPINTF), (c))
#define get_inep_int_status(a)  ((sil_rew_mem((uint32_t *)(USBD_BASE((a))+TOFF_USBD_DAEPINT)) \
								& sil_rew_mem((uint32_t *)(USBD_BASE((a))+TOFF_USBD_DAEPINTEN))) & USB_OTG_DAEPINT_IEPITB)
#define get_outep_int_status(a) (((sil_rew_mem((uint32_t *)(USBD_BASE((a))+TOFF_USBD_DAEPINT)) \
								& sil_rew_mem((uint32_t *)(USBD_BASE((a))+TOFF_USBD_DAEPINTEN))) & USB_OTG_DAEPINT_OEPITB) >> 16)

#define get_packet_data(s)      ((s[0] & 0xff) | ((s[1] & 0xff)<<8) | ((s[2] & 0xff)<<16) | ((s[3] & 0xff)<<24))

#define MIN(a, b)               (((a) < (b)) ? (a) : (b))

/*
 *  USBݡID֥åФΥޥ
 */
#define INDEX_USB(usbid)        ((uint_t)((usbid) - 1))

#define RESET_TIMEOUT           (1000*100)		/* 1ms */
#define HALT_TIMEOUT            (10*100)		/* 0.01ms */

#ifndef USBD_HS_TRDT_VALUE
 #define USBD_HS_TRDT_VALUE     9U
#endif /* USBD_HS_TRDT_VALUE */
#ifndef USBD_FS_TRDT_VALUE
 #define USBD_FS_TRDT_VALUE     5U
#endif /* USBD_FS_TRDT_VALUE */

/*
 *  ǥեFIFO
 */
#define DEFAULT_FS_GRXFSIZ      0x80
#define DEFAULT_FS_TXFIFO0_SIZ  0x40
#define DEFAULT_FS_TXFIFO1_SIZ  0x40


/*
 *  ѥåȥ
 */
#define USB_OTG_HS_MAX_PACKET_SIZE           512
#define USB_OTG_FS_MAX_PACKET_SIZE           64
#define USB_OTG_MAX_EP0_SIZE                 64

/*
 *  PHYȿ
 */
#define DSTS_ENUMSPD_HS_PHY_30MHZ_OR_60MHZ   (0 << 1)
#define DSTS_ENUMSPD_FS_PHY_30MHZ_OR_60MHZ   (1 << 1)
#define DSTS_ENUMSPD_LS_PHY_6MHZ             (2 << 1)
#define DSTS_ENUMSPD_FS_PHY_48MHZ            (3 << 1)

/*
 * ե졼।󥿡Х
 */
#define DCFG_FRAME_INTERVAL_80               0x00000000
#define DCFG_FRAME_INTERVAL_85               USB_OTG_DCFG_EOPFT_0
#define DCFG_FRAME_INTERVAL_90               USB_OTG_DCFG_EOPFT_1
#define DCFG_FRAME_INTERVAL_95               USB_OTG_DCFG_EOPFT

/*
 *  ѥåȥơ
 */
#define STS_GOUT_NAK                         1
#define STS_DATA_UPDT                        2
#define STS_XFER_COMP                        3
#define STS_SETUP_COMP                       4
#define STS_SETUP_UPDT                       6

/*
 *  PHYԡ
 */
#define HPRT0_PRTSPD_HIGH_SPEED              0
#define HPRT0_PRTSPD_FULL_SPEED              1
#define HPRT0_PRTSPD_LOW_SPEED               2

/*
 *  ۥȥ饹ɥݥȥ
 */
#define HCCHAR_CTRL                          0
#define HCCHAR_ISOC                          1
#define HCCHAR_BULK                          2
#define HCCHAR_INTR                          3


#define GRXSTS_PKTSTS_IN                     2
#define GRXSTS_PKTSTS_IN_XFER_COMP           3
#define GRXSTS_PKTSTS_DATA_TOGGLE_ERR        5
#define GRXSTS_PKTSTS_CH_HALTED              7

#define GHWCFG3_XFER_SIZE_CNTR_WIDTH         0x0000000f
#define GHWCFG3_PACKET_SIZE_CNTR_WIDTH       0x00000070
#define GHWCFG3_I2C                          0x00000100
#define GHWCFG3_DFIFO_DEPTH                  0xffff0000
#define GRXFSIZ_DEPTH                        0x0000ffff

/*
 *  åʬơ֥
 */
static const uint32_t ckdiv_table[] = {
	0x00400000,			/* USBFS prescaler select CK_PLL/1 */
	0x00000000,			/* USBFS prescaler select CK_PLL/1.5 */
	RCU_CFG0_USBFSPSC,	/* USBFS prescaler select CK_PLL/2 */
	0x00800000			/* USBFS prescaler select CK_PLL/2.5 */
};

static USB_OTG_Handle_t Husb[NUM_USBPORT];

/*
 *  ɬפʤ鶯⡼ɤꤹ
 *  ⡼ɤʤtrue֤
 */
static bool_t
usbo_force_mode_if_needed(USB_OTG_Handle_t *husb, bool_t host)
{
	uint32_t otgmode = get_usbmode(husb->base);
	uint32_t reqmode = husb->Init.usb_otg_mode;
	uint32_t gusbcfg;
	uint32_t set, clear;

	if (host && (otgmode == USB_OTG_MODE_HOST || reqmode == USB_OTG_MODE_DEVICE))
		return false;
	else if (!host && (otgmode == USB_OTG_MODE_DEVICE || reqmode == USB_OTG_MODE_HOST))
		return false;

	gusbcfg = sil_rew_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GUSBCS));
	set = host ? USB_OTG_GUSBCS_FHM : USB_OTG_GUSBCS_FDM;
	clear = host ? USB_OTG_GUSBCS_FDM : USB_OTG_GUSBCS_FHM;
	gusbcfg &= ~clear;
	gusbcfg |= set;
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GUSBCS), gusbcfg);

	dly_tsk(25);
	return true;
}

/*
 *  ⡼ɤ򥯥ꥢ
 */
static void
usbo_clear_force_mode(USB_OTG_Handle_t *husb)
{
	uint32_t gusbcfg;

	gusbcfg = sil_rew_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GUSBCS));
	gusbcfg &= ~USB_OTG_GUSBCS_FHM;
	gusbcfg &= ~USB_OTG_GUSBCS_FDM;
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GUSBCS), gusbcfg);

	/*
	 *  ѹˤԤ
	 */
	dly_tsk(25);
}

/*
 *  USB-OTG⡼ɥå
 */
static void
usbo_get_hwparam(USB_OTG_Handle_t *husb)
{
	USB_OTG_HwParam_t *hw = &husb->hw_params;
	uint32_t hwcfg3, grxfsiz;
	unsigned int width;

#ifdef TOFF_USBO_GHWCFG3
	hwcfg3 = sil_rew_mem((uint32_t *)(husb->base+TOFF_USBO_GHWCFG3));
#else
	hwcfg3 = 0x200d1e8;
#endif
	grxfsiz = sil_rew_mem((uint32_t *)(husb->base+TOFF_USBO_GRFLEN));
	syslog_2(LOG_NOTICE, "hwcfg3=%08x grxfsiz=%08x", hwcfg3, grxfsiz);

	hw->en_multiple_tx_fifo = 0;
	width = hwcfg3 & GHWCFG3_XFER_SIZE_CNTR_WIDTH;
	hw->max_transfer_size = (1 << (width + 11)) - 1;
	width = (hwcfg3 & GHWCFG3_PACKET_SIZE_CNTR_WIDTH) >> 4;
	hw->max_packet_count = (1 << (width + 4)) - 1;
	hw->total_fifo_size = (hwcfg3 & GHWCFG3_DFIFO_DEPTH) >> 16;
	hw->host_rx_fifo_size = grxfsiz & GRXFSIZ_DEPTH;

	syslog_1(LOG_NOTICE, "  max_transfer_size=%d", hw->max_transfer_size);
	syslog_1(LOG_NOTICE, "  max_packet_count=%d", hw->max_packet_count);
	syslog_1(LOG_NOTICE, "  en_multiple_tx_fifo=%d", hw->en_multiple_tx_fifo);
	syslog_1(LOG_NOTICE, "  total_fifo_size=%d", hw->total_fifo_size);
	syslog_1(LOG_NOTICE, "  host_rx_fifo_size=%d", hw->host_rx_fifo_size);
}

/*
 *  USB OTG ʥͥ⡼
 */
static ER
usbo_exit_hibernation(USB_OTG_Handle_t *husb, bool_t restore)
{
	/*
	 *  ⡼ȥåץʥ륯ꥢ
	 */
	sil_andw_mem((uint32_t *)(USBD_BASE(husb->base)+TOFF_USBD_DCTL), USB_OTG_DCTL_RWKUP);
	return E_OK;
}

/*
 *  USB_OTGϡɥ
 *  parameter1 portid USBݡID
 *  parameter2 pini   ¤ΤؤΥݥ
 *  return     NULLǤʤUSBϥɥؤΥݥ
 */
USB_OTG_Handle_t *
usbo_init(ID portid, USB_OTG_Init_t *pini)
{
	GPIO_Init_t GPIO_Init_Data;
	USB_OTG_Handle_t *husb;
	uint32_t base;
	uint32_t no, ep;
	int32_t  ckdiv;
	volatile uint32_t tmpreg = 0x00U;

	if(portid < USB1_PORTID || portid > NUM_USBPORT || pini == NULL)
		return NULL;
	if(pini->low_power_enable == 1){
		syslog_0(LOG_ERROR, "usbo_init:not support low_power mode !");
		return NULL;
	}
	ckdiv = (SYS_CLOCK / 24000000) - 2;
	if(ckdiv < 0 || (SYS_CLOCK % 24000000) != 0){
		syslog_0(LOG_ERROR, "usbo_init:not support AHB clock !");
		return NULL;
	}

	no = INDEX_USB(portid);
	husb = &Husb[no];
	memcpy(&husb->Init, pini, sizeof(USB_OTG_Init_t));

	husb->base = TADR_USB_OTGFS_BASE;

	/*
	 *  GPIO/USBݡȥåư
	 */
	sil_orw_mem((uint32_t *)ENABLE_FS_BASE1, ENABLE_FS_VAL1);
#if defined(ENABLE_FS_BASE2)
	sil_orw_mem((uint32_t *)ENABLE_FS_BASE2, ENABLE_FS_VAL2);
#endif

	/*USB DM and DP */
	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
	GPIO_Init_Data.speed     = GPIO_OSPEED_50MHZ;
	gpio_setup(TADR_FS_G1_BASE, &GPIO_Init_Data, FS_DM_PINNO);
	gpio_setup(TADR_FS_G1_BASE, &GPIO_Init_Data, FS_DP_PINNO);

#if defined(FS_ID_PINNO)
	/*USB ID */
	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.pull      = GPIO_PULLUP;
	GPIO_Init_Data.otype     = GPIO_OTYPE_OD;
	GPIO_Init_Data.speed     = GPIO_OSPEED_50MHZ;
	gpio_setup(TADR_FS_G1_BASE, &GPIO_Init_Data, FS_ID_PINNO);
#endif

#if defined(FS_VBUS_PINNO)
	/* Configure VBUS Pin(DEVICE) */
	GPIO_Init_Data.mode      = GPIO_MODE_INPUT;
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
	gpio_setup(TADR_FS_G1_BASE, &GPIO_Init_Data, FS_VBUS_PINNO);
#endif

	/*
	 *  USB FSͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN), RCU_AHBEN_USBFSEN);
	tmpreg = sil_rew_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN));

	/*
	 *  USBå
	 */
	sil_modw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_CFG0), \
		RCU_CFG0_USBFSPSC, ckdiv_table[ckdiv]);

	/*
	 *  FIFO
	 */
	usbo_get_hwparam(husb);
	husb->fifo_mem            = husb->hw_params.total_fifo_size;
	husb->pcd_rx_fifo_sz      = DEFAULT_FS_GRXFSIZ;
	husb->pcd_np_g_tx_fifo_sz = DEFAULT_FS_TXFIFO0_SIZ;
	for(ep = 1 ; ep < MAX_EPS_CHANNELS && ep < pini->dev_endpoints ; ep++)
		husb->pcd_tx_fifo_sz[ep] = DEFAULT_FS_TXFIFO1_SIZ;
	((void)(tmpreg));

	base = husb->base;
	/*
	 *  USB OTGХ߶ػ
	 */
	usbo_disableglobalint(husb);

	/*
	 *  USB OTG
	 */
	usbo_coreinit(husb);

	/*
	 *  USB OTG⡼
	 */
	usbo_setcurrentmode(husb);

	/*
	 *  ꥹPHYå
	 */
	sil_wrw_mem((uint32_t *)(USBPGC_BASE(base)), 0);

	/*
	 *  GINTΥޥ
	 */
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), 0);

	/*
	 *  GINTߥꥢ
	 */
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTF), 0xFFFFFFFF);
	return husb;
}

/*
 *  USB OTGϡɥ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_deinit(USB_OTG_Handle_t *husb)
{
	sil_orw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBRST), RCU_AHBRST_USBFSRST);
	sil_andw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBRST), RCU_AHBRST_USBFSRST);

	sil_andw_mem((uint32_t *)(TADR_RCU_BASE+TOFF_RCU_AHBEN), RCU_AHBEN_USBFSEN);
	return E_OK;
}

/*
 *  USB OTG
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_setupint(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t intmsk = USB_OTG_GINTEN_WKUPIE;

	/*
	 *  ߥڥǥ󥰥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GOTGINTF), 0xFFFFFFFF);
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTF), 0xBFFFFFFF);

	/*
	 *  ߶ػ
	 */
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), 0);

	/*
	 *  DMAߵ
	 */
	if(husb->Init.dma_enable == 0){
		intmsk |= USB_OTG_GINTEN_RXFNEIE;
	}

	/*
	 *  USB HOSTߵ
	 */
	if(husb->Init.usb_otg_mode != USB_OTG_MODE_DEVICE){
		intmsk |= (USB_OTG_GINTEN_HPIE  | USB_OTG_GINTEN_HCIE |\
				   USB_OTG_GINTEN_SOFIE | USB_OTG_GINTEN_DISCIE);
	}
	/*
	 *  USB DEVICEߵ
	 */
	if(husb->Init.usb_otg_mode != USB_OTG_MODE_HOST){
		intmsk |= (USB_OTG_GINTEN_SPIE    | USB_OTG_GINTEN_RSTIE |\
                   USB_OTG_GINTEN_ENUMFIE | USB_OTG_GINTEN_IEPIE |\
                   USB_OTG_GINTEN_OEPIE  );
	}

	/*
	 *  SOFߵ
	 */
	if(husb->Init.sof_enable){
		intmsk |= USB_OTG_GINTEN_SOFIE;
	}

	/*
	 *  SENSE ENABLEߵ
	 */
	if(husb->Init.vbus_sensing_enable == 1){
		intmsk |= (USB_OTG_GINTEN_SESIE | USB_OTG_GINTEN_OTGIE);
	}
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), intmsk);
	return E_OK;
}

/*
 *  USBꥻå
 *  parameter1 husb  USBå١ɥ쥹
 *  return     ｪλE_OK ॢȡE_TMOUT
 */
static ER
USB_CoreReset(uint32_t  base)
{
	uint32_t tick = 0;

	/*
	 *  եȥꥻå׵
	 */
	sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL), USB_OTG_GRSTCTL_CSRST);
	do{
		if(++tick > RESET_TIMEOUT)
			return E_TMOUT;
		sil_dly_nse(10*1000);
	}
	while ((sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL)) & USB_OTG_GRSTCTL_CSRST) != 0);
	return E_OK;
}

/*
 *  ѥåȥǡ񤭹
 *  parameter1 base  USB OTG쥸١ɥ쥹
 *  parameter2 src   񤭹ߥǡΥݥ
 *  parameter3 ch_ep_num ɥݥȡޤϡֹͥ
 *  parameter4 len   ǡĹ
 *  parameter5 dma   DMA⡼(0:none-DMA1:DMA)
 */
static void
usbo_writePacket(uint32_t base, uint8_t *src, uint8_t ch_ep_num, uint16_t len, uint8_t dma)
{
	uint32_t count32b = 0 , i = 0;

	if(dma == 0U){
		count32b =  (len + 3) / 4;
		for(i = 0 ; i < count32b; i++, src += 4){
			sil_wrw_mem((uint32_t *)(USBD_FIFO(base, ch_ep_num)), get_packet_data(src));
		}
	}
}

/*
 *  ѥåȥǡɤ߹
 *  parameter1 base USB OTG쥸١ɥ쥹
 *  parameter2 dest ɤ߹ߥǡΥݥ
 *  parameter3 len  ǡĹ
 */
static void *
usbo_readPacket(uint32_t base, uint8_t *dest, uint16_t len)
{
	uint32_t i=0U;
	uint32_t count32b = (len + 3U) / 4U;
	uint32_t data;

	for(i = 0 ; i < count32b; i++, dest += 4){
		data = sil_rew_mem((uint32_t *)(USBD_FIFO(base, 0)));
		dest[0] = (data & 0xff);
		dest[1] = (data >> 8) & 0xff;
		dest[2] = (data >> 16) & 0xff;
		dest[3] = (data >> 24) & 0xff;
	}
	return ((void *)dest);
}

/*
 *  USB HOSTΥȥե졼ֹФ
 *  parameter1 USBϥɥؤΥݥ
 *  return     ϥɥֹ
*/
uint32_t
usbo_getcurrentframe(USB_OTG_Handle_t *husb)
{
	return (sil_rew_mem((uint32_t *)(USBH_BASE(husb->base)+TOFF_USBH_HFINFR)) & USB_OTG_HFINFR_FRNUM);
}

/*
 *  USB OTGεǽԤ
 *  parameter1 husb  USBϥɥؤΥݥ
 *     usb_otg_modeꡧUSB_OTG_MODE_DEVICE(USB DEVICE)
 *                         USB_OTG_MODE_HOST(USB HOST)
 *                         USB_OTG_MODE_DRD(ǥ奢⡼)
 *  return     E_OK
 */
ER
usbo_setcurrentmode(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t gusbcfg = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS));

	gusbcfg &= ~(USB_OTG_GUSBCS_FHM | USB_OTG_GUSBCS_FDM);

	if(husb->Init.usb_otg_mode == USB_OTG_MODE_HOST){
		gusbcfg |= USB_OTG_GUSBCS_FHM;
	}
	else if(husb->Init.usb_otg_mode == USB_OTG_MODE_DEVICE){
		gusbcfg |= USB_OTG_GUSBCS_FDM;
	}
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), gusbcfg);
	dly_tsk(50);
	return E_OK;
}

/*
 *  USBν
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_coreinit(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t reg = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GCCFG));

	if(husb->Init.phy_itface == USB_PHY_ULPI){	/* ULPI-PHY select */
		sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), USB_OTG_GUSBCS_EMBPHY);
		sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GCCFG), USB_OTG_GCCFG_SOFOEN);

		/*
		 *  VBUS
		 */
		sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), (USB_OTG_GUSBCS_ULPIEVD | USB_OTG_GUSBCS_ULPIEOI));
		if(husb->Init.use_external_vbus == 1)
			sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), USB_OTG_GUSBCS_ULPIEVD);
		else
			sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), USB_OTG_GUSBCS_ULPIEVD);
		/*
		 *  ꥻå
		 */
		USB_CoreReset(base);
	}
	else{	/* Embedded PHY select */
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), USB_OTG_GUSBCS_EMBPHY);
		/*
		 *  ꥻå
		 */
		USB_CoreReset(base);

		reg |= USB_OTG_GCCFG_PWRON | USB_OTG_GCCFG_VBUSACEN | USB_OTG_GCCFG_VBUSBCEN;
		if(husb->Init.vbus_sensing_enable == 1)
			reg |= USB_OTG_GCCFG_VBUSIG;
		else
			reg &= ~USB_OTG_GCCFG_VBUSIG;
		if(husb->Init.sof_enable == 1)
			reg |= USB_OTG_GCCFG_SOFOEN;
		else
			reg &= ~USB_OTG_GCCFG_SOFOEN;

		sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GCCFG), reg);
		/*
		 *  Ÿ
		 */
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GCCFG), USB_OTG_GCCFG_PWRON);
	}

	/*
	 *  DMA⡼
	 */
	if(husb->Init.dma_enable == 1){
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GAHBCS), USB_OTG_GAHBCS_DMAEN);
		sil_modw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GAHBCS), \
			USB_OTG_GAHBCS_BURST, (USB_OTG_GAHBCS_BURST_0 | USB_OTG_GAHBCS_BURST_2));
	}
	return E_OK;
}

/*
 *  USBХߵ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_enableglobalint(USB_OTG_Handle_t *husb)
{
	sil_orw_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GAHBCS), USB_OTG_GAHBCS_GINTEN);
	return E_OK;
}

/*
 *  USBХ߶ػ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_disableglobalint(USB_OTG_Handle_t *husb)
{
	sil_andw_mem((uint32_t *)(USBOTG_BASE(husb->base)+TOFF_USBO_GAHBCS), USB_OTG_GAHBCS_GINTEN);
	return E_OK;
}

/*
 *  FIFOΥեå
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     ｪλE_OKॢȡE_TMOUT
 */
ER
usbo_flushTxFifo(USB_OTG_Handle_t *husb, uint32_t num)
{
	uint32_t base = husb->base;
	uint32_t tick = 0;

	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL), (USB_OTG_GRSTCTL_TXFF |(uint32_t)(num << 6)));
	do{
		if(++tick > RESET_TIMEOUT)
			return E_TMOUT;
		sil_dly_nse(10*1000);
	}while ((sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL)) & USB_OTG_GRSTCTL_TXFF) != 0);
	sil_dly_nse(3*1000);
	return E_OK;
}

/*
 *  FIFOΥեå
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     ｪλE_OKॢȡE_TMOUT
 */
ER
usbo_flushRxFifo(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t tick = 0;

	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL), USB_OTG_GRSTCTL_RXFF);
	do{
		if(++tick > RESET_TIMEOUT)
			return E_TMOUT;
		sil_dly_nse(10*1000);
	}while((sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTCTL)) & USB_OTG_GRSTCTL_RXFF) != 0);
	sil_dly_nse(3*1000);
	return E_OK;
}

/*
 *  FIFOν
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_initFiFo(USB_OTG_Handle_t *husb)
{
	unsigned int ep;
	unsigned int addr;
	uint32_t base = husb->base;
	uint32_t val;

	/*
	 *  RX/NPTX FIFO
	 */
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRFLEN), husb->pcd_rx_fifo_sz);
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_DIEP0TFLEN_HNPTFLEN),
			(husb->pcd_np_g_tx_fifo_sz<<16) | husb->pcd_rx_fifo_sz);

	/*
	 *  ٤ƤTX FIFO
	 */
	addr = husb->pcd_rx_fifo_sz + husb->pcd_np_g_tx_fifo_sz;
	for(ep = 1; ep < MAX_EPS_CHANNELS && ep < husb->Init.dev_endpoints ; ep++) {
		if(husb->pcd_tx_fifo_sz[ep] == 0){
			sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_DIEPTFLEN+(ep-1)*4), 0);
			continue;
		}
		val = addr | husb->pcd_tx_fifo_sz[ep] << 16;
		if((addr + husb->pcd_tx_fifo_sz[ep]) > husb->fifo_mem)
			syslog_1(LOG_ERROR, "insufficient fifo memory(%d)", ep);
		addr += husb->pcd_tx_fifo_sz[ep];
		sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_DIEPTFLEN+(ep-1)*4), val);
	}

	/*
	 * FIFOեå
	 */
	usbo_flushTxFifo(husb, 0x10);
	usbo_flushRxFifo(husb);
	return E_OK;
}

/*******************************************************************************
                       USB HOSTؿ
*******************************************************************************/
/*
 *  USB HOST
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_hostinit(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	bool_t forced;
	uint32_t val, i;

	/* Disable the FS/LS support mode only */
	if((husb->Init.speed == USB_SPEED_FULL) && (base != TADR_USB_OTGFS_BASE)){
		sil_orw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HCTL), USB_OTG_HCTL_SPDFSLS);
	}
	else{
		sil_andw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HCTL), USB_OTG_HCTL_SPDFSLS);
	}

	/*
	 *  HOST RX-FIFO
	 */
	forced = usbo_force_mode_if_needed(husb, true);
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRFLEN), husb->pcd_np_g_tx_fifo_sz);
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_DIEP0TFLEN_HNPTFLEN),
			(husb->pcd_np_g_tx_fifo_sz<<16) | husb->pcd_rx_fifo_sz);
	val = (husb->pcd_rx_fifo_sz + husb->pcd_np_g_tx_fifo_sz) | husb->pcd_tx_fifo_sz[1] << 16;
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_HPTFLEN), val);

	/*
	 *  FIFOեå
	 */
	usbo_flushTxFifo(husb, 0x10U); /* all Tx FIFOs */
	usbo_flushRxFifo(husb);
	if(forced)
		usbo_clear_force_mode(husb);

	/*
	 * ͥߥꥢȥޥ
	 */
	for (i = 0 ; i < husb->Init.host_channels; i++){
		sil_wrw_mem((uint32_t *)(USBHC_BASE(base, i)+TOFF_USBHC_HCHINTF), 0xFFFFFFFFU);
		sil_wrw_mem((uint32_t *)(USBHC_BASE(base, i)+TOFF_USBHC_HCHINTEN), 0U);
	}

	/*
	 *  L3ơ
	 */
	husb->lx_state = DWC2_L0;

	/*
	 *  HOST VBUS
	 */
	usbo_drivevbus(husb, 1);
	dly_tsk(200U);
	usbo_setupint(husb);
	return E_OK;
}

/*
 *  HOST׵
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_starthost(USB_OTG_Handle_t *husb)
{
	if(husb->Init.usb_otg_mode == USB_OTG_MODE_DRD
			&& husb->lx_state != DWC2_L0){
		usbo_hostinit(husb);
		dly_tsk(25);
		usbo_devdisconnect(husb);
		dly_tsk(3);
		usbo_devconnect(husb);
	}
	husb->lx_state = DWC2_L0;
	return E_OK;
}

/*
 *  HOST׵
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_stophost(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t tick = 0;
	int      i;

	usbo_disableglobalint(husb);

	/*
	 *  FIFOեå
	 */
	if(husb->Init.usb_otg_mode == USB_OTG_MODE_HOST){
		usbo_flushTxFifo(husb, 0x10U);
		usbo_flushRxFifo(husb);
	}

	/*
	 *  HOSTͥ륭塼ꥢ
	 */
	for(i = 0 ; i <= 15 ; i++){
		sil_modw_mem((uint32_t *)(USBHC_BASE(base, i)+TOFF_USBHC_HCHCTL), (USB_OTG_HCHCTL_CEN|USB_OTG_HCHCTL_EPDIR), USB_OTG_HCHCTL_CDIS);
	}

	/*
	 *  HOSTͥHALT
	 */
	for(i = 0, tick = 0 ; i <= 15 ; i++){
		sil_modw_mem((uint32_t *)(USBHC_BASE(base, i)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_EPDIR, (USB_OTG_HCHCTL_CDIS|USB_OTG_HCHCTL_CEN));
		do{
			if(++tick > HALT_TIMEOUT)
				break;
			sil_dly_nse(10*1000);
		}while((sil_rew_mem((uint32_t *)(USBHC_BASE(base, i)+TOFF_USBHC_HCHCTL)) & USB_OTG_HCHCTL_CEN) != 0);
	}

	/*
	 *  USBۥѤγߥꥢ
	 */
	sil_wrw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HACHINT), 0xFFFFFFFFU);
	sil_wrw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTF), 0xFFFFFFFFU);
	husb->lx_state = DWC2_L3;
	return E_OK;
}

/*
 *  HOSTͥν
 *  parameter1 husb   USBϥɥؤΥݥ
 *  parameter2 ch_num ֹͥ(1-15)
 *  parameter3 epnum  ɥݥֹ(1-15)
 *  return     E_OK
 */
ER
usbo_hc_init(USB_OTG_Handle_t *husb, uint8_t ch_num, uint8_t epnum)
{
	uint32_t base = husb->base;
	uint8_t  address = husb->hc[ch_num].dev_addr;
	uint8_t  speed   = husb->hc[ch_num].speed;
	uint8_t  ep_type = husb->hc[ch_num].ep_type;
	uint16_t mps     = husb->hc[ch_num].max_packet;
	uint32_t hcint   = 0;

	/*
	 *  ͥγߥꥢ
	 */
	sil_wrw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHINTF), 0xFFFFFFFFU);

	/*
	 *  ɥݥȷ˽
	 */
	switch (ep_type) {
	case EP_TYPE_CTRL:
	case EP_TYPE_BULK:
		hcint = USB_OTG_HCHINTEN_TFIE    | USB_OTG_HCHINTEN_STALLIE |\
                USB_OTG_HCHINTEN_USBERIE | USB_OTG_HCHINTEN_DTERIE  |\
                USB_OTG_HCHINTEN_DMAERIE | USB_OTG_HCHINTEN_NAKIE;

		if(epnum & 0x80U){
			hcint |= USB_OTG_HCHINTEN_BBERIE;
		}
		else{
			if(base != TADR_USB_OTGFS_BASE){
				hcint |= USB_OTG_HCHINTEN_NYETIE | USB_OTG_HCHINTEN_ACKIE;
			}
		}
		break;

	case EP_TYPE_INTR:
		hcint = USB_OTG_HCHINTEN_TFIE    | USB_OTG_HCHINTEN_STALLIE |\
                USB_OTG_HCHINTEN_USBERIE | USB_OTG_HCHINTEN_DTERIE  |\
                USB_OTG_HCHINTEN_NAKIE   | USB_OTG_HCHINTEN_DMAERIE | USB_OTG_HCHINTEN_REQOVRIE;

		if(epnum & 0x80)
			hcint |= USB_OTG_HCHINTEN_BBERIE;
		break;
	case EP_TYPE_ISOC:
			hcint = USB_OTG_HCHINTEN_TFIE    | USB_OTG_HCHINTEN_ACKIE |\
                    USB_OTG_HCHINTEN_DMAERIE | USB_OTG_HCHINTEN_REQOVRIE;

		if(epnum & 0x80)
			hcint |= USB_OTG_HCHINTEN_USBERIE | USB_OTG_HCHINTEN_BBERIE;
		break;
	}
	sil_wrw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHINTEN), hcint);

	/*
	 *  HOSTͥߥޥ
	 */
	sil_orw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HACHINTEN), (1 << ch_num));

	/*
	 *  HOSTͥߵ
	 */
	sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_HCIE);

	/*
	 *  HCCHAR쥸
	 */
	sil_wrw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), (((address << 22U) & USB_OTG_HCHCTL_DAR) |\
                             (((epnum & 0x7FU)<< 11U) & USB_OTG_HCHCTL_EPNUM)|\
                             ((((epnum & 0x80U) == 0x80U)<< 15U) & USB_OTG_HCHCTL_EPDIR)|\
                             (((speed == USB_SPEED_LOW)<< 17U) & USB_OTG_HCHCTL_LSD)|\
                             ((ep_type << 18U) & USB_OTG_HCHCTL_EPTYPE)|\
                             (mps & USB_OTG_HCHCTL_MPL)));

	if(ep_type == EP_TYPE_INTR){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_ODDFRM);
	}
	return E_OK; 
}

/*
 *  HOSTԡɤФ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     speed USB_SPEED_HIGH(High speed)
 *                   USB_SPEED_FULL(Full speed)
 *                   USB_SPEED_LOW(Low speed)
 */
uint32_t
usbo_gethostspeed(USB_OTG_Handle_t *husb)
{
	volatile uint32_t hprt0;

	hprt0 = sil_rew_mem((uint32_t*)(USBPRT_BASE(husb->base)));
	return ((hprt0 & USB_OTG_HPCS_PS) >> 17);
}

/*
 *  HOSTͥ׵
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ch_num ֹͥ(1-15)
 *  return     E_OK
 */
ER
usbo_hc_startxfer(USB_OTG_Handle_t *husb, uint8_t ch_num)
{
	USB_OTG_HCTypeDef *hc = &husb->hc[ch_num];
	uint32_t base = husb->base;
	uint8_t  is_oddframe = 0U; 
	uint16_t len_words = 0U;   
	uint16_t num_packets = 0U;
	uint16_t max_hc_pkt_count = 256U;

	if((base != TADR_USB_OTGFS_BASE) && (hc->speed == USB_SPEED_HIGH)){
		if((husb->Init.dma_enable == 0) && (hc->do_ping == 1)){
			/*
			 * Do Ping
			 */
			sil_wrw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHLEN), ((1 << 19) & USB_OTG_HCHLEN_PCNT) | USB_OTG_HCHLEN_PING);
			sil_modw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);
			return E_OK;
		}
		else if(husb->Init.dma_enable == 1){
			sil_andw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHINTEN), (USB_OTG_HCHINTEN_NYETIE | USB_OTG_HCHINTEN_ACKIE));
			hc->do_ping = 0U;
		}
	}

	/*
	 *  ѥåȿžĹη׻
	 */
	if(hc->xfer_len > 0){
		num_packets = (hc->xfer_len + hc->max_packet - 1U) / hc->max_packet;

		if(num_packets > max_hc_pkt_count){
			num_packets = max_hc_pkt_count;
			hc->xfer_len = num_packets * hc->max_packet;
		}
	}
	else{
		num_packets = 1;
	}
	if(hc->ep_is_in){
		hc->xfer_len = num_packets * hc->max_packet;
	}

	/*
	 *  HCTSIZ쥸
	 */
	sil_wrw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHLEN), (((hc->xfer_len) & USB_OTG_HCHLEN_TLEN)) |\
		((num_packets << 19U) & USB_OTG_HCHLEN_PCNT) | (((hc->data_pid) << 29U) & USB_OTG_HCHLEN_DPID));

	if(husb->Init.dma_enable){
		sil_wrw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHDMAADDR), (uint32_t)hc->xfer_buff);
	}

	/*
	 *  EVEN/ODDѥå
	 */
	is_oddframe = (sil_rew_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HFINFR)) & 0x01U) ? 0U : 1U;
	sil_andw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_ODDFRM);
	sil_orw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHCTL), (is_oddframe << 29U));

	/*
	 *  HOSTͥͭ
	 */
	sil_modw_mem((uint32_t *)(USBHC_BASE(base, hc->ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);

	if(husb->Init.dma_enable == 0){	/* none DMA⡼ */
		if((hc->ep_is_in == 0U) && (hc->xfer_len > 0U)){
			switch(hc->ep_type) {
			case EP_TYPE_CTRL:	/* NON PERIODICž */
			case EP_TYPE_BULK:
				len_words = (hc->xfer_len + 3) / 4;
				/*
				 *  FIFOĹĶ硢NPTXFEMPTYߤ
				 */
				if(len_words > (sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_HNPTFQSTAT)) & 0xFFFFU)){
					sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_NPTXFEIE);
				}
				break;
			case EP_TYPE_INTR:	/* PERIODIC ž */
			case EP_TYPE_ISOC:
				len_words = (hc->xfer_len + 3) / 4;
				/*
				 *  FIFOĹĶ硢PTXFEMPTYߤ
				 */
				if(len_words > (sil_rew_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HPTFQSTAT)) & 0xFFFFU)){ /* split the transfer */
					sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_PTXFEIE);
				}
				break;
			default:
				break;
			}

			/*
			 *  ѥåȽ񤭹
			 */
			usbo_writePacket(base, hc->xfer_buff, hc->ch_num, hc->xfer_len, 0);
		}
	}
	return E_OK;
}

/*
 *  HOSTͥHALT
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ֹͥ(1-15)
 *  return     E_OK
 */
ER
usbo_hc_halt(USB_OTG_Handle_t *husb, uint8_t ch_num)
{
	uint32_t base = husb->base;
	uint32_t tick = 0;

	/* HCCHARȥ⡼ */
	if((((sil_rew_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL)) & USB_OTG_HCHCTL_EPTYPE) >> 18) == HCCHAR_CTRL)
			|| ((((sil_rew_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL)) & USB_OTG_HCHCTL_EPTYPE) >> 18) == HCCHAR_BULK))){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS);
		if((sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_HNPTFQSTAT)) & 0xFF0000U) == 0){
			sil_andw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
			do{
				if(++tick > HALT_TIMEOUT)
					break;
				sil_dly_nse(10*1000);
			}while((sil_rew_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL)) & USB_OTG_HCHCTL_CEN) != 0);
		}
		else{
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
		}
	}
	else{	/* ¾ */
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS);
		if((sil_rew_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HPTFQSTAT)) & 0xFFFF) == 0){
			sil_andw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
			do{
				if(++tick > HALT_TIMEOUT) 
					break;
				sil_dly_nse(10*1000);
			}while((sil_rew_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL)) & USB_OTG_HCHCTL_CEN) != 0);
		}
		else{
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, ch_num)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CEN);
		}
	}
	return E_OK;
}

/*
 *  USB OTG HOSTݡȥꥻå
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_resetport(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	volatile uint32_t hprt0;

	hprt0 = sil_rew_mem((uint32_t*)(USBPRT_BASE(base)));
	hprt0 &= ~(USB_OTG_HPCS_PE | USB_OTG_HPCS_PCD | USB_OTG_HPCS_PEDC);

	sil_wrw_mem((uint32_t*)(USBPRT_BASE(base)), (USB_OTG_HPCS_PRST | hprt0));
	dly_tsk (10);				/* See Note #1 */
	sil_wrw_mem((uint32_t*)(USBPRT_BASE(base)), (~(USB_OTG_HPCS_PRST) & hprt0));
	return E_OK;
}

/*
 *  USB HOST VBUS̵ͭ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 state 1:VBUS-ACTIVE 0:VBUS-INACTIVE
 *  return     E_OK
 */
ER
usbo_drivevbus(USB_OTG_Handle_t *husb, uint8_t state)
{
	uint32_t base = husb->base;
	volatile uint32_t hprt0;

	hprt0 = sil_rew_mem((uint32_t*)(USBPRT_BASE(base)));
	hprt0 &= ~(USB_OTG_HPCS_PE | USB_OTG_HPCS_PCD | USB_OTG_HPCS_PEDC);

	if(((hprt0 & USB_OTG_HPCS_PP) == 0) && (state == 1)){
		sil_wrw_mem((uint32_t*)(USBPRT_BASE(base)), (USB_OTG_HPCS_PP | hprt0));
	}
	if(((hprt0 & USB_OTG_HPCS_PP) != 0) && (state == 0)){
		sil_wrw_mem((uint32_t*)(USBPRT_BASE(base)), ((~USB_OTG_HPCS_PP) & hprt0));
	}
	return E_OK; 
}

/*
 *  HOST EXTERNAL VBUS̵ͭ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 state 0:VBUS-ACTIVE 1:VBUS-INACTIVE
 */
void
usbo_driveextvbus(USB_OTG_Handle_t *husb, uint8_t state)
{
#if defined(FS_EVBUS_PINNO)
 	if(state == 0)
		sil_wrw_mem((uint32_t *)(TADR_EVBUS_BASE+TOFF_GPIO_BC), (1<<(FS_EVBUS_PINNO)));
	else
		sil_wrw_mem((uint32_t *)(TADR_EVBUS_BASE+TOFF_GPIO_BOP), (1<<(FS_EVBUS_PINNO)));
#endif
}

/*
 *  FSLSPClkSel
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 freq åȿ(48MHz,6MHzΤͭ)
 */
static void
usbo_initFSLSPClkSel(USB_OTG_Handle_t *husb, uint32_t freq)
{
	uint32_t base = husb->base;
	uint32_t hcfg = sil_rew_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HCTL));
	hcfg &= ~USB_OTG_HCTL_CLKSEL;
	if(freq == 48000)
		hcfg |= USB_OTG_HCTL_CLKSEL_0;
	else if(freq == 6000)
		hcfg |= USB_OTG_HCTL_CLKSEL_1;
	sil_wrw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HCTL), hcfg);

	if(freq == 48000 || freq ==  6000){
		sil_wrw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HFT), freq);
	}
}

/*
 *  HOSTͥ׵
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 chnum ֹͥ(1-15)
 */
static void
usbo_hc_in_irqhandler(USB_OTG_Handle_t *husb, uint8_t chnum)
{
    USB_OTG_HCTypeDef *hc = &husb->hc[chnum];
	uint32_t base = husb->base;

	if(hc_int_status(base, chnum, USB_OTG_HCHINTF_DMAER) != 0){
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_DMAER);
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_ACK) != 0){
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_ACK);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_STALL) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		hc->state = USBO_HC_STALL;
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_STALL);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_DTER) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
		hc->state = USBO_HC_DATATGLERR;
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_DTER);
	}

	if(hc_int_status(base, chnum, USB_OTG_HCHINTF_REQOVR) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_REQOVR);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_TF) != 0){
		if (husb->Init.dma_enable){
			hc->xfer_count = hc->xfer_len - (sil_rew_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHLEN)) & USB_OTG_HCHLEN_TLEN);
		}

		hc->state = USBO_HC_XFRC;
		hc->err_count = 0;
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_TF);

		if((hc->ep_type == EP_TYPE_CTRL) || (hc->ep_type == EP_TYPE_BULK)){
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
			usbo_hc_halt(husb, chnum);
			clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
		}
		else if(hc->ep_type == EP_TYPE_INTR){
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_ODDFRM);
			husb->hc[chnum].urb_state = URB_DONE;
			if(husb->hostchangeurbcallback != NULL)
				husb->hostchangeurbcallback(husb, chnum, husb->hc[chnum].urb_state);
		}
		hc->toggle_in ^= 1U;
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_CH) != 0){
		sil_andw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		if(hc->state == USBO_HC_XFRC){
			hc->urb_state  = URB_DONE;
		}
		else if(hc->state == USBO_HC_STALL){
			hc->urb_state  = URB_STALL;
		}
		else if((hc->state == USBO_HC_XACTERR) || (hc->state == USBO_HC_DATATGLERR)){
			if(hc->err_count++ > 3U){
				hc->err_count = 0;
				hc->urb_state = URB_ERROR;
			}
			else{
				hc->urb_state = URB_NOTREADY;
			}

			/*
			 *  ͥκƥƥ
			 */
			sil_modw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);
		}
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_CH);
		if(husb->hostchangeurbcallback != NULL)
			husb->hostchangeurbcallback(husb, chnum, husb->hc[chnum].urb_state);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_USBER) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		hc->err_count++;
		hc->state = USBO_HC_XACTERR;
#if 1	/* ROI DEBUG */
		syslog_1(LOG_NOTICE, "## SET USBO_HC_XACTERR1 [%08x] ##", sil_rew_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTF)));
#endif	/* ROI DEBUG */
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_USBER);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_NAK) != 0){
		if(hc->ep_type == EP_TYPE_INTR){
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
			usbo_hc_halt(husb, chnum);
		}
		else if((hc->ep_type == EP_TYPE_CTRL) || (hc->ep_type == EP_TYPE_BULK)){
			/*
			 *  ͥκƥƥ
			 */
			sil_modw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);
		}
		hc->state = USBO_HC_NAK;
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
	}
}

/*
 *  HOSTͥOUT׵
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 chnum ֹͥ(1-15)
 */
static void
usbo_hc_out_irqhandler(USB_OTG_Handle_t *husb, uint8_t chnum)
{
    USB_OTG_HCTypeDef *hc = &husb->hc[chnum];
	uint32_t base = husb->base;

	if(hc_int_status(base, chnum, USB_OTG_HCHINTF_DMAER) != 0){
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_DMAER);
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_ACK) != 0){
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_ACK);
		if(hc->do_ping == 1U){
			hc->state = USBO_HC_NYET;
			sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
			usbo_hc_halt(husb, chnum);
			hc->urb_state  = URB_NOTREADY;
		}
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_NYET) != 0){
		hc->state = USBO_HC_NYET;
		hc->err_count= 0;
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NYET);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_REQOVR) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_REQOVR);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_TF) != 0){
		hc->err_count = 0;
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_TF);
		hc->state = USBO_HC_XFRC;
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_STALL) != 0){
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_STALL);
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		hc->state = USBO_HC_STALL;
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_NAK) != 0){
		hc->err_count = 0;
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		hc->state = USBO_HC_NAK;
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
	}
	else if(hc_int_status(base, chnum,USB_OTG_HCHINTF_USBER) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		usbo_hc_halt(husb, chnum);
		hc->state = USBO_HC_XACTERR;
#if 1	/* ROI DEBUG */
    syslog_1(LOG_NOTICE, "## SET USBO_HC_XACTERR2 [%08x] ##", sil_rew_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTF)));
#endif	/* ROI DEBUG */
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_USBER);
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_DTER) != 0){
		sil_orw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
 		usbo_hc_halt(husb, chnum);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_NAK);
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_DTER);
		hc->state = USBO_HC_DATATGLERR;
	}
	else if(hc_int_status(base, chnum, USB_OTG_HCHINTF_CH) != 0){
		sil_andw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHINTEN), USB_OTG_HCHINTEN_CHIE);
		if(hc->state == USBO_HC_XFRC){
			hc->urb_state  = URB_DONE;
			if(hc->ep_type == EP_TYPE_BULK){
				hc->toggle_out ^= 1U; 
			}
		}
		else if (hc->state == USBO_HC_NAK){
			hc->urb_state  = URB_NOTREADY;
		}
		else if (hc->state == USBO_HC_NYET){
			hc->urb_state  = URB_NOTREADY;
			hc->do_ping = 0U;
		}
		else if (hc->state == USBO_HC_STALL){
			hc->urb_state  = URB_STALL;
		}
		else if((hc->state == USBO_HC_XACTERR) || (hc->state == USBO_HC_DATATGLERR)){
			if(hc->err_count++ > 3){
				hc->err_count = 0;
				hc->urb_state = URB_ERROR;
			}
			else{
				hc->urb_state = URB_NOTREADY;
			}

			/* re-activate the channel  */
			sil_modw_mem((uint32_t *)(USBHC_BASE(base, chnum)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);
		}
		clear_hc_int(base, chnum, USB_OTG_HCHINTF_CH);
		if(husb->hostchangeurbcallback != NULL)
			husb->hostchangeurbcallback(husb, chnum, husb->hc[chnum].urb_state);
	}
}

/*
 *  RX QUEUE LEVEL׵
 *  parameter1 husb  USBϥɥؤΥݥ
 */
static void
usbo_rxqlvl_irqhandler(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint8_t  channelnum = 0U;
	uint32_t pktsts, pktcnt;
	uint32_t temp = 0U;

	temp = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTATP));
	channelnum = temp &  USB_OTG_GRSTATRP_EPNUM;  
	pktsts = (temp &  USB_OTG_GRSTATRP_RPCKST) >> 17U;
	pktcnt = (temp &  USB_OTG_GRSTATRP_BCOUNT) >> 4U;

	switch(pktsts){
	case GRXSTS_PKTSTS_IN:
		/*
		 *  HOSTХåեѥåɤ߹
		 */
		if((pktcnt > 0) && (husb->hc[channelnum].xfer_buff != (void  *)0U)){
			usbo_readPacket(husb->base, husb->hc[channelnum].xfer_buff, pktcnt);

			/*
			 *  ޥǡ
			 */
			husb->hc[channelnum].xfer_buff += pktcnt;
			husb->hc[channelnum].xfer_count  += pktcnt;

			if((sil_rew_mem((uint32_t *)(USBHC_BASE(base, channelnum)+TOFF_USBHC_HCHLEN)) & USB_OTG_HCHLEN_PCNT) > 0){
				sil_modw_mem((uint32_t *)(USBHC_BASE(base, channelnum)+TOFF_USBHC_HCHCTL), USB_OTG_HCHCTL_CDIS, USB_OTG_HCHCTL_CEN);
				husb->hc[channelnum].toggle_in ^= 1U;
			}
		}
		break;

	case GRXSTS_PKTSTS_DATA_TOGGLE_ERR:
	case GRXSTS_PKTSTS_IN_XFER_COMP:
	case GRXSTS_PKTSTS_CH_HALTED:
	default:
		break;
	}
}

/*
 *  HOSTݡȳ׵
 *  parameter1 husb  USBϥɥؤΥݥ
 */
static void
usbo_port_irqhandler(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	volatile uint32_t hprt0, hprt0_dup;

	hprt0 = sil_rew_mem((uint32_t*)(USBPRT_BASE(base)));
	hprt0_dup = hprt0 & ~USB_OTG_HPCS_PE;

	/*
	 *  ݡȥͥȥå
	 */
	if((hprt0 & USB_OTG_HPCS_PCD) != 0){
		if((hprt0 & USB_OTG_HPCS_PCST) == USB_OTG_HPCS_PCST){
			sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_DISCIE);
			if(husb->hostconnectcallback != NULL)
				husb->hostconnectcallback(husb);
		}
	}

	/*
	 *  ݡȥ֥͡Ѳå
	 */
	if((hprt0 & USB_OTG_HPCS_PEDC) != 0){
		if((hprt0 & USB_OTG_HPCS_PE) != 0){
			if(husb->Init.phy_itface  == USB_PHY_EMBEDDED){
				if((hprt0 & USB_OTG_HPCS_PS) == (HPRT0_PRTSPD_LOW_SPEED << 17)){
					usbo_initFSLSPClkSel(husb, 6000);
				}
				else{
					usbo_initFSLSPClkSel(husb, 48000);
				}
			}
			else{
				if(husb->Init.speed == USB_SPEED_FULL){
					sil_wrw_mem((uint32_t *)(USBH_BASE(base)+TOFF_USBH_HFT), 60000U);
				}
			}

			if(husb->hostconnectcallback != NULL)
				husb->hostconnectcallback(husb);

			if(husb->Init.speed == USB_SPEED_HIGH){
				sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_DISCIE);
			}
		}
		else{
			/*
			 *  HPRTΥ꡼
			 */
			sil_andw_mem((uint32_t*)(USBPRT_BASE(base)), (USB_OTG_HPCS_PE | USB_OTG_HPCS_PCD |\
				USB_OTG_HPCS_PEDC));
			sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTEN_DISCIE);
		}
	}

	/*
	 *  ݡȳߥꥢ
	 */
	sil_wrw_mem((uint32_t*)(USBPRT_BASE(base)), hprt0_dup);
}

/*
 *  HOSTߥϥɥ
 *  parameter1 husb  USBϥɥؤΥݥ
 */
void
usbo_hcd_irqhandler(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t msk  = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN));
	uint32_t i = 0, int_status = 0, intsts = 0;

	/*
	 *  Ĥʤװγߤ̵뤹
	 */
	if((intsts = get_int_status(base, msk)) == 0){
		intsts = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTF));
		syslog_1(LOG_EMERG, "usbo_hcd_irqhandler Error Intrrupt[%08x] !", intsts);
		clear_host_int(base, intsts);
		return;
	}

	if((intsts & USB_OTG_GINTF_ISOONCIF) != 0){
		/*
		 *  ̤׵װꥢ
		 */
		clear_host_int(base, USB_OTG_GINTF_ISOONCIF);
	}
	if((intsts & USB_OTG_GINTF_ISOINCIF) != 0){
		/*
		 *  ̤׵װꥢ
		 */
		clear_host_int(base, USB_OTG_GINTF_ISOINCIF);
	}

	if((intsts & USB_OTG_GINTF_PTXFEIF) != 0){
		/*
		 *  ̤׵װꥢ
		 */
		clear_host_int(base, USB_OTG_GINTF_PTXFEIF);
	}
    if((intsts & USB_OTG_GINTF_MFIF) != 0){
		/*
		 *  ̤׵װꥢ
		 */
		clear_host_int(base, USB_OTG_GINTF_MFIF);
    }
    /*
	 *  HOSTǥͥȳ
	 */
	if((intsts & USB_OTG_GINTF_DISCIF) != 0){
		/*
		 *  HPRTΥ꡼󥢥å
		 */
		sil_andw_mem((uint32_t*)(USBPRT_BASE(base)), (USB_OTG_HPCS_PE | USB_OTG_HPCS_PCD |\
			USB_OTG_HPCS_PEDC));
		if(husb->hostdisconnectcallback != NULL)
			husb->hostdisconnectcallback(husb);
		usbo_initFSLSPClkSel(husb, 48000);
		clear_host_int(base, USB_OTG_GINTF_DISCIF);
	}
	/*
	 *  HOSTݡȳ
	 */
	if((intsts & USB_OTG_GINTF_HPIF) != 0){
		syslog_0(LOG_NOTICE, "## usbo_hcd_irqhandler USB_OTG_GINTF_HPIF ##");
		usbo_port_irqhandler(husb);
		if(husb->Init.usb_otg_mode == USB_OTG_MODE_DRD)
			sil_dly_nse(10*1000*1000);	/* աǥХǥޤǤԤ(ߤǤԤƤʤ) */
	}
	/*
	 *  Host SOF
	 */
    if((intsts & USB_OTG_GINTF_SOF) != 0){
		if(husb->hostsofcallback != NULL)
			husb->hostsofcallback(husb);
		clear_host_int(base, USB_OTG_GINTF_SOF);
	}
	/*
	 *  HOSTͥ
	 */
	if((intsts & USB_OTG_GINTF_HCIF) != 0){
		int_status = get_haint_status(base);
		for(i = 0 ; i < husb->Init.host_channels; i++){
			if(int_status & (1 << i)){
				if((sil_rew_mem((uint32_t *)(USBHC_BASE(husb->base, i)+TOFF_USBHC_HCHCTL)) &  USB_OTG_HCHCTL_EPDIR) != 0)
					usbo_hc_in_irqhandler(husb, i);
				else
					usbo_hc_out_irqhandler(husb, i);
			}
		}
		clear_host_int(base, USB_OTG_GINTF_HCIF);
	}
	/*
	 *  RX塼٥
	 */
	if((intsts & USB_OTG_GINTF_RXFNEIF) != 0){
		sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTF_RXFNEIF);
		usbo_rxqlvl_irqhandler(husb);
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTF_RXFNEIF);
	}
}

/*******************************************************************************
                       USB DEVICEؿ
*******************************************************************************/
/*
 *  USB DEVICEԡ
 *  parameter1 base  USB OTG쥸١ɥ쥹
 *  parameter2 speed device speed USB_SPEED_HIGH(High speed mode)
 *                                USB_SPEED_HIGH_IN_FULL(High speed core in Full Speed mode)
 *                                USB_SPEED_FULL(Full speed mode)
 *                                USB_SPEED_LOW(Low speed mode)
 */
static void
usbo_setDevSpeed(uint32_t base, uint8_t speed)
{
	sil_modw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCFG), USB_OTG_DCFG_DS, speed);
}

/*
 *  ǥХͥ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_devconnect(USB_OTG_Handle_t *husb)
{
	sil_andw_mem((uint32_t *)(USBD_BASE(husb->base)+TOFF_USBD_DCTL), USB_OTG_DCTL_SD);
	return E_OK;
}

/*
 *  ǥХǥͥ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_devdisconnect(USB_OTG_Handle_t *husb)
{
	sil_orw_mem((uint32_t *)(USBD_BASE(husb->base)+TOFF_USBD_DCTL), USB_OTG_DCTL_SD);
	return E_OK; 
}

/*
 *  ƥ֥ɥݥ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ep    ɥݥȹ¤ΤؤΥݥ
 *  return     E_OK
 */
ER
usbo_activateEndpoint(USB_OTG_Handle_t *husb, USB_OTG_EPTypeDef *ep)
{
	uint32_t base = husb->base;

	if(ep->is_in == 1){
		if(((sil_rew_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL))) & USB_OTG_DIEPCTL_EPACT) == 0){
			sil_andw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), \
				(USB_OTG_DIEPCTL_MPL | USB_OTG_DIEPCTL_EPTYPE | USB_OTG_DIEPCTL_TXFNUM));
			sil_orw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), \
				((ep->maxpacket & USB_OTG_DIEPCTL_MPL) | (ep->type << 18U) |\
					((ep->num) << 22U) | USB_OTG_DIEPCTL_SD0PID | USB_OTG_DIEPCTL_EPACT)); 
		}
		sil_orw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), USB_OTG_DAEPINTEN_IEPIE & (1 << (ep->num)));
	}
	else{
		if(((sil_rew_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL))) & USB_OTG_DIEPCTL_EPACT) == 0){
			sil_andw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), \
				(USB_OTG_DOEPCTL_MPL | USB_OTG_DOEPCTL_EPTYPE));
			sil_orw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), \
				((ep->maxpacket & USB_OTG_DOEPCTL_MPL) | (ep->type << 18U) |\
					USB_OTG_DOEPCTL_SD0PID | USB_OTG_DOEPCTL_EPACT));
		}
		sil_orw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), USB_OTG_DAEPINTEN_OEPIE & (1 << (ep->num + 16)));
	}
	return E_OK;
}

/*
 *  ǥƥ֥ɥݥ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ep    ɥݥȹ¤ΤؤΥݥ
 *  return     E_OK
 */
ER
usbo_deactivateEndpoint(USB_OTG_Handle_t *husb, USB_OTG_EPTypeDef *ep)
{
	uint32_t base = husb->base;

	if(ep->is_in == 1){
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DEP1INTEN), (USB_OTG_DAEPINTEN_IEPIE & (1 << (ep->num))));
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), (USB_OTG_DAEPINTEN_IEPIE & (1 << (ep->num))));
		sil_andw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), USB_OTG_DIEPCTL_EPACT);
	}
	else{
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DEP1INTEN), (USB_OTG_DAEPINTEN_OEPIE & (1 << (ep->num + 16))));
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), (USB_OTG_DAEPINTEN_OEPIE & (1 << (ep->num + 16))));
		sil_andw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), USB_OTG_DOEPCTL_EPACT);
	}
	return E_OK;
}

/*
 *  ɥݥȤ˥ơ֤
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ep    ɥݥȹ¤ΤؤΥݥ
 *  return     E_OK
 */
ER
usbo_epsetStall(USB_OTG_Handle_t *husb, USB_OTG_EPTypeDef *ep)
{
	uint32_t base = husb->base;

	if(ep->is_in == 1){
		if((sil_rew_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL)) & USB_OTG_DIEPCTL_EPEN) == 0){
			sil_andw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), USB_OTG_DIEPCTL_EPD);
		}
		sil_orw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), USB_OTG_DIEPCTL_STALL);
	}
	else
		sil_orw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), USB_OTG_DOEPCTL_STALL);
	return E_OK;
}

/*
 *  ɥݥȤΥơ֤򥯥ꥢ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ep    ɥݥȹ¤ΤؤΥݥ
 *  return     E_OK
 *
 */
ER
usbo_epclearStall(USB_OTG_Handle_t *husb, USB_OTG_EPTypeDef *ep)
{
	uint32_t base = husb->base;

	if(ep->is_in == 1){
		sil_andw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), USB_OTG_DIEPCTL_STALL);
		if(ep->type == EP_TYPE_INTR || ep->type == EP_TYPE_BULK){
			sil_orw_mem((uint32_t *)(USBIEP_BASE(base, ep->num)+TOFF_USBEI_DIEPCTL), USB_OTG_DIEPCTL_SD0PID);	/* DATA0 */
		}
	}
	else{
		sil_andw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), USB_OTG_DOEPCTL_STALL);
		if(ep->type == EP_TYPE_INTR || ep->type == EP_TYPE_BULK){
			sil_orw_mem((uint32_t *)(USBOEP_BASE(base, ep->num)+TOFF_USBEO_DOEPCTL), USB_OTG_DOEPCTL_SD0PID);	/* DATA0 */
		}
	}
	return E_OK;
}

/**
 *  USB˥ǥХɥ쥹
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 addree ǥХɥ쥹(0-255)
 *  return     E_OK
 */
ER
usbo_setDevAddress(USB_OTG_Handle_t *husb, uint8_t address)
{
	husb->connected = 1;
	sil_modw_mem((uint32_t *)(USBD_BASE(husb->base)+TOFF_USBD_DCFG), \
		USB_OTG_DCFG_DCFG_DAR, \
			((address << 4) & USB_OTG_DCFG_DCFG_DAR));
	return E_OK;
}

/*
 *  USBǥХԡɤФ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     speed ǥХԡ USB_SPEED_HIGH: High speed mode
 *                                    USB_SPEED_FULL: Full speed mode
 *                                    USB_SPEED_LOW: Low speed mode
 */
uint8_t
usbo_getDevSpeed(USB_OTG_Handle_t *husb)
{
	uint8_t  speed = 0;
	uint32_t status = sil_rew_mem((uint32_t *)(USBD_BASE(husb->base)+TOFF_USBD_DSTAT)) & USB_OTG_DSTAT_ES;

	if(status == DSTS_ENUMSPD_HS_PHY_30MHZ_OR_60MHZ){
		speed = USB_SPEED_HIGH;
	}
	else if((status == DSTS_ENUMSPD_FS_PHY_30MHZ_OR_60MHZ) || (status == DSTS_ENUMSPD_FS_PHY_48MHZ)){
		speed = USB_SPEED_FULL;
	}
	else if(status == DSTS_ENUMSPD_LS_PHY_6MHZ){
		speed = USB_SPEED_LOW;
	}
	return speed;
}

/*
 *  EP0-OUT
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 psetup åȥåץѥåȥǡؤΥݥ
 *  return     E_OK
 */
ER
usbo_ep0_outstart(USB_OTG_Handle_t *husb, uint8_t *psetup)
{
	uint32_t base = husb->base;

	sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, 0)+TOFF_USBEO_DOEPLEN), \
		((USB_OTG_DOEPLEN_PCNT & (1U << 19U)) | (3U * 8U) | USB_OTG_DOEPLEN_STPCNT));

	if(husb->Init.dma_enable == 1){
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, 0)+TOFF_USBEO_DOEPDMAADDR), (uint32_t)psetup);
 		/*
		 *  EP0ͭ
		 */
		sil_orw_mem((uint32_t *)(USBOEP_BASE(base, 0)+TOFF_USBEO_DOEPCTL), \
			(USB_OTG_DOEPCTL_EPACT | USB_OTG_DOEPCTL_EPEN));
	}
	return E_OK;
}

/*
 *  EPnž
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 ep    ɥݥȹ¤ΤؤΥݥ
 *  return     E_OK
 */
ER
usbo_epstartxfer(USB_OTG_Handle_t *husb, USB_OTG_EPTypeDef *ep)
{
	uint32_t base   = husb->base;
	uint16_t pktcnt = 0;
	uint8_t  epnum  = ep->num;
	volatile uint32_t epctl, eplen;

	if(ep->is_in == 1){	/* INɥݥ */
		epctl  = sil_rew_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPCTL));
		eplen  = sil_rew_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPLEN));
		eplen &= ~(USB_OTG_DIEPLEN_TLEN | USB_OTG_DIEPLEN_PCNT);

		/* zero length packet or endpoint 0 */
		if(ep->xfer_len == 0){
			/* set transfer packet count to 1 */
			eplen |= 1 << 19;
 		}
		else{
			/* set transfer packet count */
			if(epnum == 0){
				ep->xfer_len = MIN(ep->xfer_len, ep->maxpacket);
				eplen |= 1 << 19;
			}
			else{
				eplen |= ((ep->xfer_len - 1 + ep->maxpacket) / ep->maxpacket) << 19;
			}

			/* set endpoint transfer length */
			eplen |= ep->xfer_len;

			if(ep->type == EP_TYPE_ISOC){
				eplen |= USB_OTG_DIEPLEN_MCNT;
			}
		}
		sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPLEN), eplen);

		if(husb->Init.dma_enable == 1){
			sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPDMAADDR), ep->dma_addr);
		}

		if(ep->type == EP_TYPE_ISOC){
			if(((sil_rew_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DSTAT)) & USB_OTG_DSTAT_FNRSOF) >> 8) & 1)
				epctl |= USB_OTG_DIEPCTL_SD1PID;
			else
				epctl |= USB_OTG_DIEPCTL_SD0PID;
		}

		/*
		 *  ꥢNAKɥݥȥ֥͡
		 */
		epctl |= USB_OTG_DIEPCTL_CNAK | USB_OTG_DIEPCTL_EPEN;
		sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPCTL), epctl);

		if(ep->type != EP_TYPE_ISOC){
			/* enable the Tx FIFO empty interrupt for this endpoint */
			if(ep->xfer_len > 0){
				sil_orw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPFEINTEN), (1 << (epnum)));
			}
		}
		else
			usbo_writePacket(base, ep->xfer_buff, epnum, ep->xfer_len, husb->Init.dma_enable); 
	}
	else{	 /* OUTɥݥ */
		epctl  = sil_rew_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPCTL));
		eplen  = sil_rew_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPLEN));
		eplen &= ~(USB_OTG_DOEPLEN_TLEN | USB_OTG_DOEPLEN_PCNT);

		/* zero length packet or endpoint 0 */
		if((0U == ep->xfer_len) || (0U == epnum)){
			/* set the transfer length to max packet size */
			eplen |= ep->maxpacket;
			/* set the transfer packet count to 1 */
			eplen |= 1U << 19;
		}
		else{
			/* configure the transfer size and packet count as follows:
			 * pktcnt = N
			 * xfersize = N * maxpacket
			 */
			pktcnt = (ep->xfer_len + ep->maxpacket - 1) / ep->maxpacket;

			eplen |= pktcnt << 19;
			eplen |= pktcnt * ep->maxpacket;
		}
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPLEN), eplen);

		if(husb->Init.dma_enable == 1){
			sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPDMAADDR), (uint32_t)(ep->xfer_buff));
		}

		if(ep->type == EP_TYPE_ISOC){
			if(ep->data_pid_start){
				epctl |= USB_OTG_DOEPCTL_SD1PID;
			}
			else{
				epctl |= USB_OTG_DOEPCTL_SD0PID;
			}
		}

		/*
		 *  ꥢNAKɥݥȥ֥͡
		 */
		epctl |= USB_OTG_DOEPCTL_EPEN | USB_OTG_DOEPCTL_CNAK;
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPCTL), epctl);
	}
	return E_OK;
}

/*
 *  USBǥХ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_devinit(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	bool_t   forced;
	uint32_t i = 0U;

	/* Device mode configuration */
	sil_modw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCFG), \
		USB_OTG_DCFG_EOPFT, DCFG_FRAME_INTERVAL_80);

	if(husb->Init.phy_itface == USB_PHY_ULPI){	/* ULIP-PHY */
		if(husb->Init.speed == USB_SPEED_HIGH){
			/*
			 *  ϥԡɸ
			 */
			usbo_setDevSpeed(base, USB_SPEED_HIGH);
		}
		else{
			/*
			 * ϥե⡼
			 */
			usbo_setDevSpeed(base, USB_SPEED_HIGH_IN_FULL);
		}
	}
	else{	/* EMBEDDED-PHY */
		/*
		 *  ˥ե륹ԡ
		 */
		usbo_setDevSpeed(base, USB_SPEED_FULL);
	}

	/*
	 *  ǥХ˴ط롢٤Ƥγߤ򥯥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPINTEN), 0);
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DOEPINTEN), 0);
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINT), 0xFFFFFFFF);
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), 0);
	for(i = 0 ; i < husb->Init.dev_endpoints ; i++){
		if((sil_rew_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPCTL)) & USB_OTG_DIEPCTL_EPEN) != 0){
			sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPCTL), (USB_OTG_DIEPCTL_EPD | USB_OTG_DIEPCTL_SNAK));
		}
		else{
			sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPCTL), 0);
		}
		sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPLEN), 0);
		sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPINTF), 0xFF);

		if((sil_rew_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPCTL)) & USB_OTG_DOEPCTL_EPEN) != 0){
			sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPCTL), (USB_OTG_DOEPCTL_EPD | USB_OTG_DOEPCTL_SNAK));
		}
		else{
			sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPCTL), 0);
		}
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPLEN), 0);
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPINTF), 0xFF);
	}

	usbo_devdisconnect(husb);
	forced = usbo_force_mode_if_needed(husb, false);
	usbo_initFiFo(husb);
	if(forced)
		usbo_clear_force_mode(husb);

	if(husb->Init.usb_otg_mode == USB_OTG_MODE_DEVICE)
		usbo_setupint(husb);
	return E_OK;
}

/*
 *  USBǥХ
 *  parameter1 husb  USBϥɥؤΥݥ
 *  return     E_OK
 */
ER
usbo_stopdevice(USB_OTG_Handle_t *husb)
{
	uint32_t base = husb->base;
	uint32_t i;

	/*
	 *  ɥݥȤγߥꥢ
	 */
	for(i = 0 ; i < 15 ; i++){
		sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPINTF), 0xFF);
		sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, i)+TOFF_USBEO_DOEPINTF), 0xFF);
	}
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINT), 0xFFFFFFFF);

	/*
	 *  ɥݥȤγ
	 */
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPINTEN), 0);
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DOEPINTEN), 0);
	sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), 0);

	/*
	 *  FIFOեå
	 */
	usbo_flushRxFifo(husb);
	usbo_flushTxFifo(husb,  0x10);
	return E_OK;
}

/*
 *  LPM
 */
ER
usbo_init_lpm(USB_OTG_Handle_t *husb)
{
	return E_OK;
}

/*
 *  USB EP-INװ
 *  parameter1 base USB OTG쥸١ɥ쥹
 *  parameter2 epnum ɥݥֹ
 *  return     EP-INװ
 */
static uint32_t
usbo_devInEPInterrupt(uint32_t base , uint8_t epnum)
{
	uint32_t msk, emp;

	msk = sil_rew_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPINTEN));
	emp = sil_rew_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPFEINTEN));
	msk |= ((emp >> epnum) & 0x1) << 7;
	return sil_rew_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPINTF)) & msk;
}

/*
 *  USB EP-OUTװ
 *  parameter1 base USB OTG쥸١ɥ쥹
 *  parameter2 epnum ɥݥֹ
 *  return     EP-OUTװ
 */
static uint32_t
usbo_devOutEPInterrupt(uint32_t base, uint8_t epnum)
{
	uint32_t v;
	v = sil_rew_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPINTF));
	return v & sil_rew_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DOEPINTEN));
}


/*
 *  ΥѥåȤFIFO
 *  parameter1 husb  USBϥɥؤΥݥ
 *  parameter2 epnum ɥݥֹ
 *  return     E_OK
 */
static ER
usbo_writeemptyTxfifo(USB_OTG_Handle_t *husb, uint32_t epnum)
{
	uint32_t  base = husb->base;
	USB_OTG_EPTypeDef *ep;
	int32_t len = 0U;
	uint32_t len32b;

	ep = &husb->IN_ep[epnum];
	len = ep->xfer_len - ep->xfer_count;

	if(len > ep->maxpacket){
		len = ep->maxpacket;
	}
	len32b = (len + 3) / 4;

	while((sil_rew_mem((uint32_t *)(USBIEP_BASE(base, epnum)+TOFF_USBEI_DIEPTFSTAT)) & USB_OTG_DIEPTFSTAT_IEPTFS) > len32b
          && ep->xfer_count < ep->xfer_len && ep->xfer_len != 0){
		/*
		 * FIFOؤΥǡ񤭹
		 */
		len = ep->xfer_len - ep->xfer_count;
		if(len > ep->maxpacket){
			len = ep->maxpacket;
		}
		len32b = (len + 3) / 4;
		usbo_writePacket(base, ep->xfer_buff, epnum, len, husb->Init.dma_enable); 
		ep->xfer_buff  += len;
		ep->xfer_count += len;
	}

	if(ep->xfer_len <= ep->xfer_count){	/* ǸΥѥåȤʤEMPTYߤޥ */
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPFEINTEN), (1 << epnum));
	}
	return E_OK;
}

/*
 *  DEVICEߥϥɥ
 *  parameter1 husb  USBϥɥؤΥݥ
 */
void
usbo_pcd_irqhandler(USB_OTG_Handle_t *husb)
{
	USB_OTG_EPTypeDef *ep;
	uint32_t base = husb->base;
	uint32_t msk  = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN));
	uint32_t inr  = get_int_status(base, msk);
	uint32_t i = 0U, ep_intr = 0U, epint = 0U, epnum = 0U;
	uint32_t temp = 0;

	/*
	 *  ߥ顼å
	 */
	if(inr == 0){
		uint32_t intsts = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTF));
		syslog_1(LOG_EMERG, "usbo_pcd_irqhandler Error Intrrupt[%08x] !", intsts);
		clear_dev_int(base, intsts);
		return;
	}

    if((inr & USB_OTG_GINTF_MFIF) != 0){
		/*
		 *  顼װꥢ
		 */
		clear_dev_int(base, USB_OTG_GINTF_MFIF);
	}
	if((inr & USB_OTG_GINTF_OEPIF) != 0){
		/*
		 *  EP-OUT
		 */
		ep_intr = get_outep_int_status(base);
		for(epnum = 0 ; epnum < husb->Init.dev_endpoints ; epnum++){
			if((ep_intr & (1<<epnum)) != 0){
				epint = usbo_devOutEPInterrupt(base, epnum);
				if((epint & USB_OTG_DOEPINTF_TF) != 0){
					clear_outep_int(base, epnum, USB_OTG_DOEPINTF_TF);
					if(husb->Init.dma_enable == 1U){
						volatile uint32_t eplen = sil_rew_mem((uint32_t *)(USBOEP_BASE(base, epnum)+TOFF_USBEO_DOEPLEN));
						husb->OUT_ep[epnum].xfer_count = husb->OUT_ep[epnum].maxpacket - (eplen & USB_OTG_DOEPLEN_TLEN); 
						husb->OUT_ep[epnum].xfer_buff += husb->OUT_ep[epnum].maxpacket;
					}
					if(husb->devdataoutstagecallback != NULL)
						husb->devdataoutstagecallback(husb, epnum);
					if(husb->Init.dma_enable == 1 && epnum == 0 && husb->OUT_ep[epnum].xfer_len == 0){
						/*
						 *  DMA⡼EP0 ZERO LENGTHѥå
						 */
						usbo_ep0_outstart(husb, (uint8_t *)husb->Setup);
					}
				}
				if((epint & USB_OTG_DOEPINTF_STPF) != 0){
					/*
					 *  SETUPѥå
					 */
					if(husb->devsetupstagecallback != NULL)
						husb->devsetupstagecallback(husb);
					clear_outep_int(base, epnum, USB_OTG_DOEPINTF_STPF);
				}
			}
		}
	}
	if((inr & USB_OTG_GINTF_IEPIF) != 0){
		/*
		 *  EP-IN
		 */
		ep_intr = get_inep_int_status(base);
		for(epnum = 0 ; epnum < husb->Init.dev_endpoints ; epnum++){
			if((ep_intr & (1<<epnum)) != 0){ /* In ITR */
				epint = usbo_devInEPInterrupt(base, epnum);
				if((epint & USB_OTG_DIEPINTF_TF) != 0){
					sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPFEINTEN), (1<<epnum));
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_TF);
					if(husb->Init.dma_enable == 1){
						husb->IN_ep[epnum].xfer_buff += husb->IN_ep[epnum].maxpacket; 
					}
					if(husb->devdatainstagecallback != NULL)
						husb->devdatainstagecallback(husb, epnum);
					if(husb->Init.dma_enable == 1 && epnum == 0 && husb->IN_ep[epnum].xfer_len == 0){
						/*
						 *  DMA⡼EP0 ZERO LENGTHѥå
						 */
						usbo_ep0_outstart(husb, (uint8_t *)husb->Setup);
					}
				}
				if((epint & USB_OTG_DIEPINTF_CITO) != 0){
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_CITO);
				}
				if((epint & USB_OTG_DIEPINTF_EPTXFUD) != 0){
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_EPTXFUD);
				}
				if((epint & USB_OTG_DIEPINTF_IEPNE) != 0){
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_IEPNE);
				}
				if((epint & USB_OTG_DIEPINTF_EPDIS) != 0){
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_EPDIS);
				}
				if((epint & USB_OTG_DIEPINTF_TXFE) != 0){
					usbo_writeemptyTxfifo(husb , epnum);
					clear_inep_int(base, epnum, USB_OTG_DIEPINTF_TXFE);
				}
			}
		}
	}
    /*
	 *  SUSPEND
	 */
    if((inr & USB_OTG_GINTF_SP) != 0){
		if((sil_rew_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DSTAT)) & USB_OTG_DSTAT_SPST) != 0){
			if(husb->devsuspendcallback != NULL && husb->connected)
				husb->devsuspendcallback(husb);
			if(husb->Init.usb_otg_mode == USB_OTG_MODE_DRD && husb->hostdisconnectcallback != NULL && husb->connected){
				syslog_1(LOG_NOTICE, "usbo_pcd_irqhandler drd mode disconnect lx_state(%d) !", husb->lx_state);
				husb->connected = 0;
				husb->hostdisconnectcallback(husb);
			}
			husb->lx_state = DWC2_L2;
			if(husb->Init.low_power_enable == 1)
				sil_orw_mem((uint32_t *)(USBPGC_BASE(base)), (USB_OTG_PWRCLKCTL_SUCLK | USB_OTG_PWRCLKCTL_SHCLK));
		}
		clear_dev_int(base, USB_OTG_GINTF_SP);
	}
	/*
	 *  RESUME
	 */
	if((inr & USB_OTG_GINTF_WKUPIF) != 0){
		usbo_exit_hibernation(husb, true);		/* ʥ */
		if(husb->devresumecallback != NULL)
			husb->devresumecallback(husb);
		husb->lx_state = DWC2_L0;
		clear_dev_int(base, USB_OTG_GINTF_WKUPIF);
	}
    /*
	 *  DEVICE-SOF
	 */
	if((inr & USB_OTG_GINTF_SOF) != 0){
		if(husb->devsofcallback != NULL)
			husb->devsofcallback(husb);
		clear_dev_int(base, USB_OTG_GINTF_SOF);
	}
    /*
	 *  RXQLEVEL
	 */
	if((inr & USB_OTG_GINTF_RXFNEIF) != 0){
		uint32_t bcount;
		sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTF_RXFNEIF);
		temp   = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GRSTATP));
		bcount = (temp & USB_OTG_GRSTATRP_BCOUNT) >> 4;
		ep = &husb->OUT_ep[temp & USB_OTG_GRSTATRP_EPNUM];
		switch((temp & USB_OTG_GRSTATRP_RPCKST) >> 17U){
		case STS_DATA_UPDT:
			if(bcount != 0){
				usbo_readPacket(base, ep->xfer_buff, bcount);
				ep->xfer_buff += bcount;
				ep->xfer_count += bcount;
			}
			break;
		case STS_SETUP_UPDT:
			usbo_readPacket(base, (uint8_t *)husb->Setup, 8U);
			ep->xfer_count += bcount;
			break;
		default:
			break;
		}
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GINTEN), USB_OTG_GINTF_RXFNEIF);
	}
	/*
	 * USB-RESET
	 */
	if((inr & USB_OTG_GINTF_RST) != 0){
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCTL), USB_OTG_DCTL_RWKUP);
		usbo_flushTxFifo(husb, 0U);
		for(i = 0; i < husb->Init.dev_endpoints ; i++){
			sil_wrw_mem((uint32_t *)(USBIEP_BASE(base, i)+TOFF_USBEI_DIEPINTF), 0xFF);
			sil_wrw_mem((uint32_t *)(USBOEP_BASE(base, 0)+TOFF_USBEO_DOEPINTF), 0xFF);
		}
		sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINT), 0xFFFFFFFF);
		sil_orw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DAEPINTEN), 0x10001);
		if(husb->Init.use_dedicated_ep1){
			sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DOEP1INTEN), (USB_OTG_DOEPINTEN_STPFEN | USB_OTG_DOEPINTEN_TFEN));
			sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEP1INTEN), USB_OTG_DIEPINTEN_TFEN);
		}
		else{
			sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DOEPINTEN), (USB_OTG_DOEPINTEN_STPFEN | USB_OTG_DOEPINTEN_TFEN));
			sil_wrw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DIEPINTEN), USB_OTG_DIEPINTEN_TFEN);
		}
		/*
		 *  ǥեȥɥ쥹򥼥
		 */
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCFG), USB_OTG_DCFG_DCFG_DAR);
		/*
		 *  SETUPѥåȼ
		 */
		usbo_ep0_outstart(husb, (uint8_t *)husb->Setup);
		clear_dev_int(base, USB_OTG_GINTF_RST);
	}
	/*
	 *  ENUMERATIONϥɥ뽪λ
	 */
	if((inr & USB_OTG_GINTF_ENUMFIF) != 0){
		/*
		 *  USBƥ١ȥåȥå
		 */
		sil_andw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCTL), USB_OTG_DCTL_CGINAK);
		sil_orw_mem((uint32_t *)(USBD_BASE(base)+TOFF_USBD_DCTL), USB_OTG_DCTL_CGINAK);

		sil_andw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), USB_OTG_GUSBCS_UTT);
		if(usbo_getDevSpeed(husb) == USB_SPEED_HIGH){
			husb->Init.speed     = USB_SPEED_HIGH;
			sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), ((USBD_HS_TRDT_VALUE << 10) & USB_OTG_GUSBCS_UTT));
		}
		else{
			husb->Init.speed     = USB_SPEED_FULL;
			sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GUSBCS), ((USBD_FS_TRDT_VALUE << 10) & USB_OTG_GUSBCS_UTT));
		}
		if(husb->devresetcallback != NULL)
			husb->devresetcallback(husb);
		clear_dev_int(base, USB_OTG_GINTF_ENUMFIF);
	}
    /*
	 *  ISO IN̤
	 */
    if((inr & USB_OTG_GINTF_ISOINCIF) != 0){
		if(husb->devisoincallback != NULL)
			husb->devisoincallback(husb, epnum);
		clear_dev_int(base, USB_OTG_GINTF_ISOINCIF);
    }
	/*
	 *  ISO OUT̤
	 */
	if((inr & USB_OTG_GINTF_ISOONCIF) != 0){
		if(husb->devisooutcallback != NULL)
			husb->devisooutcallback(husb, epnum);
		clear_dev_int(base, USB_OTG_GINTF_ISOONCIF);
	}
    /*
	 *  ͥ󥤥٥ȳ
	 */
    if((inr & USB_OTG_GINTF_SESIF) != 0){
		if(husb->devconnectcallback != NULL)
			husb->devconnectcallback(husb);
		clear_dev_int(base, USB_OTG_GINTF_SESIF);
	}
    /*
	 *  ǥͥ󥤥٥ȳ
	 */
    if((inr & USB_OTG_GINTF_OTGIF) != 0){
		temp = sil_rew_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GOTGINTF));
		if((temp & USB_OTG_GOTGINTF_SESEND) == USB_OTG_GOTGINTF_SESEND){
			husb->connected = 0;
			if(husb->devdisconnectcallback != NULL)
				husb->devdisconnectcallback(husb);
		}
		sil_orw_mem((uint32_t *)(USBOTG_BASE(base)+TOFF_USBO_GOTGINTF), temp);
	}
}

/*
 *  USB-OTGߥӥ롼
 */
void
usb_otg_isr(intptr_t exinf)
{
	USB_OTG_Handle_t *husb = &Husb[INDEX_USB((uint32_t)exinf)];
	if(get_usbmode(husb->base) == USB_OTG_MODE_DEVICE){
		usbo_pcd_irqhandler(husb);
	}
	else{
		usbo_hcd_irqhandler(husb);
	}
}

/*
 *  USB-OTG WAKEUPߥӥ롼
 */
void usb_otg_wkup_isr(intptr_t exinf)
{
	USB_OTG_Handle_t *husb = &Husb[INDEX_USB((uint32_t)exinf)];
	uint32_t base = husb->base;

	syslog_1(LOG_EMERG, "usb_otg_wkup_isr(&d) !", (int)exinf);
	if((sil_rew_mem((uint32_t *)(USBD_BASE(base+TOFF_USBD_DSTAT))) & USB_OTG_DSTAT_SPST) != 0){
		if(husb->Init.low_power_enable == 1)
			sil_orw_mem((uint32_t *)(USBPGC_BASE(base)), (USB_OTG_PWRCLKCTL_SUCLK | USB_OTG_PWRCLKCTL_SHCLK));
		sil_orw_mem((uint32_t *)(USBD_BASE(base+TOFF_USBD_DCTL)), USB_OTG_DCTL_RWKUP);
		sil_dly_nse(5*1000*1000);
		sil_andw_mem((uint32_t *)(USBD_BASE(base+TOFF_USBD_DCTL)), USB_OTG_DCTL_RWKUP);
	}
}

