/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: spi_reg.c 698 2019-12-28 23:52:19Z roi $
 */
/*
 * 
 *  SPI쥬ɥ饤дؿ
 *
 */
#include "kernel.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdlib.h>
#include <string.h>
#include "device.h"
#include "spi.h"

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ
 *  parameter3  length: 
 *  return ER
 */
ER
spi_transmit(SPI_Handle_t *hspi, uint8_t *ptxData, uint16_t length)
{
	uint32_t *pTx;
	int32_t  ss;
	uint32_t i;
	ER ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	ss = hspi->Init.SsNo;
	if(hspi->hdmatx != NULL){
		pTx = malloc(length * sizeof(uint32_t));
		if(pTx == NULL)
			return E_OBJ;
		for(i = 0 ; i < length ; i++)
			pTx[i] = ptxData[i];
	}
	else
		pTx = (uint32_t *)ptxData;

#if SPI_WAIT_TIME != 0
	ercd = spi_core_transmit(hspi, ss, (uint8_t*)pTx, length);
#else
	if((ercd = spi_core_transmit(hspi, ss, (uint8_t*)pTx, length)) == E_OK){
		ercd = spi_wait(hspi, 50);
	}
#endif
	if(hspi->hdmatx != NULL)
		free(pTx);
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ
 *  parameter3  length: 
 *  return ER
 */
ER
spi_receive(SPI_Handle_t *hspi, uint8_t *pdata, uint16_t length)
{
	uint32_t *pRx;
	int32_t  ss;
	uint32_t i;
	ER  ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	ss = hspi->Init.SsNo;
	pRx = malloc(length * sizeof(uint32_t));
	if(pRx == NULL)
		return E_OBJ;
#if SPI_WAIT_TIME != 0
	ercd = spi_core_receive(hspi, ss, pRx, length);
#else
	if((ercd = spi_core_receive(hspi, ss, pRx, length)) == E_OK)
		ercd = spi_wait(hspi, 50);
#endif
	for(i = 0 ; i < length ; i++)
		pdata[i] = pRx[i];
	free(pRx);
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  ptxdata: ХåեؤΥݥ
 *  parameter3  prxdata: ХåեؤΥݥ
 *  parameter4  length: 
 *  return ER
 */
ER
spi_transrecvx(SPI_Handle_t *hspi, uint8_t *ptxdata, uint8_t *prxdata, uint16_t length)
{
	uint32_t *pTx, *pRx;
	int32_t  ss;
	uint32_t i;
	ER  ercd = E_OK;

	if(hspi == NULL)
		return E_PAR;
	ss = hspi->Init.SsNo;
	if(hspi->hdmatx != NULL){
		pRx = malloc(length * sizeof(uint32_t) * 2);
		if(pRx == NULL)
			return E_OBJ;
		pTx = pRx + length;
		for(i = 0 ; i < length ; i++)
			pTx[i] = ptxdata[i];
	}
	else{
		pRx = malloc(length * sizeof(uint32_t));
		if(pRx == NULL)
			return E_OBJ;
		pTx = (uint32_t *)ptxdata;
	}
#if SPI_WAIT_TIME != 0
	ercd = spi_core_transrecv(hspi, ss, (uint8_t *)pTx, (uint8_t *)pRx, length);
#else
	if((ercd = spi_core_transrecv(hspi, ss, (uint8_t *)pTx, (uint8_t *)pRx, length)) == E_OK)
		ercd = spi_wait(hspi, 50);
#endif
	for(i = 0 ; i < length ; i++)
		prxdata[i] = pRx[i];
	free(pRx);
	return ercd;
}

