/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: wifiuart.c,v 1.1 2020/04/18 20:06:33 roi Exp $
 */
/*
 *  STM32F7XX WIFIUARTɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include <target_syssvc.h>
#include "wifiuart.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  UARTݡID֥åФΥޥ
 */
#define INDEX_UART(uartid)      ((uint_t)((uartid) - 1))


/*
 *  CLOCK
 */
#define SOURCE_PCLK2        0
#define SOURCE_SYSCLK       1
#define SOURCE_HSI          2
#define SOURCE_LSE          3

/*
 *  ӥåĹѥ᡼
 */
#define USART_WordLength_8b 0x0000
#define USART_WordLength_9b USART_CR1_M

/*
 *  ȥåץӥåȥѥ᡼
 */ 
#define USART_StopBits_1    0x0000
#define USART_StopBits_0_5  USART_CR2_STOP_0
#define USART_StopBits_2    USART_CR2_STOP_1
#define USART_StopBits_1_5  USART_CR2_STOP

/*
 *  ѥƥѥ᡼
 */
#define USART_Parity_No     0x0000
#define USART_Parity_Even   USART_CR1_PCE
#define USART_Parity_Odd    (USART_CR1_PCE | USART_CR1_PS) 

/*
 * ϡɥեѥ᡼
 */
#define USART_HardwareFlowControl_None       0x0000
#define USART_HardwareFlowControl_RTS        USART_CR3_RTSE
#define USART_HardwareFlowControl_CTS        USART_CR3_CTSE
#define USART_HardwareFlowControl_RTS_CTS    (USART_CR3_RTSE | USART_CR3_CTSE)


#define ISR_ACK             (USART_ISR_TEACK | USART_ISR_REACK)
#define CR1_CLEAR_MASK      (USART_CR1_M | USART_CR1_PCE | USART_CR1_PS | USART_CR1_TE | USART_CR1_RE | USART_CR1_OVER8)
#define CR3_CLEAR_MASK      (USART_CR3_RTSE | USART_CR3_CTSE | USART_CR3_ONEBIT)

#define UART_TIMEOUT_VALUE  0x1FFFFFF

#ifndef QUEUE_BUFSIZE
#define QUEUE_BUFSIZE  1024
#endif

#define LIMIT1_SIZE    ((QUEUE_BUFSIZE*7)/8)
#define LIMIT2_SIZE    ((QUEUE_BUFSIZE*6)/8)

typedef struct {
	uintptr_t   base;
	uintptr_t   txbase;
	uintptr_t   rxbase;
	uint32_t    enroffset;
	uint32_t    enrvalue;
	uint32_t    rstoffset;
	uint32_t    rstvalue;
	uint32_t    txvalue;
	uint32_t    rxvalue;
	uint32_t    intno;
	uint8_t     txpin;
	uint8_t     rxpin;
	uint8_t     cshift;
	uint8_t     afvalue;
}PortData_t;

static queue_t txque[NUM_WIFIUARTPORT];
static queue_t rxque[NUM_WIFIUARTPORT];
static uint8_t txbuffer[NUM_WIFIUARTPORT][QUEUE_BUFSIZE];
static uint8_t rxbuffer[NUM_WIFIUARTPORT][QUEUE_BUFSIZE];

static const PortData_t uart_base_table[NUM_WIFIUARTPORT] = {
	{USART2_BASE, TADR_U2T_GPIO_BASE, TADR_U2R_GPIO_BASE,
	 TOFF_U2_APBNER, ENABLE_U2_PORT, TOFF_U2_APBRSTR, RESET_U2_PORT,
	 ENABLE_U2T_GPIO, ENABLE_U2R_GPIO,
	 INTNO_SIO2, TX2_PINPOS, RX2_PINPOS, U2_CLOCK_SHIFT, GPIO_U2_AF}
};

static WIFIUART_Handle_t Huart[NUM_WIFIUARTPORT];

/*
 *  GPIO AFؿ
 */
static void
setup_gpio_source(uint32_t base, uint8_t source, uint8_t select)
{
	uint32_t tmpreg, regoff;
	regoff = TOFF_GPIO_AFR0+((source>>3) * 4);
	tmpreg = (sil_rew_mem((uint32_t *)(base+regoff)) & ~(0xF << ((source & 0x7) * 4)))
				| (select << ((source & 0x7) * 4));
	sil_wrw_mem((uint32_t *)(base+regoff), tmpreg);
}

/*
 *  ߵ
 */
static void
wifiuart_enable_send(WIFIUART_Handle_t *huart)
{
	sil_orw_mem((uint32_t *)(huart->base+TOFF_USART_CR1), USART_CR1_TXEIE);
}

/*
 *  ߶ػ
 */
static void
wifiuart_disable_send(WIFIUART_Handle_t *huart)
{
	sil_andw_mem((uint32_t *)(huart->base+TOFF_USART_CR1), USART_CR1_TXEIE);
}

/*
 *  RTS
 */
static void
wifiuart_enable_rts(WIFIUART_Handle_t *huart)
{
	if(huart->Init.rtspin < 0)
		return;
	digitalWrite(huart->Init.rtspin, 0);
}

/*
 *  RTSԲ
 */
static void
wifiuart_disable_rts(WIFIUART_Handle_t *huart)
{
	if(huart->Init.rtspin < 0)
		return;
	digitalWrite(huart->Init.rtspin, 1);
}

/*
 *  CTSå
 */
static bool_t
wifiuart_sens_cts(WIFIUART_Handle_t *huart)
{
	if(huart->Init.ctspin < 0)
		return true;
	return (bool_t)(digitalRead(huart->Init.ctspin) == 0);
}

/*
 *  WIFI-UARTߥϥɥ
 */
void
wifiuart_isr(intptr_t exinf)
{
	WIFIUART_Handle_t *huart = &Huart[INDEX_UART(exinf)];
	uint32_t base = huart->base;
	uint32_t cr1, cr3, isr;
	uint32_t tmp;
	queue_t  *que;
	int      bsize;

	cr1 = sil_rew_mem((uint32_t *)(base+TOFF_USART_CR1));
	cr3 = sil_rew_mem((uint32_t *)(base+TOFF_USART_CR3));
	isr = sil_rew_mem((uint32_t *)(base+TOFF_USART_ISR));

	if((isr & USART_ISR_PE) != 0 && (cr3 & USART_CR3_EIE) != 0){
		sil_wrw_mem((uint32_t *)(base+TOFF_USART_ICR), USART_ICR_PECF);
	}
	if((isr & USART_ISR_FE) != 0 && (cr3 & USART_CR3_EIE) != 0){
		sil_wrw_mem((uint32_t *)(base+TOFF_USART_ICR), USART_ICR_FECF);
	}
	if((isr & USART_ISR_NE) != 0 && (cr3 & USART_CR3_EIE) != 0){
		sil_wrw_mem((uint32_t *)(base+TOFF_USART_ICR), USART_ICR_NCF);
	}
	if((isr & USART_ISR_ORE) != 0 && (cr3 & USART_CR3_EIE) != 0){
		sil_wrw_mem((uint32_t *)(base+TOFF_USART_ICR), USART_ICR_ORECF);
	}
	if((cr1 & USART_CR1_RXNEIE) != 0 && (isr & USART_ISR_RXNE) != 0){
		que = huart->rque;
		bsize = que->top - que->tail;
		if(bsize < 0)
			bsize += que->size;
		if(bsize < (que->size-1)){
			tmp = sil_rew_mem((uint32_t *)(base+TOFF_USART_RDR)) & 0xFF;
			que->buffer[que->top++] = (uint8_t)tmp;
			if(que->top >= que->size)
				que->top = 0;
			if(huart->rxcallback != NULL)
				huart->rxcallback(huart);
		}
		if(bsize >= LIMIT1_SIZE)
			wifiuart_disable_rts(huart);
	}
	if((cr1 & USART_CR1_TXEIE) != 0 && (isr & USART_ISR_TXE) != 0){
		que = huart->tque;
		if(wifiuart_sens_cts(huart) && que->top != que->tail){
			sil_wrw_mem((uint32_t *)(base+TOFF_USART_TDR), (uint32_t)que->buffer[que->tail++]);
			if(que->tail >= que->size)
				que->tail -= que->size;
			if(que->top == que->tail)
				wifiuart_disable_send(huart);
		}
		else
			wifiuart_disable_send(huart);
	}
}

/*
 *  ĥUART
 */
WIFIUART_Handle_t *
wifiuart_init(ID portid, WIFIUART_Init_t *ini)
{
	WIFIUART_Handle_t *huart;
	const PortData_t  *pd;
	uint32_t no = INDEX_UART(portid);
	uint32_t base, txbase, rxbase;
	uint32_t txpin, rxpin, tmp;

	if(portid < 0 || portid > NUM_WIFIUARTPORT || ini == NULL)
		return NULL;

	txque[no].size = QUEUE_BUFSIZE;
	txque[no].top  = 0;
	txque[no].tail = 0;
	txque[no].buffer = txbuffer[no];
	rxque[no].size = QUEUE_BUFSIZE;
	rxque[no].top  = 0;
	rxque[no].tail = 0;
	rxque[no].buffer = rxbuffer[no];

	pd = &uart_base_table[no];
	base   = (uint32_t)pd->base;
	txbase = (uint32_t)pd->txbase;
	rxbase = (uint32_t)pd->rxbase;
	txpin  = (uint32_t)pd->txpin;
	rxpin  = (uint32_t)pd->rxpin;
	huart  = &Huart[no];
	memcpy((void *)&huart->Init, ini, sizeof(WIFIUART_Init_t));
	huart->base = base;
	huart->no   = no;
	huart->tque = &txque[no];
	huart->rque = &rxque[no];

	if(ini->rtspin >= 0)
		wifipinMode(ini->rtspin, GPIO_MODE_OUTPUT, ini->pull, ini->otype);
	if(ini->ctspin >= 0)
		wifipinMode(ini->ctspin, GPIO_MODE_INPUT, ini->pull, ini->otype);

	sil_andw_mem((uint32_t *)(base+TOFF_USART_CR1), USART_CR1_UE);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR1), 0x0);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR2), 0x0);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR3), 0x0);

	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset));
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset), tmp | pd->rstvalue);
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset), tmp & ~(pd->rstvalue));

	if(pd->cshift == 0){	/* USART1ξ硢SYSLOCK */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_DCKCFGR2), RCC_DCKCFGR2_USART1SEL, RCC_DCKCFGR2_USART1SEL_0);
	}
	else if(pd->cshift == 10){	/* F723/USART6ξ硢SYSLOCK */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_DCKCFGR2), RCC_DCKCFGR2_USART6SEL, RCC_DCKCFGR2_USART6SEL_0);
	}

	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+pd->enroffset), pd->enrvalue);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), pd->txvalue);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), pd->rxvalue);

	sil_modw_mem((uint32_t *)(txbase+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (txpin*2)), (GPIO_MODE_AF << (txpin*2)));
	sil_modw_mem((uint32_t *)(txbase+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (txpin*2)), (GPIO_SPEED_HIGH << (txpin*2)));
	sil_modw_mem((uint32_t *)(txbase+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << txpin), (GPIO_OTYPE_PP << txpin));
	sil_modw_mem((uint32_t *)(txbase+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (txpin*2)), (GPIO_PULLUP << (txpin*2)));
	sil_modw_mem((uint32_t *)(rxbase+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (rxpin*2)), (GPIO_MODE_AF << (rxpin*2)));
	sil_modw_mem((uint32_t *)(rxbase+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (rxpin*2)), (GPIO_SPEED_HIGH << (rxpin*2)));
	sil_modw_mem((uint32_t *)(rxbase+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << rxpin), (GPIO_OTYPE_PP << rxpin));
	sil_modw_mem((uint32_t *)(rxbase+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (rxpin*2)), (GPIO_PULLUP << (rxpin*2)));
	setup_gpio_source(txbase, txpin, pd->afvalue);
	setup_gpio_source(rxbase, rxpin, pd->afvalue);

	tmp = sil_rew_mem((uint32_t *)(base+TOFF_USART_CR2)) & ~USART_CR2_STOP;
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR2), tmp | USART_StopBits_1);
	tmp = sil_rew_mem((uint32_t *)(base+TOFF_USART_CR1)) & ~CR1_CLEAR_MASK;
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR1), tmp | USART_WordLength_8b | USART_Parity_No | USART_CR1_RE | USART_CR1_TE);
	tmp = sil_rew_mem((uint32_t *)(base+TOFF_USART_CR3)) & ~CR3_CLEAR_MASK;
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR3), tmp | USART_HardwareFlowControl_None);

	switch((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_DCKCFGR2)) >> pd->cshift) & 3){
	case SOURCE_HSI:
		tmp = HSI_VALUE;
		break;
	case SOURCE_SYSCLK:
		tmp = SysFreHCLK;
		break;
	case SOURCE_LSE:
		tmp = LSE_VALUE;
		break;
	case SOURCE_PCLK2:
	default:
		if(pd->cshift == 10)	/* USART6,1ξ PCLK2 */
			tmp = SysFrePCLK2;
		else
			tmp = SysFrePCLK1;
		break;
	}
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_BRR), tmp/ini->bps);
	sil_orw_mem((uint32_t *)(base+TOFF_USART_CR1), USART_CR1_UE);

	/*
	 * RX/TX ACKåȤޤԤ
	 */
	tmp = 0;
    while((sil_rew_mem((uint32_t *)(base+TOFF_USART_ISR)) & ISR_ACK) != ISR_ACK){
		/* Check for the Timeout */
		if(tmp >=  UART_TIMEOUT_VALUE){
			/* Disable TXE, RXNE, PE and ERR (Frame error, noise error, overrun error) interrupts for the interrupt process */
			sil_andw_mem((uint32_t *)(base+TOFF_USART_CR1), (USART_CR1_TXEIE | USART_CR1_RXNEIE | USART_CR1_PEIE));
			sil_andw_mem((uint32_t *)(base+TOFF_USART_CR3), USART_CR3_EIE);
			return NULL;
		}
		sil_dly_nse(1000);
		tmp++;
	}

	sil_orw_mem((uint32_t *)(base+TOFF_USART_CR1), USART_CR1_PEIE);
	sil_orw_mem((uint32_t *)(base+TOFF_USART_CR3), USART_CR3_EIE);
	sil_orw_mem((uint32_t *)(base+TOFF_USART_CR1), USART_CR1_RXNEIE);

	/*
	 *  ꥢI/OߤΥޥ롥
	 */
	ena_int(pd->intno);
	wifiuart_enable_rts(huart);
	return huart;
}

/*
 *  WIFI-UART̵
 */
ER
wifiuart_deinit(WIFIUART_Handle_t *huart)
{
	const PortData_t  *pd;
	uint32_t base, tmp;

	if(huart == NULL)
		return E_PAR;
	/*
	 *  ϡɥΥꥻå
	 */
	pd = &uart_base_table[huart->no];
	base = huart->base;
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_ICR), 0xFFFFFFFF);
	sil_andw_mem((uint32_t *)(base+TOFF_USART_CR1), USART_CR1_UE);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR1), 0x0);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR2), 0x0);
	sil_wrw_mem((uint32_t *)(base+TOFF_USART_CR3), 0x0);
	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset));
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset), tmp | pd->rstvalue);
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+pd->rstoffset), tmp & ~(pd->rstvalue));
	return E_OK;
}

/*
 *  WIFI-UART
 */
int
wifiuart_recev(WIFIUART_Handle_t *huart, unsigned char *buf, int len)
{
	queue_t  *que;
	int      bsize, rlen = 0;
	uint32_t i;

	if(huart == NULL)
		return 0;
	que = huart->rque;
	while(len > rlen){
		bsize = que->top - que->tail;
		if(bsize == 0)
			break;
		else if(bsize < 0)
			bsize = que->size - que->tail;
		if(bsize > (len - rlen))
			bsize = len - rlen;
		for(i = 0 ; i < bsize ; i++)
			*buf++ = que->buffer[que->tail++];
		if(que->tail >= que->size)
			que->tail -= que->size;
		rlen += bsize;
	}
	return rlen;
}

/*
 *  WIFI-UART
 */
int
wifiuart_recsize(WIFIUART_Handle_t *huart)
{
	queue_t  *que;
	int      bsize;

	if(huart == NULL)
		return 0;
	que = huart->rque;
	bsize = que->top - que->tail;
	if(bsize < 0)
		bsize += que->size;
	if(bsize < LIMIT2_SIZE)
		wifiuart_enable_rts(huart);
	return bsize;
}

/*
 *  WIFI-UART
 */
int
wifiuart_send(WIFIUART_Handle_t *huart, unsigned char *buf, int len)
{
	queue_t  *que;
	int      bsize, i;

	if(huart == NULL)
		return 0;
	que = (queue_t *)huart->tque;
	for(i = 0 ; i < len ; i++){
		bsize = que->top - que->tail;
		if(bsize < 0)
			bsize += que->size;
		if(bsize < (que->size-1)){
			que->buffer[que->top++] = *buf++;
			if(que->top >= que->size)
				que->top -= que->size;
		}
		else{
			if(wifiuart_sens_cts(huart))
				wifiuart_enable_send(huart);
			dly_tsk(20);
		}
	}
	if(i != 0 && wifiuart_sens_cts(huart))
		wifiuart_enable_send(huart);
	return i;
}

/*
 *  CTS󥹥å
 */
bool_t
wifiuart_cts_sens(WIFIUART_Handle_t *huart)
{
	queue_t  *que;

	if(huart == NULL)
		return false;
	que = huart->tque;
	if(que->top == que->tail || !wifiuart_sens_cts(huart))
		return false;
	wifiuart_enable_send(huart);
	return true;
}

/*
 *  GPIOؿ
 */
void
wifipinMode(uint8_t no, uint32_t mode, uint32_t pull, uint32_t otype)
{
	Arduino_PortControlBlock *pgcb = getGpioTable(DIGITAL_PIN, no);
	GPIO_Init_t GPIO_Init_Data;

	if(pgcb == NULL)
		return;
	pinClock(no);
	GPIO_Init_Data.mode  = mode;
	GPIO_Init_Data.pull  = pull;
	GPIO_Init_Data.otype = otype;
	GPIO_Init_Data.speed = GPIO_SPEED_FAST;
	gpio_setup(pgcb->giobase, &GPIO_Init_Data, pgcb->giopin);
}

