/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2019-01-21 18:55:48Z roi $
 */
/*
 * STM32G0xxѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#if !defined(HSE_VALUE)
#define HSE_VALUE               8000000		/* Value of the External oscillator in Hz */
#endif /* HSE_VALUE */
#if !defined(HSI_VALUE)
#define HSI_VALUE               16000000	/* Value of the Internal oscillator in Hz*/
#endif /* HSI_VALUE */

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

uint_t dipsw_value;
void (*exti_func[NUM_EXTI_FUNC])(void) = { 0 };


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVT           0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	0,							/* index 4 */
	TADR_GPIOF_BASE				/* index 5 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t iocurrent = 1<<pin;
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	/*
	 *  GPIO⡼
	 */
	/* 륿͡ȡե󥯥󡦥⡼ */
	temp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4));
	temp &= ~((uint32_t)0xF << ((uint32_t)(pin & (uint32_t)0x07) * 4)) ;
	if(init->mode == GPIO_MODE_AF)
		temp |= ((uint32_t)(init->alternate) << (((uint32_t)pin & (uint32_t)0x07) * 4));
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4), temp);

	/*  ϥ⡼ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (pin * 2)), ((init->mode & GPIO_MODE) << (pin * 2)));

	/*  ϥ⡼ */
	if(init->mode == GPIO_MODE_OUTPUT || init->mode == GPIO_MODE_AF){
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (pin * 2)), (init->speed << (pin * 2)));
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << pin), (init->otype << pin));
	}

	/*  ץ륢åסץ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (pin * 2)), (init->pull << (pin * 2)));

	/*
	 *  EXTI⡼
	 */
	if((init->mode & EXTI_MODE) == EXTI_MODE){
		/*  SYSCFGå */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR2), RCC_APBENR2_SYSCFGEN);
		temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR2));

		temp = sil_rew_mem((uint32_t *)((TADR_EXTI_BASE+TOFF_EXTI_EXTICR+(pin / 1)) & ~3));
		temp &= ~(0x0F << (8 * (pin & 0x03)));
		temp |= (index << (8 * (pin & 0x03)));
		sil_wrw_mem((uint32_t *)((TADR_EXTI_BASE+TOFF_EXTI_EXTICR+(pin / 1)) & ~3), temp);

		if((init->mode & GPIO_MODE_IT) == GPIO_MODE_IT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);

		if((init->mode & GPIO_MODE_EVT) == GPIO_MODE_EVT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);

		if((init->mode & RISING_EDGE) == RISING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);

		if((init->mode & FALLING_EDGE) == FALLING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
	}
}


/*
 *  DMACؿ
 */
#define DMA_TRS_TIMEOUT 2000	/* 2 */

#define DMAmux_Channel(h)   (TADR_DMAMUX1_BASE+((h)->chid2)*4)
#define DMAmux_RreqGEN(h)   (TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GEN+((h)->genreqno)*4)

static const uint32_t dma_base[] = {
	TADR_DMA1_CH1_BASE,
	TADR_DMA1_CH2_BASE,
	TADR_DMA1_CH3_BASE,
	TADR_DMA1_CH4_BASE,
	TADR_DMA1_CH5_BASE,
	TADR_DMA1_CH6_BASE,
	TADR_DMA1_CH7_BASE
};

#define NUM_CHDMA       ((sizeof(dma_base))/sizeof(uint32_t))


static DMA_Handle_t *pDmaHandler[NUM_CHDMA];

/*
 *  CHANNEL DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0;
	uint32_t i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid  = i % 7;
	hdma->chid2 = i;
	pDmaHandler[i] = hdma;
	hdma->base = hdma->cbase & ~0x3FF;
	index = (i % 7) * 4;
	hdma->chnum = (((hdma->base & 0xFF) - 8) / 20) & 0x1F;

	/* DMAꥯ */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHBENR), RCC_AHBENR_DMAEN);

	if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
	    hdma->Init.Request = DMA_REQUEST_MEM2MEM;

	/* DMAMUX˥ڥեͥ */
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID);
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->Init.Request > 0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		index = hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		hdma->genreqno = index - 1;
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}
	else
		hdma->genreqno = -1;

	/* DMA-CCR쥸 */
	tmp = sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR));

	/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
	tmp &= ((uint32_t)~(DMA_CCR_DIR  | DMA_CCR_MEM2MEM | DMA_CCR_PINC  | \
                        DMA_CCR_MINC | DMA_CCR_PSIZE   | DMA_CCR_MSIZE | \
                        DMA_CCR_CIRC | DMA_CCR_PL));

	/* DMA⡼ɤƥݥꥣ */
	tmp |=   hdma->Init.Direction        |
          hdma->Init.PeriphInc           | hdma->Init.MemInc           |
          hdma->Init.PeriphDataAlignment | hdma->Init.MemDataAlignment |
          hdma->Init.Mode                | hdma->Init.Priority;

	/* DMA⡼ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), tmp);

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid = i % 7;
	pDmaHandler[i] = NULL;
	index = (i % 7) * 4;

	/* DMA֥͡ */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), 0);

	/* ƥե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_ISR_GIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));

	/* DMAꥯȥꥢ */
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), 0);

	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	/* ɬפʤꥯȥͥ졼Υѥ᡼Ԥ */
	if(hdma->Init.Request >  0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	hdma->status = DMA_STATUS_BUSY;

	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->genreqno >= 0)	/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));

	/* ǡĹ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), DataLength);

	if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
		/* ꤫ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), DstAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), SrcAddress);
	}
	else{	/* ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), SrcAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), DstAddress);
	}

	/* ߥ֥͡ */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), (DMA_CCR_TCIE | DMA_CCR_HTIE | DMA_CCR_TEIE));

    /* DMAMUX 󥯥ʥ */
	if((sil_rew_mem((uint32_t *)(DMAmux_Channel(hdma))) & DMAMUX_CxCR_SE) != 0){
		/* DMAMUX 󥯥ʥСߤ */
		sil_orw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
    }
	if(hdma->genreqno >= 0){
		/* DMAMUXꥯȥͥ졼  overrunߤͭ */
		sil_orw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
	}

	/* DMA */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	return E_OK;
}

/*
 *  STREAM DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	uint32_t tick = 0;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	/* DMAԤ */
	while((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_EN) != 0){
	    /* Check for the Timeout */
		if(tick > DMA_TRS_TIMEOUT){
			/* ॢȥ顼 */
			hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
			return E_TMOUT;
		}
    	dly_tsk(1);
	}
	hdma->status = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	uint32_t index = hdma->chid * 4;

	/*
	 *  ž顼
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TEIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TEIE) != 0){
			/* ž顼ߥꥢ */
			sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TEIE);
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* ž顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
			/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ž顼֤ */
			hdma->ErrorCode |= DMA_ERROR_TE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}

	/*
	 *  ϡžλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_HTIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_HTIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* ϡžߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_HTIE);
			}
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* ϡžλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ϡžơѹ */
			hdma->status = DMA_STATUS_READY_HMEM0;
			/* ϡžλХå */
			if(hdma->xferhalfcallback != NULL)
				hdma->xferhalfcallback(hdma);
		}
	}

	/*
	 *  žλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TCIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TCIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* žߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TCIE);
			}
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* žλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ơλ */
			hdma->ErrorCode = DMA_ERROR_NONE;
			hdma->status = DMA_STATUS_READY_MEM0;
			/* žλХå */
			if(hdma->xfercallback != NULL)
				hdma->xfercallback(hdma);
		}
	}
}

/*
 *  STREAM DMA ߥӥ롼
 */
void
channel_dma_isr(intptr_t exinf)
{
	uint32_t index = (uint32_t)exinf;

	if(index == IRQ_VECTOR_DMA1_CH1){
		if(pDmaHandler[0] != NULL)
			dma_inthandler(pDmaHandler[0]);
	}
	else if(index == IRQ_VECTOR_DMA1_CH_LOW){
		if(pDmaHandler[1] != NULL)
			dma_inthandler(pDmaHandler[1]);
		if(pDmaHandler[2] != NULL)
			dma_inthandler(pDmaHandler[2]);
	}
	else if(index == IRQ_VECTOR_DMA1_CH_HIGH){
		if(pDmaHandler[3] != NULL)
			dma_inthandler(pDmaHandler[3]);
		if(pDmaHandler[4] != NULL)
			dma_inthandler(pDmaHandler[4]);
		if(pDmaHandler[5] != NULL)
			dma_inthandler(pDmaHandler[5]);
		if(pDmaHandler[6] != NULL)
			dma_inthandler(pDmaHandler[6]);
	}
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint32_t mcount, mcounter, wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;
	if(hini->Window > 127 || hini->Window < 64)
		return E_PAR;
	if(hini->Counter > 127 || hini->Counter < 64)
		return E_PAR;

	mcount    = hini->MCounter;
	mcounter  = 0;
	wdogcount = hini->Counter;
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR1), RCC_APBENR1_WWDGEN);
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	if(mcount == 0)
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_WDGTB | hini->Window));
	else
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_EWI | WWDG_CFR_WDGTB | hini->Window));
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), (WWDG_CR_WDGA | wdogcount));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR1), RCC_APBENR1_WWDGEN);
	wdogcount = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDG_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDG_CFR)) & WWDG_CFR_EWI) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDG_SR)) & WWDG_SR_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDG_SR), WWDG_SR_EWIF);
		if(mcounter++ < mcount)
			sil_wrw_mem((uint32_t *)(base+TOFF_WWDG_CR), wdogcount);
	}
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    clkbit;
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
	{ RCC_IOPENR_IOPAEN, TADR_GPIOA_BASE, PINPOSITION5, GPIO_PULLUP}
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))

/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_IOPENR), pled->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (1 << pled->pinpos)<<16);
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_IDR)) & (1<<(pled->pinpos))) != 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t reg1, reg2, i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		reg1 = reg2 = 0;
		if((led_data & (1<<i)) != 0)
			reg2 = 1 << pled->pinpos;
		else
			reg1 = 1 << pled->pinpos;

		/* 񤭹ߥǡƽ񤭹 */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (reg1<<16) | reg2);
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


#define TADR_PSW_BASE   TADR_GPIOC_BASE
#define PSWGIOPEN       RCC_IOPENR_GPIOCEN
#define	PSW_PIN         PINPOSITION13

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	GPIO_Init_t GPIO_Init_Data;
	volatile uint32_t temp;

	/* ϥݡȤ */
	/* Enable the BUTTON Clock */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_IOPENR), PSWGIOPEN);
	temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_IOPENR));
	(void)(temp);
	/* ⡼ */
	GPIO_Init_Data.mode      = GPIO_MODE_IT_FALLING;
	/* ץ륢å ץ */
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	gpio_setup(TADR_PSW_BASE, &GPIO_Init_Data, PSW_PIN);
}

/*
 * PUSHåХåؿ
 */
void setup_sw_func(intptr_t exinf)
{
	exti_func[PSW_PIN] = (void (*)(void))exinf;
}

/*
 * EXTI0ߥϥɥ
 */
void
exti0_handler(void)
{
	uint32_t irstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1));
	uint32_t ifstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1));
	uint32_t istatus  = irstatus | ifstatus;
	uint32_t i;

	if((irstatus & EXTI0_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1), irstatus & EXTI0_MASK);
	if((ifstatus & EXTI0_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1), ifstatus & EXTI0_MASK);
	for(i = 0 ; i < NUM_EXTI0_FUNC ; i++){
		if((istatus & (1<<i)) != 0){
			if(exti_func[i] != NULL)
				(exti_func[i])();
		}
	}
}

/*
 * EXTI2ߥϥɥ
 */
void
exti2_handler(void)
{
	uint32_t irstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1));
	uint32_t ifstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1));
	uint32_t istatus  = irstatus | ifstatus;
	uint32_t i;

	if((irstatus & EXTI2_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1), irstatus & EXTI2_MASK);
	if((ifstatus & EXTI2_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1), ifstatus & EXTI2_MASK);
	for(i = 0 ; i < NUM_EXTI2_FUNC ; i++){
		if((istatus & (1<<(i+EXTI2_BASENO))) != 0){
			if(exti_func[i+EXTI2_BASENO] != NULL)
				(exti_func[i+EXTI2_BASENO])();
		}
	}
}

/*
 * EXTI4-10ߥϥɥ
 */
void
exti4_handler(void)
{
	uint32_t irstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1));
	uint32_t ifstatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1));
	uint32_t istatus  = irstatus | ifstatus;
	uint32_t i;

	if((irstatus & EXTI4_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1), irstatus & EXTI4_MASK);
	if((ifstatus & EXTI4_MASK) != 0)
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1), ifstatus & EXTI4_MASK);
	for(i = 0 ; i < NUM_EXTI4_FUNC ; i++){
		if((istatus & (1<<(i+EXTI4_BASENO))) != 0){
			if(exti_func[i+EXTI4_BASENO] != NULL)
				(exti_func[i+EXTI4_BASENO])();
		}
	}
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

