/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: i2c.c 698 2019-12-20 18:06:50Z roi $
 */
/*
 *
 *  I2Cɥ饤дؿ(STM32G4XX)
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "i2c.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  I2CݡID֥åФΥޥ
 */
#define INDEX_I2C(i2cid)        ((uint_t)((i2cid) - 1))

/*
 *  I2CAF
 */
#define GPIO_AF4_I2C1          ((uint8_t)0x04)  /* I2C1 Alternate Function mapping */
#define GPIO_AF4_I2C2          ((uint8_t)0x04)  /* I2C2 Alternate Function mapping */
#define GPIO_AF4_I2C3          ((uint8_t)0x04)  /* I2C3 Alternate Function mapping */

#define I2C_TIMEOUT_FLAG            35			/* 35 ms */
#define I2C_TIMEOUT_BUSY_FLAG       10000		/* 10 s  */

#define TIMING_CLEAR_MASK           0xF0FFFFFF	/* I2C TIMING clear register Mask */

/*
 *  ž
 */
#define MAX_NBYTE_SIZE              255

/*
 *  I2C_RELOAD⡼
 */
#define  I2C_RELOAD_MODE            I2C_CR2_RELOAD
#define  I2C_AUTOEND_MODE           I2C_CR2_AUTOEND
#define  I2C_SOFTEND_MODE           0x00000000

/*
 *  I2C_XFEROPTION
 */
#define I2C_FIRST_FRAME             I2C_SOFTEND_MODE
#define I2C_FIRST_AND_NEXT_FRAME    (I2C_RELOAD_MODE | I2C_SOFTEND_MODE)
#define I2C_NEXT_FRAME              (I2C_RELOAD_MODE | I2C_SOFTEND_MODE)
#define I2C_FIRST_AND_LAST_FRAME    I2C_AUTOEND_MODE
#define I2C_LAST_FRAME              I2C_AUTOEND_MODE
#define I2C_NO_OPTION_FRAME         0xFFFF0000

/*
 *  I2C_START_STOP_MODE
 */
#define  I2C_NO_STARTSTOP           (0x00000000U)
#define  I2C_GENERATE_STOP          I2C_CR2_STOP
#define  I2C_GENERATE_START_READ    (I2C_CR2_START | I2C_CR2_RD_WRN)
#define  I2C_GENERATE_START_WRITE   I2C_CR2_START

#define SlaveAddr_SHIFT             7
#define SlaveAddr_MSK               0x06

#define I2C_INT_TX                  (I2C_CR1_TCIE   | I2C_CR1_TXIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ERRIE)
#define I2C_INT_RX                  (I2C_CR1_TCIE   | I2C_CR1_RXIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ERRIE)
#define I2C_INT_LISTEN              (I2C_CR1_ADDRIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ERRIE)
#define I2C_INT_ALL                 (I2C_INT_TX | I2C_INT_RX | I2C_INT_LISTEN)

/*
 *  I2Cݡơ֥
 */
static const I2C_PortControlBlock i2c_pcb[NUM_I2CPORT] = {
  {	TADR_I2C1_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB2ENR),   RCC_AHB2ENR_GPIOBEN, RCC_AHB2ENR_GPIOBEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR1),  RCC_APB1ENR1_I2C1EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_I2C1RST,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	8, GPIO_AF4_I2C1, 9, GPIO_AF4_I2C1 },

  {	TADR_I2C2_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB2ENR),   RCC_AHB2ENR_GPIOAEN, RCC_AHB2ENR_GPIOAEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR1),  RCC_APB1ENR1_I2C2EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_I2C2RST,
	TADR_GPIOA_BASE, TADR_GPIOA_BASE,
	8, GPIO_AF4_I2C2, 9, GPIO_AF4_I2C2 },

  {	TADR_I2C3_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB2ENR),   RCC_AHB2ENR_GPIOFEN, RCC_AHB2ENR_GPIOFEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR1),  RCC_APB1ENR1_I2C3EN,
	(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_I2C3RST,
	TADR_GPIOF_BASE, TADR_GPIOF_BASE,
	3, GPIO_AF4_I2C3, 4, GPIO_AF4_I2C3 }
};


/*
 * I2C ϥɥ
 */
static I2C_Handle_t I2cHandle[NUM_I2CPORT];

/*
 *  I2Cž
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 Size:       ž
 *  parameter4 Mode:       ž⡼
 *  parameter5 Request:    Ⱦ֥ꥯ
 *  return                 ʤ
 */
static void
i2c_transfconfig(I2C_Handle_t *hi2c, uint16_t DevAddress, uint8_t Size, uint32_t Mode, uint32_t Request)
{
	uint32_t tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2));

	/*
	 *  ⡼ɥꥢ
	 */
	tmp &= ~(I2C_CR2_SADD | I2C_CR2_NBYTES | I2C_CR2_RELOAD | I2C_CR2_AUTOEND | I2C_CR2_RD_WRN | I2C_CR2_START | I2C_CR2_STOP);

	/*
	 *  žԤ
	 */
	tmp |= (uint32_t)((DevAddress & I2C_CR2_SADD) | ((Size << 16) & I2C_CR2_NBYTES) | Mode | Request);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), tmp);
}

/*
 *  I2Cǡ쥸եå
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  return              ʤ
 */
static void
i2c_flushtxdr(I2C_Handle_t *hi2c)
{
	/*
	 *  TXISե饰ʤ顢TXDR˥򥻥å
	 */
	if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_TXIS) != 0)
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), 0);

	/*
	 *  TXEե饰դʤ顢EMPTYå
	 */
	if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_TXE) == 0)
		sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR), I2C_ISR_TXE);
}

/*
 *  ISR쥸åԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Flag:    Ԥե饰
 *  parameter3 Timeout: Ԥ(ms)
 *  parameter4 pTimeout:Ԥݥ
 *  return              ｪλE_OK
 */
static ER
i2c_isrflagsetwait(I2C_Handle_t *hi2c, uint32_t Flag, uint32_t Timeout, uint32_t *pTimeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

	*pTimeout = Timeout;
	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & Flag) == 0){
		if(ctime == 0 || tick > ctime){
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	*pTimeout -= tick / 1000;
	return E_OK;
}

/*
 *  I2C ACKǧ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Timeout: Ԥ(ms)
 *  parameter3 pTimeout:Ԥݥ
 *  return              ｪλE_OK
 */
static ER
i2c_checkacknowledgefailed(I2C_Handle_t *hi2c, uint32_t Timeout, uint32_t *pTimeout)
{
	uint32_t tick = 0;

	if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_NACKF) != 0){
		/*
		 *  NACKʤSTOPե饰Ԥ
		 */
		while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_STOPF) == 0){
			if(tick > (Timeout * 1000)){
				hi2c->status= I2C_STATUS_READY;
				*pTimeout -= tick * 1000;
				return E_TMOUT;
			}
			tick++;
			sil_dly_nse(1000);
		}
		*pTimeout -= tick * 1000;

		/*
		 *  NACKե饰ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);

		/*
		 *  STOPե饰ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);

		/*
		 *  쥸ꥢ
		 */
		i2c_flushtxdr(hi2c);

		/*
		 *  ž쥸ꥢ
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), ((I2C_CR2_SADD | I2C_CR2_HEAD10R | I2C_CR2_NBYTES | I2C_CR2_RELOAD | I2C_CR2_RD_WRN)));

		hi2c->ErrorCode = I2C_ERROR_AF;
		hi2c->status= I2C_STATUS_READY;
		return E_SYS;
	}
	return E_OK;
}

/*
 *  I2C TXISե饰åԤ
 *  parameter2 Timeout: Ԥ(ms)
 *  parameter3 pTimeout:Ԥݥ
 *  return              ｪλE_OK
 */
static ER
i2c_waitsettxisflag(I2C_Handle_t *hi2c, uint32_t Timeout, uint32_t *pTimeout)
{
	ER ercd = E_OK;
	uint32_t tick = 0;

	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_TXIS) == 0){
	    /*
		 *  NACKǧ
		 */
	    if((ercd = i2c_checkacknowledgefailed(hi2c, Timeout, pTimeout)) != E_OK)
			return ercd;
		if(tick > (Timeout * 1000)){
			hi2c->ErrorCode |= I2C_ERROR_TIMEOUT;
			hi2c->status = I2C_STATUS_READY;
			return E_TMOUT;
		}
		tick++;
		sil_dly_nse(1000);
	}
	*pTimeout -= tick / 1000;
	return E_OK;
}

/*
 *  BUS READYԤ
 *  parameter1 hi2c:    I2CϥɥؤΥݥ
 *  parameter2 Timeout: Ԥ(ms)
 *  return              ｪλE_OK
 */
static ER
i2c_busreadywait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;
	uint32_t ctime = Timeout * 1000;

    while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_BUSY) != 0){
		if(ctime == 0 || tick > ctime){
			hi2c->status= I2C_STATUS_READY;
			return E_TMOUT;
		}
        sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  I2CžԤ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 Timeout:ॢȻ(ms)
 */
static ER
i2c_transwait(I2C_Handle_t *hi2c, uint32_t Timeout)
{
	uint32_t tick = 0;

	while(hi2c->status != I2C_STATUS_READY){
		if(tick >= Timeout)
			return E_TMOUT;
		if(hi2c->Init.semid != 0)
			twai_sem(hi2c->Init.semid, 1);
		else
			dly_tsk(1);
		tick++;
	}
	if(hi2c->ErrorCode == I2C_ERROR_NONE)
		return E_OK;
	else
		return E_SYS;
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterreadaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	if(MemAddSize == 0)
		return E_OK;

	i2c_transfconfig(hi2c, DevAddress, MemAddSize, I2C_SOFTEND_MODE, I2C_GENERATE_START_WRITE);

	/*
	 *  TXISե饰Ԥ
	 */
	if((ercd = i2c_waitsettxisflag(hi2c, Timeout, &Timeout)) != E_OK){
		if(hi2c->ErrorCode == I2C_ERROR_AF)
			return E_SYS;
		else
			return ercd;
	}

	/*
	 *  ꥢɥ쥹
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		/*
		 *  8ӥåȥꥢɥ쥹
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), (MemAddress & 0xff));
	}
	else{
		/*
		 *  16ӥåȥꥢɥ쥹
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), ((MemAddress >> 8) & 0xff));

		/*
		 *  TXISե饰Ԥ
		 */
		if((ercd = i2c_waitsettxisflag(hi2c, Timeout, &Timeout)) != E_OK){
			if(hi2c->ErrorCode == I2C_ERROR_AF)
				return E_SYS;
			else
				return ercd;
		}
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), (MemAddress & 0xff));
	}

	/*
	 *  TCե饰Ԥ
	 */
	return i2c_isrflagsetwait(hi2c, I2C_ISR_TC, Timeout, &Timeout);
}

/*
 *  ޥ ǥХɥ쥹ȥꥢɥ쥹
 *  parameter1 hi2c:  I2CϥɥؤΥݥ
 *  parameter2 DevAddress: ǥХɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 Timeout:    ॢȻ(ms)
 *  return                 ｪλE_OK
 */
static ER
i2c_masterwriteaddress(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint32_t Timeout)
{
	ER ercd = E_OK;

	if(MemAddSize == 0)
		return E_OK;

	i2c_transfconfig(hi2c,DevAddress,MemAddSize, I2C_RELOAD_MODE, I2C_GENERATE_START_WRITE);

	/*
	 *  TXISե饰Ԥ
	 */
	if((ercd = i2c_waitsettxisflag(hi2c, Timeout, &Timeout)) != E_OK){
		if(hi2c->ErrorCode == I2C_ERROR_AF)
			return E_SYS;
		else
			return ercd;
	}

	/*
	 *  ꥢɥ쥹
	 */
	if(MemAddSize == I2C_MEMADD_SIZE_8BIT){
		/*
		 *  8ӥåȥꥢɥ쥹
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), (MemAddress & 0xff));
	}
	else{
		/*
		 *  16ӥåȥꥢɥ쥹
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), ((MemAddress >> 8) & 0xff));

		/*
		 *  TXISե饰Ԥ
		 */
		if((ercd = i2c_waitsettxisflag(hi2c, Timeout, &Timeout)) != E_OK){
			if(hi2c->ErrorCode == I2C_ERROR_AF)
				return E_SYS;
			else
				return ercd;
		}

		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), (MemAddress & 0xff));
	}

	/*
	 *  TCRե饰Ԥ
	 */
	return i2c_isrflagsetwait(hi2c, I2C_ISR_TCR, Timeout, &Timeout);
}

/*
 *  I2CǥХν
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
I2C_Handle_t *
i2c_init(ID port, I2C_Init_t *ii2c)
{
	GPIO_Init_t GPIO_Init_Data;
	const I2C_PortControlBlock *ipcb;
	I2C_Handle_t *hi2c;
	uint32_t no;

	if(port < I2C1_PORTID || port > NUM_I2CPORT)
		return NULL;
	if(ii2c == NULL)
		return NULL;

	no = INDEX_I2C(port);
	ipcb = &i2c_pcb[no];
	hi2c = &I2cHandle[no];
	if(hi2c->status != I2C_STATUS_RESET)
		return NULL;
	memcpy(&hi2c->Init, ii2c, sizeof(I2C_Init_t));
	hi2c->base  = ipcb->base;
	hi2c->i2cid = port;

	/*
	 *  SCL/SDAԥ/I2Cå
	 */
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit1);
	sil_orw_mem((uint32_t *)ipcb->gioclockbase, ipcb->gioclockbit2);
	sil_orw_mem((uint32_t *)ipcb->i2cclockbase, ipcb->i2cclockbit);

	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	GPIO_Init_Data.otype     = GPIO_OTYPE_OD;
	GPIO_Init_Data.speed     = GPIO_SPEED_FAST;
	GPIO_Init_Data.alternate = ipcb->sclaf;
	gpio_setup(ipcb->giobase1, &GPIO_Init_Data, ipcb->sclpin);

	GPIO_Init_Data.alternate = ipcb->sdaaf;
	gpio_setup(ipcb->giobase2, &GPIO_Init_Data, ipcb->sdapin);


	hi2c->status = I2C_STATUS_BUSY;

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	/*
	 *  I2Cߥ󥰥쥸
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TIMINGR), (hi2c->Init.Timing &  TIMING_CLEAR_MASK));

	/*
	 *  I2CǥХɥ쥹
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR1), I2C_OAR1_OA1EN);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR1), (hi2c->Init.AddressingMode | hi2c->Init.OwnAddress1));

	/*
	 *  I2Cޥ⡼
	 */
	if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_10BIT)
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_ADD10);
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_AUTOEND | I2C_CR2_NACK));

	/*
	 *  I2CǥХɥ쥹
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR2), I2C_DUALADDRESS_ENABLE);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR2), (hi2c->Init.DualAddressMode | hi2c->Init.OwnAddress2 | hi2c->Init.OwnAddress2Masks));

	/*
	 *  I2C CR1
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (hi2c->Init.GeneralCallMode | hi2c->Init.NoStretchMode));

	/*
	 *  I2Cư
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_READY;
	return hi2c;
}

/*
 *  I2CǥХ̵
 *  parameter1 hi2c  I2CϥɥؤΥݥ
 *  return           ｪλE_OK
 */
ER
i2c_deinit(I2C_Handle_t *hi2c)
{
	const I2C_PortControlBlock *ipcb;

	if(hi2c == NULL){
		return E_PAR;
	}

	hi2c->status = I2C_STATUS_BUSY;
	ipcb = &i2c_pcb[INDEX_I2C(hi2c->i2cid)];

	/*
	 *  I2C
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_CR1_PE);

	/*
	 *  I2Cåꥻå
	 */
	sil_orw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);
	sil_andw_mem((uint32_t *)ipcb->i2cresetbase, ipcb->i2cresetbit);

	/*
	 *  I2Cå
	 */
	sil_andw_mem((uint32_t *)ipcb->i2cclockbase, ipcb->i2cclockbit);

	hi2c->ErrorCode = I2C_ERROR_NONE;
	hi2c->status = I2C_STATUS_RESET;
	return E_OK;
}

/*
 *  I2C졼ּ
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slaveread(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

    /*
	 *  ɥ쥹Υåͭ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_NACK);

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;
    hi2c->XferCount2 = Size;
	hi2c->XferOptions = I2C_NO_OPTION_FRAME;

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (I2C_INT_RX | I2C_INT_LISTEN));
	ercd = i2c_transwait(hi2c, 5000);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2C졼
 *  parameter1 hi2c:   I2CϥɥؤΥݥ
 *  parameter2 pData:  ɤ߽ФХåեؤΥݥ
 *  parameter3 Size:   ɤ߽Ф
 *  return             ｪλʤE_OK
 */
ER
i2c_slavewrite(I2C_Handle_t *hi2c, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;

	if(pData == NULL || Size == 0)
		return E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

    /*
	 *  ɥ쥹Υåͭ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_NACK);

    hi2c->status = I2C_STATUS_BUSY_TX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;
    hi2c->XferCount2 = Size;
    hi2c->XferOptions = I2C_NO_OPTION_FRAME;

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (I2C_INT_TX | I2C_INT_LISTEN));
	ercd = i2c_transwait(hi2c, 5000);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ꡼
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ɤ߽ФХåեؤΥݥ
 *  parameter6 Size:       ɤ߽Ф
 *  return                 ｪλʤE_OK
 */
ER
i2c_memread(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;
	uint32_t xfermode;

	if(pData == NULL || Size == 0)
		return E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return E_PAR;

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

    hi2c->status = I2C_STATUS_BUSY_RX;
    hi2c->ErrorCode = I2C_ERROR_NONE;

    hi2c->pBuffPtr = pData;
    hi2c->XferSize = Size;
    hi2c->XferCount = Size;
    if(hi2c->XferCount > MAX_NBYTE_SIZE){
		hi2c->XferCount2 = MAX_NBYTE_SIZE;
		xfermode = I2C_RELOAD_MODE;
	}
	else{
		hi2c->XferCount2 = hi2c->XferCount;
		xfermode = I2C_AUTOEND_MODE;
	}

	/* 졼֥ɥ쥹ɤ߽Фɥ쥹 */
	if((ercd = i2c_masterreadaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
        return ercd;
	}

    /* Set NBYTES to write and reload if hi2c->XferCount > MAX_NBYTE_SIZE and generate RESTART */
	i2c_transfconfig(hi2c,DevAddress, hi2c->XferCount2, xfermode, I2C_GENERATE_START_READ);

    /*
	 *  ɤ߽Фߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_RX);
	ercd = i2c_transwait(hi2c, 100);

	/*
	 *  ɤ߽Фå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǡ饤
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 MemAddress: ꥢɥ쥹
 *  parameter4 MemAddSize: ꥢɥ쥹
 *  parameter5 pData:      ߥХåեؤΥݥ
 *  parameter6 Size:       ߥ
 *  return                 ｪλʤE_OK
 */
ER
i2c_memwrite(I2C_Handle_t *hi2c, uint16_t DevAddress, uint16_t MemAddress, uint16_t MemAddSize, uint8_t *pData, uint16_t Size)
{
	ER ercd = E_OK;
	uint32_t xfermode;

	if(pData == NULL || Size == 0)
		return  E_PAR;
	if(hi2c->Init.AddressingMode != I2C_ADDRESSINGMODE_7BIT && MemAddSize != 0)
		return  E_PAR;

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		wai_sem(hi2c->Init.semlock);

	if(hi2c->status != I2C_STATUS_READY || i2c_busreadywait(hi2c, I2C_TIMEOUT_BUSY_FLAG) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return E_OBJ;
	}

	hi2c->status = I2C_STATUS_BUSY_TX;
	hi2c->ErrorCode = I2C_ERROR_NONE;

	hi2c->pBuffPtr = pData;
	hi2c->XferSize = Size;
	hi2c->XferCount = Size;
    if(hi2c->XferCount > MAX_NBYTE_SIZE){
		hi2c->XferCount2 = MAX_NBYTE_SIZE;
		xfermode = I2C_RELOAD_MODE;
	}
	else{
		hi2c->XferCount2 = hi2c->XferCount;
		xfermode = I2C_AUTOEND_MODE;
	}

	/* 졼֥ɥ쥹Ƚߥɥ쥹 */
	if((ercd = i2c_masterwriteaddress(hi2c, DevAddress, MemAddress, MemAddSize, I2C_TIMEOUT_FLAG)) != E_OK){
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
		return ercd;
	}

    /* Set NBYTES to write and reload if hi2c->XferCount > MAX_NBYTE_SIZE and generate RESTART */
	if(MemAddSize == 0)
	    i2c_transfconfig(hi2c, DevAddress, hi2c->XferSize, xfermode, I2C_GENERATE_START_WRITE);
	else
	    i2c_transfconfig(hi2c, DevAddress, hi2c->XferCount2, xfermode, I2C_NO_STARTSTOP);

    /*
	 *  񤭹߳ߵ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_TX);
	ercd = i2c_transwait(hi2c, 500);

	/*
	 *  񤭹ߥå
	 */
	if(hi2c->Init.semlock != 0)
		sig_sem(hi2c->Init.semlock);
	return ercd;
}

/*
 *  I2Cǥå
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 DevAddress: 졼֥ɥ쥹
 *  parameter3 Trials:     ȥ饤
 *  parameter4 Timeout:    ॢ
 *  return                 ǥʤE_OK
 */
ER
i2c_ready(I2C_Handle_t *hi2c, uint16_t DevAddress, uint32_t Trials, uint32_t Timeout)
{
	volatile uint32_t I2C_Trials = 0;
	uint32_t tmp1 = 0;

	if(hi2c->status == I2C_STATUS_READY){
		if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_BUSY) != 0)
			return E_OBJ;

		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			wai_sem(hi2c->Init.semlock);

		hi2c->status    = I2C_STATUS_BUSY;
		hi2c->ErrorCode = I2C_ERROR_NONE;

		do{
			/* START */
			if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_7BIT)
				tmp1 = ((DevAddress & I2C_CR2_SADD) | I2C_CR2_START | I2C_CR2_AUTOEND) & (~I2C_CR2_RD_WRN);
			else
				tmp1 = ((DevAddress & I2C_CR2_SADD) | I2C_CR2_ADD10 | I2C_CR2_START) & (~I2C_CR2_RD_WRN);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), tmp1);

			/* START STOP/NACKԤ */
			if(i2c_isrflagsetwait(hi2c, (I2C_ISR_STOPF | I2C_ISR_NACKF), Timeout, &Timeout) != E_OK){
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_TMOUT;
			}


			/* NACKե饰å */
			if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_NACKF) == 0){
				/*
				 *  STOPե饰ΩĤޤԤ
				 */
				if(i2c_isrflagsetwait(hi2c, I2C_ISR_STOPF, Timeout, &Timeout) != E_OK)
					return E_TMOUT;

				/*
				 *  STOPե饰ꥢ
				 */
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);
				hi2c->status = I2C_STATUS_READY;
				/*
				 *  å
				 */
				if(hi2c->Init.semlock != 0)
					sig_sem(hi2c->Init.semlock);
				return E_OK;
			}
			else{	/* Υå */
				/*
				 *  STOPե饰ΩĤޤԤ
				 */
				if(i2c_isrflagsetwait(hi2c, I2C_ISR_STOPF, Timeout, &Timeout) != E_OK)
					return E_TMOUT;

				/*
				 *  NACKե饰ꥢ
				 */
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);

				/*
				 *  STOPե饰ꥢ
				 */
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);
			}
			if(I2C_Trials++ == Trials){
				/*
				 *  ׵
				 */
				sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_STOP);

				/*
				 *  STOPե饰ΩĤޤԤ
				 */
				if(i2c_isrflagsetwait(hi2c, I2C_ISR_STOPF, Timeout, &Timeout) != E_OK)
					return E_TMOUT;

				/*
				 *  STOPե饰ꥢ
				 */
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);
			}
		}while(I2C_Trials < Trials);

		hi2c->status = I2C_STATUS_READY;
		/*
		 *  å
		 */
		if(hi2c->Init.semlock != 0)
			sig_sem(hi2c->Init.semlock);
	    return E_TMOUT;
	}
	else
		return E_OBJ;
}

/*
 *  I2Cޥ󥷥ˤ뽪λ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  return                 ʤ
 */
static void
i2c_intmastersequentialcomplete(I2C_Handle_t *hi2c)
{
	/* No Generate Stop, to permit restart mode */
	/* The stop will be done at the end of transfer, when I2C_AUTOEND_MODE enable */
	if(hi2c->status == I2C_STATUS_BUSY_TX){
		/*
		 *  ʤнλ
		 */
		hi2c->status = I2C_STATUS_READY;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_TX);
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
	}
	else{
		/*
		 *  ʤнλ
		 */
		hi2c->status = I2C_STATUS_READY;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_RX);
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
	if(hi2c->Init.semid != 0)
		isig_sem(hi2c->Init.semid);
}

/*
 *  I2CޥSTOPե饰ˤ뽪λ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 isr:        ISR쥸
 *  return                 ʤ
 */
static void
i2c_intmastercomplete(I2C_Handle_t *hi2c, uint32_t isr)
{
	/*
	 *  STOPե饰ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);

	/*
	 *  ž쥸ꥢ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), ((I2C_CR2_SADD | I2C_CR2_HEAD10R | I2C_CR2_NBYTES | I2C_CR2_RELOAD | I2C_CR2_RD_WRN)));

	/*
	 *  ץꥯꥢ
	 */
	hi2c->XferOptions   = I2C_NO_OPTION_FRAME;

	/*
	 *  NACKʤ饨顼
	 */

	if((isr & I2C_ISR_NACKF) != 0){
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);
		hi2c->ErrorCode |= I2C_ERROR_AF;
	}

	/*
	 *  쥸եå
	 */
	i2c_flushtxdr(hi2c);

	/*
	 *  ߥޥ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (I2C_INT_TX | I2C_INT_RX));

	/*
	 *  顼ȯʤХ顼ղ
	 */
	if(hi2c->ErrorCode != I2C_ERROR_NONE){
		hi2c->ErrorCode |= I2C_ERROR_SIZE;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);
	}
	else if(hi2c->status == I2C_STATUS_BUSY_TX){
		/*
		 *  ʤнλ
		 */
		hi2c->status = I2C_STATUS_READY;
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
		if(hi2c->Init.semid != 0)
			isig_sem(hi2c->Init.semid);
	}
	else if(hi2c->status == I2C_STATUS_BUSY_RX){
		/*
		 *  ʤнλ
		 */
		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
		if(hi2c->Init.semid != 0)
			isig_sem(hi2c->Init.semid);
	}
}

/*
 *  I2C졼LISTENλ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 isr:        ISR쥸
 *  return                 ʤ
 */
static void
i2c_intlistencomplete(I2C_Handle_t *hi2c, uint32_t isr)
{
	/*
	 *  I2C򥯥ꥢ
	 */
	hi2c->XferOptions = I2C_NO_OPTION_FRAME;
	hi2c->status = I2C_STATUS_READY;

	/* Store Last receive data if any */
	if((isr & I2C_ISR_RXNE) != 0){
	    /*
		 *  ǡС
		 */
	    (*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_RXDR));

		/*
		 *  뤷ƤʤСNon-Acknowledge顼Ȥ
		 */

	    if(hi2c->XferCount2 > 0){
			hi2c->XferCount2--;
			hi2c->XferCount--;
			hi2c->ErrorCode |= I2C_ERROR_AF;
		}
	}

	/*
	 *  ߶ػ
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);

	/*
	 *  NACKե饰ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);

	/*
	 *  LISTENХå
	 */
	if(hi2c->listencallback != NULL)
		hi2c->listencallback(hi2c);
}

/*
 *  I2C졼֥ɥ쥹λ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 isr:        ISR쥸
 *  return                 ʤ
 */
static void
i2c_intaddrcomplete(I2C_Handle_t *hi2c, uint32_t isr)
{
	uint8_t  transferdirection = 0;
	uint16_t slaveaddrcode = 0;
	uint16_t ownadd1code = 0;
	uint16_t ownadd2code = 0;

	/*
	 *  LISTENʤСɥ쥹Ф
	 */
	if((hi2c->status & I2C_STATUS_LISTEN) == I2C_STATUS_LISTEN){
		transferdirection = (sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_DIR) >> 16;
		slaveaddrcode     = (sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR)) & I2C_ISR_ADDCODE) >> 16;
		ownadd1code       = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR1)) & I2C_OAR1_OA1;
		ownadd2code       = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_OAR2)) & I2C_OAR2_OA2;

		if(hi2c->Init.AddressingMode == I2C_ADDRESSINGMODE_10BIT){
			/*
			 *  10ӥåȥɥ쥹⡼
			 */
			if((slaveaddrcode & SlaveAddr_MSK) == ((ownadd1code >> SlaveAddr_SHIFT) & SlaveAddr_MSK)){
				slaveaddrcode = ownadd1code;
				hi2c->AddrEventCount++;
				if(hi2c->AddrEventCount == 2){
					/*
					 *  ٥ȥ󥿡ꥢ
					 */
					hi2c->AddrEventCount = 0;

					/*
					 *  ADDRե饰ꥢ
					 */
					sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_ADDRCF);

				}
				/*
				 *  ɥ쥹Хå
				 */
				if(hi2c->addrcallback != NULL)
					hi2c->addrcallback(hi2c, transferdirection, slaveaddrcode);
			}
			else{
				slaveaddrcode = ownadd2code;

				/*
				 *  LISTEN߶ػ
				 */
				sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_LISTEN);

				/*
				 *  ɥ쥹Хå
				 */
				if(hi2c->addrcallback != NULL)
					hi2c->addrcallback(hi2c, transferdirection, slaveaddrcode);
			}
		}
		else{
			/*
			 *  7ӥåȥɥ쥹⡼ɡLISTEN߶ػ
			 */
			sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_LISTEN);

			/*
			 *  ɥ쥹Хå
			 */
			if(hi2c->addrcallback != NULL)
				hi2c->addrcallback(hi2c, transferdirection, slaveaddrcode);
		}
	}
	else{
		/*
		 *  ADDRե饰ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_ADDRCF);
	}
}

/*
 *  I2C졼֥󥷥ˤ뽪λ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  return                 ʤ
  */
static void
i2c_intslavesequencecomplete(I2C_Handle_t *hi2c)
{
	if(hi2c->status == I2C_STATUS_BUSY_TX_LISTEN){
		/*
		 *  ʤ齪λ
		 */
		hi2c->status = I2C_STATUS_LISTEN;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_TX);
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
	}
	else if(hi2c->status == I2C_STATUS_BUSY_RX_LISTEN){
		/*
		 *  ʤ齪λ
		 */
	    hi2c->status = I2C_STATUS_LISTEN;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_RX);
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
}

/*
 *  I2C졼STOPե饰ˤ뽪λ
 *  parameter1 hi2c:       I2CϥɥؤΥݥ
 *  parameter2 isr:        ISR쥸
 *  return                 ʤ
 */
static void
i2c_intslavecomplete(I2C_Handle_t *hi2c, uint32_t isr)
{
	/*
	 *  STOPե饰򥯥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_STOPCF);

	/*
	 *  ADDRե饰򥯥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_ADDRCF);

	/*
	 *  
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);

	/*
	 *  ɥ쥹Υåͭ
	 */
	sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_NACK);

	/*
	 *  ž쥸ꥻå
	 */
	sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), (I2C_CR2_SADD | I2C_CR2_HEAD10R | I2C_CR2_NBYTES | I2C_CR2_RELOAD | I2C_CR2_RD_WRN));

	/*
	 *  쥸եå
	 */
	i2c_flushtxdr(hi2c);

	if(hi2c->XferCount != 0){
		/*
		 *  žλǤʤСNon-Acknowledge顼
		 */
		hi2c->ErrorCode |= I2C_ERROR_AF;
	}

	/*
	 *  ׵᤬м
	 */
	if((isr & I2C_ISR_RXNE) != 0){
		(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_RXDR));
		if((hi2c->XferCount > 0U)){
			hi2c->XferCount2--;
			hi2c->XferCount--;
		}
	}

	if(hi2c->ErrorCode != I2C_ERROR_NONE){
	    /*
		 *  顼ȯƤХ顼ä
		 */
		hi2c->ErrorCode |= I2C_ERROR_SIZE;
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);

	    if(hi2c->status == I2C_STATUS_LISTEN){
			/*
			 *  LISTENʤСLISTENλԤ
			 */
			i2c_intlistencomplete(hi2c, isr);
		}
	}
	else if(hi2c->XferOptions != I2C_NO_OPTION_FRAME){
		/*
		 *  ץ꤬硢LISTENλ
		 */
	    hi2c->XferOptions = I2C_NO_OPTION_FRAME;
	    hi2c->status = I2C_STATUS_READY;
		if(hi2c->listencallback != NULL)
			hi2c->listencallback(hi2c);
	}
	else if(hi2c->status == I2C_STATUS_BUSY_RX){
		/*
		 *  ʤ齪λ
		 */
		hi2c->status = I2C_STATUS_READY;
		if(hi2c->readcallback != NULL)
			hi2c->readcallback(hi2c);
	}
	else{
		/*
		 *  ʤ齪λ
		 */
		hi2c->status = I2C_STATUS_READY;
		if(hi2c->writecallback != NULL)
			hi2c->writecallback(hi2c);
	}
}


/*
 *  I2C EVߥϥɥ
 */
void
i2c_ev_handler(I2C_Handle_t *hi2c)
{
	uint32_t isr = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR));
	uint32_t cr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1));
	uint16_t devaddress = 0;
	volatile uint32_t tmp;

	/*
	 *  ޥ⡼
	 */
	if(hi2c->Init.OwnAddress1 == 0){
		if((isr & I2C_ISR_NACKF) != 0 && (cr1 & I2C_CR1_NACKIE) != 0){
			/*
			 *  NACK顼ե饰ꥢ
			 */
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);
		    hi2c->ErrorCode |= I2C_ERROR_AF;

		    /*
			 *  쥸եå
			 */
		    i2c_flushtxdr(hi2c);
		}
		else if((isr & I2C_ISR_RXNE) != 0 && (cr1 & I2C_CR1_RXIE) != 0){
			/*
			 *  ǡ
			 */
			*hi2c->pBuffPtr++ = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_RXDR));
			hi2c->XferCount--;
			hi2c->XferCount2--;
		}
		else if((isr & I2C_ISR_TXIS) != 0 && (cr1 & I2C_CR1_TXIE) != 0){
			/*
			 *  ǡ
			 */
			tmp = (uint8_t)(*hi2c->pBuffPtr++);
			sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), tmp);
			hi2c->XferCount--;
			hi2c->XferCount2--;
		}
		else if((isr & I2C_ISR_TCR) != 0 && (cr1 & I2C_CR1_TCIE) != 0){
			if(hi2c->XferCount2 == 0 && hi2c->XferCount != 0){
				devaddress = (sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2)) & I2C_CR2_SADD);
				if(hi2c->XferCount > MAX_NBYTE_SIZE){
					hi2c->XferCount2 = MAX_NBYTE_SIZE;
					i2c_transfconfig(hi2c, devaddress, hi2c->XferCount2, I2C_RELOAD_MODE, I2C_NO_STARTSTOP);
				}
				else{
					hi2c->XferCount2 = hi2c->XferCount;
					if(hi2c->XferOptions != I2C_NO_OPTION_FRAME)
						i2c_transfconfig(hi2c, devaddress, hi2c->XferCount2, hi2c->XferOptions, I2C_NO_STARTSTOP);
					else
						i2c_transfconfig(hi2c, devaddress, hi2c->XferCount2, I2C_AUTOEND_MODE, I2C_NO_STARTSTOP);
				}
			}
			else{
				/*
				 *  AUTOEND⡼ɤʤｪλ
				 */
				if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2)) & I2C_CR2_AUTOEND) != 0)
					i2c_intmastersequentialcomplete(hi2c);
				else{
					/*
					 *  ۾ｪλ
					 */
					hi2c->ErrorCode |= I2C_ERROR_SIZE;
					sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);
				}
			}
		}
		else if((isr & I2C_ISR_TC) != 0 && (cr1 & I2C_CR1_TCIE) != 0){
			if(hi2c->XferCount == 0){
				if((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2)) & I2C_CR2_AUTOEND) != I2C_CR2_AUTOEND){
					if(hi2c->XferOptions == I2C_NO_OPTION_FRAME){
						/*
						 *  ץ󤬤ʤSTOP׵
						 */
						sil_orw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR2), I2C_CR2_STOP);
					}
					else
						i2c_intmastersequentialcomplete(hi2c);
				}
			}
			else{
				/*
				 *  TCե饰顼
				 */
				hi2c->ErrorCode |= I2C_ERROR_SIZE;
				sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), I2C_INT_ALL);
			}
		}
		if((isr & I2C_ISR_STOPF) != 0 && (cr1 & I2C_CR1_STOPIE) != 0){
			/*
			 *  ߥե饰齪λ
			 */
			i2c_intmastercomplete(hi2c, isr);
		}
	}

	/*
	 *  졼֥⡼
	 */
	else{
		if((isr & I2C_ISR_NACKF) != 0 && (cr1 & I2C_CR1_NACKIE) != 0){
			/*
			 *  NACKȯ
			 */
			if(hi2c->XferCount == 0){
				/*
				 *  žλ
				 */
				if((hi2c->XferOptions == I2C_FIRST_AND_LAST_FRAME || hi2c->XferOptions == I2C_LAST_FRAME) && \
			        (hi2c->status == I2C_STATUS_LISTEN)){
					/* Call I2C Listen complete process */
					i2c_intlistencomplete(hi2c, isr);
				}
				else if(hi2c->XferOptions != I2C_NO_OPTION_FRAME && hi2c->status == I2C_STATUS_BUSY_TX_LISTEN){
					/*
					 *  NACKե饰ꥢ
					 */
					sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);

					/* Flush TX register */
					i2c_flushtxdr(hi2c);

					/* Last Byte is Transmitted */
					/* Call I2C Slave Sequential complete process */
					i2c_intslavesequencecomplete(hi2c);
				}
				else{
					/*
					 *  NACKե饰ꥢ
					 */
					sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);
				}
			}
			else{
				/*
				 *  žλƤʤ硢Non-Acknowledge顼
				 */
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_NACKCF);
				hi2c->ErrorCode |= I2C_ERROR_AF;
			}
		}
		else if((isr & I2C_ISR_RXNE) != 0 && (cr1 & I2C_CR1_RXIE) != 0){
			/*
			 *  
			 */
			if(hi2c->XferCount > 0){
				(*hi2c->pBuffPtr++) = (uint8_t)sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_RXDR));
				hi2c->XferCount2--;
				hi2c->XferCount--;
			}
			if(hi2c->XferCount == 0 && hi2c->XferOptions != I2C_NO_OPTION_FRAME){
				/*
				 *  I2C졼֥󥹽λ
				 */
				i2c_intslavesequencecomplete(hi2c);
			}
		}
		else if((isr & I2C_ISR_ADDR) != 0 && (cr1 & I2C_CR1_ADDRIE) != 0){
			/*
			 *  ADDRե饰
			 */
			i2c_intaddrcomplete(hi2c, isr);
		}
		else if((isr & I2C_ISR_TXIS) != 0 && (cr1 & I2C_CR1_TXIE) != 0){
			/*
			 *  
			 */
		    if(hi2c->XferCount > 0){
				/* Write data to TXDR */
				tmp = (*hi2c->pBuffPtr++);
				sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_TXDR), tmp);
				hi2c->XferCount--;
				hi2c->XferCount2--;
			}
			else{
				if(hi2c->XferOptions == I2C_NEXT_FRAME || hi2c->XferOptions == I2C_FIRST_FRAME){
					/*
					 *  I2C졼󥹽λ
					 */
 					i2c_intslavesequencecomplete(hi2c);
				}
			}
		}
		if((isr & I2C_ISR_STOPF) != 0 && (cr1 & I2C_CR1_STOPIE) != 0){
			/*
			 *  ߥե饰齪λ
			 */
			i2c_intslavecomplete(hi2c, isr);
		}
	}
}

/*
 *  I2C ERߥϥɥ
 */
void
i2c_er_handler(I2C_Handle_t *hi2c)
{
	uint32_t isr = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR));
	uint32_t cr1 = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1));

	/*
	 *  I2C Х顼
	 */
	if(((isr & I2C_ISR_BERR) != 0) && ((cr1 & I2C_CR1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_BERR;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_BERRCF);
	}
	isr = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR));

	/*
	 *  I2C С󡦥顼
	 */
	if(((isr & I2C_ISR_OVR) != 0) && ((cr1 & I2C_CR1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_OVR;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_OVRCF);
	}
	isr = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR));

	/*
	 *  I2C ӥơ󡦥顼
	 */
	if(((isr & I2C_ISR_ARLO) != 0) && ((cr1 & I2C_CR1_ERRIE) != 0)){
		hi2c->ErrorCode |= I2C_ERROR_ARLO;
		sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_ICR), I2C_ICR_ARLOCF);
	}
	isr = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_ISR));

	if(hi2c->ErrorCode != I2C_ERROR_NONE){
		hi2c->status = I2C_STATUS_READY;

		/*
		 *  ̿ػ
		 */
		sil_andw_mem((uint32_t *)(hi2c->base+TOFF_I2C_CR1), (I2C_CR1_TCIE |
			 I2C_CR1_TXIE | I2C_CR1_RXIE | I2C_CR1_ERRIE | I2C_CR1_STOPIE | I2C_CR1_ADDRIE));

		if(hi2c->errorcallback != NULL)
			hi2c->errorcallback(hi2c);
	}
}


/*
 *  I2C EVߥӥ롼
 */
void
i2c_ev_isr(intptr_t exinf)
{
	i2c_ev_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

/*
 *  I2C ERߥӥ롼
 */
void
i2c_er_isr(intptr_t exinf)
{
	i2c_er_handler(&I2cHandle[INDEX_I2C((uint32_t)exinf)]);
}

