/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 699 2020-11-04 15:10:05Z roi $
 */
/*
 * STM32H7xxѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define byte2bcd(value) ((((value)/10)<<4) | ((value) % 10))
#define bcd2byte(value) ((((value)>>4) * 10) + ((value) & 0xF))

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

uint_t dipsw_value;
void (*exti_func[16])(void) = { 0 };


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVT           0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	TADR_GPIOE_BASE,			/* index 4 */
	TADR_GPIOF_BASE,			/* index 5 */
	TADR_GPIOG_BASE,			/* index 6 */
	TADR_GPIOH_BASE,			/* index 7 */
	TADR_GPIOI_BASE,			/* index 8 */
	TADR_GPIOJ_BASE,			/* index 9 */
	TADR_GPIOK_BASE				/* index 10 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t iocurrent = 1<<pin;
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	/*
	 *  GPIO⡼
	 */
	/* 륿͡ȡե󥯥󡦥⡼ */
	temp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4));
	temp &= ~((uint32_t)0xF << ((uint32_t)(pin & (uint32_t)0x07) * 4)) ;
	if(init->mode == GPIO_MODE_AF)
		temp |= ((uint32_t)(init->alternate) << (((uint32_t)pin & (uint32_t)0x07) * 4));
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4), temp);

	/*  ϥ⡼ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_MODER), (GPIO_MODER_MODER0 << (pin * 2)), ((init->mode & GPIO_MODE) << (pin * 2)));

	/*  ϥ⡼ */
	if(init->mode == GPIO_MODE_OUTPUT || init->mode == GPIO_MODE_AF){
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR0 << (pin * 2)), (init->speed << (pin * 2)));
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT_0 << pin), (init->otype << pin));
	}

	/*  ץ륢åסץ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR0 << (pin * 2)), (init->pull << (pin * 2)));

	/*
	 *  EXTI⡼
	 */
	if((init->mode & EXTI_MODE) == EXTI_MODE){
		/*  SYSCFGå */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB4ENR), RCC_APB4ENR_SYSCFGEN);
		temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB4ENR));

		temp = sil_rew_mem((uint32_t *)(TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR1+(pin & 0x0C)));
		temp &= ~(0x0F << (4 * (pin & 0x03)));
		temp |= (index << (4 * (pin & 0x03)));
		sil_wrw_mem((uint32_t *)(TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR1+(pin & 0x0C)), temp);

		if((init->mode & GPIO_MODE_IT) == GPIO_MODE_IT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);

		if((init->mode & GPIO_MODE_EVT) == GPIO_MODE_EVT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);

		if((init->mode & RISING_EDGE) == RISING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);

		if((init->mode & FALLING_EDGE) == FALLING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
	}
}


/*
 *  ɥ쥹Då奤Хǡȴؿ
 *  parameter1  addr    address (aligned to CACHE_LINE_SIZE(32)-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void
invalidatedcache_by_addr(uint8_t *addr, int32_t dsize)
{
	int32_t  op_size = dsize + (CACHE_LINE_SIZE-1);
	uint32_t op_addr = (((uint32_t)addr) & ~(CACHE_LINE_SIZE-1));

	asm("dsb 0xF":::"memory");
	while (op_size > 0) {
		sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_DCIMVAC), op_addr);
		op_addr += CACHE_LINE_SIZE;
		op_size -= CACHE_LINE_SIZE;
	}
	asm("dsb 0xF":::"memory");
	asm("isb 0xF":::"memory");
}

/*
 *  ɥ쥹Dåեåؿ
 *  parameter1  addr    address (aligned to CACHE_LINE_SIZE(32)-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void flushdcache_by_addr(uint8_t *addr, int32_t dsize)
{
	int32_t  op_size = dsize + (CACHE_LINE_SIZE-1);
	uint32_t op_addr = (((uint32_t)addr) & ~(CACHE_LINE_SIZE-1));

	asm("dsb 0xF":::"memory");
	while (op_size > 0) {
		sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_DCCMVAC), op_addr);
		op_addr += CACHE_LINE_SIZE;
		op_size -= CACHE_LINE_SIZE;
	}
	asm("dsb 0xF":::"memory");
	asm("isb 0xF":::"memory");
}

/*
 *  ɥ쥹Dåեå&Хǡȴؿ
 *  parameter1  addr    address (aligned to CACHE_LINE_SIZE(32)-byte boundary)
 *  parameter2  dsize   size of memory block (in number of bytes)
 */
void flushinvalidatedcache_by_addr(uint8_t *addr, int32_t dsize)
{
	int32_t  op_size = dsize + (CACHE_LINE_SIZE-1);
	uint32_t op_addr = (((uint32_t)addr) & ~(CACHE_LINE_SIZE-1));

	asm("dsb 0xF":::"memory");
	while (op_size > 0) {
		sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_DCCIMVAC), op_addr);
		op_addr += CACHE_LINE_SIZE;
		op_size -= CACHE_LINE_SIZE;
	}
	asm("dsb 0xF":::"memory");
	asm("isb 0xF":::"memory");
}


/*
 *  DMACؿ
 */

#define NUM_STMDMA      16
#define NUM_BDMA        8
#define NUM_DMAINDEX    5
#define INDEX_TC        0
#define INDEX_HT        1
#define INDEX_TE        2
#define INDEX_FE        3
#define INDEX_DME       4

#define DMA_TRS_TIMEOUT 2000	/* 2 */

static const uint32_t dma_base[NUM_STMDMA+NUM_BDMA] = {
	TADR_DMA1_STM0_BASE,
	TADR_DMA1_STM1_BASE,
	TADR_DMA1_STM2_BASE,
	TADR_DMA1_STM3_BASE,
	TADR_DMA1_STM4_BASE,
	TADR_DMA1_STM5_BASE,
	TADR_DMA1_STM6_BASE,
	TADR_DMA1_STM7_BASE,
	TADR_DMA2_STM0_BASE,
	TADR_DMA2_STM1_BASE,
	TADR_DMA2_STM2_BASE,
	TADR_DMA2_STM3_BASE,
	TADR_DMA2_STM4_BASE,
	TADR_DMA2_STM5_BASE,
	TADR_DMA2_STM6_BASE,
	TADR_DMA2_STM7_BASE,

	TADR_BDMA_CH0_BASE,
	TADR_BDMA_CH1_BASE,
	TADR_BDMA_CH2_BASE,
	TADR_BDMA_CH3_BASE,
	TADR_BDMA_CH4_BASE,
	TADR_BDMA_CH5_BASE,
	TADR_BDMA_CH6_BASE,
	TADR_BDMA_CH7_BASE
};

static const uint32_t dma_flag[NUM_STMDMA][NUM_DMAINDEX] = {
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 1-0 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 1-1 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 1-2 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 1-3 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 1-4 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 1-5 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 1-6 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 1-7 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 2-0 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 2-1 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 2-2 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 2-3 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 2-4 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 2-5 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 2-6 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 }	/* 2-7 */
};

//static const uint8_t flagBitshiftOffset[8] = {0, 6, 16, 22, 0, 6, 16, 22};

static const uint32_t dma_faddr[NUM_STMDMA/4] = {
	(TADR_DMA1_BASE+TOFF_DMAI_LISR),
	(TADR_DMA1_BASE+TOFF_DMAI_HISR),
	(TADR_DMA2_BASE+TOFF_DMAI_LISR),
	(TADR_DMA2_BASE+TOFF_DMAI_HISR)
};

static DMA_Handle_t *pDmaHandler[NUM_STMDMA+NUM_BDMA];

#define get_DMAmuxChannel(h)        ((h)->muxbase + ((h)->sdid % NUM_STMDMA) * 4)
#define get_stream_statusmask(h)    (1 << (((h)->sdid % NUM_STMDMA) & 0x1F))


/*
 *  STREAM DMA FLAG ALL CLEAR
 */
static void
stream_flag_clear(DMA_Handle_t *hdma)
{
	uint32_t i, value;

	for(i = value = 0 ; i < NUM_DMAINDEX ; i++)
		value |= dma_flag[hdma->sdid][i];
	sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), value);
}

/*
 *  STREAM DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0;
	uint32_t i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < (NUM_STMDMA+NUM_BDMA) ; i++){
		if(dma_base[i] == hdma->base)
			break;
	}
	if(i == (NUM_STMDMA+NUM_BDMA))
		return E_PAR;
	hdma->sdid = i;
	pDmaHandler[i] = hdma;

	if(i < NUM_STMDMA){		/* Stream DMA1 or 2 */
		hdma->muxbase = TADR_DMAMUX1_BASE;

		/* DMA-CR쥸 */
		tmp = sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR));

		/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
		tmp &= ((uint32_t)~(DMA_SxCR_MBURST | DMA_SxCR_PBURST | \
                      DMA_SxCR_PL    | DMA_SxCR_MSIZE  | DMA_SxCR_PSIZE  | \
                      DMA_SxCR_MINC  | DMA_SxCR_PINC   | DMA_SxCR_CIRC   | DMA_SxCR_PFCTRL | \
                      DMA_SxCR_DIR   | DMA_SxCR_CT     | DMA_SxCR_DBM));

		/* DMA⡼ɤƥݥꥣ */
		tmp |=  hdma->Init.Direction | hdma->Init.PeriphInc | \
		        hdma->Init.MemInc    | hdma->Init.PeriphDataAlignment | \
		        hdma->Init.MemDataAlignment | hdma->Init.Mode | \
		        hdma->Init.Priority;

		/* FIFO⡼ɤʤС */
		if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE)
			tmp |=  hdma->Init.MemBurst | hdma->Init.PeriphBurst;

		/* DMA⡼ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), tmp);

		/* DMA-FCR쥸 */
		tmp = sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR));

		/* FIFO⡼ɤƥݥꥣ˥ꥢ */
		tmp &= (uint32_t)~(DMA_SxFCR_DMDIS | DMA_SxFCR_FTH);

		/* FIFO⡼ɤƥݥꥣ */
		tmp |= hdma->Init.FIFOMode;

		/* FIFO⡼ɤʤ饹åۡɤ */
		if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE){
			tmp |= hdma->Init.FIFOThreshold;

			/* Check compatibility between FIFO threshold level and size of the memory burst */
			/* for INCR4, INCR8, INCR16 */
			if(hdma->Init.MemBurst != DMA_MBURST_SINGLE
					&& hdma->Init.MemDataAlignment == DMA_MDATAALIGN_BYTE){
				switch (hdma->Init.FIFOThreshold){
				case DMA_FIFO_THRESHOLD_1QUARTERFULL:
				case DMA_FIFO_THRESHOLD_3QUARTERSFULL:
					if((hdma->Init.MemBurst & DMA_SxCR_MBURST_1) == DMA_SxCR_MBURST_1)
						return E_PAR;
					break;
				case DMA_FIFO_THRESHOLD_HALFFULL:
					if(hdma->Init.MemBurst == DMA_MBURST_INC16)
						return E_PAR;
					break;
				default:
					break;
				}
			}
		}

		/* DMA-FCR쥸 */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), tmp);

		/* ƥե饰򥯥ꥢ */
		stream_flag_clear(hdma);
	}
	else{			/* BDMA */
		hdma->muxbase = TADR_DMAMUX2_BASE;

		/* BDMA-CCR쥸 */
		tmp = sil_rew_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR));

		/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
		tmp &= ((uint32_t)~(BDMA_CCR_PL    | BDMA_CCR_MSIZE   | BDMA_CCR_PSIZE  | \
		              BDMA_CCR_MINC  | BDMA_CCR_PINC    | BDMA_CCR_CIRC   | \
		              BDMA_CCR_DIR   | BDMA_CCR_MEM2MEM | BDMA_CCR_DBM    | \
		              BDMA_CCR_CT));

		/* DMA⡼ɤƥݥꥣ */
		if(hdma->Init.Direction == DMA_MEMORY_TO_PERIPH)
			tmp |= BDMA_CCR_DIR;		/* memory to periph */
		else if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
			tmp |= BDMA_CCR_MEM2MEM;	/* memory to memory */
		else
			tmp |= 0x00000000;			/* periph to memory */
		tmp |=  (hdma->Init.PeriphInc | hdma->Init.MemInc | \
		        hdma->Init.PeriphDataAlignment | hdma->Init.MemDataAlignment | \
		        (hdma->Init.Mode & ~DMA_SxCR_PFCTRL)) >> 3;
		tmp |= hdma->Init.Priority >> 4;

		/* DMA⡼ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), tmp);

		/* ƥե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), BDMA_IFCR_CGIF0 << ((hdma->sdid-NUM_STMDMA) * 4));
	}

	if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
		hdma->Init.Request = DMA_REQUEST_MEM2MEM;

	/* DMAMUX˥ڥեͥ */
	sil_wrw_mem((uint32_t *)(get_DMAmuxChannel(hdma)+TOFF_DMAMUX_CHANNEL_CR), (hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID));
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));

	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase + TOFF_DMAMUX_REQUEST_GEN + (((request-1) * 4))), 0x00000000);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase + TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request-1)));
	}

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < (NUM_STMDMA+NUM_BDMA) ; i++){
		if(dma_base[i] == hdma->base)
			break;
	}
	if(i == (NUM_STMDMA+NUM_BDMA))
		return E_PAR;
	hdma->sdid = i;

	if(i < NUM_STMDMA){		/* Stream DMA1 or 2 */
		/* DMA֥͡ */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_NDTR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M1AR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), 0x00000021);

		/* ƥե饰򥯥ꥢ */
		stream_flag_clear(hdma);
	}
	else{			/* BDMA */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CNDTR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CPAR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CM0AR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CM1AR), 0x00000000);
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), BDMA_IFCR_CGIF0 << ((hdma->sdid-NUM_STMDMA) * 4));
	}

	/*
	 *  DMAMUXͥν
	 */
	if(hdma->muxbase != 0){
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));
	}
	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN + (((request - 1) * 4))), 0x00000000);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request - 1)));
	}
	hdma->muxbase = 0;

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  STREAM DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	if(hdma->sdid < NUM_STMDMA){		/* Stream DMA1 or 2 */
		/* DMA */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE | DMA_SxCR_HTIE | DMA_SxCR_TEIE | DMA_SxCR_DMEIE);
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);

		/* ƥե饰򥯥ꥢ */
		stream_flag_clear(hdma);
		hdma->status = DMA_STATUS_BUSY;

		/* DBMӥåȥꥢ */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_DBM);
		/* ǡĹ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_NDTR), DataLength);

		if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
			/* ꤫ǥХ */
			flushdcache_by_addr((uint8_t *)SrcAddress, DataLength*4);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), DstAddress);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), SrcAddress);
		}
		else{	/* ǥХ */
			flushinvalidatedcache_by_addr((uint8_t *)DstAddress, DataLength*4);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), SrcAddress);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), DstAddress);
		}

		/* ߥ֥͡ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE | DMA_SxCR_HTIE | DMA_SxCR_TEIE | DMA_SxCR_DMEIE);
		if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE)
			sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);
	}
	else{
		/* DMA */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), BDMA_CCR_EN);
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), (BDMA_CCR_TCIE | BDMA_CCR_HTIE | BDMA_CCR_TEIE));

		/* ե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), BDMA_IFCR_CGIF0 << ((hdma->sdid-NUM_STMDMA) * 4));
		hdma->status = DMA_STATUS_BUSY;

		/* ǡĹ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CNDTR), DataLength);

		if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
			/* ꤫ǥХ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CPAR), DstAddress);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CM0AR), SrcAddress);
		}
		else{	/* ǥХ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CPAR), SrcAddress);
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CM0AR), DstAddress);
		}

		/* ߥ֥͡ */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), (BDMA_CCR_TCIE | BDMA_CCR_HTIE | BDMA_CCR_TEIE));
	}

	/*
	 *  DMAMUX SynchronizationͭʤСsync overrunߤͭ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CR)) & DMAMUX_CxCR_SE) != 0)
		sil_orw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CR), DMAMUX_CxCR_SOIE);

	/*
	 *  ꥯȥͥ졼ʤСgen overrunߤͭ
	 */
	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7)
		sil_orw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN), DMAMUX_RGxCR_OIE);

	/* DMA */
	if(hdma->sdid < NUM_STMDMA)		/* Stream DMA1 or 2 */
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
	else
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), BDMA_CCR_EN);
	return E_OK;
}

/*
 *  STREAM DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	uint32_t tick = 0;

	if(hdma->sdid < NUM_STMDMA){		/* Stream DMA1 or 2 */
		/* DMA */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);

		/* DMAԤ */
		while((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_EN) != 0){
		    /* Check for the Timeout */
			if(tick > DMA_TRS_TIMEOUT){
				/* ॢȥ顼 */
				hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
				return E_TMOUT;
			}
	    	dly_tsk(1);
		}
	}
	else{		/* BDMA */
		/* DMA */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), (BDMA_CCR_TCIE | BDMA_CCR_HTIE | BDMA_CCR_TEIE));
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), BDMA_CCR_EN);

		/*
		 *  DMAMUX sync overrunߤ
		 */
		sil_andw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CR), DMAMUX_CxCR_SOIE);

		/* ե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), BDMA_IFCR_CGIF0 << ((hdma->sdid-NUM_STMDMA) * 4));

		/* DMAԤ */
		while((sil_rew_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR)) & BDMA_CCR_EN) != 0){
			if(tick > DMA_TRS_TIMEOUT){
				/* ॢȥ顼 */
				hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
				return E_TMOUT;
			}
	    	dly_tsk(1);
		}
	}

	/*
	 *  DMAMUXΥСե饰򥯥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));

	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_andw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN), DMAMUX_RGxCR_OIE);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request - 1)));
	}
	hdma->status = 0;
	return E_OK;
}

/*
 *  STREAM DMAΥ١ɥ쥹
 */
uintptr_t
dma_stream_base(int32_t sid)
{
	if(sid < 0 || sid >= NUM_STMDMA)
		return 0;
	if(sid < DMA2STM0_SID)
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMA1EN);
	else
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMA2EN);
	sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR));
	return (uintptr_t)dma_base[sid];
}

/*
 *  STREAM DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	/*
	 *  ž顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_TE]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_TEIE) != 0){
			/* ž顼ߥꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TEIE);
			/* ž顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_TE]);
			/* ž顼֤ */
			hdma->ErrorCode |= DMA_ERROR_TE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  FIFO顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_FE]) != 0){
	    if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR)) & DMA_SxFCR_FEIE) != 0){
			/* FIFO顼ߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);
			/* FIFO顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_FE]);
			/* FIFO顼 */
				hdma->ErrorCode |= DMA_ERROR_FE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  쥯ȡꡦ顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_DME]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DMEIE) != 0){
			/* DME顼ߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_DMEIE);
			/* DME顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_DME]);
			/* DME顼 */
			hdma->ErrorCode |= DMA_ERROR_DME;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  ϡžλ
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_HT]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_HTIE) != 0){
			/*
			 *  ϡžե饰򥯥ꥢ
			 */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_HT]);
			/* ޥХåե⡼ */
			if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DBM) != 0){
				/* ꣰ */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CT) == 0){
					/* ϡե꣰ξ */
					hdma->status = DMA_STATUS_READY_HMEM0;
					if(hdma->xferhalfcallback != NULL)
						hdma->xferhalfcallback(hdma);
				}
				/* ꣱ */
				else{
					/* ϡե꣱ξ */
					hdma->status = DMA_STATUS_READY_HMEM1;
					if(hdma->xferm1halfcallback != NULL)
						hdma->xferm1halfcallback(hdma);
				}
			}
			else{
				/*
				 *  åǤʤХϡճ
				 */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CIRC) == 0)
					sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_HTIE);
				/* ϡžơѹ */
				hdma->status = DMA_STATUS_READY_HMEM0;
				/* ϡžλХå */
				if(hdma->xferhalfcallback != NULL)
					hdma->xferhalfcallback(hdma);
			}
		}
	}
	/*
	 *  žλ
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_TC]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_TCIE) != 0){
			/*
			 *  žλե饰򥯥ꥢ
			 */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_TC]);
			/* ޥХåե⡼ */
			if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DBM) != 0){
				/* ꣱ */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CT) == 0){
					hdma->status = DMA_STATUS_READY_MEM1;
					/* 1žλХå */
					if(hdma->xferm1callback != NULL)
						hdma->xferm1callback(hdma);
				}
				/* ꣰ */
				else{
					hdma->status = DMA_STATUS_READY_MEM0;
					/* žλХå */
					if(hdma->xfercallback != NULL)
						hdma->xfercallback(hdma);
				}
			}
			else{
				/*
				 *  åǤʤдλ
				 */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CIRC) == 0){
					sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE);
					hdma->ErrorCode = DMA_ERROR_NONE;
				}

				/* ơλ */
				hdma->status = DMA_STATUS_READY_MEM0;
				/* žλХå */
				if(hdma->xfercallback != NULL)
					hdma->xfercallback(hdma);
			}
		}
	}
}

/*
 *  BDMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
bdma_inthandler(DMA_Handle_t *hdma)
{
	uint32_t ccr = sil_rew_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR));
	uint32_t isr = sil_rew_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_ISR));
	uint32_t index = ((hdma->sdid - NUM_STMDMA) * 4) & 0x1F;

	/*
	 *  ϡžλ
	 */
	if((isr & (BDMA_ISR_HTIF0 << index)) != 0 && (ccr & BDMA_CCR_HTIE) != 0){
		/*
		 *  ϡžե饰򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), (BDMA_IFCR_CHTIF0 << index));

		if((ccr & BDMA_CCR_DBM) != 0){	/* DMA Memory to Memory */
			if((ccr & BDMA_CCR_CT) == 0){	/* Memory1ž */
				hdma->status = DMA_STATUS_READY_HMEM1;
				if(hdma->xferm1halfcallback != NULL)
					hdma->xferm1halfcallback(hdma);
			}
			else{			/* Memory0ž */
				hdma->status = DMA_STATUS_READY_HMEM0;
				if(hdma->xferhalfcallback != NULL)
					hdma->xferhalfcallback(hdma);
			}
		}
		else{		/* DMA Memory and Device */
			/*
			 *  åǤʤХϡճ
			 */
			if((ccr & BDMA_CCR_CIRC) == 0)
				sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), BDMA_CCR_HTIE);

			/*
			 *  ϡճߥХå
			 */
			hdma->status = DMA_STATUS_READY_HMEM0;
			if(hdma->xferhalfcallback != NULL)
				hdma->xferhalfcallback(hdma);
		}
	}
	/*
	 *  žλ
	 */
	else if((isr & (BDMA_ISR_TCIF0 << index)) != 0 && (ccr & BDMA_CCR_TCIE) != 0){
		/*
		 *  žλե饰򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), (BDMA_IFCR_CTCIF0 << index));

		if((ccr & BDMA_CCR_DBM) != 0){	/* DMA Memory to Memory */
			if((ccr & BDMA_CCR_CT) == 0){	/* Memory1ž */
				hdma->status = DMA_STATUS_READY_MEM1;
				if(hdma->xferm1callback != NULL)
					hdma->xferm1callback(hdma);
			}
			else{			/* Memory0ž */
				hdma->status = DMA_STATUS_READY_MEM0;
				if(hdma->xfercallback != NULL)
					hdma->xfercallback(hdma);
			}
		}
		else{		/* DMA Memory and Device */
			/*
			 *  åǤʤдλ
			 */
			if((ccr & BDMA_CCR_CIRC) == 0){
				sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), (BDMA_CCR_TCIE |BDMA_CCR_TEIE));
				hdma->status = 0;
			}
			/*
			 *  žλߥХå
			 */
			hdma->status = DMA_STATUS_READY_MEM0;
			if(hdma->xfercallback != NULL)
				hdma->xfercallback(hdma);
		}
	}
	/*
	 *  ž顼Ƚ
	 */
	else if((isr & (BDMA_ISR_TEIF0 << index)) != 0 && (ccr & BDMA_CCR_TEIE) != 0){
		/*
		 *  ž顼ξ硢߶ػ
		 */
		sil_andw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), (BDMA_CCR_TCIE | BDMA_CCR_HTIE |BDMA_CCR_TEIE));

		/*
		 *  Хߥե饰򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(TADR_BDMA_BASE+TOFF_BDMA_IFCR), (BDMA_IFCR_CGIF0 << index));

		/*
		 *  ž顼
		 */
		hdma->ErrorCode |= DMA_ERROR_TE;
		hdma->status = DMA_STATUS_READY_ERROR;
		if(hdma->errorcallback != NULL)
			hdma->errorcallback(hdma);
    }
}

/*
 *  STREAM DMA ߥӥ롼
 */
void
stream_dma_isr(intptr_t exinf)
{
	dma_inthandler(pDmaHandler[(uint32_t)exinf]);
}

/*
 *  BDMA ߥӥ롼
 */
void
bdma_dma_isr(intptr_t exinf)
{
	bdma_inthandler(pDmaHandler[(uint32_t)exinf+NUM_STMDMA]);
}


static DMA2D_Handle_t *pDma2dHandler;

/*
 *  DMA2Dߴؿ
 */
void dma2d_inthandler(DMA2D_Handle_t *hdma2d)
{
	uint32_t isr = sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_ISR));
	/*
	 *  ž顼Ƚ
	 */
	if((isr & DMA2D_ISR_TEIF) != 0){
		if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR)) & DMA2D_CR_TEIE) != 0){
			/* ž顼Disable */
			sil_andw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_TEIE);
			hdma2d->ErrorCode |= DMA2D_ERROR_TE;	/* ž顼 */
			hdma2d->status = DMA2D_STATUS_ERROR;
			/* ž顼ߥꥢ */
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_IFCR), DMA2D_IFSR_CTEIF);
			/* ž顼Хå */
			if(hdma2d->errorcallback != NULL)
				hdma2d->errorcallback(hdma2d);
			isig_sem(DMA2DTRNSEM);
		}
	}
	/*
	 *  ե졼󥨥顼Ƚ
	 */
	if((isr & DMA2D_ISR_CEIF) != 0){
		if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR)) & DMA2D_CR_CEIE) != 0){
			/* ե졼󥨥顼Disable */
			sil_andw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_CEIE);
			/* ե졼󥨥顼ߥꥢ */
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_IFCR), DMA2D_IFSR_CCEIF);
			hdma2d->ErrorCode |= DMA2D_ERROR_CE;	/* ե졼󥨥顼 */
			hdma2d->status = DMA2D_STATUS_ERROR;
			/* ե졼󥨥顼Хå */
			if(hdma2d->errorcallback != NULL)
				hdma2d->errorcallback(hdma2d);
			isig_sem(DMA2DTRNSEM);
		}
	}
	/*
	 *  žλȽ
	 */
	if((isr & DMA2D_ISR_TCIF) != 0){
		if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR)) & DMA2D_CR_TCIE) != 0){ 
			/* žλDisable */
			sil_andw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_TCIE);
			/* žλߥꥢ */  
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_IFCR), DMA2D_IFSR_CTCIF);
			hdma2d->ErrorCode = DMA2D_ERROR_NONE;	/* 顼ꥻå */
			hdma2d->status = DMA2D_STATUS_TCOMP;
			/* žλХå */
			if(hdma2d->xfercallback != NULL)
				hdma2d->xfercallback(hdma2d);
			isig_sem(DMA2DTRNSEM);
		}
	}
	/*
	 *  CLUTǥžȽ
	 */
	if((isr & DMA2D_ISR_CTCIF) != 0){
		if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR)) & DMA2D_CR_CTCIE) != 0){ 
			/* CLUTžλDisable */
			sil_andw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_CTCIE);
			/* žλߥꥢ */
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_IFCR), DMA2D_IFSR_CCTCIF);
			hdma2d->ErrorCode = DMA2D_ERROR_NONE;	/* 顼ꥻå */
			hdma2d->status = DMA2D_STATUS_CCOMP;
			/* žλХå */
			if(hdma2d->xfercallback != NULL)
				hdma2d->xfercallback(hdma2d);
			isig_sem(DMA2DTRNSEM);
		}
	}
}

/*
 *  DMA2Dν
 *  parameter1 hdma2d    DMA2DϥɥؤΥݥ
 *  return               ｪλʤE_OK
 */
ER
dma2d_init(DMA2D_Handle_t *hdma2d)
{
	volatile uint32_t tmp = 0;
	static bool_t init = false;

	if(hdma2d == NULL)
		return E_PAR;

	wai_sem(DMA2DSEM);	/* ޥեå */
	pDma2dHandler = hdma2d;
	if(!init){
		/* å */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB3ENR), RCC_AHB3ENR_DMA2DEN);
		tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB3ENR));
		(void)(tmp);
		init = true;
	}

	/* DMA2D CR쥸 */
	sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_MODE, hdma2d->Init.Mode);
	/* DMA2D OPFCC쥸 */
	sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OPFCCR), DMA2D_OPFCCR_CM, hdma2d->Init.ColorMode);
	/* DMA2D OOR쥸 */
	sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OOR), DMA2D_OOR_LO, hdma2d->Init.OutputOffset);

	/* DMA2D OPFCCR쥸 AI/RBSWAP */
	sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OPFCCR), DMA2D_OPFCCR_AI, hdma2d->Init.AlphaInverted);
	sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OPFCCR), DMA2D_OPFCCR_RBS, hdma2d->Init.RedBlueSwap);

	/* 顼ꥻå */
	hdma2d->ErrorCode = DMA2D_ERROR_NONE;
	return E_OK;
}

/*
 *  DMA2D쥤䡼
 *  parameter1 hdma2d    DMA2DϥɥؤΥݥ
 *  parameter2 LayerIdx  DMA2D쥤䡼ǥå
 *  return               ｪλʤE_OK
 */
ER
dma2d_configlayer(DMA2D_Handle_t *hdma2d, uint32_t LayerIdx)
{
	DMA2D_LayerCfg_t *pLayerCfg = &hdma2d->LayerCfg[LayerIdx];
	uint32_t tmp = 0;

	if(hdma2d->Init.Mode != DMA2D_R2M && hdma2d->Init.Mode != DMA2D_M2M){
		if(pLayerCfg->AlphaMode != DMA2D_NO_MODIF_ALPHA
			&& pLayerCfg->AlphaMode != DMA2D_REPLACE_ALPHA && pLayerCfg->AlphaMode != DMA2D_COMBINE_ALPHA){
			sig_sem(DMA2DSEM);
			return E_PAR;
		}
	}

	/* 쥤䡼0(background)DMA2D */
	if(LayerIdx == 0){
		/* DMA2D BGPFCCR쥸 */
		tmp = sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_BGPFCCR));
		tmp &= (uint32_t)~(DMA2D_BGPFCCR_CM | DMA2D_BGPFCCR_AM | DMA2D_BGPFCCR_ALPHA); 
		if (pLayerCfg->InputColorMode == CM_A4 || pLayerCfg->InputColorMode == CM_A8)
			tmp |= (pLayerCfg->InputColorMode | (pLayerCfg->AlphaMode << 16) | ((pLayerCfg->InputAlpha) & 0xFF000000));
		else
			tmp |= (pLayerCfg->InputColorMode | (pLayerCfg->AlphaMode << 16) | (pLayerCfg->InputAlpha << 24));
		sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_BGPFCCR), tmp);

		/* DMA2D BGOR쥸 */
		sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_BGOR), DMA2D_BGOR_LO, pLayerCfg->InputOffset);
		/* DMA2D BGCOLR쥸 */
		if ((pLayerCfg->InputColorMode == CM_A4) || (pLayerCfg->InputColorMode == CM_A8)){
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_BGCOLR), (pLayerCfg->InputAlpha & 0x00FFFFFF));
		}
	}
	/* 쥤䡼1(foreground)DMA2D */
	else{
		/* DMA2D FGPFCCR쥸 */
		tmp = sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGPFCCR));
		tmp &= (uint32_t)~(DMA2D_FGPFCCR_CM | DMA2D_FGPFCCR_AM | DMA2D_FGPFCCR_ALPHA); 
		if (pLayerCfg->InputColorMode == CM_A4 || pLayerCfg->InputColorMode == CM_A8)
			tmp |= (pLayerCfg->InputColorMode | (pLayerCfg->AlphaMode << 16) | ((pLayerCfg->InputAlpha) & 0xFF000000));
		else
			tmp |= (pLayerCfg->InputColorMode | (pLayerCfg->AlphaMode << 16) | (pLayerCfg->InputAlpha << 24));
		sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGPFCCR), tmp);

		/* DMA2D FGOR쥸 */
		sil_modw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGOR), DMA2D_FGOR_LO, pLayerCfg->InputOffset);
		/* DMA2D FGCOLR쥸 */
		if ((pLayerCfg->InputColorMode == CM_A4) || (pLayerCfg->InputColorMode == CM_A8)){
			sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGCOLR), (pLayerCfg->InputAlpha & 0x00FFFFFF));
		}
	}
	return E_OK;
}

/*
 *  DMA2Dž
 *  parameter1 hdma2d    DMA2DϥɥؤΥݥ
 *  parameter2 pdata     Хåեɥ쥹
 *  parameter3 DstAddr   ǥƥ͡Хåեɥ쥹
 *  parameter4 Width     žǡ
 *  parameter5 Height    žǡ⤵
 *  return               ｪλʤE_OK
 */
ER
dma2d_start(DMA2D_Handle_t *hdma2d, uint32_t pdata, uint32_t DstAddress, uint32_t Width,  uint32_t Height)
{
	if(hdma2d == NULL)
		return E_PAR;
	/* DMA2D Disable */
	sil_andw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_START);

	/* DMA2Dǡ */
	sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_NLR), (Width << 16) | Height);

	/* DMA2DΥǥƥ͡󥢥ɥ쥹 */
	sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OMAR), DstAddress);

	/* DMA2D⡼ */
	if (hdma2d->Init.Mode == DMA2D_R2M){
		uint32_t tmp = 0;
		uint32_t tmp1 = 0;
		uint32_t tmp2 = 0;
		uint32_t tmp3 = 0;
		uint32_t tmp4 = 0;

		tmp1 = pdata & DMA2D_OCOLR_ALPHA_1;
		tmp2 = pdata & DMA2D_OCOLR_RED_1;
		tmp3 = pdata & DMA2D_OCOLR_GREEN_1;
		tmp4 = pdata & DMA2D_OCOLR_BLUE_1;

		/* Prepare the value to be wrote to the OCOLR register according to the color mode */
		if (hdma2d->Init.ColorMode == DMA2D_ARGB8888){
			tmp = (tmp3 | tmp2 | tmp1| tmp4);
		}
		else if (hdma2d->Init.ColorMode == DMA2D_RGB888){
			tmp = (tmp3 | tmp2 | tmp4);  
		}
		else if (hdma2d->Init.ColorMode == DMA2D_RGB565){
			tmp2 = (tmp2 >> 19);
			tmp3 = (tmp3 >> 10);
			tmp4 = (tmp4 >> 3 );
			tmp  = ((tmp3 << 5) | (tmp2 << 11) | tmp4); 
		}
 		else if (hdma2d->Init.ColorMode == DMA2D_ARGB1555){
			tmp1 = (tmp1 >> 31);
			tmp2 = (tmp2 >> 19);
			tmp3 = (tmp3 >> 11);
			tmp4 = (tmp4 >> 3 );
			tmp  = ((tmp3 << 5) | (tmp2 << 10) | (tmp1 << 15) | tmp4);
		}
		else{ /* DMA2D_CMode = DMA2D_ARGB4444 */
			tmp1 = (tmp1 >> 28);
			tmp2 = (tmp2 >> 20);
			tmp3 = (tmp3 >> 12);
			tmp4 = (tmp4 >> 4 );
			tmp  = ((tmp3 << 4) | (tmp2 << 8) | (tmp1 << 12) | tmp4);
		}
		/* Write to DMA2D OCOLR register */
		sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_OCOLR), tmp);
	}
	else{ /* M2M, M2M_PFC or M2M_Blending DMA2D Mode */
		/* DMA2Dɥ쥹 */
		sil_wrw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGMAR), pdata);
	}
	/*
	 *  žߤ
	 */
	sil_orw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), (DMA2D_CR_TEIE | DMA2D_CR_CEIE | DMA2D_CR_TCIE));

	/* DMA2D Enable */
	sil_orw_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR), DMA2D_CR_START);
	hdma2d->status = DMA2D_STATUS_BUSY;
	return E_OK;
}

/*
 *  DMA2DžԤ
 *  parameter1 hdma2d    DMA2DϥɥؤΥݥ
 *  parameter2 Timeout   Ԥ(MS)
 *  return               ｪλʤE_OK
 */
ER
dma2d_waittransfar(DMA2D_Handle_t *hdma2d, uint32_t Timeout)
{
	ER       ercd = E_OK;
	uint32_t time = 0;

	/*
	 *  ž꤬ΥݡԤ
	 */
	if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_CR)) & DMA2D_CR_START) != 0){
		while(hdma2d->status == DMA2D_STATUS_BUSY){
			if(time > Timeout){	/* ॢȻ֤å */
				hdma2d->ErrorCode |= DMA2D_ERROR_TIMEOUT;	/* ॢȥ顼 */
				sig_sem(DMA2DSEM);
				return E_TMOUT;
			}
			twai_sem(DMA2DTRNSEM, 1);
			time++;
		}
		if(hdma2d->status == DMA2D_STATUS_ERROR)
			ercd = E_OBJ;
	}
	/*
	 *  CLUTǥ󥰤νԤ
	 */
	time = 0;
	if((sil_rew_mem((uint32_t *)(hdma2d->base+TOFF_DMA2D_FGPFCCR)) & DMA2D_FGPFCCR_START) != 0){
		while(hdma2d->status == DMA2D_STATUS_BUSY){
			if(time > Timeout){	/* ॢȻ֤å */
				hdma2d->ErrorCode |= DMA2D_ERROR_TIMEOUT;	/* ॢȥ顼 */
				sig_sem(DMA2DSEM);
				return E_TMOUT;
			}
			twai_sem(DMA2DTRNSEM, 1);
			time++;
		}
		if(hdma2d->status == DMA2D_STATUS_ERROR)
			ercd = E_OBJ;
	}
	sig_sem(DMA2DSEM);	/* ޥեå */
	return ercd;
}

/*
 *  DMA2Dߥϥɥ
 */
void dma2d_handler(void)
{
	dma2d_inthandler(pDma2dHandler);
}


#define RTC_ASYNCH_PREDIV       0x7F		/* LSE RTCå */
#define RTC_SYNCH_PREDIV        0x00FF

#define RTC_TR_RESERVED_MASK    (RTC_TR_SU | RTC_TR_ST | RTC_TR_MNU | RTC_TR_MNT | \
								 RTC_TR_HU | RTC_TR_HT | RTC_TR_PM)
#define RTC_DR_RESERVED_MASK    (RTC_DR_DU | RTC_DR_DT  | RTC_DR_MU | RTC_DR_MT | \
								 RTC_DR_WDU | RTC_DR_YU | RTC_DR_YT)
#define RTC_INIT_MASK           ((uint32_t)0xFFFFFFFF)  
#define RTC_RSF_MASK            ((uint32_t)0xFFFFFF5F)

#define RTC_TIMEOUT_VALUE       (1000*1000)
#define LSI_TIMEOUT_VALUE       (100*1000)
#define RCC_DBP_TIMEOUT_VALUE   (100*1000)
#define RCC_LSE_TIMEOUT_VALUE   (5000*1000)

uint32_t rtcerrorcode;

static void (*rtcalarmA_callback)(void);	/* DMA transfer complete Memory1 callback */
static void (*rtcalarmB_callback)(void);	/* DMA transfer error callback */
static void (*rtcwakeup_callback)(void);	/* rtc wakeup callback */

/*
 *  RTCȥ⡼
 */
static ER
rtc_requestInitMode(void)
{
	uint32_t tick = 0;

	/*
	 * ˥⡼Ƚ
	 */
	if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INITF) == 0){
	    /*
		 *  ˥⡼
		 */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), RTC_INIT_MASK);
	    while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INITF) == 0){
			if(tick > RTC_TIMEOUT_VALUE){
				return E_TMOUT;
			}
		}
		sil_dly_nse(1000);
		tick++;
	}
	return E_OK;
}

/*
 *  RTCѽ
 */
void
rtc_init(intptr_t exinf)
{
	uint32_t tick = 0;
	volatile uint32_t tmpreg;

	rtcerrorcode = 0;
	/*
	 *  LSI򥪥դ
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR), RCC_CSR_LSION);
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CSR)) & RCC_CSR_LSIRDY) != 0){
		if( tick > LSI_TIMEOUT_VALUE){
			rtcerrorcode |= RTC_ERROR_LSI;
			return;
		}
		sil_dly_nse(1000);
		tick++;
	}

	/*
	 *  LSE򥪥󤹤
	 */
    /*
	 *  ХååץɥᥤФ륢
	 */
	sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_DBP);
	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_DBP) == 0){
		if(tick > RCC_DBP_TIMEOUT_VALUE){
			rtcerrorcode |= RTC_ERROR_LSE;
			return;
		}
		sil_dly_nse(1000);
		tick++;
	}
    
    /*
	 *  LSEON / LSEBYPꥻå
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSEBYP);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSEON);

	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & RCC_BDCR_LSERDY) != 0){
		if(tick > RCC_LSE_TIMEOUT_VALUE){
			rtcerrorcode |= RTC_ERROR_LSE;
			return;
		}
		sil_dly_nse(1000);
		tick++;
	}
    
    /*
	 *  LSE
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSEBYP);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_LSEON);

	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & RCC_BDCR_LSERDY) == 0){
		if(tick > RCC_LSE_TIMEOUT_VALUE){
			rtcerrorcode |= RTC_ERROR_LSE;
			return;
		}
		sil_dly_nse(1000);
		tick++;
	}

    /*
	 *  RTCå
	 */
	if((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & RCC_BDCR_RTCSEL) != RCC_BDCR_RTCSEL_0){
		/*
		 *  åŸ
		 */
		tmpreg = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & ~(RCC_BDCR_RTCSEL));

		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_BDRST);
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_BDRST);

		sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), tmpreg);

		if((tmpreg & RCC_BDCR_LSERDY) != 0){
			tick = 0;
			/* Wait till LSE is ready */  
			while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR)) & RCC_BDCR_LSERDY) == 0){
				if(tick > RCC_LSE_TIMEOUT_VALUE){
					rtcerrorcode |= RTC_ERROR_RTC;
					return;
				}
				sil_dly_nse(1000);
				tick++;
			}
		}
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_RTCPRE);
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_RTCSEL_0);
	}

	/*
	 *  RTCΥå
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_BDCR), RCC_BDCR_RTCEN);

	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	/*
	 *  RTC
	 */
	if(rtc_requestInitMode() != E_OK){
		rtcerrorcode |= RTC_ERROR_RTC;
	    /* ץƥ */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
		return;
	}
	else{ 
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), (RTC_CR_FMT | RTC_CR_OSEL | RTC_CR_POL));

		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_PRER), RTC_SYNCH_PREDIV);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_PRER), (RTC_ASYNCH_PREDIV << 16));

		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), RTC_ISR_INIT);

	    /* ץƥ */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
		return;
	}
}

/*
 *  RTCλؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_set_time(struct tm2 *pt)
{
	uint32_t timetmp = 0;
	uint32_t datetmp = 0;

	if(pt == NULL)
		return E_PAR;
	if(rtcerrorcode != 0)
		return E_SYS;

	SVC_PERROR(wai_sem(RTCSEM));

	datetmp = (((uint32_t)byte2bcd(pt->tm_year - 30) << 16) |
				((uint32_t)byte2bcd(pt->tm_mon) << 8) |
				((uint32_t)byte2bcd(pt->tm_mday)) |
				((uint32_t)pt->tm_wday << 13));
    timetmp = (uint32_t)(((uint32_t)byte2bcd(pt->tm_hour) << 16) |
				((uint32_t)byte2bcd(pt->tm_min) << 8) |
				((uint32_t)byte2bcd(pt->tm_sec)));

	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	/*
	 *  ⡼
	 */
	if(rtc_requestInitMode() != E_OK){
		/* ץƥ */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
		SVC_PERROR(sig_sem(RTCSEM));
		return E_TMOUT;
	}
	else{
		/*
		 *  ա
		 */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_DR), (datetmp & RTC_DR_RESERVED_MASK));
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_TR), (timetmp & RTC_TR_RESERVED_MASK));

		/*  ⡼ɽλ */
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_BKP);
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), RTC_ISR_INIT);

		/*
		 *  Ʊ
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR)) & RTC_CR_BYPSHAD) == 0){
			uint32_t tick = 0;

			sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), RTC_RSF_MASK);
    		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_RSF) == 0){
				if(tick > (RTC_TIMEOUT_VALUE/1000)){
					SVC_PERROR(sig_sem(RTCSEM));
					return E_TMOUT;
				}
				dly_tsk(1);
				tick++;
			}
		}

		/* ץƥ */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
	}
	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTCλФؿ
 *
 *  PONIXtm¤ΤѤ
 *  PONIXΥ󥯥롼ɤʤθƱܤtm2ɥ饤ФȤ롣
 */
ER
rtc_get_time(struct tm2 *pt)
{
	uint32_t timetmp = 0;
	uint32_t datetmp = 0;

	if(pt == NULL)
		return E_PAR;
	if(rtcerrorcode != 0)
		return E_SYS;

	SVC_PERROR(wai_sem(RTCSEM));
	/*
	 *  
	 */
	timetmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_TR)) & RTC_TR_RESERVED_MASK;
	pt->tm_hour = (uint8_t)bcd2byte((uint8_t)((timetmp & (RTC_TR_HT | RTC_TR_HU)) >> 16));
	pt->tm_min = (uint8_t)bcd2byte((uint8_t)((timetmp & (RTC_TR_MNT | RTC_TR_MNU)) >>8));
	pt->tm_sec = (uint8_t)bcd2byte((uint8_t)(timetmp & (RTC_TR_ST | RTC_TR_SU)));
  
	/*
	 *  ռ
	 */
	datetmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_DR)) & RTC_DR_RESERVED_MASK;
	pt->tm_year = (uint8_t)bcd2byte((uint8_t)((datetmp & (RTC_DR_YT | RTC_DR_YU)) >> 16)) + 30;
	pt->tm_mon = (uint8_t)bcd2byte((uint8_t)((datetmp & (RTC_DR_MT | RTC_DR_MU)) >> 8));
	pt->tm_mday = (uint8_t)bcd2byte((uint8_t)(datetmp & (RTC_DR_DT | RTC_DR_DU)));
	pt->tm_wday = (uint8_t)((datetmp & (RTC_DR_WDU)) >> 13); 

	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC顼
 *  parameter1 : parm: Pointer to Alarm structure
 *  parameter2 : ptm: Pointer to struct tm2
 *  return ER
 */
ER
rtc_setalarm(RTC_Alarm_t *parm, struct tm2 *ptm)
{
	uint32_t tick = 0;
	uint32_t tmparm = 0, subsecond = 0;
	uint32_t tmp, day;

	if(parm == NULL || ptm == NULL)
		return E_PAR;
	if(rtcerrorcode != 0)
		return E_SYS;

	if(parm->dayselect == ALARMDAYSEL_DATE)
		day = ptm->tm_mday;
	else
		day = ptm->tm_wday;

	/*
	 *  ALARM-AB쥸ͤ
	 */
	SVC_PERROR(wai_sem(RTCSEM));
    tmparm = ((byte2bcd(ptm->tm_hour) << 16) | (byte2bcd(ptm->tm_min) << 8) |
              (byte2bcd(ptm->tm_sec)) | (byte2bcd(day) << 24) |
              ((uint32_t)parm->dayselect) | (parm->alarmmask)); 
	subsecond = (uint32_t)(parm->subsecond | parm->subsecondmask);

	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	if(parm->alarm == RTC_ALARM_A){
		/*
		 *  ALARM-Aꡢ쥸
		 */
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRAE);
		tmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR));
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), (~(RTC_ISR_ALRAF | RTC_ISR_INIT) | tmp));
		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_ALRAWF) == 0){
			if(tick > (RTC_TIMEOUT_VALUE/1000)){
			    /* ץƥ */
				sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
				SVC_PERROR(sig_sem(RTCSEM));
				return E_TMOUT;
			}
			dly_tsk(1);
			tick++;
		}

		/*
		 *  ALARM-A
		 */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMAR), tmparm);
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMASSR), subsecond);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRAE);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRAIE);
		rtcalarmA_callback = parm->callback;
	}
	else{
		/*
		 *  ALARM-Bꡢ쥸
		 */
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRBE);
		tmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR));
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), (~(RTC_ISR_ALRBF | RTC_ISR_INIT) | tmp));
		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_ALRBWF) == 0){
			if(tick > (RTC_TIMEOUT_VALUE/1000)){
			    /* ץƥ */
				sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
				SVC_PERROR(sig_sem(RTCSEM));
				return E_TMOUT;
			}
			dly_tsk(1);
			tick++;
		}

		/*
		 *  ALARM-B
		 */
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMBR), tmparm);
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMBSSR), subsecond);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRBE);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRBIE);
		rtcalarmB_callback = parm->callback;
	}

	/*
	 * RTC ALARMEXTI
	 */
	sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), EXTI_IMR1_IM17);
	sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), EXTI_RTSR1_TR17);

    /* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC顼
 *  parameter1 : Alarm: 顼
 *  return ER
 */
ER
rtc_stopalarm(uint32_t Alarm)
{
	uint32_t tick = 0;

	if(rtcerrorcode != 0)
		return E_SYS;
	SVC_PERROR(wai_sem(RTCSEM));
	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	if(Alarm == RTC_ALARM_A){
		/*
		 *  ALARM-Aߥ֥͡
		 */
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRAE);
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRAIE);
		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_ALRAWF) == 0){
			if(tick > (RTC_TIMEOUT_VALUE/1000)){
			    /* ץƥ */
				sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
				SVC_PERROR(sig_sem(RTCSEM));
				return E_TMOUT;
			}
			dly_tsk(1);
			tick++;
		}
		rtcalarmA_callback = NULL;
	}
	else{
		/*
		 *  ALARM-Bߥ֥͡
		 */
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRBE);
		sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_ALRBIE);
		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_ALRBWF) == 0){
			if( tick > RTC_TIMEOUT_VALUE){
			    /* ץƥ */
				sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
				SVC_PERROR(sig_sem(RTCSEM));
				return E_TMOUT;
			}
			dly_tsk(1);
			tick++;
		}
		rtcalarmB_callback = NULL;
	}
    /* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK; 
}

/*
 *  RTC顼ͼ
 *  parameter1 : parm: Pointer to Alarm structure
 *  parameter2 : ptm: Pointer to struct tm2
 *  return ER
 */
ER
rtc_getalarm(RTC_Alarm_t *parm, struct tm2 *ptm, uint32_t Alarm)
{
	uint32_t tmparm = 0;
	uint32_t subsecond = 0;

	if(parm == NULL || ptm == NULL)
		return E_PAR;
	if(rtcerrorcode != 0)
		return E_SYS;

	SVC_PERROR(wai_sem(RTCSEM));
	if(Alarm == RTC_ALARM_A){
		/*
		  ALARM-A 쥸
		 */
		parm->alarm = RTC_ALARM_A;
		tmparm = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMAR));
		subsecond = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMASSR)) & RTC_ALRMASSR_SS;
	}
	else{
		/*
		  ALARM-B 쥸
		 */
		parm->alarm = RTC_ALARM_B;
		tmparm = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMBR));
		subsecond = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ALRMBSSR)) & RTC_ALRMBSSR_SS;
	}

	/*
	 *  쥸ѥ᡼Ѵ
	 */
	ptm->tm_hour = bcd2byte((uint8_t)((tmparm & (RTC_ALRMAR_HT | RTC_ALRMAR_HU)) >> 16));
	ptm->tm_min  = bcd2byte((uint8_t)((tmparm & (RTC_ALRMAR_MNT | RTC_ALRMAR_MNU)) >> 8));
	ptm->tm_sec = bcd2byte((uint8_t)(tmparm & (RTC_ALRMAR_ST | RTC_ALRMAR_SU)));
	parm->subsecond = (uint32_t) subsecond;
	parm->dayselect = (uint32_t)(tmparm & RTC_ALRMAR_WDSEL);
	if(parm->dayselect == ALARMDAYSEL_DATE)
		ptm->tm_mday = bcd2byte((uint8_t)((tmparm & (RTC_ALRMAR_DT | RTC_ALRMAR_DU)) >> 24));
	else
		ptm->tm_wday = bcd2byte((uint8_t)((tmparm & (RTC_ALRMAR_DT | RTC_ALRMAR_DU)) >> 24));

	parm->alarmmask = (uint32_t)(tmparm & ALARMMASK_ALL);
	SVC_PERROR(sig_sem(RTCSEM));
	return E_OK;
}

/*
 *  RTC WAKEUP޽
 *  parameter1 : parm: Pointer to wakeup structure
 *  parameter2 : func: callback function
 *  return ER
 */
ER
rtc_wakeup_init(RTC_Wakeup_t *parm, void (*func)(void))
{
	uint32_t prer, isr;

	if(parm == NULL)
		return E_PAR;
	parm->wakeuptimerPrescaler = (4 - (sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR)) & RTC_CR_WUCKSEL));
	prer = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_PRER));
	parm->asynchPrescaler = ((prer >> 16) & 0x7f) + 1;
	parm->synchPrescaler  = prer & 0xffff;

	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_WUTE);
	isr  = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INIT;
	isr |= ~(RTC_ISR_WUTF | RTC_ISR_INIT);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), isr);
	sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), EXTI_PR1_PR19);
	sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_WUTIE);
	/*
	 *  EXTI WAKEUPLINE
	 */
	sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), EXTI_RTSR1_TR19);
	sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), EXTI_IMR1_IM19);

    /* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);

	rtcwakeup_callback = func;
	return E_OK;
}

/*
 *  RTC WAKEUP
 *  parameter1 : parm: WAKEUP
 *  return ER
 */
ER
rtc_setup_wakeuptime(uint32_t time)
{
	uint32_t timeout = LSI_TIMEOUT_VALUE;
	/*
	 *  WAUPTIMER
	 */
	if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR)) & RTC_CR_WUTE) != 0){
		while((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_WUTWF) == 0){
			timeout--;
			if(timeout == 0)
				return E_TMOUT;
			sil_dly_nse(1000);
		}
	}
	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);

	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_WUTE);

	if(time != 0 && time < 0xFFFF){
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), EXTI_PR1_PR19);
		sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WUTR), time);
		sil_orw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_WUTE);
	}
    /* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
	return E_OK;
}

/*
 *  SSRμФ
 */
uint32_t
rtc_get_ssr(void)
{
	return sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_SSR));
}

/*
 *  RTCߥϥɥ
 */
void
rtc_handler(void)
{
	uint32_t tmp;

	if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_CR_ALRAE) != 0){
		/*
		 * ALARM-A߳ǧ
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR)) & RTC_CR_ALRAIE) != 0){
			/*
			 *  ALARM-AХå
			 */
			if(rtcalarmA_callback != NULL)
				rtcalarmA_callback();

			/* ץƥ */
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);
			/*
			 *  ALARM-Aߥꥢ
			 */
			tmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INIT;
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), (~(RTC_ISR_ALRAF | RTC_ISR_INIT) | tmp));
		    /* ץƥ */
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
		}
	}
	if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_CR_ALRBE) != 0){
		/*
		 * ALARM-B߳ǧ
		 */
		if((sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR)) & RTC_CR_ALRBIE) != 0){
			/*
			 *  ALARM-BХå
			 */
			if(rtcalarmB_callback != NULL)
				rtcalarmB_callback();

			/* ץƥ */
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);
			/*
			 *  ALARM-Bߥꥢ
			 */
			tmp = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INIT;
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), (~(RTC_ISR_ALRBF | RTC_ISR_INIT) | tmp));
		    /* ץƥ */
			sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);
		}
	}
	/*
	 *  EXTI RTCڥǥ󥰥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), EXTI_PR1_PR17);
}

/*
 *  WAKEUPߥϥɥ
 */
void
wakeup_handler(void)
{
	uint32_t isr;

	/* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xCA);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0x53);
	/*
	 *  WAKEUP TIMER
	 */
	sil_andw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_CR), RTC_CR_WUTE);

	/*
 	 *  WUTF flagꥢ
	 */
	isr  = sil_rew_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR)) & RTC_ISR_INIT;
	isr |= ~(RTC_ISR_WUTF | RTC_ISR_INIT);
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_ISR), isr);

    /* ץƥ */
	sil_wrw_mem((uint32_t *)(TADR_RTC_BASE+TOFF_RTC_WPR), 0xFF);

	/*
	 *  EXTIγߥꥢ
	 */
	sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), EXTI_PR1_PR19);

	/*
	 *  Хåؿ¹
	 */
	if(rtcwakeup_callback != NULL)
		rtcwakeup_callback();
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint32_t mcount, mcounter, wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;
	if(hini->Window > 127 || hini->Window < 64)
		return E_PAR;
	if(hini->Counter > 127 || hini->Counter < 64)
		return E_PAR;

	mcount    = hini->MCounter;
	mcounter  = 0;
	wdogcount = hini->Counter;
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB3ENR), RCC_APB3ENR_WWDG1EN);
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	if(mcount == 0)
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_WDGTB | hini->Window));
	else
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_EWI | WWDG_CFR_WDGTB | hini->Window));
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), (WWDG_CR_WDGA | wdogcount));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB3ENR), RCC_APB3ENR_WWDG1EN);
	wdogcount = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDG_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDG_CFR)) & WWDG_CFR_EWI) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDG_SR)) & WWDG_SR_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDG_SR), WWDG_SR_EWIF);
		if(mcounter++ < mcount)
			sil_wrw_mem((uint32_t *)(base+TOFF_WWDG_CR), wdogcount);
	}
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    clkbit;
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
	{ RCC_AHB4ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION0, GPIO_NOPULL},
#if NUCLEO_ZI1
	{ RCC_AHB4ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION7, GPIO_NOPULL},
#else
	{ RCC_AHB4ENR_GPIOEEN, TADR_GPIOE_BASE, PINPOSITION1, GPIO_NOPULL},
#endif
	{ RCC_AHB4ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION14, GPIO_NOPULL}
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))

/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB4ENR), pled->clkbit);
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (1 << pled->pinpos)<<16);
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_IDR)) & (1<<(pled->pinpos))) != 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t reg1, reg2, i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		reg1 = reg2 = 0;
		if((led_data & (1<<i)) != 0)
			reg2 = 1 << pled->pinpos;
		else
			reg1 = 1 << pled->pinpos;

		/* 񤭹ߥǡƽ񤭹 */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (reg1<<16) | reg2);
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


#define TADR_PSW_BASE   TADR_GPIOC_BASE
#define PSWGIOPEN       RCC_AHB4ENR_GPIOCEN
#define	PSW_PIN         PINPOSITION13

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	GPIO_Init_t GPIO_Init_Data;
	volatile uint32_t temp;

	/* ϥݡȤ */
	/* Enable the BUTTON Clock */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB4ENR), PSWGIOPEN);
	temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB4ENR));
	(void)(temp);
	/* ⡼ */
	GPIO_Init_Data.mode      = GPIO_MODE_IT_FALLING;
	/* ץ륢å ץ */
	GPIO_Init_Data.pull      = GPIO_NOPULL;
	gpio_setup(TADR_PSW_BASE, &GPIO_Init_Data, PSW_PIN);
}

/*
 * PUSHåХåؿ
 */
void setup_sw_func(intptr_t exinf)
{
	exti_func[PSW_PIN] = (void (*)(void))exinf;
}

/*
 * EXTI15-10ߥϥɥ
 */
void
exti15_handler(void)
{
	/* EXTI line interrupt detected */
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1));
	uint32_t i;
	for(i = 0 ; i < NUM_EXTI15_FUNC ; i++){
		if((istatus & (1<<(i+EXTI15_BASENO))) != 0){
			if(exti_func[i+EXTI15_BASENO] != NULL)
				(exti_func[i+EXTI15_BASENO])();
		}
	}
    sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), istatus);
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

