/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: spi.c 699 2019-12-20 17:56:39Z roi $
 */
/*
 * 
 *  SPIɥ饤дؿ
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "spi.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#ifndef SPI_GPIO_SPEED
#define SPI_GPIO_SPEED  GPIO_SPEED_HIGH
#endif

#ifndef SPI_REQUEST_FIFO_THREAD
#define SPI_REQUEST_FIFO_THREAD 0x00000000
#endif

#ifndef SPI_REQUEST_AUTOSUSMODE
#define SPI_REQUEST_AUTOSUSMODE SPI_MASTER_RX_AUTOSUSP_DISABLE
#endif

#define SPI_IER_TXVALUE (SPI_IER_UDRIE | SPI_IER_TIFREIE | SPI_IER_MODFIE)
#define SPI_IER_RXVALUE (SPI_IER_OVRIE | SPI_IER_TIFREIE | SPI_IER_MODFIE)

/*
 *  SPIOݡID֥åФΥޥ
 */
#define INDEX_SPI(spiid)        ((uint_t)((spiid) - 1))

#define GPIO_AF5_SPI1          ((uint8_t)0x05)  /* SPI1 Alternate Function mapping        */
#define GPIO_AF5_SPI2          ((uint8_t)0x05)  /* SPI2 Alternate Function mapping        */
#define GPIO_AF6_SPI3          ((uint8_t)0x06)  /* SPI3 Alternate Function mapping        */

#define SPI_TIMEOUT_VALUE  (10*1000)

static const SPI_PortControlBlock spi_pcb[NUM_SPIPORT] = {
  {	TADR_SPI1_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB4ENR), RCC_AHB4ENR_GPIOAEN, RCC_AHB4ENR_GPIOAEN,
	(TADR_RCC_BASE+TOFF_RCC_APB2ENR), RCC_APB2ENR_SPI1EN,
	TADR_GPIOA_BASE, TADR_GPIOA_BASE, TADR_GPIOA_BASE,
	5, 6, 7, GPIO_AF5_SPI1,
	DMA_REQUEST_SPI1_TX, DMA_REQUEST_SPI1_RX },

  {	TADR_SPI2_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB4ENR), RCC_AHB4ENR_GPIOBEN, RCC_AHB4ENR_GPIOBEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1LENR), RCC_APB1LENR_SPI2EN,
	TADR_GPIOB_BASE, TADR_GPIOB_BASE, TADR_GPIOB_BASE,
	13, 14, 15, GPIO_AF5_SPI2,
	DMA_REQUEST_SPI2_TX, DMA_REQUEST_SPI2_RX },

  {	TADR_SPI3_BASE,
	(TADR_RCC_BASE+TOFF_RCC_AHB4ENR), RCC_AHB4ENR_GPIOCEN, RCC_AHB4ENR_GPIOCEN,
	(TADR_RCC_BASE+TOFF_RCC_APB1LENR), RCC_APB1LENR_SPI3EN,
	TADR_GPIOC_BASE, TADR_GPIOC_BASE, TADR_GPIOC_BASE,
	10, 11, 12, GPIO_AF6_SPI3,
	DMA_REQUEST_SPI3_TX, DMA_REQUEST_SPI3_RX },
};

static SPI_Handle_t SpiHandle[NUM_SPIPORT];
static DMA_Handle_t hdma_tx[NUM_SPIPORT];
static DMA_Handle_t hdma_rx[NUM_SPIPORT];


/*
 *  SPI DMAХåؿ
 */
static void
spi_dmatransmit_func(DMA_Handle_t *hdma)
{
	SPI_Handle_t *hspi = (SPI_Handle_t *)hdma->localdata;

    if(hdma->Init.Mode != DMA_CIRCULAR)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_EOTIE);
	if(hspi->txxfercallback != NULL)
		hspi->txxfercallback(hdma);
}

/*
 *  SPI DMAХåؿ
 */
static void
spi_dmareceive_func(DMA_Handle_t *hdma)
{
	SPI_Handle_t *hspi = (SPI_Handle_t *)hdma->localdata;

    if(hdma->Init.Mode != DMA_CIRCULAR)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_EOTIE);
	if(hspi->rxxfercallback != NULL)
		hspi->rxxfercallback(hdma);
}

/*
 *  SPI DMAХåؿ
 */
static void
spi_dmatransrecv_func(DMA_Handle_t *hdma)
{
	SPI_Handle_t *hspi = (SPI_Handle_t *)hdma->localdata;

    if(hdma->Init.Mode != DMA_CIRCULAR)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_EOTIE);
	if(hspi->rxxfercallback != NULL)
		hspi->rxxfercallback(hdma);
}

/*
 *  SPIžλԤ
 */
ER
spi_inwait(SPI_Handle_t *hspi, uint32_t timeout)
{
	ER ercd = E_OK;
	int tick = timeout;

	while((hspi->status & SPI_STATUS_BUSY) != 0 && tick > 0){
		if(hspi->Init.semid != 0)
    		twai_sem(hspi->Init.semid, 1);
		else
			dly_tsk(1);
		tick--;
	}
	if(hspi->hdmarx != NULL)
		dma_end(hspi->hdmarx);
	if(hspi->hdmatx != NULL)
		dma_end(hspi->hdmatx);
	if(hspi->ErrorCode != 0)
		ercd = E_OBJ;
	else if(tick == 0)
		ercd = E_TMOUT;
	hspi->TxXferCount = 0;
	hspi->RxXferCount = 0;
	return ercd;
}

/*
 *  SPI DMA顼Хåؿ
 */
static
void spi_dmaerror_func(DMA_Handle_t *hdma)
{
	SPI_Handle_t *hspi = (SPI_Handle_t *)hdma->localdata;

	hspi->TxXferCount = 0;
	hspi->RxXferCount = 0;
	hspi->status= SPI_STATUS_READY;
	hspi->ErrorCode |= SPI_ERROR_DMA;
	syslog_2(LOG_ERROR, "SPI DMA Error handle[%08x] ErrorCode[%08x] !", hdma, hdma->ErrorCode);
	sil_dly_nse(1000*500);	/* 顼ϥ󥰥åײ0.5msecԤ */
}

/*
 *  SPI
 *  parameter1  port: SPIݡֹ
 *  parameter2  spii: SPI깽¤ΤؤΥݥ
 *  return SPIϥɥؤΥݥ󥿡NULLǥ顼
 */
SPI_Handle_t *
spi_init(ID port, SPI_Init_t *spii)
{
	GPIO_Init_t GPIO_Init_Data;
	SPI_Handle_t *hspi;
	const SPI_PortControlBlock *spcb;
	DMA_Handle_t *hdma;
	uint32_t no;
	volatile uint32_t tmpreg;

	if(port < SPI1_PORTID || port > NUM_SPIPORT)
		return NULL;
	if(spii == NULL)
		return NULL;

	no = INDEX_SPI(port);
	hspi = &SpiHandle[no];
	if(hspi->status != SPI_STATUS_RESET)
		return NULL;
	spcb = &spi_pcb[no];
	memcpy(&hspi->Init, spii, sizeof(SPI_Init_t));
	hspi->base = spcb->base;
	hspi->port = port;

	/*
	 *  SPI å
	 */
	sil_orw_mem((uint32_t *)spcb->gioclockbase, spcb->gioclockbit);
	tmpreg = sil_rew_mem((uint32_t *)spcb->gioclockbase);
	sil_orw_mem((uint32_t *)spcb->gioclockbase, spcb->gioclockbit2);
	tmpreg = sil_rew_mem((uint32_t *)spcb->gioclockbase);
	sil_orw_mem((uint32_t *)spcb->spiclockbase, spcb->spiclockbit);
	tmpreg = sil_rew_mem((uint32_t *)spcb->spiclockbase);
    ((void)(tmpreg));
  
	/*
	 *  SPI GPIOԥ
	 */
	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	if(spii->GpioClkPull >= 0 && spii->GpioClkPull <= GPIO_PULLDOWN)
		GPIO_Init_Data.pull  = spii->GpioClkPull;
	else
		GPIO_Init_Data.pull  = GPIO_PULLUP;
	GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
	GPIO_Init_Data.speed     = SPI_GPIO_SPEED;
	GPIO_Init_Data.alternate = spcb->apvalue;
	gpio_setup(spcb->giobase1, &GPIO_Init_Data, spcb->sckpin);
	if(spii->GpioDataPull >= 0 && spii->GpioDataPull <= GPIO_PULLDOWN)
		GPIO_Init_Data.pull  = spii->GpioDataPull;
	else
		GPIO_Init_Data.pull  = GPIO_PULLUP;
	gpio_setup(spcb->giobase2, &GPIO_Init_Data, spcb->misopin);
	gpio_setup(spcb->giobase3, &GPIO_Init_Data, spcb->mosipin);

	/*
	 *  SPIѼDMP
	 */
	hdma = &hdma_tx[no];
	if((hdma->base = dma_stream_base(spii->TxDMAStream)) != 0){
		hdma->Init.Request             = spcb->dmatxrequest;
		hdma->Init.Direction           = DMA_MEMORY_TO_PERIPH;
		hdma->Init.PeriphInc           = DMA_PINC_DISABLE;
		hdma->Init.MemInc              = DMA_MINC_ENABLE;
		hdma->Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
		hdma->Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
		hdma->Init.Mode                = DMA_NORMAL;
		hdma->Init.Priority            = DMA_PRIORITY_LOW;
		hdma->Init.FIFOMode            = DMA_FIFOMODE_DISABLE;
		hdma->Init.FIFOThreshold       = DMA_FIFO_THRESHOLD_FULL;
		hdma->Init.MemBurst            = DMA_MBURST_INC4;
		hdma->Init.PeriphBurst         = DMA_PBURST_INC4;

		dma_init(hdma);
		hspi->hdmatx    = hdma;
		hdma->localdata = hspi;
	}
	else
		hspi->hdmatx    = NULL;

	/*
	 *  SPIѼDMP
	 */
	hdma = &hdma_rx[no];
	if((hdma->base = dma_stream_base(spii->RxDMAStream)) != 0){
		hdma->Init.Request             = spcb->dmarxrequest;
		hdma->Init.Direction           = DMA_PERIPH_TO_MEMORY;
		hdma->Init.PeriphInc           = DMA_PINC_DISABLE;
		hdma->Init.MemInc              = DMA_MINC_ENABLE;
		hdma->Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
		hdma->Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
		hdma->Init.Mode                = DMA_NORMAL;
		hdma->Init.Priority            = DMA_PRIORITY_HIGH;
		hdma->Init.FIFOMode            = DMA_FIFOMODE_DISABLE;
		hdma->Init.FIFOThreshold       = DMA_FIFO_THRESHOLD_FULL;
		hdma->Init.MemBurst            = DMA_MBURST_INC4;
		hdma->Init.PeriphBurst         = DMA_PBURST_INC4; 

		dma_init(hdma);
		hspi->hdmarx    = hdma;
		hdma->localdata = hspi;
	}
	else
		hspi->hdmarx    = NULL;

	if(spi_reset(hspi) == E_OK)
		return hspi;
	else
		return NULL;
}

/*
 *  SPIλ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  return ER
 */
ER
spi_deinit(SPI_Handle_t *hspi)
{
	if(hspi == NULL)
		return E_PAR;

	if(hspi->hdmatx != NULL){
		dma_deinit(hspi->hdmatx);
		hspi->hdmatx = NULL;
	}
	if(hspi->hdmarx != NULL){
		dma_deinit(hspi->hdmarx);
		hspi->hdmarx = NULL;
	}

	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

	hspi->ErrorCode = SPI_ERROR_NONE;
	hspi->status = SPI_STATUS_RESET;
	return E_OK;
}

/*
 *  SPI⥸塼Υꥻå
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  return ER
 */
ER
spi_reset(SPI_Handle_t *hspi)
{
	uint32_t crc_length = 0;

	/*
	 *  SPIǥ֥
	 */
	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

	if(hspi->Init.CRC == SPI_CRC_ENABLE){
		if(hspi->port > SPI3_PORTID && hspi->Init.CRCLength > SPI_CRC_LENGTH_16BIT)
			return E_PAR;

		/* Align the CRC Length on the data size */
		if(hspi->Init.CRCLength == SPI_CRC_LENGTH_DATASIZE)
			crc_length = hspi->Init.DataSize;
		else
			crc_length = hspi->Init.CRCLength;

		/* Verify that the CRC Length is higher than DataSize */
		if(hspi->Init.DataSize > crc_length)
			return E_PAR;
	}

	/*
	 *  SPI CR1/CFG1/CFG2쥸
	 */
	if(hspi->Init.NSS == SPI_NSS_SOFT
		 && hspi->Init.Mode == SPI_MODE_MASTER
			&& hspi->Init.NSSPolarity == SPI_NSS_POLARITY_LOW){
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SSI);
	}
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), (hspi->Init.Prescaler | \
		hspi->Init.CRC | (crc_length << 16) | SPI_REQUEST_FIFO_THREAD | hspi->Init.DataSize));
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG2), (hspi->Init.NSSPMode | \
		hspi->Init.TIMode      | hspi->Init.NSSPolarity | hspi->Init.NSS | \
		hspi->Init.CLKPolarity | hspi->Init.CLKPhase    | hspi->Init.SignBit | \
		hspi->Init.Mode        | hspi->Init.MasterInterDataIdleness | hspi->Init.Direction | \
		hspi->Init.MasterSSIdleness | hspi->Init.IOSwap));

	/*
	 *  SPI CRC
	 */
	if(hspi->Init.CRC == SPI_CRC_ENABLE){
		/*
		 *  TXCRCѥ
		 */
		if(hspi->Init.TxCRCInitializationPattern == SPI_CRC_INITIALIZATION_ALL_ONE_PATTERN)
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_TCRCINI);
		else
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_TCRCINI);
		/*
		 *  RXCRCѥ
		 */
		if(hspi->Init.RxCRCInitializationPattern == SPI_CRC_INITIALIZATION_ALL_ONE_PATTERN)
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_RCRCINI);
		else
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_RCRCINI);

		/*
		 *  Enable 33/17 bits CRC computation
		 */
	    if((hspi->port > SPI3_PORTID && crc_length == SPI_CRC_LENGTH_16BIT) || \
				(hspi->port <= SPI3_PORTID && crc_length == SPI_CRC_LENGTH_32BIT))
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_CRC33_17);
		else
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_CRC33_17);

		/*
		 *  CRC polynomial
		 */
		sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_CRCPOLY), hspi->Init.CRCPolynomial);
	}

	/*
	 *  SPIξ硢I2S򥪥
	 */
	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_I2SCFGR), SPI_I2SCFGR_I2SMOD);

	/*
	 *  졼֥⡼ɤξ硢Ԥ
	 */
	if(hspi->Init.Mode == SPI_MODE_SLAVE){
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_UDRCFG, SPI_CFG1_UDRCFG_1);
		/* Set Default Underrun configuration */
		if(hspi->Init.CRC == SPI_CRC_DISABLE)
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_UDRDET_0);
	}
	/*
	 *  ޥ⡼AFCNTRԤ
	 */
	else{
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG2), SPI_CFG2_AFCNTR, (hspi->Init.MasterKeepIOState & SPI_CFG2_AFCNTR));
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_MASRX, SPI_REQUEST_AUTOSUSMODE);
	}

	hspi->ErrorCode = SPI_ERROR_NONE;
	hspi->status = SPI_STATUS_READY;
	return E_OK;
}

/*
 *  SPI-DMACIRCLAR⡼ɤꤹ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  no:   DMA 0-TX, 1-RX
 */
ER
spi_dmacirclarmode(SPI_Handle_t *hspi, uint8_t no)
{
	DMA_Handle_t *hdma;

	if(hspi == NULL || no > 1)
		return E_PAR;
	if(no == 0)
		hdma = hspi->hdmatx;
	else
		hdma = hspi->hdmarx;
	hdma->Init.Mode = DMA_SxCR_CIRC;
	if(hdma->muxbase == TADR_DMAMUX1_BASE)
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_CIRC);
	else
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_BDMA_CCR), BDMA_CCR_CIRC);
	return E_OK;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ(4Хȥ饤)
 *  parameter3  length: (CRCǤ+1)
 *  return ER
 */
ER
spi_transmit(SPI_Handle_t *hspi, uint8_t *pdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || pdata == NULL || length == 0)
		return E_PAR;
	if((((uint32_t)pdata) & 3) != 0)
		return E_PAR;
	if(hspi->hdmatx == NULL)
		return E_OBJ;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

	/* Configure communication */
	hspi->xmode       = SPI_XMODE_TX;
	hspi->status      = SPI_STATUS_BUSY;
	hspi->ErrorCode   = SPI_ERROR_NONE;

	hspi->pTxBuffPtr  = pdata;
	hspi->TxXferSize  = length;
	hspi->TxXferCount = length;
	hspi->RxXferSize  = 0;
	hspi->RxXferCount = 0;

	/*
	 *  饤ȥå
	 */
	if(hspi->Init.DataSize > SPI_DATASIZE_16BIT && hspi->hdmatx->Init.MemDataAlignment != DMA_MDATAALIGN_WORD)
		return E_PAR;
	if(hspi->Init.DataSize > SPI_DATASIZE_8BIT && hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_BYTE)
		return E_PAR;

    /*
	 *  1饤̿
	 */
	if(hspi->Init.Direction == SPI_DIRECTION_1LINE){
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_HDDIR);
	}

	/*
	 *  DMAȥ
	 */
	hspi->hdmatx->xfercallback = spi_dmatransmit_func;
	hspi->hdmatx->errorcallback = spi_dmaerror_func;

	/*
	 *  DMAž
	 */
	if(hspi->Init.DataSize <= SPI_DATASIZE_8BIT){
		if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_HALFWORD)
			hspi->TxXferCount = (hspi->TxXferCount + 1) >> 1;
		else if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD)
			hspi->TxXferCount = (hspi->TxXferCount + 3) >> 2;
	}
	else if(hspi->Init.DataSize <= SPI_DATASIZE_16BIT){
		if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD)
			hspi->TxXferCount = (hspi->TxXferCount + 1) >> 1;
	}
	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_TXDMAEN);

	/*
	 *  DMAȥ
	 */
	dma_start(hspi->hdmatx, (uint32_t)hspi->pTxBuffPtr, (uint32_t)(hspi->base+TOFF_SPI_TXDR), hspi->TxXferCount);

	/*
	 *  žǡ
	 */
	if(hspi->hdmatx->Init.Mode == DMA_CIRCULAR)
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, 0);
	else
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, length);

	/*
	 * Tx DMA
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_TXDMAEN);

	/*
	 *  SPI
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_TXVALUE);
	if(hspi->Init.CRC == SPI_CRC_ENABLE)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_CRCEIE);

	/*
	 *  SPIͭ
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

    /*
	 * SPIޥ
	 */
	if(hspi->Init.Mode == SPI_MODE_MASTER)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_CSTART);

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  pdata: ХåեؤΥݥ(4Хȥ饤)
 *  parameter3  length: (CRCǤ+1)
 *  return ER
 */
ER
spi_receive(SPI_Handle_t *hspi, uint8_t *pdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || pdata == NULL || length == 0)
		return E_PAR;
	if((((uint32_t)pdata) & 3) != 0)
		return E_PAR;
	if(hspi->hdmarx == NULL)
		return E_OBJ;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

    /*
	 *  ž
	 */
	hspi->xmode       = SPI_XMODE_RX;
    hspi->status      = SPI_STATUS_BUSY;
    hspi->ErrorCode   = SPI_ERROR_NONE;
    hspi->pRxBuffPtr  = pdata;
    hspi->RxXferSize  = length;
    hspi->RxXferCount = length;
    hspi->TxXferSize   = 0;
    hspi->TxXferCount  = 0;

	/*
	 *  饤ȥå
	 */
	if(hspi->Init.DataSize > SPI_DATASIZE_16BIT && hspi->hdmatx->Init.MemDataAlignment != DMA_MDATAALIGN_WORD)
		return E_PAR;
	if(hspi->Init.DataSize > SPI_DATASIZE_8BIT && hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_BYTE)
		return E_PAR;

    /*
	 *  1饤̿
	 */
    if(hspi->Init.Direction == SPI_DIRECTION_1LINE){
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_HDDIR);
    }
    else if(hspi->Init.Direction == SPI_DIRECTION_2LINES){
	    hspi->status = SPI_STATUS_READY;
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		if(hspi->Init.Mode == SPI_MODE_MASTER){
			/*
			 *  Ԥ
			 */
			memset(pdata, 0xff, length);
			return spi_transrecv(hspi, pdata, pdata, length);
		}
		else
			return E_PAR;
    }

	/*
	 *  DMAž
	 */
	if(hspi->Init.DataSize <= SPI_DATASIZE_8BIT){
		if(hspi->hdmarx->Init.MemDataAlignment == DMA_MDATAALIGN_HALFWORD)
			hspi->RxXferCount = (hspi->RxXferCount + 1) >> 1;
		else if(hspi->hdmarx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD)
			hspi->RxXferCount = (hspi->RxXferCount + 3) >> 2;
	}
	else if (hspi->Init.DataSize <= SPI_DATASIZE_16BIT){
		if(hspi->hdmarx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD)
			hspi->RxXferCount = (hspi->RxXferCount + 1) >> 1;
	}
	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_RXDMAEN);

    /*
	 *  DMAХåؿ
	 */
	hspi->hdmarx->xfercallback = spi_dmareceive_func;
    hspi->hdmarx->errorcallback = spi_dmaerror_func;

	/*
	 *  DMAȥ
	 */
    dma_start(hspi->hdmarx, (uint32_t)(hspi->base+TOFF_SPI_RXDR), (uint32_t)hspi->pRxBuffPtr, hspi->RxXferCount);

	/*
	 *  žǡ
	 */
	if(hspi->hdmatx->Init.Mode == DMA_CIRCULAR)
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, 0);
	else
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, length);

	/*
	 * Rx DMA
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_RXDMAEN);

	/*
	 *  SPI
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_RXVALUE);
	if(hspi->Init.CRC == SPI_CRC_ENABLE)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_CRCEIE);

	/*
	 *  SPIͭ
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

    /*
	 * SPIޥ
	 */
	if(hspi->Init.Mode == SPI_MODE_MASTER)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_CSTART);

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPI¹Դؿ
 *  parameter1  hspi: SPIϥɥؤΥݥ
 *  parameter2  ptxdata: ХåեؤΥݥ(4Хȥ饤)
 *  parameter3  prxdata: ХåեؤΥݥ(4Хȥ饤)
 *  parameter4  length: (CRCǤ+1)
 *  return ER
 */
ER
spi_transrecv(SPI_Handle_t *hspi, uint8_t *ptxdata, uint8_t *prxdata, uint16_t length)
{
	ER ercd = E_OK;

	if(hspi == NULL || prxdata == NULL || length == 0)
		return E_PAR;
	if((((uint32_t)ptxdata) & 3) != 0 || (((uint32_t)prxdata) & 3) != 0)
		return E_PAR;
	if(hspi->Init.Direction != SPI_DIRECTION_2LINES && hspi->Init.Mode != SPI_MODE_MASTER)
		return E_PAR;
	if(hspi->hdmatx == NULL || hspi->hdmarx == NULL)
		return E_OBJ;

	if(hspi->Init.semlock != 0)
		wai_sem(hspi->Init.semlock);
	if(hspi->status != SPI_STATUS_READY){
		if(hspi->Init.semlock != 0)
			sig_sem(hspi->Init.semlock);
		return E_OBJ;
	}

	if(hspi->xmode != SPI_XMODE_RX)
		hspi->xmode = SPI_XMODE_TXRX;
	hspi->status = SPI_STATUS_BUSY;

	/*
	 *  ž
	 */
	hspi->ErrorCode   = SPI_ERROR_NONE;
	hspi->pTxBuffPtr  = (uint8_t*)ptxdata;
	hspi->TxXferSize  = length;
	hspi->TxXferCount = length;
	hspi->pRxBuffPtr  = (uint8_t*)prxdata;
	hspi->RxXferSize  = length;
	hspi->RxXferCount = length;

	/*
	 *  饤ȥå
	 */
	if(hspi->Init.DataSize > SPI_DATASIZE_16BIT && hspi->hdmatx->Init.MemDataAlignment != DMA_MDATAALIGN_WORD)
		return E_PAR;
	if(hspi->Init.DataSize > SPI_DATASIZE_8BIT && hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_BYTE)
		return E_PAR;

	/*
	 *  DMAž
	 */
	if(hspi->Init.DataSize <= SPI_DATASIZE_8BIT){
		if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_HALFWORD){
			hspi->TxXferCount = (hspi->TxXferCount + 1) >> 1;
			hspi->RxXferCount = (hspi->RxXferCount + 1) >> 1;
		}
		else if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD){
			hspi->TxXferCount = (hspi->TxXferCount + 3) >> 2;
			hspi->RxXferCount = (hspi->RxXferCount + 3) >> 2;
		}
	}
	else if(hspi->Init.DataSize <= SPI_DATASIZE_16BIT){
		if(hspi->hdmatx->Init.MemDataAlignment == DMA_MDATAALIGN_WORD){
			hspi->TxXferCount = (hspi->TxXferCount + 1) >> 1;
			hspi->RxXferCount = (hspi->RxXferCount + 1) >> 1;
		}
	}

	/*
	 *  DMAХåؿ
	 */
	if(hspi->xmode == SPI_XMODE_RX)
		hspi->hdmarx->xfercallback = spi_dmareceive_func;
	else
		hspi->hdmarx->xfercallback = spi_dmatransrecv_func;
	hspi->hdmarx->errorcallback = spi_dmaerror_func;
	sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), (SPI_CFG1_TXDMAEN | SPI_CFG1_RXDMAEN));

	/*
	 *  DMAȥ
	 */
	dma_start(hspi->hdmarx, (uint32_t)(hspi->base+TOFF_SPI_RXDR), (uint32_t)hspi->pRxBuffPtr, hspi->RxXferCount);

	/*
	 *  SPIDMA
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_RXDMAEN);

	/*
	 *  DMAХå
	 */
	hspi->hdmatx->errorcallback = spi_dmaerror_func;

	/*
	 *  DMAȥ
	 */
	dma_start(hspi->hdmatx, (uint32_t)hspi->pTxBuffPtr, (uint32_t)(hspi->base+TOFF_SPI_TXDR), hspi->TxXferCount);

	/*
	 *  žǡ
	 */
	if(hspi->hdmatx->Init.Mode == DMA_CIRCULAR)
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, 0);
	else
		sil_modw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR2), SPI_CR2_TSIZE, length);

	/*
	 *  SPIDMA
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_TXDMAEN);

	/*
	 *  SPI
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), (SPI_IER_TXVALUE | SPI_IER_RXVALUE));
	if(hspi->Init.CRC == SPI_CRC_ENABLE)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_CRCEIE);

	/*
	 *  SPIͭ
	 */
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

    /*
	 * SPIޥ
	 */
	if(hspi->Init.Mode == SPI_MODE_MASTER)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_CSTART);

#if SPI_WAIT_TIME != 0
	ercd = spi_inwait(hspi, SPI_WAIT_TIME * length);

	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPIžλԤ
 */
ER
spi_wait(SPI_Handle_t *hspi, uint32_t timeout)
{
	ER ercd = E_OK;

#if SPI_WAIT_TIME == 0
	if(hspi == NULL)
		return E_PAR;
	ercd = spi_inwait(hspi, timeout);
	if(hspi->Init.semlock != 0)
		sig_sem(hspi->Init.semlock);
#endif
	return ercd;
}

/*
 *  SPIߥӥ롼
 */
void
spi_handler(SPI_Handle_t *hspi)
{
	uint32_t ier = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_IER));
	uint32_t isr = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_SR));

	if(((isr & ier) & SPI_SR_EOT) != 0){
		/*
		 *  ߥꥢ
		 */
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_EOTC);
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_TXTFC);
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_SUSPC);

		/*
		 *  EOT߶ػ
		 */
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), SPI_IER_EOTIE);

		/*
		 *  SPI
		 */
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CR1), SPI_CR1_SPE);

		/*
		 *  SPI߶ػ
		 */
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_IER), (SPI_IER_TXVALUE | SPI_IER_RXVALUE | SPI_IER_CRCEIE));

		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_TXDMAEN);
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CFG1), SPI_CFG1_RXDMAEN);

		hspi->TxXferCount = 0;
		hspi->RxXferCount = 0;
		hspi->status = SPI_STATUS_READY;
	}

	/*
	 *  ʳΥ顼Ƚ
	 */
	if(hspi->xmode != SPI_XMODE_RX){
		if(((isr & ier) & SPI_SR_UDR) != 0){
			hspi->ErrorCode |= SPI_ERROR_UDR;
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_UDRC);
		}
	}

	/*
	 *  ʳΥ顼Ƚ
	 */
	if(hspi->xmode != SPI_XMODE_TX){
		if(((isr & ier) & SPI_SR_OVR) != 0){
			hspi->ErrorCode |= SPI_ERROR_OVR;
			sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_OVRC);
		}
	}
    /*
	 *  CRCΥ顼Ƚ
	 */
	if(((isr & ier) & SPI_SR_CRCE) != 0){
		hspi->ErrorCode |= SPI_ERROR_CRC;
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_CRCEC);
	}

	/*
	 *  SPI̤Υ顼Ƚ
	 */
	if(((isr & ier) & SPI_SR_MODF) != 0){
		hspi->ErrorCode |= SPI_ERROR_MODF;
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_MODFC);
	}
	if(((isr & ier) & SPI_SR_TIFRE) != 0){
		hspi->ErrorCode |= SPI_ERROR_FRE;
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_IFCR), SPI_IFCR_TIFREC);
	}
	if(hspi->Init.semid != 0)
		isig_sem(hspi->Init.semid);
}


/*
 *  SPIߥӥ롼
 */
void spi_isr(intptr_t exinf)
{
  spi_handler(&SpiHandle[INDEX_SPI((uint32_t)exinf)]);
}

