/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2020-09-25 14:32:33Z roi $
 */
/*
 * STM32MP1xxѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVT           0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	TADR_GPIOE_BASE,			/* index 4 */
	TADR_GPIOF_BASE,			/* index 5 */
	TADR_GPIOG_BASE,			/* index 6 */
	TADR_GPIOH_BASE,			/* index 7 */
	TADR_GPIOI_BASE,			/* index 8 */
	TADR_GPIOJ_BASE,			/* index 9 */
	TADR_GPIOK_BASE,			/* index 10 */
	TADR_GPIOZ_BASE				/* index 11 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t iocurrent = 1<<pin;
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	/*
	 *  GPIO⡼
	 */
	target_periph_lock(base, LOCK_RESOURCE_TIMEOUT);

	/* 륿͡ȡե󥯥󡦥⡼ */
	temp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4));
	temp &= ~((uint32_t)0xF << ((uint32_t)(pin & (uint32_t)0x07) * 4)) ;
	if(init->mode == GPIO_MODE_AF)
		temp |= ((uint32_t)(init->alternate) << (((uint32_t)pin & (uint32_t)0x07) * 4));
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4), temp);

	/*  ϥ⡼ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (pin * 2)), ((init->mode & GPIO_MODE) << (pin * 2)));

	/*  ϥ⡼ */
	if(init->mode == GPIO_MODE_OUTPUT || init->mode == GPIO_MODE_AF){
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (pin * 2)), (init->speed << (pin * 2)));
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << pin), (init->otype << pin));
	}

	/*  ץ륢åסץ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (pin * 2)), (init->pull << (pin * 2)));
	target_periph_unlock(base);

	/*
	 *  EXTI⡼
	 */
	if((init->mode & EXTI_MODE) == EXTI_MODE){
		target_periph_lock(TADR_EXTI_BASE, LOCK_RESOURCE_TIMEOUT);

		temp = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EXTICR+(pin & 0x0C)));
		temp &= ~(0x0F << (4 * (pin & 0x03)));
		temp |= (index << (4 * (pin & 0x03)));
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EXTICR+(pin & 0x0C)), temp);

		if((init->mode & GPIO_MODE_IT) == GPIO_MODE_IT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_C2IMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_C2IMR1), iocurrent);

		if((init->mode & GPIO_MODE_EVT) == GPIO_MODE_EVT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_C2EMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_C2EMR1), iocurrent);

		if((init->mode & RISING_EDGE) == RISING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);

		if((init->mode & FALLING_EDGE) == FALLING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		target_periph_unlock(TADR_EXTI_BASE);
	}
}


/*
 *  DMACؿ
 */
#define NUM_STMDMA      16
#define NUM_DMAINDEX    5
#define INDEX_TC        0
#define INDEX_HT        1
#define INDEX_TE        2
#define INDEX_FE        3
#define INDEX_DME       4

#define DMA_TRS_TIMEOUT 2000	/* 2 */

#define get_DMAmuxChannel(h)        ((h)->muxbase + ((h)->sdid % NUM_STMDMA) * 4)
#define get_stream_statusmask(h)    (1 << (((h)->sdid % NUM_STMDMA) & 0x1F))

static const uint32_t dma_base[NUM_STMDMA] = {
	TADR_DMA1_STM0_BASE,
	TADR_DMA1_STM1_BASE,
	TADR_DMA1_STM2_BASE,
	TADR_DMA1_STM3_BASE,
	TADR_DMA1_STM4_BASE,
	TADR_DMA1_STM5_BASE,
	TADR_DMA1_STM6_BASE,
	TADR_DMA1_STM7_BASE,
	TADR_DMA2_STM0_BASE,
	TADR_DMA2_STM1_BASE,
	TADR_DMA2_STM2_BASE,
	TADR_DMA2_STM3_BASE,
	TADR_DMA2_STM4_BASE,
	TADR_DMA2_STM5_BASE,
	TADR_DMA2_STM6_BASE,
	TADR_DMA2_STM7_BASE
};

static const uint32_t dma_flag[NUM_STMDMA][NUM_DMAINDEX] = {
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 1-0 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 1-1 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 1-2 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 1-3 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 1-4 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 1-5 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 1-6 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 1-7 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 2-0 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 2-1 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 2-2 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 },	/* 2-3 */
	{	0x00000020, 0x00000010, 0x00000008, 0x00000001, 0x00000004 },	/* 2-4 */
	{	0x00000800, 0x00000400, 0x00000200, 0x00000040, 0x00000100 },	/* 2-5 */
	{	0x00200000, 0x00100000, 0x00080000, 0x00010000, 0x00040000 },	/* 2-6 */
	{	0x08000000, 0x04000000, 0x02000000, 0x00400000, 0x01000000 }	/* 2-7 */
};

static const uint32_t dma_faddr[NUM_STMDMA/4] = {
	(TADR_DMA1_BASE+TOFF_DMAI_LISR),
	(TADR_DMA1_BASE+TOFF_DMAI_HISR),
	(TADR_DMA2_BASE+TOFF_DMAI_LISR),
	(TADR_DMA2_BASE+TOFF_DMAI_HISR)
};

static DMA_Handle_t *pDmaHandler[NUM_STMDMA];

/*
 *  STREAM DMA FLAG ALL CLEAR
 */
static void
stream_flag_clear(DMA_Handle_t *hdma)
{
	uint32_t i, value;

	for(i = value = 0 ; i < NUM_DMAINDEX ; i++)
		value |= dma_flag[hdma->sdid][i];
	sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), value);
}

/*
 *  CHANNEL DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0;
	uint32_t i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < (NUM_STMDMA) ; i++){
		if(dma_base[i] == hdma->base)
			break;
	}
	if(i == NUM_STMDMA)
		return E_PAR;
	hdma->sdid = i;
	pDmaHandler[i] = hdma;

	hdma->muxbase = TADR_DMAMUX1_BASE;

	if(i < DMA2STM0_SID)
		sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_AHB2ENSETR), RCC_MC_AHB2ENSETR_DMA1EN);
	else
		sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_AHB2ENSETR), RCC_MC_AHB2ENSETR_DMA2EN);
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_AHB2ENSETR), RCC_MC_AHB2ENSETR_DMAMUXEN);

	/* DMA-CR쥸 */
	tmp = sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR));

	/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
	tmp &= ((uint32_t)~(DMA_SxCR_MBURST | DMA_SxCR_PBURST | \
                     DMA_SxCR_PL    | DMA_SxCR_MSIZE  | DMA_SxCR_PSIZE  | \
                     DMA_SxCR_MINC  | DMA_SxCR_PINC   | DMA_SxCR_CIRC   | DMA_SxCR_PFCTRL | \
                     DMA_SxCR_DIR   | DMA_SxCR_CT     | DMA_SxCR_DBM));

	/* DMA⡼ɤƥݥꥣ */
	tmp |=  hdma->Init.Direction | hdma->Init.PeriphInc | \
	        hdma->Init.MemInc    | hdma->Init.PeriphDataAlignment | \
	        hdma->Init.MemDataAlignment | hdma->Init.Mode | \
	        hdma->Init.Priority;

	/* FIFO⡼ɤʤС */
	if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE)
		tmp |=  hdma->Init.MemBurst | hdma->Init.PeriphBurst;

	/* DMA⡼ */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), tmp);

	/* DMA-FCR쥸 */
	tmp = sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR));

	/* FIFO⡼ɤƥݥꥣ˥ꥢ */
	tmp &= (uint32_t)~(DMA_SxFCR_DMDIS | DMA_SxFCR_FTH);

	/* FIFO⡼ɤƥݥꥣ */
	tmp |= hdma->Init.FIFOMode;

	/* FIFO⡼ɤʤ饹åۡɤ */
	if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE){
		tmp |= hdma->Init.FIFOThreshold;

		/* Check compatibility between FIFO threshold level and size of the memory burst */
		/* for INCR4, INCR8, INCR16 */
		if(hdma->Init.MemBurst != DMA_MBURST_SINGLE
				&& hdma->Init.MemDataAlignment == DMA_MDATAALIGN_BYTE){
			switch (hdma->Init.FIFOThreshold){
			case DMA_FIFO_THRESHOLD_1QUARTERFULL:
			case DMA_FIFO_THRESHOLD_3QUARTERSFULL:
				if((hdma->Init.MemBurst & DMA_SxCR_MBURST_1) == DMA_SxCR_MBURST_1)
					return E_PAR;
				break;
			case DMA_FIFO_THRESHOLD_HALFFULL:
				if(hdma->Init.MemBurst == DMA_MBURST_INC16)
					return E_PAR;
				break;
			default:
				break;
			}
		}
	}

	/* DMA-FCR쥸 */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), tmp);

	/* ƥե饰򥯥ꥢ */
	stream_flag_clear(hdma);

	if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
		hdma->Init.Request = DMA_REQUEST_MEM2MEM;

	/* DMAMUX˥ڥեͥ */
	sil_wrw_mem((uint32_t *)(get_DMAmuxChannel(hdma)+TOFF_DMAMUX_CHANNEL_CCR), (hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID));
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));

	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase + TOFF_DMAMUX_REQUEST_GEN + (((request-1) * 4))), 0x00000000);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase + TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request-1)));
	}

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_STMDMA ; i++){
		if(dma_base[i] == hdma->base)
			break;
	}
	if(i == NUM_STMDMA)
		return E_PAR;
	hdma->sdid = i;

	/* DMA֥͡ */
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), 0x00000000);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_NDTR), 0x00000000);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), 0x00000000);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), 0x00000000);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M1AR), 0x00000000);
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), 0x00000021);

	/* ƥե饰򥯥ꥢ */
	stream_flag_clear(hdma);

	/*
	 *  DMAMUXͥν
	 */
	if(hdma->muxbase != 0){
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CCR), 0x00000000);
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));
	}
	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN + (((request - 1) * 4))), 0x00000000);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request - 1)));
	}
	hdma->muxbase = 0;

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHAANEL DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE | DMA_SxCR_HTIE | DMA_SxCR_TEIE | DMA_SxCR_DMEIE);
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);

	/* ƥե饰򥯥ꥢ */
	stream_flag_clear(hdma);
	hdma->status = DMA_STATUS_BUSY;

	/* DBMӥåȥꥢ */
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_DBM);
	/* ǡĹ */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_NDTR), DataLength);

	if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
		/* ꤫ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), DstAddress);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), SrcAddress);
	}
	else{	/* ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_PAR), SrcAddress);
		sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMAS_M0AR), DstAddress);
	}

	/* ߥ֥͡ */
	sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE | DMA_SxCR_HTIE | DMA_SxCR_TEIE | DMA_SxCR_DMEIE);
	if(hdma->Init.FIFOMode == DMA_FIFOMODE_ENABLE)
		sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);

	/*
	 *  DMAMUX SynchronizationͭʤСsync overrunߤͭ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CCR)) & DMAMUX_CxCR_SE) != 0)
		sil_orw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_CCR), DMAMUX_CxCR_SOIE);

	/*
	 *  ꥯȥͥ졼ʤСgen overrunߤͭ
	 */
	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7)
		sil_orw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN), DMAMUX_RGxCR_OIE);

	/* DMA */
	sil_orw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);
	return E_OK;
}

/*
 *  CHANNEL DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	uint32_t tick = 0;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_EN);

	/* DMAԤ */
	while((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_EN) != 0){
	    /* Check for the Timeout */
		if(tick > DMA_TRS_TIMEOUT){
			/* ॢȥ顼 */
			hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
			return E_TMOUT;
		}
    	dly_tsk(1);
	}

	/*
	 *  DMAMUXΥСե饰򥯥ꥢ
	 */
	sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_CHANNEL_STS_CFR), get_stream_statusmask(hdma));

	if(hdma->Init.Request >= DMA_REQUEST_GENERATOR0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR7){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		uint32_t request =  hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		/*
		 *  DMAMUX request generatorꥻå
		 */
		sil_andw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GEN), DMAMUX_RGxCR_OIE);

		/*
		 *  DMAMUX request generator overrun flag򥯥ꥢ
		 */
		sil_wrw_mem((uint32_t *)(hdma->muxbase+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (request - 1)));
	}
	hdma->status = 0;
	return E_OK;
}

/*
 *  STREAM DMAΥ١ɥ쥹
 */
uintptr_t
dma_stream_base(int32_t sid)
{
	if(sid < 0 || sid >= NUM_STMDMA)
		return 0;
	return (uintptr_t)dma_base[sid];
}

/*
 *  STREAM DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	/*
	 *  ž顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_TE]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_TEIE) != 0){
			/* ž顼ߥꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TEIE);
			/* ž顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_TE]);
			/* ž顼֤ */
			hdma->ErrorCode |= DMA_ERROR_TE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  FIFO顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_FE]) != 0){
	    if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR)) & DMA_SxFCR_FEIE) != 0){
			/* FIFO顼ߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_FCR), DMA_SxFCR_FEIE);
			/* FIFO顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_FE]);
			/* FIFO顼 */
				hdma->ErrorCode |= DMA_ERROR_FE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  쥯ȡꡦ顼
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_DME]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DMEIE) != 0){
			/* DME顼ߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_DMEIE);
			/* DME顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_DME]);
			/* DME顼 */
			hdma->ErrorCode |= DMA_ERROR_DME;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}
	/*
	 *  ϡžλ
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_HT]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_HTIE) != 0){
			/*
			 *  ϡžե饰򥯥ꥢ
			 */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_HT]);
			/* ޥХåե⡼ */
			if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DBM) != 0){
				/* ꣰ */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CT) == 0){
					/* ϡե꣰ξ */
					hdma->status = DMA_STATUS_READY_HMEM0;
					if(hdma->xferhalfcallback != NULL)
						hdma->xferhalfcallback(hdma);
				}
				/* ꣱ */
				else{
					/* ϡե꣱ξ */
					hdma->status = DMA_STATUS_READY_HMEM1;
					if(hdma->xferm1halfcallback != NULL)
						hdma->xferm1halfcallback(hdma);
				}
			}
			else{
				/*
				 *  åǤʤХϡճ
				 */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CIRC) == 0)
					sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_HTIE);
				/* ϡžơѹ */
				hdma->status = DMA_STATUS_READY_HMEM0;
				/* ϡžλХå */
				if(hdma->xferhalfcallback != NULL)
					hdma->xferhalfcallback(hdma);
			}
		}
	}
	/*
	 *  žλ
	 */
	if((sil_rew_mem((uint32_t *)(dma_faddr[hdma->sdid/4])) & dma_flag[hdma->sdid][INDEX_TC]) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_TCIE) != 0){
			/*
			 *  žλե饰򥯥ꥢ
			 */
			sil_wrw_mem((uint32_t *)(dma_faddr[hdma->sdid/4]+8), dma_flag[hdma->sdid][INDEX_TC]);
			/* ޥХåե⡼ */
			if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_DBM) != 0){
				/* ꣱ */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CT) == 0){
					hdma->status = DMA_STATUS_READY_MEM1;
					/* 1žλХå */
					if(hdma->xferm1callback != NULL)
						hdma->xferm1callback(hdma);
				}
				/* ꣰ */
				else{
					hdma->status = DMA_STATUS_READY_MEM0;
					/* žλХå */
					if(hdma->xfercallback != NULL)
						hdma->xfercallback(hdma);
				}
			}
			else{
				/*
				 *  åǤʤдλ
				 */
				if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR)) & DMA_SxCR_CIRC) == 0){
					sil_andw_mem((uint32_t *)(hdma->base+TOFF_DMAS_CR), DMA_SxCR_TCIE);
					hdma->ErrorCode = DMA_ERROR_NONE;
				}

				/* ơλ */
				hdma->status = DMA_STATUS_READY_MEM0;
				/* žλХå */
				if(hdma->xfercallback != NULL)
					hdma->xfercallback(hdma);
			}
		}
	}
}

/*
 *  STREAM DMA ߥӥ롼
 */
void
stream_dma_isr(intptr_t exinf)
{
	dma_inthandler(pDmaHandler[(uint32_t)exinf]);
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint32_t mcount, mcounter, wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;
	if(hini->Window > 127 || hini->Window < 64)
		return E_PAR;
	if(hini->Counter > 127 || hini->Counter < 64)
		return E_PAR;

	mcount    = hini->MCounter;
	mcounter  = 0;
	wdogcount = hini->Counter;
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_APB1ENSETR), RCC_MC_APB1ENSETR_WWDG1EN);
	sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CR), wdogcount);
	if(mcount == 0)
		sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CFR), (hini->Prescaler | hini->Window));
	else
		sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CFR), (WWDG_CFR_EWI | hini->Prescaler | hini->Window));
	sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CR), (WWDG_CR_WDGA | wdogcount));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CR), wdogcount);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_APB1ENCLRR), RCC_MC_APB1ENCLRR_WWDG1EN);
	wdogcount = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG1_BASE+TOFF_WWDG_CR), wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDG1_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDG_CFR)) & WWDG_CFR_EWI) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDG_SR)) & WWDG_SR_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDG_SR), WWDG_SR_EWIF);
		if(mcounter++ < mcount)
			sil_wrw_mem((uint32_t *)(base+TOFF_WWDG_CR), wdogcount);
	}
}


/*
 *  OpenAMPIPCCɥ饤
 */
#define TX_MR_SHIFT     (16)
/*
 * CHANNEL MASK
 */
#define IPCC_ALL_RX_BUF (IPCC_MR_CH1OM | IPCC_MR_CH2OM | IPCC_MR_CH3OM | \
						 IPCC_MR_CH4OM | IPCC_MR_CH5OM | IPCC_MR_CH6OM)
#define IPCC_ALL_TX_BUF (IPCC_MR_CH1FM | IPCC_MR_CH2FM | IPCC_MR_CH3FM | \
						 IPCC_MR_CH4FM | IPCC_MR_CH5FM | IPCC_MR_CH6FM)

/*
 *  IPCCȥID֥åФΥޥ
 */
#define INDEX_IPCC(ipccid)      ((uint_t)((ipccid) - 1))

/*
 *  IPCCǡǼΰ
 */
static IPCC_Handle_t *shipcc[NUM_IPCCCONTROL];

/*
 *  IPCC
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  return ER
 */
ER
ipcc_init(IPCC_Handle_t *hipcc)
{
	uint32_t cbase;
	uint32_t no, i;

	if(hipcc == NULL)
		return E_PAR;
	if(hipcc->Init.controlID > NUM_IPCCCONTROL || hipcc->Init.controlID == 0)
		return E_PAR;

	no = INDEX_IPCC(hipcc->Init.controlID);
	hipcc->base  = (uint32_t)TADR_IPCC_BASE;
	if(hipcc->Init.controlID == IPCC1_CONTROLID){
		hipcc->cbase = hipcc->base + TOFF_IPCC1;
		hipcc->obase = hipcc->base + TOFF_IPCC2;
	}
	else{
		hipcc->cbase = hipcc->base + TOFF_IPCC2;
		hipcc->obase = hipcc->base + TOFF_IPCC1;
	}
	cbase = hipcc->cbase;

	/*
	 *  IPCCå
	 */
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_MC_AHB3ENSETR), RCC_MC_AHB3ENSETR_IPCCEN);

	/*
	 *  IPCC쥸
	 */
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_CR), 0);
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_MR), (IPCC_ALL_TX_BUF | IPCC_ALL_RX_BUF));
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_SCR), IPCC_ALL_RX_BUF);

    /*
	 *  
	 */
	if(hipcc->Init.inttx >= 0){
		sil_orw_mem((uint32_t *)(cbase+TOFF_IPCC_CR), IPCC_CR_TXFIE);
		ena_int(hipcc->Init.inttx);
	}
	if(hipcc->Init.intrx >= 0){
		sil_orw_mem((uint32_t *)(cbase+TOFF_IPCC_CR), IPCC_CR_RXOIE);
		ena_int(hipcc->Init.intrx);
	}

	/*
	 *  Хåؿ
	 */
	for(i = 0 ; i < NUM_IPCC_CH ; i++){
		hipcc->recvcallback[i] = NULL;
		hipcc->sendcallback[i] = NULL;
	}

	/* Reset all callback notification request */
	hipcc->rstatus = 0;

	shipcc[no] = hipcc;
	return E_OK;
}

/*
 *  IPCCλؿ
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  return ER
 */
ER
ipcc_deinit(IPCC_Handle_t *hipcc)
{
	uint32_t cbase;
	uint32_t no, i;

	if(hipcc == NULL)
		return E_PAR;
	if(hipcc->Init.controlID > NUM_IPCCCONTROL || hipcc->Init.controlID == 0)
		return E_PAR;

	no = INDEX_IPCC(hipcc->Init.controlID);
	if(shipcc[no] == NULL)
		return E_OBJ;
	cbase = hipcc->cbase;

	/*
	 *  IPCC쥸
	 */
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_CR), 0);
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_MR), (IPCC_ALL_TX_BUF | IPCC_ALL_RX_BUF));
	sil_wrw_mem((uint32_t *)(cbase+TOFF_IPCC_SCR), IPCC_ALL_RX_BUF);

	/*
	 *  
	 */
	if(hipcc->Init.inttx >= 0)
		dis_int(hipcc->Init.inttx);
	if(hipcc->Init.intrx >= 0)
		dis_int(hipcc->Init.intrx);

	/*
	 *  Хåؿ
	 */
	for(i = 0 ; i < NUM_IPCC_CH ; i++){
		hipcc->recvcallback[i] = NULL;
		hipcc->sendcallback[i] = NULL;
	}

	/* Reset all callback notification request */
	hipcc->rstatus = 0;
	shipcc[no] = NULL;

	return E_OK;
}

/*
 *  IPCCͥ롦Хåؿ
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  parameter2  cid:   IPCCͥID
 *  parameter3  mode:  ⡼
 *  parameter4  cb:    Хåؿ
 *  return ER
 */
ER
ipcc_activate_callback(IPCC_Handle_t *hipcc, uint8_t cid, uint8_t mode, void(cb)(IPCC_Handle_t*, uint32_t, uint8_t))
{
	if(hipcc == NULL)
		return E_PAR;
	if(cid >= NUM_IPCC_CH)
		return E_PAR;
	if(shipcc[INDEX_IPCC(hipcc->Init.controlID)] == NULL)
		return E_OBJ;

	/*
	 *  Хå
	 */
	if(mode == IPCC_TX_CHANNEL){
		hipcc->sendcallback[cid] = cb;
		hipcc->rstatus |= (IPCC_MR_CH1FM << cid);
	}
	else{
		hipcc->recvcallback[cid] = cb;
		hipcc->rstatus |= (IPCC_MR_CH1OM << cid);
		/*
		 *  ͥߵ
		 */
		sil_andw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), (IPCC_MR_CH1OM << cid));
	}
	return E_OK;
}

/*
 *  IPCCͥ롦Хåߴؿ
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  parameter2  cid:   IPCCͥID
 *  parameter3  mode:  ⡼
 *  return ER
 */
ER
ipcc_deactivate_callback(IPCC_Handle_t *hipcc, uint8_t cid, uint8_t mode)
{
	if(hipcc == NULL)
		return E_PAR;
	if(cid >= NUM_IPCC_CH)
		return E_PAR;
	if(shipcc[INDEX_IPCC(hipcc->Init.controlID)] == NULL)
		return E_OBJ;

	/*
	 *  Хåꥯꥢ
	 */
	if(mode == IPCC_TX_CHANNEL){
		hipcc->sendcallback[cid] = NULL;
		hipcc->rstatus &= ~(IPCC_MR_CH1FM << cid);
		/*
		 *  ͥ߶ػ
		 */
		sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), (IPCC_MR_CH1FM << cid));
	}
	else{
		hipcc->recvcallback[cid] = NULL;
		hipcc->rstatus &= ~(IPCC_MR_CH1OM << cid);
		/*
		 *  ͥ߶ػ
		 */
		sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), (IPCC_MR_CH1OM << cid));
	}
	return E_OK;
}

/*
 *  IPCCͥ롦ơؿ
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  parameter2  cid:   IPCCͥID
 *  parameter3  mode:  ⡼
 *  return ֤ޤER
 */
ER_BOOL
ipcc_channel_occupied(IPCC_Handle_t *hipcc, uint8_t cid, uint8_t mode)
{
	uint32_t channel_state;
	uint32_t status;

	if(hipcc == NULL)
		return E_PAR;
	if(cid >= NUM_IPCC_CH)
		return E_PAR;
	if(shipcc[INDEX_IPCC(hipcc->Init.controlID)] == NULL)
		return E_OBJ;

	status = sil_rew_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_SR));
	if(mode == IPCC_TX_CHANNEL)
		channel_state = status & (IPCC_SR_CH1F << cid);
	else
		channel_state = status & (IPCC_SR_CH1F << cid);
	return (channel_state != 0);
}

/*
 *  ⡼ȤCPUؤ
 *  parameter1  hipcc: IPCCϥɥؤΥݥ
 *  parameter2  cid:   ͥID
 *  parameter3  mode:  ⡼
 *  return ER
 */
ER
ipcc_notify_cpu(IPCC_Handle_t *hipcc, uint8_t cid, uint8_t mode)
{
  uint32_t mask;

	if(hipcc == NULL)
		return E_PAR;
	if(cid >= NUM_IPCC_CH)
		return E_PAR;
	if(shipcc[INDEX_IPCC(hipcc->Init.controlID)] == NULL)
		return E_OBJ;

	if(mode == IPCC_TX_CHANNEL){
		sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_SCR), (IPCC_SCR_CH1S << cid));
		mask = IPCC_MR_CH1FM << cid;
		/*
		 *  ͥߵ
		 */
	    if((hipcc->rstatus & mask) == mask)
			sil_andw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), (IPCC_MR_CH1FM << cid));
	}
	else{
		sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_SCR), (IPCC_SCR_CH1C << cid));
		mask = IPCC_MR_CH1OM << cid;
		/*
		 *  ͥߵ
		 */
	    if((hipcc->rstatus & mask) == mask)
			sil_andw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), (IPCC_MR_CH1OM << cid));
	}
	return E_OK;
}

/*
 *  IPCCߥӥ롼
 */
void
ipcc_rx_isr(intptr_t exinf)
{
	IPCC_Handle_t *hipcc = shipcc[INDEX_IPCC((int)exinf)];
	uint32_t irqmask, channel, i;
	syslog_1(LOG_NOTICE, "ipcc_rx_isr(%d)", (int)exinf);

	if(hipcc == NULL)
		return;

	/*
	 *  RXͥߥӥåȤФ
	 */
	irqmask = ~(sil_rew_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR))) & IPCC_ALL_RX_BUF;
	irqmask = irqmask & sil_rew_mem((uint32_t *)(hipcc->obase+TOFF_IPCC_SR));

	for(i = 0 ; irqmask != 0UL && i < NUM_IPCC_CH ; i++){
		channel = 1 << i;
		if ((irqmask & channel) != 0U){
			sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), channel);
			if(hipcc->recvcallback[i] != NULL)
				hipcc->recvcallback[i](hipcc, i, IPCC_RX_CHANNEL);
			irqmask = irqmask & ~(channel);
		}
	}
}

/*
 *  IPCCӥ롼
 */
void
ipcc_tx_isr(intptr_t exinf)
{
	IPCC_Handle_t *hipcc = shipcc[INDEX_IPCC((int)exinf)];
	uint32_t irqmask, channel, i;

	if(hipcc == NULL)
		return;

	/*
	 *  TXͥߥӥåȤФ
	 */
	irqmask = ~(sil_rew_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR))) & IPCC_ALL_TX_BUF;
	irqmask = irqmask & ~(sil_rew_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_SR)) << TX_MR_SHIFT);

	for(i = 0 ; irqmask != 0UL && i < NUM_IPCC_CH ; i++){
		channel = 1 << (TX_MR_SHIFT + i);
		if((irqmask & channel) != 0U){
			sil_orw_mem((uint32_t *)(hipcc->cbase+TOFF_IPCC_MR), channel);
			if(hipcc->sendcallback[i] != NULL)
				hipcc->sendcallback[i](hipcc, i, IPCC_TX_CHANNEL);
			irqmask =  irqmask & ~(channel);
		}
	}
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
	{ TADR_GPIOH_BASE, PINPOSITION7, GPIO_PULLUP}
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))


/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (1 << pled->pinpos)<<16);
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_IDR)) & (1<<(pled->pinpos))) != 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t reg1, reg2, i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		reg1 = reg2 = 0;
		if((led_data & (1<<i)) != 0)
			reg2 = 1 << pled->pinpos;
		else
			reg1 = 1 << pled->pinpos;

		/* 񤭹ߥǡƽ񤭹 */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (reg1<<16) | reg2);
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


uint_t dipsw_value;
void (*sw_func)(int arg);

static const gio_conftype sw_confdata[] = {
	{ TADR_GPIOA_BASE, PSW1_PIN, GPIO_NOPULL},
	{ TADR_GPIOA_BASE, PSW2_PIN, GPIO_NOPULL}
};
#define NUM_SW    (sizeof(sw_confdata)/sizeof(gio_conftype))

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	const gio_conftype *psw = &sw_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_SW ; psw++, i++){
		/* ϥݡȤ */
		/* ⡼ */
		GPIO_Init_Data.mode  = GPIO_MODE_IT_FALLING;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull  = psw->pinpp;
        /* ԡ */
		GPIO_Init_Data.speed = GPIO_SPEED_HIGH;
		gpio_setup(psw->giobase, &GPIO_Init_Data, psw->pinpos);
	}
}

/*
 *  PUSHåХåؿ
 */
void
setup_sw_func(intptr_t exinf)
{
	sw_func = (void (*)(int))exinf;
}

/*
 *  PUSHåФ
 */
uint32_t
switch_push_sense(void)
{
	const gio_conftype *psw = &sw_confdata[0];
	uint32_t sw = 0;
	uint32_t i;

	for(i = 0 ; i < NUM_SW ; psw++, i++){
		if((sil_rew_mem((uint32_t *)(psw->giobase+TOFF_GPIO_IDR)) & (1<<psw->pinpos)) == 0)
			sw |= 1 << i;
	}
	return sw;
}

/*
 *  PUSHåӥ
 */
void
sw_isr(intptr_t exinf)
{
	uint32_t istatus;
	uint8_t  pin = (uint8_t)exinf;

	istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1));
	if((istatus & (1 << pin)) != 0){
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RPR1), (1 << pin));
		if(sw_func != NULL)
			sw_func((int)exinf);
	}
	istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1));
	if((istatus & (1 << pin)) != 0){
		sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FPR1), (1 << pin));
		if(sw_func != NULL)
			sw_func((int)exinf);
	}
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

