/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: device.c 698 2020-08-20 20:20:53Z roi $
 */
/*
 * STM32WBxxѥǥХɥ饤
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <target_syssvc.h>
#include "kernel_cfg.h"
#include "device.h"


/*
 *  SILؿΥޥ
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))


uint_t dipsw_value;
void (*exti_func[NUM_EXTI_FUNC])(void) = { 0 };

extern uint32_t sysclock_init_value(void);


/*
 *  GPIO⡼ɤ
 */
#define GPIO_MODE               0x00000003
#define EXTI_MODE               0x10000000
#define GPIO_MODE_IT            0x00010000
#define GPIO_MODE_EVT           0x00020000
#define RISING_EDGE             0x00100000
#define FALLING_EDGE            0x00200000
#define GPIO_OUTPUT_TYPE        0x00000010

static const uint32_t gpio_index[] = {
	TADR_GPIOA_BASE,			/* index 0 */
	TADR_GPIOB_BASE,			/* index 1 */
	TADR_GPIOC_BASE,			/* index 2 */
	TADR_GPIOD_BASE,			/* index 3 */
	TADR_GPIOE_BASE,			/* index 4 */
	0,							/* index 5 */
	0,							/* index 6 */
	TADR_GPIOH_BASE				/* index 7 */
};

#define NUM_OF_GPIOPORT (sizeof(gpio_index)/sizeof(uint32_t))

/*
 *  GPIOνؿ
 */
void
gpio_setup(uint32_t base, GPIO_Init_t *init, uint32_t pin)
{
	uint32_t iocurrent = 1<<pin;
	uint32_t temp = 0x00;
	uint32_t index;

	for(index = 0 ; index < NUM_OF_GPIOPORT ; index++){
		if(gpio_index[index] == base)
			break;
	}
	if(index == NUM_OF_GPIOPORT)
		return;

	/*
	 *  GPIO⡼
	 */
	/* 륿͡ȡե󥯥󡦥⡼ */
	temp = sil_rew_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4));
	temp &= ~((uint32_t)0xF << ((uint32_t)(pin & (uint32_t)0x07) * 4)) ;
	if(init->mode == GPIO_MODE_AF)
		temp |= ((uint32_t)(init->alternate) << (((uint32_t)pin & (uint32_t)0x07) * 4));
	sil_wrw_mem((uint32_t *)(base+TOFF_GPIO_AFR0+(pin>>3)*4), temp);

	/*  ϥ⡼ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_MODER), (GPIO_MODER_MODER2 << (pin * 2)), ((init->mode & GPIO_MODE) << (pin * 2)));

	/*  ϥ⡼ */
	if(init->mode == GPIO_MODE_OUTPUT || init->mode == GPIO_MODE_AF){
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OSPEEDR), (GPIO_OSPEEDER_OSPEEDR2 << (pin * 2)), (init->speed << (pin * 2)));
		sil_modw_mem((uint32_t *)(base+TOFF_GPIO_OTYPER), (GPIO_OTYPER_OT << pin), (init->otype << pin));
	}

	/*  ץ륢åסץ */
	sil_modw_mem((uint32_t *)(base+TOFF_GPIO_PUPDR), (GPIO_PUPDR_PUPDR2 << (pin * 2)), (init->pull << (pin * 2)));

	/*
	 *  EXTI⡼
	 */
	if((init->mode & EXTI_MODE) == EXTI_MODE){
		temp = sil_rew_mem((uint32_t *)((TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR+(pin / 1)) & ~3));
		temp &= ~(0x0F << (8 * (pin & 0x03)));
		temp |= (index << (8 * (pin & 0x03)));
		sil_wrw_mem((uint32_t *)((TADR_SYSCFG_BASE+TOFF_SYSCFG_EXTICR+(pin / 1)) & ~3), temp);

		if((init->mode & GPIO_MODE_IT) == GPIO_MODE_IT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_IMR1), iocurrent);

		if((init->mode & GPIO_MODE_EVT) == GPIO_MODE_EVT)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_EMR1), iocurrent);

		if((init->mode & RISING_EDGE) == RISING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_RTSR1), iocurrent);

		if((init->mode & FALLING_EDGE) == FALLING_EDGE)
			sil_orw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
		else
			sil_andw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_FTSR1), iocurrent);
	}
}


/*
 *  ƥ९å֤
 *  return SYSCLK(Hz)
 */
uint32_t
get_sysclock(uint8_t *sws)
{
	uint32_t cfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));

	if(sws != NULL)
		*sws = (cfgr & RCC_CFGR_SWS) >> 2;
	return sysclock_init_value();

}


/*
 *  DMACؿ
 */
#define DMA_TRS_TIMEOUT 2000	/* 2 */

#define DMAmux_Channel(h)   (TADR_DMAMUX1_BASE+((h)->chid2)*4)
#define DMAmux_RreqGEN(h)   (TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GEN+((h)->genreqno)*4)

static const uint32_t dma_base[] = {
	TADR_DMA1_CH1_BASE,
	TADR_DMA1_CH2_BASE,
	TADR_DMA1_CH3_BASE,
	TADR_DMA1_CH4_BASE,
	TADR_DMA1_CH5_BASE,
	TADR_DMA1_CH6_BASE,
	TADR_DMA1_CH7_BASE,
	TADR_DMA2_CH1_BASE,
	TADR_DMA2_CH2_BASE,
	TADR_DMA2_CH3_BASE,
	TADR_DMA2_CH4_BASE,
	TADR_DMA2_CH5_BASE,
	TADR_DMA2_CH6_BASE,
	TADR_DMA2_CH7_BASE
};

#define NUM_CHDMA       ((sizeof(dma_base))/sizeof(uint32_t))
#define NUM_DMACHANNEL  (NUM_CHDMA/2)

static DMA_Handle_t *pDmaHandler[NUM_CHDMA];

/*
 *  CHANNEL DMAؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_init(DMA_Handle_t *hdma)
{
	uint32_t tmp = 0;
	uint32_t i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid  = i % NUM_DMACHANNEL;
	hdma->chid2 = i;
	pDmaHandler[i] = hdma;
	hdma->base = hdma->cbase & ~0x3FF;
	index = (i % NUM_DMACHANNEL) * 4;
	hdma->chnum = (((hdma->base & 0xFF) - 8) / 20) & 0x1F;

	/* DMAꥯ */
	if(i < NUM_DMACHANNEL)
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMA1EN);
	else
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMA2EN);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB1ENR), RCC_AHB1ENR_DMAMUX1EN);

	if(hdma->Init.Direction == DMA_MEMORY_TO_MEMORY)
	    hdma->Init.Request = DMA_REQUEST_MEM2MEM;

	/* DMAMUX˥ڥեͥ */
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID);
	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->Init.Request > 0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* ɬפʤDMAMUXꥯͥ졼νԤ*/
		index = hdma->Init.Request & DMAMUX_CxCR_DMAREQ_ID;
		hdma->genreqno = index - 1;
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}
	else
		hdma->genreqno = -1;

	/* DMA-CCR쥸 */
	tmp = sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR));

	/* DMA⡼ɥӥåȤƥݥꥣ˥ꥢ */
	tmp &= ((uint32_t)~(DMA_CCR_DIR  | DMA_CCR_MEM2MEM | DMA_CCR_PINC  | \
                        DMA_CCR_MINC | DMA_CCR_PSIZE   | DMA_CCR_MSIZE | \
                        DMA_CCR_CIRC | DMA_CCR_PL));

	/* DMA⡼ɤƥݥꥣ */
	tmp |=   hdma->Init.Direction        |
          hdma->Init.PeriphInc           | hdma->Init.MemInc           |
          hdma->Init.PeriphDataAlignment | hdma->Init.MemDataAlignment |
          hdma->Init.Mode                | hdma->Init.Priority;

	/* DMA⡼ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), tmp);

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAλؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_deinit(DMA_Handle_t *hdma)
{
	int i, index;

	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	for(i = 0 ; i < NUM_CHDMA ; i++){
		if(dma_base[i] == hdma->cbase)
			break;
	}
	if(i == NUM_CHDMA)
		return E_PAR;
	hdma->chid = i % NUM_DMACHANNEL;
	pDmaHandler[i] = NULL;
	index = (i % NUM_DMACHANNEL) * 4;

	/* DMA֥͡ */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), 0);
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), 0);

	/* ƥե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_ISR_GIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
	sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));

	/* DMAꥯȥꥢ */
	sil_wrw_mem((uint32_t *)(DMAmux_Channel(hdma)), 0);

	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	/* ɬפʤꥯȥͥ졼Υѥ᡼Ԥ */
	if(hdma->Init.Request >  0 && hdma->Init.Request <= DMA_REQUEST_GENERATOR3){
		/* DMAMUXꥯȥͥ졼 쥸ꥻå */
		sil_wrw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), 0);
		/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
	}

	/* 顼֤򥯥ꥢ */
	hdma->ErrorCode = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAϴؿ
 *  parameter1  hdma:       DMAϥɥؤΥݥ
 *  parameter2  SrcAddress: ɥ쥹
 *  parameter3  DstAddress: ǥƥ͡󥢥ɥ쥹
 *  parameter4  DataLength: žĹ
 *  return ER
 */
ER
dma_start(DMA_Handle_t *hdma, uint32_t SrcAddress, uint32_t DstAddress, uint32_t DataLength)
{
	/* ѥ᡼å */
	if(hdma == NULL)
		return E_PAR;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);
	hdma->status = DMA_STATUS_BUSY;

	/* DMAMUX 󥯥ʥСե饰򥯥ꥢ */
	sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

	if(hdma->genreqno >= 0)	/* DMAMUXꥯȥͥ졼 Сե饰򥯥ꥢ */
		sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));

	/* ǡĹ */
	sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CNDTR), DataLength);

	if((hdma->Init.Direction) == DMA_MEMORY_TO_PERIPH){
		/* ꤫ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), DstAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), SrcAddress);
	}
	else{	/* ǥХ */
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CPAR), SrcAddress);
		sil_wrw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CMAR), DstAddress);
	}

	/* ߥ֥͡ */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), (DMA_CCR_TCIE | DMA_CCR_HTIE | DMA_CCR_TEIE));

    /* DMAMUX 󥯥ʥ */
	if((sil_rew_mem((uint32_t *)(DMAmux_Channel(hdma))) & DMAMUX_CxCR_SE) != 0){
		/* DMAMUX 󥯥ʥСߤ */
		sil_orw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
    }
	if(hdma->genreqno >= 0){
		/* DMAMUXꥯȥͥ졼  overrunߤͭ */
		sil_orw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
	}

	/* DMA */
	sil_orw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	return E_OK;
}

/*
 *  CHANNEL DMAߴؿ
 *  parameter1  hdma  : DMAϥɥؤΥݥ
 *  return ER
 */
ER
dma_end(DMA_Handle_t *hdma)
{
	uint32_t tick = 0;

	/* DMA */
	sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_EN);

	/* DMAԤ */
	while((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_EN) != 0){
	    /* Check for the Timeout */
		if(tick > DMA_TRS_TIMEOUT){
			/* ॢȥ顼 */
			hdma->ErrorCode |= DMA_ERROR_TIMEOUT;
			return E_TMOUT;
		}
    	dly_tsk(1);
	}
	hdma->status = DMA_ERROR_NONE;
	return E_OK;
}

/*
 *  CHANNEL DMAΥ١ɥ쥹
 */
uintptr_t
dma_channel_base(int32_t cid)
{
	if(cid < 0 || cid >= NUM_CHDMA)
		return 0;
	return (uintptr_t)dma_base[cid];
}

/*
 *  CHANNEL DMA߽ؿ
 *  parameter1  hdma: DMAϥɥؤΥݥ
 */
void
dma_inthandler(DMA_Handle_t *hdma)
{
	uint32_t index = hdma->chid * 4;

	/*
	 *  ž顼
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TEIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TEIE) != 0){
			/* ž顼ߥꥢ */
			sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TEIE);
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* ž顼ե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTEIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
			/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ž顼֤ */
			hdma->ErrorCode |= DMA_ERROR_TE;
			hdma->status = DMA_STATUS_READY_ERROR;
			/* 顼Хå */
			if(hdma->errorcallback != NULL)
				hdma->errorcallback(hdma);
		}
	}

	/*
	 *  ϡžλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_HTIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_HTIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* ϡžߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_HTIE);
			}
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* ϡžλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CHTIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ϡžơѹ */
			hdma->status = DMA_STATUS_READY_HMEM0;
			/* ϡžλХå */
			if(hdma->xferhalfcallback != NULL)
				hdma->xferhalfcallback(hdma);
		}
	}

	/*
	 *  žλ
	 */
	if((sil_rew_mem((uint32_t *)(hdma->base+TOFF_DMA_ISR)) & (DMA_ISR_TCIF << index)) != 0){
		if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_TCIE) != 0){
			if((sil_rew_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR)) & DMA_CCR_CIRC) == 0){
				/* žߤ򥯥ꥢ */
				sil_andw_mem((uint32_t *)(hdma->cbase+TOFF_DMACH_CCR), DMA_CCR_TCIE);
			}
			/* DMAMUX 󥯥Сߤ򥯥ꥢ */
			sil_andw_mem((uint32_t *)(DMAmux_Channel(hdma)), DMAMUX_CxCR_SOIE);
			/* žλե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(hdma->base+TOFF_DMA_IFCR), (DMA_IFCR_CTCIF << index));
			/* DMAMUX 󥯥Сե饰򥯥ꥢ */
			sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_CHANNEL_STS_CFR), (1 << hdma->chnum));

			if(hdma->genreqno >= 0){
				/* DMAMUXꥯȥͥ졼 overrunߤ̵ */
				sil_andw_mem((uint32_t *)(DMAmux_RreqGEN(hdma)), DMAMUX_RGxCR_OIE);
				/* DMAMUXꥯȥͥ졼 overrun flag򥯥ꥢ */
				sil_wrw_mem((uint32_t *)(TADR_DMAMUX1_BASE+TOFF_DMAMUX_REQUEST_GENSTS_RGCFR), (1 << (hdma->genreqno & 3)));
			}

			/* ơλ */
			hdma->ErrorCode = DMA_ERROR_NONE;
			hdma->status = DMA_STATUS_READY_MEM0;
			/* žλХå */
			if(hdma->xfercallback != NULL)
				hdma->xfercallback(hdma);
		}
	}
}

/*
 *  CHANNEL DMA ߥӥ롼
 */
void
channel_dma_isr(intptr_t exinf)
{
	dma_inthandler(pDmaHandler[(uint32_t)exinf]);
}


/*
 *  WINDOW WATCH-DOGǡ¸ΰ
 */
static uint32_t mcount, mcounter, wdogcount;

/*
 *  WINDOW WATCH-DOG
 *  parameter1  port: WDOGݡֹ
 *  parameter2  hini: WDOGϥɥΥݥ
 *  return ｪλ E_OK
 */
ER
wdog_init(ID portid, WDOG_Init_t *hini)
{
	if(hini == NULL || portid != WDOG1_PORTID)
		return E_PAR;
	if(hini->Window > 127 || hini->Window < 64)
		return E_PAR;
	if(hini->Counter > 127 || hini->Counter < 64)
		return E_PAR;

	mcount    = hini->MCounter;
	mcounter  = 0;
	wdogcount = hini->Counter;
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_WWDGEN);
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	if(mcount == 0)
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (hini->Prescaler | hini->Window));
	else
		sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CFR), (WWDG_CFR_EWI | hini->Prescaler | hini->Window));
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), (WWDG_CR_WDGA | wdogcount));
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGλ
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_deinit(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_WWDGEN);
	wdogcount = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGꥻå
 *  parameter1  port: WDOGݡֹ
 *  return ER
 */
ER
wdog_reset(ID portid)
{
	if(portid != WDOG1_PORTID || wdogcount == 0)
		return E_PAR;
	sil_wrw_mem((uint32_t *)(TADR_WWDG_BASE+TOFF_WWDG_CR), wdogcount);
	mcounter = 0;
	return E_OK;
}

/*
 *  WINDOW WATCH-DOGߥϥɥ
 */
void
wdog_handler(void)
{
	uint32_t base = TADR_WWDG_BASE;

	if((sil_rew_mem((uint32_t *)(base+TOFF_WWDG_CFR)) & WWDG_CFR_EWI) != 0
			&& (sil_rew_mem((uint32_t *)(base+TOFF_WWDG_SR)) & WWDG_SR_EWIF) != 0){
		sil_andw_mem((uint32_t *)(base+TOFF_WWDG_SR), WWDG_SR_EWIF);
		if(mcounter++ < mcount)
			sil_wrw_mem((uint32_t *)(base+TOFF_WWDG_CR), wdogcount);
	}
}


/*
 *  IPCCǡǼΰ
 */
static IPCC_Handle_t *shipcc;

static void
ipcc_callback_handler(void)
{
	IPCC_Handle_t *hipcc = shipcc;

	hipcc->callback();
	ipcc_setflag_channel(hipcc->cbmode, hipcc->cbcid);
}

/*
 *  IPCC
 */
void
ipcc_init(IPCC_Handle_t *hipcc)
{
	uint32_t base = TADR_IPCC_BASE;
	uint32_t tmp;

	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB3ENR), RCC_AHB3ENR_IPCCEN);
	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB3ENR));
	(void)(tmp);

	sil_orw_mem((uint32_t *)(base+TOFF_IPCC_C1CR), IPCC_C1CR_RXOIE);
	sil_orw_mem((uint32_t *)(base+TOFF_IPCC_C1CR), IPCC_C1CR_TXFIE);

	ena_int(IRQ_VECTOR_IPCC_C1_RX);
	ena_int(IRQ_VECTOR_IPCC_C1_TX);

	hipcc->base = base;
	hipcc->rstatus = 0;
	shipcc = hipcc;
}

/*
 *  IPCCͭ
 */
void
ipcc_enable(void)
{
	IPCC_Handle_t *hipcc = shipcc;

	if(hipcc == NULL)
		return;
	asm("sev");
	asm("wfe");
	sil_orw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR4), PWR_CR4_C2BOOT);
}

/*
 *  IPCCͥͭ
 */
void
ipcc_enable_channel(uint8_t mode, uint8_t cid, bool_t flag)
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t   base, channel, shift;

	if(hipcc == NULL || cid >= NUM_IPCC_CH)
		return;
	base = hipcc->base;
	channel = 1 << cid;
	shift = (mode & 1) * 16;
	if(flag)
		sil_wrw_mem((uint32_t *)(base+TOFF_IPCC_C1SCR), (channel << shift));
	sil_andw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), (channel << shift));
	if(mode == IPCC_RX_CHANNEL){
		if(flag)
			hipcc->rstatus &= ~channel;
		else
			hipcc->rstatus |= channel;
	}
	syslog_3(LOG_DEBUG, "ipcc_enable_channel mode(%d) cid(%d) hipcc->rstatus[%04x]", mode, cid, hipcc->rstatus);
}

/*
 *  IPCCͥͭ(2ͥ)
 */
void
ipcc_enable_channel2(uint8_t mode, uint8_t cid1, uint8_t cid2)
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t   base, shift;

	if(hipcc == NULL || cid1 >= NUM_IPCC_CH || cid2 >= NUM_IPCC_CH)
		return;
	base = hipcc->base;
	shift = (mode & 1) * 16;
	sil_andw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), ((1 << cid1) << shift));
	sil_andw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), ((1 << cid2) << shift));
}

/*
 *  IPCCͥ륻åȥե饰
 */
void
ipcc_setflag_channel(uint8_t mode, uint8_t cid)
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t   base, channel, shift;

	if(hipcc == NULL || cid >= NUM_IPCC_CH)
		return;
	base = hipcc->base;
	channel = 1 << cid;
	shift = (mode & 1) * 16;
	sil_wrw_mem((uint32_t *)(base+TOFF_IPCC_C1SCR), (channel << shift));
}

/*
 *  IPCCͥ륳Хå
 */
void
ipcc_callback_channel(uint8_t mode, uint8_t cid, void(*cb)(void))
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t   base, channel, shift;

	if(hipcc == NULL || cid >= NUM_IPCC_CH)
		return;
	base = hipcc->base;
	channel = 1 << cid;
	shift = (mode & 1) * 16;
	hipcc->cbmode   = mode;
	hipcc->cbcid    = cid;
	hipcc->callback = cb;
	if((sil_rew_mem((uint32_t *)(base+TOFF_IPCC_C1TOC2SR)) & channel) == channel){
		if(mode == IPCC_TX_CHANNEL)
			hipcc->sendcallback[cid] = ipcc_callback_handler;
		else
			hipcc->recvcallback[cid] = ipcc_callback_handler;
		sil_andw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), (channel << shift));
	}
	else
		ipcc_callback_handler();
}

/*
 *  IPCCߥϥɥ
 */
void
ipcc_rx_handler(void)
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t  base, channel, c2toc1sr, c1mr, i;

	if(hipcc == NULL)
		return;
	base = hipcc->base;
	c2toc1sr = sil_rew_mem((uint32_t *)(base+TOFF_IPCC_C2TOC1SR));
	c1mr = sil_rew_mem((uint32_t *)(base+TOFF_IPCC_C1MR));
	for(i = 0 ; i < NUM_IPCC_CH ; i++){
		channel = 1 << i;
		if((c2toc1sr & channel) == channel && ((~c1mr) & (channel << 0)) != 0){
			if((hipcc->rstatus & channel) != 0)
				sil_wrw_mem((uint32_t *)(base+TOFF_IPCC_C1SCR), channel);
			else
				sil_orw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), channel);
			if(hipcc->recvcallback[i] != NULL)
				hipcc->recvcallback[i]();
			break;
		}
	}
}

/*
 *  IPCCߥϥɥ
 */
void
ipcc_tx_handler(void)
{
	IPCC_Handle_t *hipcc = shipcc;
	uint32_t  base, channel, c1toc2sr, c1mr, i;

	if(hipcc == NULL)
		return;
	base = hipcc->base;
	c1toc2sr = sil_rew_mem((uint32_t *)(base+TOFF_IPCC_C1TOC2SR));
	c1mr = sil_rew_mem((uint32_t *)(base+TOFF_IPCC_C1MR));
	for(i = 0 ; i < NUM_IPCC_CH ; i++){
		channel = 1 << i;
		if((c1toc2sr & channel) != channel && ((~c1mr) & (channel << 16)) != 0){
			sil_orw_mem((uint32_t *)(base+TOFF_IPCC_C1MR), channel << 16);
			if(hipcc->sendcallback[i] != NULL)
				hipcc->sendcallback[i]();
			break;
		}
	}
}


/*
 *  ѥ⡼
 */
#define PWR_MODE_STOP0          0x000000000
#define PWR_MODE_STOP1          (PWR_CR1_LPMS_0)
#define PWR_MODE_STOP2          (PWR_CR1_LPMS_1)
#define PWR_MODE_STANDBY        (PWR_CR1_LPMS_1 | PWR_CR1_LPMS_0)
#define PWR_MODE_SHUTDOWN       (PWR_CR1_LPMS_2)

static const uint32_t HW_LPM_MODE[5] = {PWR_MODE_STOP0, PWR_MODE_STOP1, PWR_MODE_STOP2, \
	PWR_MODE_STANDBY, PWR_MODE_SHUTDOWN};

/*
 *  LPM⡼ѹ
 */
void
lpm_changemode(uint8_t config)
{
	if(config >= LPM_STANDBYMODE)
		sil_wrw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_SCR), PWR_SCR_CWUF);

	if(config == LPM_SLEEPMODE){	/* sleep */
		sil_andw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_SCR), SCB_SCR_SLEEPDEEP_Msk);
	}
	else{	/* stop ޤ off */
		sil_modw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_LPMS, HW_LPM_MODE[config-LPM_STOPMODE0]);
		sil_orw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_SCR), SCB_SCR_SLEEPDEEP_Msk);
	}
	Asm("wfi");
}


/*
 *  GET EUI-64ǡ
 */
uint32_t
get_eui64data(uint8_t off)
{
	return sil_rew_mem((uint32_t *)(UID64_BASE+off*4));
}


/*
 *  FLASHɥ饤
 */
#define FLASH_TIMEOUT_VALUE             1000000		/* FLASH Execution Timeout, 1 s */
#define FLASH_FLAG_SR_ERROR             (FLASH_SR_OPERR   | FLASH_SR_PROGERR | FLASH_SR_WRPERR | \
                                         FLASH_SR_PGAERR  | FLASH_SR_SIZERR  | FLASH_SR_PGSERR | \
                                         FLASH_SR_MISERR  | FLASH_SR_FASTERR | FLASH_SR_RDERR  | \
                                         FLASH_SR_OPTVERR)
#define FLASH_SIZE                      (((uint32_t)(*((uint16_t *)FLASHSIZE_BASE)) & (0x07FF)) << 10)
#define FLASH_NB_DOUBLE_WORDS_IN_ROW    64

#define FLASH_KEY1                      0x45670123U  /* Flash key1 */
#define FLASH_KEY2                      0xCDEF89ABU  /* Flash key2: used with FLASH_KEY1 */

/*
 *  FLASH꽪λԤؿ
 */
static ER
flash_endoperation(uint32_t Timeout)
{
	uint32_t sr;
	uint32_t tick = 0;

	/*
	 *  BUSYʤԤ
	 */
	while((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR)) & FLASH_SR_BSY) != 0){
		if(tick++ >= Timeout)
			return E_TMOUT;
		sil_dly_nse(1000);
	}

	/*
	 *  SR쥸Ф
	 */
	sr = sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR));

	/*
	 *  ڥ졼֤ʤХꥢ
	 */
	if((sr & FLASH_SR_EOP) != 0)
		sil_wrw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR), FLASH_SR_EOP);
	sr = sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR));

	/*
	 *  顼֤򥯥ꥢ
	 */
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ECCR)) & FLASH_ECCR_ECCC) != 0)
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ECCR), FLASH_ECCR_ECCC);
 	if((sr & FLASH_FLAG_SR_ERROR) != 0)
		sil_wrw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR), (sr & FLASH_FLAG_SR_ERROR));

	sr = sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR));
 	if((sr & FLASH_FLAG_SR_ERROR) != 0)
		return E_SYS;

	/*
	 *  CONFIG-BUSYʤԤ
	 */
	while((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR)) & FLASH_SR_CFGBSY) != 0){
		if(tick++ >= Timeout)
			return E_TMOUT;
		sil_dly_nse(1000);
	}
	return E_OK;
}

/*
 *  饤ȥץƥΰ
 *  parameter1  pwrp:     FLASH 񤭹߶ػ߾¤ΤΥݥ
 *  return      E_OK
 */
ER
flash_getwrparea(FLASH_WRPAREA_t *pwrp)
{
	uint32_t wrp1r;

	if(pwrp->WRPArea == OB_WRPAREA_BANK1_AREAA)
		wrp1r = sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_WRP1AR));
	else if(pwrp->WRPArea == OB_WRPAREA_BANK1_AREAB)
		wrp1r = sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_WRP1BR));
	else
		return E_PAR;
	pwrp->WRPStartOffset = wrp1r & FLASH_WRP1AR_WRP1A_STRT;
	pwrp->WRPEndOffset   = (wrp1r & FLASH_WRP1AR_WRP1A_END) >> 16;
	return E_OK;
}

/*
 *  FLASH UNLOCK
 */
ER
flash_unlock(void)
{
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR)) & FLASH_CR_LOCK) == 0)
		return E_OK;
	sil_wrw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_KEYR), FLASH_KEY1);
	sil_wrw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_KEYR), FLASH_KEY2);
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR)) & FLASH_CR_LOCK) != 0)
		return E_SYS;
	else
		return E_OK;
}

/*
 *  FLASH LOCK
 */
ER
flash_lock(void)
{
	sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), FLASH_CR_LOCK);
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR)) & FLASH_CR_LOCK) == 0)
		return E_SYS;
	else
		return E_OK;
}

/*
 *  FLASH-ROMǡõ
 *  parameter1  pErase:     õ׵ṽ¤ΤΥݥ
 *  parameter2  pPageError: ڡõξ硢顼ȯڡֹ
 *  return      E_OK
 */
ER
flash_erase(FLASH_Erase_t *pErase, uint32_t *PageError)
{
	ER ercd = E_OK;
	uint32_t i;

	if(pErase == NULL)
		return E_PAR;

	if((ercd = flash_endoperation(FLASH_TIMEOUT_VALUE)) != E_OK)
		return ercd;

	if(pErase->TypeErase == FLASH_TYPEERASE_MASS){
		/*
		 *  ޥ졼
		 */
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), (FLASH_CR_MER | FLASH_CR_STRT));
		ercd = flash_endoperation(FLASH_TIMEOUT_VALUE);
	}
	else if(pErase->TypeErase == FLASH_TYPEERASE_PAGES){
		if(PageError == NULL)
			return E_PAR;
		*PageError = 0xFFFFFFFFU;

		for(i = 0 ; i < pErase->NbPages ; i++){
			/*
			 *  ڡõ
			 */
			sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), FLASH_CR_PNB, ((pErase->Page + i) << 3));
			sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), (FLASH_CR_PER | FLASH_CR_STRT));

			/*
			 *  ڡõԤ
			 */
			if((ercd = flash_endoperation(FLASH_TIMEOUT_VALUE)) != E_OK){
				*PageError = pErase->Page + i;
				break;
			}
		}

		/*
		 *  졼λʤ顢ڥ졼ӥåȤ򥪥դ
		 */
		sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), (FLASH_CR_PER | FLASH_CR_PNB));
	}
	else
		return E_PAR;

	/*
	 *  FLASHå奯ꥢ
	 */
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_ICEN) != 0){
		sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_ICEN);
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_ICRST);
		sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_ICRST);
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_ICEN);
	}
	if((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_DCEN) != 0){
		sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_DCEN);
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_DCRST);
		sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_DCRST);
		sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_DCEN);
	}
	return ercd;
}

/*
 *  FLASH񤭹
 *  parameter1  type:   õ׵ṽ¤ΤΥݥ
 *  parameter2  addr:   ڡõξ硢顼ȯڡֹ
 *  parameter3  data:   type:FLASH_TYPEPROGRAM_DOUBLEWORDξ硢񤭹ߥǡ
 *                      type:FLASH_TYPEPROGRAM_FASTξ硢񤭹ߥǡΰΥݥ
 *  return      E_OK
 */
ER
flash_program(uint32_t type, uint32_t addr, uint64_t data)
{
	ER ercd = E_OK;

	if((addr & 7) != 0)
		return E_PAR;

	if((ercd = flash_endoperation(FLASH_TIMEOUT_VALUE)) != E_OK)
		return ercd;

	if(type == FLASH_TYPEPROGRAM_DOUBLEWORD){
		if((addr >= FLASH_BASE && addr <= (FLASH_BASE + FLASH_SIZE - 8)) ||
			(addr >= OTP_AREA_BASE && addr <= (OTP_AREA_END_ADDR - 7))){
			/*
			 *  PGӥåȥå
			 */
			sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), FLASH_CR_PG);
			/*
			 *  եȥɽ񤭹
			 */
			*(uint32_t *)addr = (uint32_t)data;
			Asm("isb 0xF":::"memory");

			/*
			 *  ɥɽ񤭹
			 */
			*(uint32_t *)(addr + 4) = (uint32_t)(data >> 32);
		}
		else
			return E_PAR;
    }
    else if(type == FLASH_TYPEPROGRAM_FAST){
		if(addr >= FLASH_BASE && addr <= (FLASH_BASE + FLASH_SIZE - 8)){
			volatile uint32_t *d = (volatile uint32_t *)addr;
			volatile uint32_t *s = (volatile uint32_t *)((uint32_t)data);
			uint32_t tick = 0;
			uint32_t i;

			/*
			 *  FSTPGӥåȥå
			 */
			sil_orw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), FLASH_CR_FSTPG);
			loc_cpu();

			/*
			 *  ǡ񤭹
			 */
			for(i = 0 ; i < (2 * FLASH_NB_DOUBLE_WORDS_IN_ROW) ; i++)
				*d++ = *s++;

			/*
			 *  BSYӥåȤΤԤ
			 */
			while((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_SR)) & FLASH_SR_BSY) != 0){
				if(tick++ >= FLASH_TIMEOUT_VALUE){
					ercd = E_TMOUT;
					break;
				}
				sil_dly_nse(1000);
			}
			unl_cpu();
		}
		else
			return E_PAR;
    }
	else
		return E_PAR;
	/*
	 *  񤭹߽λԤ
	 */
	if(ercd == E_OK)
		ercd = flash_endoperation(FLASH_TIMEOUT_VALUE);

	/*
	 *  ץӥåȤȤ
	 */
	sil_andw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_CR), type);
	return ercd;
}


/*
 *  LED³ݡ
 *
 *  ĥI/OܡɤLED1-3ϥץޥ֥ϥݡ0
 *  ˥С𤷤³Ƥ롥
 */

typedef struct gio_confstruct
{
	uint32_t    clkbit;
	uint32_t    giobase;
	uint32_t    pinpos;
	uint32_t    pinpp;
} gio_conftype;

static const gio_conftype led_confdata[] = {
#if BOARDNO == 1
	{ RCC_AHB2ENR_GPIOAEN, TADR_GPIOA_BASE, PINPOSITION4, GPIO_PULLUP},
#else
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION5, GPIO_PULLUP},
#endif
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION0, GPIO_PULLUP},
	{ RCC_AHB2ENR_GPIOBEN, TADR_GPIOB_BASE, PINPOSITION1, GPIO_PULLUP}
};
#define NUM_LED   (sizeof(led_confdata)/sizeof(gio_conftype))

/*
 *  LED³ݡȽ
 */ 
void
led_init(intptr_t exinf)
{
	const gio_conftype *pled = &led_confdata[0];
	GPIO_Init_t GPIO_Init_Data;
	uint32_t i;

	for(i = 0 ; i < NUM_LED ; pled++, i++){
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), pled->clkbit);

		/* ⡼ */
		GPIO_Init_Data.mode      = GPIO_MODE_OUTPUT;
		/* ץ륢å ץ */
		GPIO_Init_Data.pull      = pled->pinpp;
		/* ϥ⡼ */
		GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
        /* ԡ */
		GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
		gpio_setup(pled->giobase, &GPIO_Init_Data, pled->pinpos);

		/*  */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (1 << pled->pinpos)<<16);
	}
}

/*
 *  LED³ݡɤ߽Ф
 */
uint_t
led_in(void)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t data, i;

	for(i = 0, data = 0 ; i < NUM_LED ; pled++, i++){
		if((sil_rew_mem((uint32_t *)(pled->giobase+TOFF_GPIO_IDR)) & (1<<(pled->pinpos))) != 0)
			data |= (1<<i);
	}
	return data;
}

/*
 *  LED³ݡȽ񤭹
 */ 
void
led_out(unsigned int led_data)
{
	const gio_conftype *pled = &led_confdata[0];
	uint32_t reg1, reg2, i;

	/* ͤLED³ӥåȰʳѹʤ */
	for(i = 0 ; i < NUM_LED ; pled++, i++){
		reg1 = reg2 = 0;
		if((led_data & (1<<i)) != 0)
			reg2 = 1 << pled->pinpos;
		else
			reg1 = 1 << pled->pinpos;

		/* 񤭹ߥǡƽ񤭹 */
		sil_wrw_mem((uint32_t *)(pled->giobase+TOFF_GPIO_BSRR), (reg1<<16) | reg2);
	}
}

/*
 * LEDȥåθꡦɤ߹ߴؿ
 */

/*
 *  LEDξ¸ѿ
 */	
unsigned int LedState;


/*
 *  LED
 */
void 
set_led_state(unsigned int led, unsigned char state){
	if (state == ON) {
		LedState = LedState | led;
	} else {
		LedState = LedState & ~led;
	}
	led_out(LedState);
}


#if BOARDNO == 1
#define TADR_PSW1_BASE  TADR_GPIOA_BASE
#define PSW1GIOPEN      RCC_AHB2ENR_GPIOAEN
#define	PSW1_PIN        PINPOSITION10
#else
#define TADR_PSW1_BASE  TADR_GPIOC_BASE
#define PSW1GIOPEN      RCC_AHB2ENR_GPIOCEN
#define	PSW1_PIN        PINPOSITION4
#define TADR_PSW2_BASE  TADR_GPIOD_BASE
#define PSW2GIOPEN      RCC_AHB2ENR_GPIODEN
#define	PSW2_PIN        PINPOSITION0
#define TADR_PSW3_BASE  TADR_GPIOD_BASE
#endif
#define	PSW3_PIN        PINPOSITION1
#define EXTI0_BASENO    0
#define EXTI1_BASENO    1
#define EXTI4_BASENO    4

/*
 * PUSHå³ݡȽ
 */
void
switch_push_init(intptr_t exinf)
{
	GPIO_Init_t GPIO_Init_Data;
	volatile uint32_t temp;

	/* ϥݡȤ */
	/* Enable the BUTTON Clock */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), PSW1GIOPEN);
	temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR));
#if defined(TADR_PSW2_BASE) || defined(TADR_PSW3_BASE)
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR), PSW2GIOPEN);
	temp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_AHB2ENR));
#endif
	(void)(temp);

	/* ⡼ */
	GPIO_Init_Data.mode      = GPIO_MODE_IT_FALLING;
	/* ץ륢å ץ */
	GPIO_Init_Data.pull      = GPIO_PULLUP;
	GPIO_Init_Data.speed     = GPIO_SPEED_LOW;
	gpio_setup(TADR_PSW1_BASE, &GPIO_Init_Data, PSW1_PIN);

	/* ⡼ */
#ifdef TADR_PSW2_BASE
#ifndef USE_SWITCH_INT
	GPIO_Init_Data.mode      = GPIO_MODE_INPUT;
#endif
	gpio_setup(TADR_PSW2_BASE, &GPIO_Init_Data, PSW2_PIN);
#endif
#ifdef TADR_PSW3_BASE
	GPIO_Init_Data.mode      = GPIO_MODE_INPUT;
	gpio_setup(TADR_PSW3_BASE, &GPIO_Init_Data, PSW3_PIN);
#endif
}

/*
 * PUSHåХåؿ
 */
void setup_sw_func(intptr_t exinf)
{
	exti_func[PSW1_PIN] = (void (*)(void))exinf;
}

/*
 * EXTI4ޤ10ߥϥɥ
 */
void
exti_handler(void)
{
	uint32_t istatus = sil_rew_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1));
	uint32_t i;

	for(i = 0 ; i < NUM_EXTI_FUNC ; i++){
		if((istatus & (1<<i)) != 0){
			sil_wrw_mem((uint32_t *)(TADR_EXTI_BASE+TOFF_EXTI_PR1), istatus & (1<<i));
			if(exti_func[i] != NULL)
				(exti_func[i])();
		}
	}
}

/*
 *  PUSHåФ
 */
uint32_t
switch_push_sense(void)
{
	uint32_t sw = 0;

	if((sil_rew_mem((uint32_t *)(TADR_PSW1_BASE+TOFF_GPIO_IDR)) & (1<<PSW1_PIN)) == 0)
		sw |= PSW1;
#ifdef TADR_PSW2_BASE
	if((sil_rew_mem((uint32_t *)(TADR_PSW2_BASE+TOFF_GPIO_IDR)) & (1<<PSW2_PIN)) == 0)
		sw |= PSW2;
#endif
#ifdef TADR_PSW3_BASE
	if((sil_rew_mem((uint32_t *)(TADR_PSW3_BASE+TOFF_GPIO_IDR)) & (1<<PSW3_PIN)) == 0)
		sw |= PSW3;
#endif
	return sw;
}


/*
 * DIPSWμФ
 */
uint_t
switch_dip_sense(void)
{
	return dipsw_value;
}

