/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2021 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: pwm.c 2416 2021-01-24 17:18:59Z roi $
 */

/* 
 *  PWMɥ饤дؿ
 *
 */
#include "kernel.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "device.h"
#include "pwm.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  TIMݡID֥åФΥޥ
 */
#define INDEX_TIM(timid)        ((uint_t)((timid) - 1))

#define GPIO_AF1_TIM            0x01

/*
 *  TIM Capture/Compare Channel
 */
#define TIM_CCx_ENABLE                   TIM_CCER_CC1E		/* Input or output channel is enabled */
#define TIM_CCx_DISABLE                  0x00000000U		/* Input or output channel is disabled */
#define TIM_CCxN_ENABLE                  TIM_CCER_CC1NE		/* Complementary output channel is enabled */
#define TIM_CCxN_DISABLE                 0x00000000U		/* Complementary output channel is enabled */

/*
 *  CCxCCxNޥ
 */
#define TIM_CCER_CCxE_MASK  ((uint32_t)(TIM_CCER_CC1E | TIM_CCER_CC2E | TIM_CCER_CC3E | TIM_CCER_CC4E))
#define TIM_CCER_CCxNE_MASK ((uint32_t)(TIM_CCER_CC1NE | TIM_CCER_CC2NE | TIM_CCER_CC3NE))


static const TIM_PortControlBlock tim_pcb[NUM_TIMPORT] = {
  {	TADR_TIM1_BASE,
	(TADR_RCC_BASE+TOFF_RCC_APB2ENR), RCC_APB2ENR_TIM1EN },

  {	TADR_TIM2_BASE,
	(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_TIM2EN },
};


/*
 *  PWMؿ
 */
static ER
pwm_configchannel(TIM_Handle_t *htim, TIM_OC_Init_t *sConfig, uint32_t Channel)
{
	uint32_t shift1 = (Channel & TIM_CHANNELID_ALL) * 4;
	uint32_t shift2 = (Channel & 0x01) * 8;
	uint32_t shift0 = shift1 / 2;
	uint32_t moff   = (Channel & 0x02) * 2;
	uint32_t tmpccmrx, tmpccer, tmpcr2;

	/*
	 *  Channel CCxEꥻå
	 */
	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_CCER), (TIM_CCER_CC1E << shift1));

	/*
	 *  TIMx CCER, CR2, CCMR쥸ͼ
	 */
	tmpccer = sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CCER));
	tmpcr2 = sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CR2));
	tmpccmrx = sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CCMR1+moff));

	/*
	 *  Output Compare Modeӥåȥꥻå
	 */
	tmpccmrx &= ~(TIM_CCMR1_OC1M << shift2);
	tmpccmrx &= ~(TIM_CCMR1_CC1S << shift2);
	/*
	 *  Output Compare Mode
	 */
	tmpccmrx |= sConfig->OCMode << shift2;

	/*
	 *  Output Polarity level
	 */
	tmpccer &= ~(TIM_CCER_CC1P << shift1);
	tmpccer |= sConfig->OCPolarity << shift1;

	if(htim->base == TADR_TIM1_BASE){
		if(Channel <= TIM_CHANNEL3_ID){
			/*
			 *  Output N Polarity levelꥻå
			 */
			tmpccer &= ~(TIM_CCER_CC1NP << shift1);
			/*
			 * Output N State
			 */
			tmpccer &= ~(TIM_CCER_CC1NE << shift1);
		}

		/*
		 *  Output Compare and Output Compare N ꥻå
		 */
		tmpcr2 &= ~(TIM_CR2_OIS1 << shift0);
		tmpcr2 &= ~(TIM_CR2_OIS1N << shift0);
	}

	/*
	 *  TIMx CR2, CCMR쥸
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_CR2), tmpcr2);
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_CCMR1+moff), tmpccmrx);

	/*
	 *  TIMx PULSE
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_CCR1+shift1), sConfig->Pulse);

	/*
	 *  TIMx CCER
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_CCER), tmpccer);

    /*
	 *  ͥץ
	 */
	sil_orw_mem((uint32_t *)(htim->base+TOFF_TIM_CCMR1+moff), (TIM_CCMR1_OC1PE << shift2));

	/*
	 *  Output Fast mode
	 */
	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_CCMR1+moff), (TIM_CCMR1_OC1FE << shift2));
	sil_orw_mem((uint32_t *)(htim->base+TOFF_TIM_CCMR1+moff), (sConfig->OCFastMode << shift2));

	return E_OK;
}

/*
 *  TIM Capture Compare Channel x̵ͭ
 */
static void
pwm_ccx_channelcmd(TIM_Handle_t *htim, uint32_t Channel, uint32_t ChannelState)
{
	uint32_t tmp = TIM_CCER_CC1E << ((Channel & TIM_CHANNELID_ALL) * 4); /* 0x1FU = 31 bits max shift */

	/*
	 *  CCxEӥåȥꥻå
	 */
	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_CCER), tmp);

	/*
	 *  CCxEӥåȥåץǡ
	 */
	sil_orw_mem((uint32_t *)(htim->base+TOFF_TIM_CCER), (ChannelState << ((Channel & TIM_CHANNELID_ALL) * 4)));
}

/*
 *  TIM-PWM
 *  parameter1  port: TIMݡֹ
 *  parameter2  htim: TIMϥɥؤΥݥ
 *  return ER
 */
ER
pwm_init(ID port, TIM_Handle_t *htim)
{
	const TIM_PortControlBlock *tpcb;
	TIM_Init_t *pini;
	uint32_t no;
	uint32_t tmpreg, tmpcr1;
	uint32_t i;

	/* Check the TIM handle allocation */
	if (htim == NULL)
		return E_PAR;

	if(port < TIM1_PORTID || port > NUM_TIMPORT)
		return E_PAR;

	no = INDEX_TIM(port);
	tpcb = &tim_pcb[no];
	htim->base = tpcb->base;
	pini = &htim->Init;

	/*
	 *  TIMå
	 */
	sil_orw_mem((uint32_t *)(tpcb->timclockbase), tpcb->timclockbit);
	tmpreg = sil_rew_mem((uint32_t *)(tpcb->timclockbase));
	(void)tmpreg;

	/*
	 *  TIM-CR1(󥿥⡼ɡåʬȥ)
	 */
	tmpcr1 = sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CR1));
    tmpcr1 &= ~(TIM_CR1_DIR | TIM_CR1_CMS);
    tmpcr1 |= pini->CounterMode;
    tmpcr1 &= ~TIM_CR1_CKD;
    tmpcr1 |= (uint32_t)pini->ClockDivision;
	tmpcr1 &= ~TIM_CR1_ARPE;
	tmpcr1 |= pini->AutoReloadPreload;
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_CR1), tmpcr1);

	/*
	 *  
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_ARR), pini->Period);

	/*
	 *  ץꥹ
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_PSC), pini->Prescaler);

	if(htim->base == TADR_TIM1_BASE){
		/*
		 *  Repetition Counter value
		 */
		sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_RCR), pini->RepetitionCounter);
	}

	/*
	 *  åץ
	 */
	sil_wrw_mem((uint32_t *)(htim->base+TOFF_TIM_EGR), TIM_EGR_UG);

	/*
	 *  ֤ͥꥻåȤ
	 */
	for(i = 0 ; i < NUM_TIM_CHANNEL ; i++)
		htim->ChannelState[i] = TIM_CHANNEL_STATE_RESET;

	/*
	 *  ȥꥬȥץåȥ⡼ɤꥻå
	 */
	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_CR2), (TIM_CR2_MMS | TIM_CR2_MMS2));

	/*
	 *  Master/slave mode̵
	 */
	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_SMCR), TIM_SMCR_MSM);

	return E_OK;
}

/*
 *  PWMϴؿ
 *  parameter1  htim:ޡϥɥ
 *  parameter2  sconfing: PWM깽¤ΤؤΥݥ
 *  parameter3  Channel:  ͥID
 *  return ER
 */
ER
pwm_gpio_channel(TIM_Handle_t *htim, uint32_t Channel, uint32_t gbase, uint8_t pin)
{
	GPIO_Init_t GPIO_Init_Data = {0};

	if(htim == NULL)
		return E_PAR;
	if(Channel >= NUM_TIM_CHANNEL)
		return E_PAR;
	GPIO_Init_Data.mode      = GPIO_MODE_AF;
	GPIO_Init_Data.otype     = GPIO_OTYPE_PP;
	GPIO_Init_Data.pull      = GPIO_PULLUP;
	GPIO_Init_Data.speed     = GPIO_SPEED_HIGH;
	GPIO_Init_Data.alternate = GPIO_AF1_TIM;
	gpio_setup(gbase, &GPIO_Init_Data, pin);
	htim->ChannelState[Channel] = TIM_CHANNEL_STATE_READY;
	htim->ActChannel |= 1 << Channel;
	return E_OK;
}

/*
 *  PWMϴؿ
 *  parameter1  htim:ޡϥɥ
 *  parameter2  sconfing: PWM깽¤ΤؤΥݥ
 *  parameter3  Channel:  ͥID
 *  return ER
 */
ER
pwm_start(TIM_Handle_t *htim, TIM_OC_Init_t *sConfig, uint32_t Channel)
{
	ER ercd;

	if(htim == NULL)
		return E_PAR;

	/*
	 *  ޡ֤ǧ
	 */
	if(htim->ChannelState[Channel] != TIM_CHANNEL_STATE_READY)
		return E_OBJ;

	/*
	 *  PWMͥ
	 */
	if((ercd = pwm_configchannel(htim, sConfig, Channel)) != E_OK)
		return ercd;

	/*
	 *  BUSY֤
	 */
	htim->ChannelState[Channel] = TIM_CHANNEL_STATE_BUSY;

	/* Enable the Capture compare channel */
	pwm_ccx_channelcmd(htim, Channel, TIM_CCx_ENABLE);

	if(htim->base == TADR_TIM1_BASE){
		/*
		 *  ᥤ󥢥ȥץåȳ
		 */
		sil_orw_mem((uint32_t *)(htim->base+TOFF_TIM_BDTR), TIM_BDTR_MOE);
	}

	/*
	 *  ޡͭ
	 */
	sil_orw_mem((uint32_t *)(htim->base+TOFF_TIM_CR1), TIM_CR1_CEN);
	return E_OK;
}

/*
 *  PWMߴؿ
 *  parameter1  htim:ޡϥɥ
 *  parameter2  Channel:  ͥID
 *  return ER
 */
ER
pwm_stop(TIM_Handle_t *htim, uint32_t Channel)
{
	if(htim == NULL)
		return E_PAR;

  /* Disable the Capture compare channel */
	pwm_ccx_channelcmd(htim, Channel, TIM_CCx_DISABLE);

	if(htim->base == TADR_TIM1_BASE){
		/*
		 *  ᥤ󥢥ȥץå
		 */
		if((sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CCER)) & (TIM_CCER_CCxE_MASK | TIM_CCER_CCxNE_MASK)) == 0){
        	sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_BDTR), TIM_BDTR_MOE);
		}
	}

	/*
	 *  ޡ
	 */
	if((sil_rew_mem((uint32_t *)(htim->base+TOFF_TIM_CCER)) & (TIM_CCER_CCxE_MASK | TIM_CCER_CCxNE_MASK)) == 0){
        sil_andw_mem((uint32_t *)(htim->base+TOFF_TIM_CR1), TIM_CR1_CEN);
	}

	/*
	 *  ޡ֤ͥǥ
	 */
	htim->ChannelState[Channel] = TIM_CHANNEL_STATE_READY;
	return E_OK;
}

