/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: wakeup_timer.c 2416 2019-05-30 21:32:59Z roi $
 */

/* 
 *  WAKEUP TIMER DRIVERץ
 */

#if 1
#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include <string.h>
#include <target_syssvc.h>
#else
#include "app_common.h"
#include <sil.h>
#include "hw_conf.h"
#endif
#include "device.h"
#include "rtc.h"
#include "wakeup_timer.h"

/*
 *  ӥΥ顼Υ
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))

#ifndef NUM_WAKEUP_TIMER
#define NUM_WAKEUP_TIMER        8
#endif
#ifndef RTC_HANDLER_MAX_DELAY
#define RTC_HANDLER_MAX_DELAY	320	/* 1ms x 37Khz / (RTC_ASYNCH_PRESCALER+1) */
#endif

#define SSR_RESET	            0xFFFFFFFF
#define TIMER_LIST_EMPTY        0xFFFF

/*
 *  WAKEUP TIMER
 */
#define WAKEUP_TIMER_FREE       0
#define WAKEUP_TIMER_ASSIGNED   1
#define WAKEUP_TIMER_USED       2

/*
 *  WAKEUP TIMER֥⡼
 */
#define WAKEUP_TIMER_OVERPASS   0
#define WAKEUP_TIMER_ENOUGH     1

#if NUM_WAKEUP_TIMER > 0

typedef enum
{
  SSR_Read_Requested,
  SSR_Read_Not_Requested
}RequestReadSSR_t;


typedef struct _wakeuptimerhandle wakeuptimerhandle_t;
struct _wakeuptimerhandle{
	wakeuptimerhandle_t      *pprev;
	wakeuptimerhandle_t      *pnext;
	void                     *pfTimerCallBack;
	uint32_t                 CounterInit;
	uint32_t                 CountLeft;
	uint32_t                 pID;
	uint8_t                  status;
	uint8_t                  mode;
	uint8_t                  tmID;
	uint8_t                  dummy;
};




static wakeuptimerhandle_t wakeupTimerContext[NUM_WAKEUP_TIMER];
static volatile wakeuptimerhandle_t *pCurrentRunningTimer;
static volatile wakeuptimerhandle_t *pPreviousRunningTimer;
static volatile uint32_t SSRValueOnLastSetup;
static volatile uint16_t MaxWakeupTimerSetup;
static volatile uint8_t  WakeupTimerMode;
static void     (*timercallback)(uint32_t pid, uint8_t tid, void *func);

static RTC_Wakeup_t rtc_wup;

__attribute__((weak)) void HW_TS_RTC_CountUpdated_AppNot(void);


/*
 *  SSR쥸ɤ߽Ф
 */
static uint32_t
ReadSSR(void)
{
	uint32_t result = SSR_RESET;
	/*
	 * It needs to be checked the reading of the SSR register is stable
	 */
	while(result != rtc_get_ssr()){
		result = rtc_get_ssr();
	}
	return (uint32_t)result;
}

/*
 *  в֤Ф
 */
static uint16_t
TimeElapsed(void)
{
	uint32_t time_elapsed = 0;

	if(SSRValueOnLastSetup != SSR_RESET){
		time_elapsed = ReadSSR();
		if (SSRValueOnLastSetup >= time_elapsed)
			time_elapsed = SSRValueOnLastSetup - time_elapsed;
		else{
			uint32_t wrap_counter = rtc_wup.synchPrescaler - time_elapsed;
			time_elapsed = SSRValueOnLastSetup + wrap_counter;
		}

		/*
		 *  WAKUP TICKͤ
		 */
		time_elapsed = time_elapsed * rtc_wup.asynchPrescaler;
		time_elapsed = time_elapsed >> rtc_wup.wakeuptimerPrescaler;
	}
	return (uint16_t)time_elapsed;
}


/*
 *  ޥϥɥΥ
 *  parameter1 pcurr:󥯤륿ޥϥɥ
 *  parameter2 type: 󥯥
 *  parameter3 RequestReadSSR: Request to read the SSR register or not
 */
static void
UnlinkTimerHandle(wakeuptimerhandle_t *pcurr, uint8_t type, RequestReadSSR_t RequestReadSSR)
{
	wakeuptimerhandle_t *pnext, *pprev;

	pnext = pcurr->pnext;
	if(pcurr == pCurrentRunningTimer){
		pPreviousRunningTimer = pcurr;
		pCurrentRunningTimer  = pnext;
		if(pnext != NULL)
			pnext->pprev = NULL;
	}
	else{
		pprev = pcurr->pprev;

		if((pprev = pcurr->pprev) != NULL)
			pprev->pnext = pnext;
		if(pnext != NULL)
			pnext->pprev = pprev;
	}

	/*
	 *  ޥϥɥ饯ꥢ
	 */
	pcurr->status = type;
	pcurr->pnext = NULL;
	pcurr->pprev = NULL;

	if(pCurrentRunningTimer == NULL && RequestReadSSR == SSR_Read_Requested){
		SSRValueOnLastSetup = SSR_RESET;
	}
}

/*
 *  ޥϥɥΥ
 *  parameter1 pcurr:󥯤륿ޥϥɥ
 *  return в
 */
static uint16_t
linkTimerHandle(wakeuptimerhandle_t *pcurr)
{
	uint32_t timeleft;
	uint16_t time_elapsed = 0;
	uint8_t timer_id_lookup;
	wakeuptimerhandle_t *pnext, *pnextnext = NULL;

	if((pnext = (wakeuptimerhandle_t *)pCurrentRunningTimer) == NULL){
		/*
		 *  WAKEUPޤ̤¹Ԥξ
		 */
		pPreviousRunningTimer = NULL;
		pCurrentRunningTimer  = pcurr;
		pcurr->pnext = NULL;
		pcurr->pprev = NULL;
		SSRValueOnLastSetup = SSR_RESET;
	}
	else{
		/*
		 *  ¹ξ硢в֤Ф
		 */
		time_elapsed = TimeElapsed();
		pcurr->CountLeft += time_elapsed;
		timeleft = pcurr->CountLeft;

		if(pnext != NULL && pnext->CountLeft <= timeleft){
			/*
			 *  ľǥॢåפ륿ޤõ
			 */
			timer_id_lookup = pnext->tmID;
			while((pnextnext = pnext->pnext) != NULL && pnextnext->CountLeft <= timeleft){
				timer_id_lookup = pnextnext->tmID;
				pnext = pnextnext;
			}

			/*
			 *  pnextθ˥ơ֥
			 */
#if 1	/* ROI DEBUG */
			syslog_2(LOG_NOTICE, "## pnext[%08x] p[timer_id_lookup][%08x] ##", pnext, &wakeupTimerContext[timer_id_lookup]);
#endif	/* ROI DEBUG */
			if((pnextnext = pnext->pnext) != NULL)
				pnextnext->pprev = pcurr;
			pcurr->pnext = pnextnext;
			pcurr->pprev = pnext;
			pnext->pnext = pcurr;
		}
		else{
			/*
			 *  Ƭ˥󥯤
			 */
			if(pnext != NULL)
				pnext->pprev = pcurr;
			pPreviousRunningTimer = pnext;
			pcurr->pnext = (wakeuptimerhandle_t *)pnext;
			pcurr->pprev = NULL;
			pCurrentRunningTimer = pcurr;
		}
	}
	return time_elapsed;
}

/*
 *  WAKEUP TIMER
 */
static void
RescheduleTimer(void)
{
	uint32_t timecountleft = 0;
	uint16_t wakeup_ticks;
	uint16_t time_elapsed;
	volatile wakeuptimerhandle_t *pcurr = pCurrentRunningTimer;


	if(pcurr != NULL)
		timecountleft = pcurr->CountLeft;

	time_elapsed = TimeElapsed();

	if(timecountleft < time_elapsed){	/* в֤᤮Ƥ */
		wakeup_ticks = 0;
		WakeupTimerMode = WAKEUP_TIMER_ENOUGH;
	}
	else{		/* в */
		if(timecountleft > (time_elapsed + MaxWakeupTimerSetup)){
			/*
			 *  ֤礭
			 */
			wakeup_ticks = MaxWakeupTimerSetup;
			WakeupTimerMode = WAKEUP_TIMER_OVERPASS;
		}
		else{
			wakeup_ticks = timecountleft - time_elapsed;
			WakeupTimerMode = WAKEUP_TIMER_ENOUGH;
		}
	}

	/*
	 *  ¾ΥꥹȤ
	 */
	while(pcurr != NULL){
		if(pcurr->CountLeft < time_elapsed)
			pcurr->CountLeft = 0;
		else
			pcurr->CountLeft -= time_elapsed;
		pcurr = pcurr->pnext;
	}

	/*
	 *  WAKEUPֻ
	 */
	if(wakeup_ticks == 0)
		SSRValueOnLastSetup = ReadSSR();
	else{
		if(wakeup_ticks > 1 || rtc_wup.wakeuptimerPrescaler != 1)
			wakeup_ticks--;

		rtc_setup_wakeuptime(wakeup_ticks);

	    /**
	     * Update the value here after the WUTWF polling that may take some time
	     */
	    SSRValueOnLastSetup = ReadSSR();

	    HW_TS_RTC_CountUpdated_AppNot();
	}
}


/*
 *  WAKEUP TIMER߽
 *  parameter1 tid:  ID
 *  parameter2 type: ߥ׻
 */
static void
timer_stop(uint8_t tid, uint8_t type)
{
	if(wakeupTimerContext[tid].status == WAKEUP_TIMER_USED){
		UnlinkTimerHandle(&wakeupTimerContext[tid], type, SSR_Read_Requested);
		if(pCurrentRunningTimer == NULL)
			rtc_setup_wakeuptime(-1);
		else if(pPreviousRunningTimer != pCurrentRunningTimer)
			RescheduleTimer();
	}
}

/*
 *  WAKEUP TIMER
 *  parameter1 tid:  ID
 *  parameter2 ticks:٥Ȼ(wakeup timer ticks)
 */
static void
timer_start(uint8_t tid, uint32_t ticks)
{
	wakeuptimerhandle_t *pcurr = &wakeupTimerContext[tid];
	uint16_t time_elapsed;

	if(pcurr->status == WAKEUP_TIMER_USED){
		timer_stop(tid, WAKEUP_TIMER_ASSIGNED);
	}

	pcurr->status      = WAKEUP_TIMER_USED;
	pcurr->CountLeft   = ticks;
	pcurr->CounterInit = ticks;
	time_elapsed = linkTimerHandle(pcurr);

	if(pPreviousRunningTimer != pCurrentRunningTimer)
		RescheduleTimer();
	else
		pcurr->CountLeft -= time_elapsed;
}

/*
 *  RTC WAKEUPߤΥХåؿ
 */
static void
rtc_wakeup_handler(void)
{
	volatile wakeuptimerhandle_t *pcurr = pCurrentRunningTimer;
#if 1	/* ROI DEBUG */
  static int count = 0;
  if((count % 100) == 0)
    syslog_1(LOG_NOTICE, "## rtc_wakeup_handler(%d) ##", count);
  count++;
#endif	/* ROI DEBUG */


	if(pcurr != NULL && pcurr->status == WAKEUP_TIMER_USED){
		/*
		 *  RUNNING TIMERͭξ硢߽Ԥ
		 */
		if(WakeupTimerMode != WAKEUP_TIMER_OVERPASS){
			if(pcurr->mode == WAKEUP_TIMER_REPEAT){
				UnlinkTimerHandle((wakeuptimerhandle_t *)pcurr, WAKEUP_TIMER_ASSIGNED, SSR_Read_Not_Requested);
				timer_start(pcurr->tmID, pcurr->CounterInit);

			}
			else
				timer_stop(pcurr->tmID, WAKEUP_TIMER_ASSIGNED);
			/*
			 *  ޥХåؿƤӽФ
			 */
			if(timercallback != NULL)
				timercallback(pcurr->pID, pcurr->tmID, pcurr->pfTimerCallBack);
		}
		else
			RescheduleTimer();
	}
	else
		rtc_wakeup_int_clear();
}

/*
 *  WAKEUP TIMER
 *  param1 func:Хåؿݥ
 */
void
wakeup_timer_init(void *func)
{
	uint32_t i;
	uint32_t localmaxwakeuptimersetup;

	rtc_wakeup_init(&rtc_wup, rtc_wakeup_handler);
	localmaxwakeuptimersetup = ((((rtc_wup.synchPrescaler - 1) * rtc_wup.asynchPrescaler) - RTC_HANDLER_MAX_DELAY) >> rtc_wup.wakeuptimerPrescaler);
 syslog_1(LOG_NOTICE, "## RTC_HANDLER_MAX_DELAY(%d) ##", RTC_HANDLER_MAX_DELAY);

	if(localmaxwakeuptimersetup >= 0xFFFF){
		MaxWakeupTimerSetup = 0xFFFF;
	}
	else{
		MaxWakeupTimerSetup = (uint16_t)localmaxwakeuptimersetup;
	}

	WakeupTimerMode = WAKEUP_TIMER_ENOUGH;
	SSRValueOnLastSetup = SSR_RESET;

	/*
	 *  WAKEUPޥϥɥ
	 */
	for(i = 0 ; i < NUM_WAKEUP_TIMER ; i++){
		wakeupTimerContext[i].status = WAKEUP_TIMER_FREE;
		wakeupTimerContext[i].tmID   = i;
	}

	timercallback = func;
	pCurrentRunningTimer = NULL;
}

/*
 *  WAKEUP TIMER SETUP
 *  parameter1 pid:  ץID
 *  parameter2 ptid: IDåΰؤΥݥ
 *  parameter3 tmode:ޥ⡼
 *  parameter4 func: Хåؿݥ
 *  return ｪλʤE_OK
 */
ER
wakeup_timer_setup(uint32_t pid, uint8_t *pTimerId, uint8_t tmode, void *func)
{
	wakeuptimerhandle_t *pcurr;
	uint8_t	        i = 0;

	loc_cpu();
	for(i = 0, pcurr = &wakeupTimerContext[0] ; i < NUM_WAKEUP_TIMER ; i++, pcurr++){
		if(pcurr->status == WAKEUP_TIMER_FREE)
			break;
	}

	if(i != NUM_WAKEUP_TIMER){
		pcurr->status = WAKEUP_TIMER_ASSIGNED;
		unl_cpu();

		pcurr->pID  = pid;
		pcurr->mode = tmode;
		pcurr->pfTimerCallBack = func;
		*pTimerId = i;
		return E_OK;
	}
	else{
		unl_cpu();
		return E_OBJ;
	}
}

/*
 *  WAKEUP TIMER
 *  parameter1 tid:  ID
 *  return ｪλʤE_OK
 */
ER
wakeup_timer_delete(uint8_t tid)
{
	if(tid >= NUM_WAKEUP_TIMER)
		return E_PAR;
	loc_cpu();
	timer_stop(tid, WAKEUP_TIMER_FREE);
	unl_cpu();
	return E_OK;
}

/*
 *  WAKEUP TIMER
 *  parameter1 tid:  ID
 *  return ｪλʤE_OK
 */
ER
wakeup_timer_stop(uint8_t tid)
{
	if(tid >= NUM_WAKEUP_TIMER)
		return E_PAR;
	loc_cpu();
 	timer_stop(tid, WAKEUP_TIMER_ASSIGNED);
	unl_cpu();
	return E_OK;
}

/*
 *  WAKEUP TIMER
 *  parameter1 tid:  ID
 *  parameter2 ticks:٥Ȼ(ms)
 *  return ｪλʤE_OK
 */
ER
wakeup_timer_start(uint8_t tid, uint32_t ticks)
{
	if(tid >= NUM_WAKEUP_TIMER)
		return E_PAR;
	loc_cpu();
	timer_start(tid, ticks);
	unl_cpu();
	return E_OK;
}

uint16_t
wakeup_timer_ReadLeftTicksToCount(void)
{
	int32_t wut;
	uint16_t return_value, auro_reload_value, elapsed_time_value;

	loc_cpu();

	wut = rtc_get_wut();
	if(wut > 0){
		auro_reload_value = (uint16_t)wut;

		elapsed_time_value = TimeElapsed();

		if(auro_reload_value > elapsed_time_value){
			return_value = auro_reload_value - elapsed_time_value;
		}
		else{
			return_value = 0;
		}
	}
	else{
		return_value = TIMER_LIST_EMPTY;
	}

	unl_cpu();

	return (return_value);
}

#endif

