/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: wifiuart.c,v 1.1 2020/04/17 13:18:33 roi Exp $
 */
/*
 * 
 *  STM32F7XX WIFI-ESP32(LWIP)ɥ饤дؿ
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include <string.h>
#include "device.h"
#include "wifi.h"
#include "kernel_cfg.h"

#define MIN(a, b) ((a) > (b) ? (b) : (a))

/*
 *  WIFICݡID֥åФΥޥ
 */
#define INDEX_WIFI(emacid)      ((uint_t)((emacid) - 1))

/*
 *  Хåեݥ󥿤Фޥ
 */
#define get_workbuf(h, n)       ((h)->Init.databuffer + ((h)->Init.paketbufsize + sizeof(WIFI_Head_t)) * (n))

/*
 *  WIFI ϥɥ
 */
static WIFI_Handle_t WifiHandle[NUM_WIFI_PORT];


static uint8_t *
buf_length2char(uint8_t *d, int len)
{
	int     i;
	uint8_t buf[16];

	for(i = 0 ; i < 16 && len != 0 ; i++){
		buf[i] = (len % 10) + '0';
		len /= 10;
	}
	while(i > 0){
		*d++ = buf[--i];
	}
	return d;
}

/*
 *  UARTХåؿ
 */
static void
wifiuart_rxcallback(WIFIUART_Handle_t *huart)
{
	isig_sem(WUART1_SEM);
}

static int
wifiuart_recv(WIFIUART_Handle_t *huart, uint8_t *buf, int len)
{
	int rlen = 0;

	while(rlen < len){
		wifiuart_cts_sens(huart);
		rlen = wifiuart_recsize(huart);
		if(rlen < len)
			twai_sem(WUART1_SEM, 100);
	}
	return wifiuart_recev(huart, buf, len);
}

/*
 *  DESCRIPTERΰν
 *  parameter1 hmac  EMACϥɥؤΥݥ
 */
static void
wifi_bd_reset(WIFI_Handle_t *hwifi)
{
	WIFI_Head_t *h;
	uint8_t     *p = hwifi->Init.databuffer;
	uint_t      i;

	for(i = 0 ; i < hwifi->Init.wbufcount ; i++){
		h = (WIFI_Head_t *)p;
		h->status = 0;
		h->length = 0;
		p += (hwifi->Init.paketbufsize + sizeof(WIFI_Head_t));
	}
	hwifi->rx_top   = 0;
	hwifi->rx_tail  = 0;
	hwifi->rx_count = 0;
}

/*
 *  EMACǥХν
 *  parameter1 portid EMACΥݡID
 *  parameter2 ini    ¤ΤؤΥݥ
 *  return            ϥɥؤΥݥ
 */
WIFI_Handle_t *
wifi_init(ID portid, WIFI_Init_t *ini)
{
	WIFIUART_Init_t Init;
	WIFI_Handle_t *hwifi;
	uint32_t no;

	if(portid == 0 || portid > NUM_WIFI_PORT)
		return NULL;

	no = INDEX_WIFI(portid);
	hwifi = &WifiHandle[no];

	memset(hwifi, 0, sizeof(WIFI_Handle_t));
	memcpy(&hwifi->Init, ini, sizeof(WIFI_Init_t));

	if(ini->bps_pin >= 0){
		wifipinMode(ini->bps_pin, GPIO_MODE_OUTPUT, ini->pull, ini->otype);
		digitalWrite(ini->bps_pin, ini->bps_select);
	}
	if(ini->rst_pin >= 0){
		wifipinMode(ini->rst_pin, GPIO_MODE_OUTPUT, ini->pull, ini->otype);
		digitalWrite(ini->rst_pin, 0);
	}

	if(ini->bps_select != WIFIUART_460800BPS)
		Init.bps = 115200;
	else
		Init.bps = 460800;
	Init.rtspin = ini->rts_pin;
	Init.ctspin = ini->cts_pin;
	Init.pull   = ini->pull;
	Init.otype  = ini->otype;
	hwifi->huart = wifiuart_init(WIFIUART1_PORTID, &Init);
	if(hwifi->huart == NULL){
		syslog_0(LOG_ERROR, "## WIFI UART INIT ERROR ##");
		return NULL;
	}

	/*
	 *  Хåեꥻå
	 */
	wifi_bd_reset(hwifi);

	hwifi->macid = portid;
	return hwifi;
}

/*
 *  WIFIλ
 *  parameter1 hmac  EMACϥɥؤΥݥ
 *  return     E_OKｪλ
 */
ER
wifi_deinit(WIFI_Handle_t *hwifi)
{
	if(hwifi == NULL || hwifi->macid == 0 || hwifi->macid > NUM_WIFI_PORT)
		return E_OBJ;
	wifiuart_deinit(hwifi->huart);
	hwifi->macid = 0;
	return E_OK;
}

/*
 *  WIFI
 *  parameter1 hmac  EMACϥɥؤΥݥ
 *  parameter2 pktp  ѥåȹ¤(LWIP)
 *  return     E_OKｪλ
 */
ER
wifi_send(WIFI_Handle_t *hwifi, struct wifi_pkt *pktp)
{
	uint8_t  head[16], *buf;
	ER       ercd = E_OK;

	if(hwifi == NULL || pktp == NULL)
		return E_PAR;

	head[0] = 'W';
	head[1] = 'D';
	buf = buf_length2char(&head[2], pktp->tot_len);
	*buf++ = ':';
	wifiuart_send(hwifi->huart, head, (buf-head));

	while(pktp != NULL){
		if(pktp->len == 0)
			break;

		/*
		 *  ǡХåե˽񤭹
		 */
		wifiuart_send(hwifi->huart, pktp->payload, pktp->len);
		pktp = pktp->next;
	}
	return ercd;
}

/*
 *  ǡĹμФ
 *  parameter1 hwifi  WIFIϥɥؤΥݥ
 *  return     0ǤʤСѥåĹ
 */
int
wifi_recv_length(WIFI_Handle_t *hwifi)
{
	WIFI_Head_t *h;
	int      num, i;
	uint8_t  *p;

	if(hwifi->rx_count == 0)
		return 0;
	for(i = hwifi->rx_tail, num = hwifi->rx_count ; num > 0 ; num--){
		p = get_workbuf(hwifi, i);
		h = (WIFI_Head_t *)p;
		if((h->status & DESC_RX_ACTIVE) != 0)
			return h->length;
		if(++i >= hwifi->Init.wbufcount)
			i = 0;
	}
	return 0;
}

/*
 *  ѥåȤμ
 *  parameter1 hwifi  WIFIϥɥؤΥݥ
 *  parameter2 pktp  ǡǼѥåȹ¤(LWIP)
 *  return     E_OKʤСѥåĹ
 */
ER
wifi_recv(WIFI_Handle_t *hwifi, struct wifi_pkt *pktp)
{
	WIFI_Head_t *h;
	int      actlen, i;
	volatile uint8_t  *b, *pdata, *p;
	uint32_t len, tlen, num;

	if(pktp != NULL){
		b = pktp->payload;
		len = pktp->len;
	}
	else{
		b = NULL;
		len = 0;
	}

	/*
	 *  ѥåȤγǧʤϥ顼λ
	 */
	if(hwifi->rx_count == 0)
		return E_NOMEM;

	while(hwifi->rx_count > 0){
		p = get_workbuf(hwifi, hwifi->rx_tail);
		h = (WIFI_Head_t *)p;
		num = hwifi->rx_tail;
		if(++hwifi->rx_tail >= hwifi->Init.wbufcount)
			hwifi->rx_tail = 0;
		hwifi->rx_count--;
		if((h->status & DESC_RX_ACTIVE) != 0){
			h->status &= ~DESC_RX_ACTIVE;
			/*
			 *  ѥåȳǼΰ褬ϼѥåȤ
			 */
			if(b != NULL){
				/* ǡ */
				pdata = p + sizeof(WIFI_Head_t);
				actlen = h->length;
				syslog_4(LOG_DEBUG, "WIFI:wifi_recv num(%d) b[%08x] pdata[%08x] actlen[%08x] !", num, b, pdata, actlen);
				syslog_4(LOG_DEBUG, "  12[%02x] 13[%02x] 14[%02x] 15[%02x]", pdata[12], pdata[13], pdata[14], pdata[15]);
				while(actlen > 0){
					tlen = MIN(len, actlen);
					for(i = 0 ; i < tlen ; i++)
						*b++ = *pdata++;
					actlen -= tlen;
					len    -= tlen;
					if(len == 0 /*&& actlen > 0*/){
						if((pktp = pktp->next) == NULL){
							b = NULL;
							break;
						}
						b = pktp->payload;
						if((len = pktp->len) == 0){
							b = NULL;
							break;
						}
					}
				}
			}
			return E_OK;
		}
	}
	return E_NOMEM;
}

/*
 *  LINK֤
 *  parameter1 hmac  EMACϥɥؤΥݥ
 *  return     LINK(0̤³)
 */
uint32_t
wifi_link_detect(WIFI_Handle_t *hwifi)
{
	return hwifi->linkvalue & LINK_CONNECT;
}

/*
 *  WIFIꥻåȤ
 *  parameter1 hmac   EMACϥɥؤΥݥ
 *  return     E_OKｪλ
 */
ER
wifi_reset(WIFI_Handle_t *hwifi)
{
	int len;
	uint8_t buf[8];

	hwifi->linkvalue &= ~LINK_READY;

	/*
	 *  WIFI⥸塼ꥻå
	 */
	wifipinMode(hwifi->Init.rst_pin, GPIO_MODE_OUTPUT, GPIO_PULLUP, GPIO_OTYPE_PP);
	digitalWrite(hwifi->Init.rst_pin, 0);
	dly_tsk(100);
	wifipinMode(hwifi->Init.rst_pin, GPIO_MODE_INPUT, GPIO_NOPULL, GPIO_OTYPE_PP);
	dly_tsk(500);

	/*
	 *  Хåեեå
	 */
	while((len = wifiuart_recsize(hwifi->huart)) > 0){
		len = MIN(len, 8);
		wifiuart_recev(hwifi->huart, buf, len);
	}

	/*
	 *  Хåեꥻå
	 */
	wifi_bd_reset(hwifi);
	return E_OK;
}

WIFI_Handle_t *
wifi_gethandler(ID portid)
{
	if(portid == 0 || portid > NUM_WIFI_PORT)
		return NULL;
	return &WifiHandle[INDEX_WIFI(portid)];
}

/*
 *  ESP32 BLE SPP
 */
ER
wifi_ble_send(WIFI_Handle_t *hwifi, uint8_t *pdata, uint32_t length)
{
	uint8_t *buf, head[16];

	if(hwifi == NULL || pdata == NULL || length == 0)
		return E_PAR;
	if((hwifi->linkvalue & LINK_BLE_CONNECT) == 0)
		return E_OBJ;
	head[0] = 'B';
	head[1] = 'D';
	buf = buf_length2char(&head[2], length);
	*buf++ = ':';
	wifiuart_send(hwifi->huart, head, (buf-head));

	/*
	 *  ǡХåե˽񤭹
	 */
	wifiuart_send(hwifi->huart, pdata, length);
	return E_OK;
}

void
wifi_dispatch(WIFI_Handle_t *hwifi)
{
	WIFIUART_Handle_t *huart = hwifi->huart;
	WIFI_Head_t *h;
	uint_t i;
	uint32_t event = 0;
	int len, len2, clen;
	uint8_t *p, buf[32];

	huart->rxcallback  = wifiuart_rxcallback;

	while(1){
		while(wifiuart_recsize(huart) == 0){
			wifiuart_cts_sens(huart);
			twai_sem(WUART1_SEM, 100);
		}
		wifiuart_recv(huart, buf, 1);
		if(buf[0] != 'W' && buf[0] != 'M' && buf[0] != 'B')
			continue;
		len = wifiuart_recv(huart, &buf[1], 1);
		clen = 0;
		event = 0;
		while(len > 0){
			len = wifiuart_recv(huart, &buf[2], 1);
			if(buf[2] >= '0' && buf[2] <= '9')
				clen = clen * 10 + (buf[2] - '0');
			else
				break;
		}
		if(buf[0] == 'W'){
			switch(buf[1]){
			case 'S':
				hwifi->linkvalue |= LINK_START;
				if(clen > 0)
					wifiuart_recv(huart, hwifi->hwaddr, clen);
				clen = 0;
				syslog_3(LOG_NOTICE, "mac[%02x][%02x][%02x]", hwifi->hwaddr[0], hwifi->hwaddr[1], hwifi->hwaddr[2]);
				syslog_3(LOG_NOTICE, "mac[%02x][%02x][%02x]", hwifi->hwaddr[3], hwifi->hwaddr[4], hwifi->hwaddr[5]);
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, WIFI_START);
				event = START_EVENT_DETECT;
				break;
			case 'C':
				hwifi->linkvalue |= LINK_CONNECT;
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, WIFI_CONNECT);
				event = PHY_EVENT_DETECT;
				break;
			case 'N':
				hwifi->linkvalue &= ~LINK_CONNECT;
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, WIFI_DISCONNECT);
				event = PHY_EVENT_DETECT;
				break;
			case 'E':
				hwifi->linkvalue &= ~LINK_START;
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, WIFI_STOP);
				break;
			case 'D':
				if(hwifi->rx_count < hwifi->Init.wbufcount){
					p = get_workbuf(hwifi, hwifi->rx_top);
					h = (WIFI_Head_t *)p;
					wifiuart_recv(huart, &p[sizeof(WIFI_Head_t)], clen);
					h->length  = clen;
					h->status |= DESC_RX_ACTIVE;
					hwifi->rx_count++;
					syslog_2(LOG_DEBUG, "WD count(%d) top(%d)", hwifi->rx_count, hwifi->rx_top);
					if(++hwifi->rx_top >= hwifi->Init.wbufcount)
						hwifi->rx_top = 0;
					clen = 0;
					event = RECV_EVENT_DETECT;
				}
				else
					syslog_1(LOG_ERROR, "wifi:lost packet(%d) !", clen);
				break;
			default:
				break;
			}
			if(hwifi->wifievent != NULL && event != 0)
				hwifi->wifievent(hwifi, event);
			for(i = 0 ; i < clen ; i++)
				wifiuart_recv(huart, &buf[0], 1);
		}
		if(buf[0] == 'B'){
			switch(buf[1]){
			case 'C':
				hwifi->linkvalue |= LINK_BLE_CONNECT;
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, BLE_CONNECT);
				break;
			case 'N':
				hwifi->linkvalue &= ~LINK_BLE_CONNECT;
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, BLE_DISCONNECT);
				break;
			case 'D':
				hwifi->blecount = clen;
				if(hwifi->Init.blebuffer != NULL){
					wifiuart_recv(huart, hwifi->Init.blebuffer, clen);
					clen = 0;
				}
				if(hwifi->uartevent != NULL)
					hwifi->uartevent(hwifi, BLE_RECV);
				break;
			default:
				break;
			}
			for(i = 0 ; i < clen ; i++)
				wifiuart_recv(huart, &buf[0], 1);
		}
		if(buf[0] == 'M'){
			switch(buf[1]){
			case 'R':
				if(clen >= 2){
					wifiuart_recv(huart, hwifi->version, 2);
					clen -= 2;
				}
				if(hwifi->Init.bledevice != NULL){
					hwifi->linkvalue |= LINK_BLE_START;
					buf[0] = 'B';
					buf[1] = 'S';
					len = strlen(hwifi->Init.bledevice);
					p = buf_length2char(&buf[2], len);
					*p++ = ':';
					wifiuart_send(huart, buf, (p-buf));
					wifiuart_send(huart, (uint8_t *)hwifi->Init.bledevice, len);
				}
				if(hwifi->Init.ssid != NULL){
					hwifi->linkvalue |= LINK_READY;
					buf[0] = 'W';
					buf[1] = 'S';
					len = strlen(hwifi->Init.ssid);
					if(hwifi->Init.password == NULL)
						len2 = 0;
					else
						len2 = strlen(hwifi->Init.password);
					p = buf_length2char(&buf[2], len+len2+1);
					*p++ = ':';
					wifiuart_send(huart, buf, (p-buf));
					wifiuart_send(huart, (uint8_t *)hwifi->Init.ssid, len);
					wifiuart_send(huart, (uint8_t *)",", 1);
					if(len2 > 0)
						wifiuart_send(huart, (uint8_t *)hwifi->Init.password, len2);
					if(hwifi->uartevent != NULL)
						hwifi->uartevent(hwifi, MODULE_READY);
				}
				break;
			default:
				break;
			}
			for(i = 0 ; i < clen ; i++)
				wifiuart_recv(huart, &buf[0], 1);
		}
	}
}

