/**
  ******************************************************************************
  * @file    hw.h
  * @author  MCD Application Team
  * @brief   Hardware
  ******************************************************************************
   * @attention
  *
  * <h2><center>&copy; Copyright (c) 2019 STMicroelectronics.
  * All rights reserved.</center></h2>
  *
  * This software component is licensed by ST under BSD 3-Clause license,
  * the "License"; You may not use this file except in compliance with the
  * License. You may obtain a copy of the License at:
  *                        opensource.org/licenses/BSD-3-Clause
  *
  ******************************************************************************
 */


/* Define to prevent recursive inclusion -------------------------------------*/
#ifndef __HW_H
#define __HW_H

#ifdef __cplusplus
extern "C" {
#endif

  /* Includes ------------------------------------------------------------------*/
#if 1	/* ROI DEBUG */
#include "stm32wbxx.h"
#include "stm32wbxx_hal.h"

  /******************************************************************************
   * HW TimerServer
   ******************************************************************************/
  /* Exported types ------------------------------------------------------------*/
  /**
   * This setting is used when standby mode is supported.
   * hw_ts_InitMode_Limited should be used when the device restarts from Standby Mode. In that case, the Timer Server does
   * not re-initialized its context. Only the Hardware register which content has been lost is reconfigured
   * Otherwise, hw_ts_InitMode_Full should be requested (Start from Power ON) and everything is re-initialized.
   */

#include "wakeup_timer.h"

  /**
   * When a Timer is created as a SingleShot timer, it is not automatically restarted when the timeout occurs. However,
   * the timer is kept reserved in the list and could be restarted at anytime with HW_TS_Start()
   *
   * When a Timer is created as a Repeated timer, it is automatically restarted when the timeout occurs.
   */
  typedef enum
  {
    hw_ts_SingleShot = WAKEUP_TIMER_SINGLE,
    hw_ts_Repeated = WAKEUP_TIMER_REPEAT
  } HW_TS_Mode_t;

#define HW_TS_Create    wakeup_timer_setup
#define HW_TS_Delete    wakeup_timer_delete
#define HW_TS_Stop      wakeup_timer_stop
#define HW_TS_Start     wakeup_timer_start

#define LL_HSEM_1StepLock(i, s)      hsem_1steplock((s))
#define LL_HSEM_ReleaseLock(i, s, p) hsem_releaselock((s), (p))

#endif	/* ROI DEBUG */

  /******************************************************************************
   * HW IPCC
   ******************************************************************************/
#if 0	/* ROI DEBUG */
  void HW_IPCC_Enable( void );
  void HW_IPCC_Init( void );
  void HW_IPCC_Rx_Handler( void );
  void HW_IPCC_Tx_Handler( void );

  void HW_IPCC_BLE_Init( void );
  void HW_IPCC_BLE_SendCmd( void );
  void HW_IPCC_MM_SendFreeBuf( void (*cb)( void ) );
#else	/* ROI DEBUG */

/* TX */
#define IPCC_TX_BLE_CMD_CID                             0
#define IPCC_TX_SYSTEM_CMD_RSP_CID                      1
#define IPCC_TX_THREAD_OT_CMD_RSP_CID                   2
#define IPCC_TX_MM_RELEASE_BUFFER_CID                   3
#define IPCC_TX_THREAD_CLI_CMD_CID                      4
#define IPCC_TX_HCI_ACL_DATA_CID                        5

/* RX */
#define IPCC_RX_BLE_EVENT_CID                           0
#define IPCC_RX_SYSTEM_EVENT_CID                        1
#define IPCC_RX_THREAD_NOTIFICATION_ACK_CID             2
#define IPCC_RX_TRACES_CID                              3
#define IPCC_RX_THREAD_CLI_NOTIFICATION_ACK_CID         4

#define HW_IPCC_Enable  ipcc_enable

#define HW_IPCC_BLE_Init()          ipcc_enable_channel(IPCC_RX_CHANNEL, IPCC_RX_BLE_EVENT_CID, false)
#define HW_IPCC_BLE_SendCmd()       ipcc_setflag_channel(IPCC_TX_CHANNEL, IPCC_TX_BLE_CMD_CID)
#define HW_IPCC_MM_SendFreeBuf(cb)  ipcc_callback_channel(IPCC_TX_CHANNEL, IPCC_TX_MM_RELEASE_BUFFER_CID, (cb))
#endif	/* ROI DEBUG */
  void HW_IPCC_BLE_RxEvtNot( void );
#if 0	/* ROI DEBUG */
  void HW_IPCC_BLE_SendAclData( void );
#else	/* ROI DEBUG */
#define HW_IPCC_BLE_SendAclData()   ipcc_enable_channel(IPCC_TX_CHANNEL, IPCC_TX_HCI_ACL_DATA_CID, true)
#endif	/* ROI DEBUG */
  void HW_IPCC_BLE_AclDataAckNot( void );

#if 0	/* ROI DEBUG */
  void HW_IPCC_SYS_Init( void );
  void HW_IPCC_SYS_SendCmd( void );
#else	/* ROI DEBUG */
#define HW_IPCC_SYS_Init()          ipcc_enable_channel(IPCC_RX_CHANNEL, IPCC_RX_SYSTEM_EVENT_CID, false)
#define HW_IPCC_SYS_SendCmd()       ipcc_enable_channel(IPCC_TX_CHANNEL, IPCC_TX_SYSTEM_CMD_RSP_CID, true)
#endif	/* ROI DEBUG */
  void HW_IPCC_SYS_CmdEvtNot( void );
  void HW_IPCC_SYS_EvtNot( void );

#if 0	/* ROI DEBUG */
  void HW_IPCC_THREAD_Init( void );
  void HW_IPCC_OT_SendCmd( void );
  void HW_IPCC_CLI_SendCmd( void );
  void HW_IPCC_THREAD_SendAck( void );
#else	/* ROI DEBUG */
#define HW_IPCC_THREAD_Init()       ipcc_enable_channel2(IPCC_RX_CHANNEL, IPCC_RX_THREAD_NOTIFICATION_ACK_CID, IPCC_RX_THREAD_CLI_NOTIFICATION_ACK_CID)
#define HW_IPCC_OT_SendCmd()        ipcc_enable_channel(IPCC_TX_CHANNEL, IPCC_TX_THREAD_OT_CMD_RSP_CID, true)
#define HW_IPCC_CLI_SendCmd()       ipcc_setflag_channel(IPCC_TX_CHANNEL, IPCC_TX_THREAD_CLI_CMD_CID)
#define HW_IPCC_THREAD_SendAck()    ipcc_enable_channel(IPCC_RX_CHANNEL, IPCC_RX_THREAD_NOTIFICATION_ACK_CID, true)
#endif	/* ROI DEBUG */
  void HW_IPCC_OT_CmdEvtNot( void );
  void HW_IPCC_CLI_CmdEvtNot( void );
  void HW_IPCC_THREAD_EvtNot( void );
#if 0	/* ROI DEBUG */
  void HW_IPCC_THREAD_CliSendAck( void );
#else	/* ROI DEBUG */
#define HW_IPCC_THREAD_CliSendAck() ipcc_enable_channel(IPCC_RX_CHANNEL, IPCC_RX_THREAD_CLI_NOTIFICATION_ACK_CID, true)
#endif	/* ROI DEBUG */
  void HW_IPCC_THREAD_CliEvtNot( void );


  void HW_IPCC_LLDTESTS_Init( void );
  void HW_IPCC_LLDTESTS_SendCliCmd( void );
  void HW_IPCC_LLDTESTS_ReceiveCliRsp( void );
  void HW_IPCC_LLDTESTS_SendCliRspAck( void );
  void HW_IPCC_LLDTESTS_ReceiveM0Cmd( void );
  void HW_IPCC_LLDTESTS_SendM0CmdAck( void );


  void HW_IPCC_BLE_LLD_Init( void );
  void HW_IPCC_BLE_LLD_SendCliCmd( void );
  void HW_IPCC_BLE_LLD_ReceiveCliRsp( void );
  void HW_IPCC_BLE_LLD_SendCliRspAck( void );
  void HW_IPCC_BLE_LLD_ReceiveM0Cmd( void );
  void HW_IPCC_BLE_LLD_SendM0CmdAck( void );
  void HW_IPCC_BLE_LLD_SendCmd( void );
  void HW_IPCC_BLE_LLD_ReceiveRsp( void );
  void HW_IPCC_BLE_LLD_SendRspAck( void );

  
#if 0	/* ROI DEBUG */
  void HW_IPCC_TRACES_Init( void );
#else	/* ROI DEBUG */
#define HW_IPCC_TRACES_Init()       ipcc_enable_channel(IPCC_RX_CHANNEL, IPCC_RX_TRACES_CID, false)
#endif	/* ROI DEBUG */
  void HW_IPCC_TRACES_EvtNot( void );
  
  void HW_IPCC_MAC_802_15_4_Init( void );
  void HW_IPCC_MAC_802_15_4_SendCmd( void );
  void HW_IPCC_MAC_802_15_4_SendAck( void );
  void HW_IPCC_MAC_802_15_4_CmdEvtNot( void );
  void HW_IPCC_MAC_802_15_4_EvtNot( void );

  void HW_IPCC_ZIGBEE_Init( void );

  void HW_IPCC_ZIGBEE_SendM4RequestToM0(void); /* M4 Request to M0 */
  void HW_IPCC_ZIGBEE_RecvAppliAckFromM0(void); /* Request ACK from M0 */

  void HW_IPCC_ZIGBEE_RecvM0NotifyToM4(void); /* M0 Notify to M4 */
  void HW_IPCC_ZIGBEE_SendM4AckToM0Notify(void); /* Notify ACK from M4 */
  void HW_IPCC_ZIGBEE_RecvM0RequestToM4(void); /* M0 Request to M4 */
  void HW_IPCC_ZIGBEE_SendM4AckToM0Request(void); /* Request ACK from M4 */

#if 1	/* ROI DEBUG */
#define LL_FLASH_GetUDN()           get_eui64data(0)
#define LL_FLASH_GetSTCompanyID()   ((get_eui64data(1) >> 8) & 0x00FFFFFF)
#define LL_FLASH_GetDeviceID()      (get_eui64data(1) & 0x000000FF)
#endif	/* ROI DEBUG */

#ifdef __cplusplus
}
#endif

#endif /*__HW_H */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
