/*
 *  TOPPERS/ASP/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: malloc.c 698 2020-03-31 13:05:40Z roi $
 */
/*
 * 
 *  ꥢǽ
 *
 */
#include <kernel.h>
#include <t_syslog.h>
#include <stdlib.h>
#include <string.h>
#include "kernel_cfg.h"

#ifndef CACHE_LINE_SIZE
#define CACHE_LINE_SIZE	    32	/* default cacle line size */
#endif

#if defined(TLSF_SEM)	/* TLSF */
#include "tlsf.h"

/*
 *  TLSF2.4.6ˤϡȡե꡼٤֤ⷫ
 *  ΰƬХȤ˲꤬.
 *  TLSF¦νʼݾڤǤʤᡢmalloc/freeؿ
 *  бԤ.
 */
#define TLSF_PTR_AREA_SIZE  8

static void *memory_pool;

/*
 *  ס
 */
void
heap_init(intptr_t exinf)
{
	intptr_t *param = (intptr_t *)exinf;
	uint32_t *p;
	int i, size;
	memory_pool = (void*)param[0];
	p = (uint32_t *)param[0];
	size = param[1] / sizeof(uint32_t);
	for(i = 0 ; i < size ; i++)
		*p++ = 0;
	init_memory_pool(param[1], (void*)memory_pool);
}

/*
 *  MALLOC
 */
void *
malloc(size_t __size)
{
	void *mempool, *addr;
	mempool = (void *)memory_pool;
	if(mempool == NULL)
		return NULL;
	wai_sem(TLSF_SEM);
	addr = (void*)malloc_ex(__size+TLSF_PTR_AREA_SIZE, mempool);
	sig_sem(TLSF_SEM);
	if(addr != NULL)
		return addr+TLSF_PTR_AREA_SIZE;
	else
		return addr;
}

/*
 *  CALLOC
 */
void *
calloc(size_t __nmemb, size_t __size)
{
	void *mempool, *addr;
	mempool = (void *)memory_pool;
	if(mempool == NULL)
		return NULL;
	wai_sem(TLSF_SEM);
	addr = (void*)malloc_ex((__nmemb * __size)+TLSF_PTR_AREA_SIZE, mempool);
	sig_sem(TLSF_SEM);
	if(addr != NULL){
		memset(addr+TLSF_PTR_AREA_SIZE, 0, __nmemb * __size);
		return addr+TLSF_PTR_AREA_SIZE;
	}
	else
		return addr;
}

/*
 *  FREE
 */
void
free(void *ptr)
{
	void *mempool = memory_pool;
	if(ptr == NULL)
		return;
	if(mempool != NULL){
		wai_sem(TLSF_SEM);
		free_ex(ptr-TLSF_PTR_AREA_SIZE, mempool);
		sig_sem(TLSF_SEM);
	}
}

#elif(LOCAL_MALLOC)		/* LOCAL MALLOC */

typedef struct _LMAP LMAP_t;
struct _LMAP {
  unsigned long lsize;		/* ꥵ */
  LMAP_t *      laddr;		/* ꥹȥݥ */
};

static LMAP_t lmap;

/*
 *  륢
 */
void
heap_init(intptr_t exinf)
{
	intptr_t *param = (intptr_t *)exinf;
	LMAP_t *plmap;

	plmap = (LMAP_t *)param[0];
	plmap->lsize = (unsigned long)param[1];
	plmap->laddr = NULL;
	lmap.lsize   = plmap->lsize;
	lmap.laddr   = plmap;
}

/*
 *  MALLOC
 */
void *
malloc(size_t __size)
{
	LMAP_t *ph = &lmap;
	LMAP_t *p;

	__size = (__size+3+sizeof(unsigned int)) & ~3;
	if(__size < sizeof(LMAP_t))
		__size = sizeof(LMAP_t);

	loc_cpu();
	while((p = ph->laddr) != 0){
		if(p->lsize >= __size && p->lsize < (__size+sizeof(LMAP_t))){
			ph->laddr = p->laddr;
			unl_cpu();
			return (void *)(&p->laddr);
		}
		else if(p->lsize >= (__size+sizeof(LMAP_t))){
			LMAP_t *pw = (LMAP_t *)((char *)p+__size);
			pw->lsize = p->lsize-__size;
			pw->laddr = p->laddr;
			ph->laddr = pw;
			p->lsize  = __size;
			unl_cpu();
			return (void *)(&p->laddr);
		}
		ph = p;
	}
	unl_cpu();
	return NULL;
}

/*
 *  CALLOC
 */
void *
calloc(size_t __nmemb, size_t __size)
{
	return malloc(__nmemb * __size);
}

/*
 *  FREE
 */
void
free(void *ptr)
{
	LMAP_t *ph = &lmap;
	LMAP_t *p;
	char   *t, *pf;

	if(ptr == NULL)
		return;
	p  = (LMAP_t *)((char *)ptr-sizeof(unsigned long));
	t  = (char *)p+p->lsize;
	loc_cpu();
	for(;;){
		if(ph->laddr == NULL){		/* ʤ */
			p->laddr  = NULL;
			ph->laddr = p;
			break;
		}
		else if(t == (char *)ph->laddr){
			p->lsize += (ph->laddr)->lsize;
			p->laddr  = (ph->laddr)->laddr;
			ph->laddr = p;
			break;
		}
		else if(t < (char *)ph->laddr){
			p->laddr = ph->laddr;
			ph->laddr = p;
			break;
		}
		ph = ph->laddr;
		pf = (char *)ph+ph->lsize;
		if(pf > (char *)p){			/* ΰ襪Сå */
			unl_cpu();
			syslog_0(LOG_EMERG, "free:FATAL:irrguler free memory !");
			return;
		}
		else if(pf == (char *)p){
			ph->lsize += p->lsize;
			if(t > (char *)ph->laddr){	/* ΰ襪Сå */
				unl_cpu();
				syslog_0(LOG_EMERG, "free:FATAL:irrguler free memory !");
				return;
			}
			else if(t == (char *)ph->laddr){
				ph->lsize += (ph->laddr)->lsize;
				ph->laddr  = (ph->laddr)->laddr;
			}
			break;
		}
	}
	unl_cpu();
}

#else		/* LIBC MALLOC */

static uintptr_t *heap_param;

/*
 *  LIBC
 */
void
heap_init(intptr_t exinf)
{
	heap_param = (uintptr_t *)exinf;
}

void
__malloc_lock(void * reent)
{
	loc_cpu();
}

void
__malloc_unlock(void * reent)
{
	unl_cpu();
}

void *
_sbrk(int incr)
{ 
	static char *heap_end = NULL;
	char        *prev_heap_end;

	if(incr < 0 || incr > heap_param[1]){
		syslog_1(LOG_ERROR, "_sbrk: incr[%08x] parameter error !", incr);
		return (void *)-1;
	}

	if(heap_end == NULL)
		heap_end = (char *)heap_param[0];

	prev_heap_end  = heap_end;
	heap_end      += incr;

	if(heap_end < (char *)(heap_param[0]+heap_param[1]))
		return (void *) prev_heap_end;
	else{
		heap_end = prev_heap_end;
		syslog_1(LOG_ERROR, "_sbrk: incr[%08x] allocation error !", incr);
		return (void *)-1;
	}
}

#endif

/*
 *  CACHE ALINE MALLOC
 */
void *
malloc_cache(uint32_t len)
{
	void * addr = malloc(len+CACHE_LINE_SIZE);
	uint32_t *p;
	intptr_t uaddr;

	if(addr == NULL)
		return addr;
	p = addr;
	uaddr = (intptr_t)addr;
	if((uaddr & (CACHE_LINE_SIZE-1)) != 0)
		uaddr &= ~(CACHE_LINE_SIZE-1);
	uaddr += CACHE_LINE_SIZE;
	p = (uint32_t *)uaddr;
	*(p-1) = (intptr_t)addr;
	return p;
}

/*
 *  CACHE ALINE FREE
 */
void
free_cache(void *addr)
{
	intptr_t *p = addr;

	if(addr == NULL)
		return;
	addr = (void*)(*(p-1));
	free(addr);
}

