/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  $Id: glcd_disp.c 2416 2019-04-09 18:27:17Z roi $
 */

/* 
 *  GRAPHIC-LCDƥɽץ
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "glcd_disp.h"
#include "font24.c"
#include "font20.c"
#include "font16.c"
#include "font12.c"
#include "font8.c"

/*
 *  ʸ
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x0: ʸ Xɸ
 *  param3  y0: ʸ Yɸ
 *  param4  c:  ʸ᡼ؤΥݥ
 */
#ifndef MAX_FONT_IMAGE
static void
DrawChar(LCD_DrawProp_t *pDrawProp, uint16_t x0, uint16_t y0, const uint8_t *c)
{
	LCD_Handler_t *hlcd = pDrawProp->hlcd;
	sFONT    *pFont = pDrawProp->pFont;
	uint32_t h, w, stride;
	uint8_t  *pchar;

	stride = ((pFont->Width + 7)/8);
	for(h = 0 ; h < pFont->Height ; h++){
		pchar = ((uint8_t *)c + stride * h);
		for(w = 0; w < pFont->Width ; w++){
			if(pchar[w/8] & (1 << (7 - (w % 8)))){
				lcd_drawPixel(hlcd, (x0 + w), (y0 + h), pDrawProp->TextColor);
			}
			else{
				lcd_drawPixel(hlcd, (x0 + w), (y0 + h), pDrawProp->BackColor);
			}
		}
	}
}
#else
static void
DrawChar(LCD_DrawProp_t *pDrawProp, uint16_t x0, uint16_t y0, const uint8_t *c)
{
	LCD_Handler_t *hlcd = pDrawProp->hlcd;
	sFONT    *pFont = pDrawProp->pFont;
	uint32_t h, w, stride;
	const uint8_t  *pchar;
	color_t  image[MAX_FONT_IMAGE];

	stride = ((pFont->Width + 7)/8);
	for(h = 0 ; h < pFont->Height ; h++){
		pchar = ((uint8_t *)c + stride * h);
		for(w = 0; w < pFont->Width ; w++){
			if(pchar[w/8] & (1 << (7 - (w % 8)))){
				image[w] = pDrawProp->TextColor;
			}
			else{
				image[w] = pDrawProp->BackColor;
			}
		}
		lcd_drawImageHLine(hlcd, x0, (y0 + h), pFont->Width, image);
	}
}
#endif

/*
 *  ʸɽ饤󥯥ꥢ
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  饤ֹ
 */
void
lcd_clearStringLine(LCD_DrawProp_t *pDrawProp, uint16_t Line)
{
	sFONT    *pFont = pDrawProp->pFont;
	lcd_fillRect(pDrawProp->hlcd, 0, (Line * pFont->Height), pDrawProp->hlcd->_width, pFont->Height, pDrawProp->BackColor);
}

/*
 *  ʸ
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x0: ʸ Xɸ
 *  param3  y0: ʸ Yɸ
 *  param4  襢(ϰϤ0x200x7Eޤ)
 */
void
lcd_DisplayChar(LCD_DrawProp_t *pDrawProp, uint16_t x0, uint16_t y0, uint8_t Ascii)
{
	sFONT    *pFont = pDrawProp->pFont;
	DrawChar(pDrawProp, x0, y0, &pFont->table[(Ascii-' ') * pFont->Height * ((pFont->Width + 7) / 8)]);
}

/*
 *  ʸ
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param1  x0: ʸ Xɸ
 *  param2  y0: ʸ Yɸ
 *  param3  Text: ʸؤΥݥ
 */
void
lcd_DisplayStringAt(LCD_DrawProp_t *pDrawProp, uint16_t x0, uint16_t y0, uint8_t *Text, Line_ModeTypdef Mode)
{
	sFONT    *pFont = pDrawProp->pFont;
	uint16_t ref_column = 1, i = 0;
	uint32_t size = 0, xsize = 0; 
	uint8_t  *ptr = (uint8_t *)Text;

	/* Get the text size */
	while (*ptr++) size++;

	/* Characters number per line */
	xsize = (pDrawProp->hlcd->_width / pFont->Width);

	switch(Mode){
	case CENTER_MODE:
		ref_column = x0 + ((xsize - size) * pFont->Width) / 2;
		break;
	case LEFT_MODE:
		ref_column = x0;
		break;
	case RIGHT_MODE:
		ref_column =  - x0 + ((xsize - size) * pFont->Width);
		break;
	default:
		ref_column = x0;
		break;
	}

	/*
	 *  ϥ֤򥹥꡼˼
	 */
	if((ref_column < 1) || (ref_column >= 0x8000)){
		ref_column = 1;
	}

	/*
	 *  ʸȤäʸ褹
	 */
	while((*Text != 0) & (((pDrawProp->hlcd->_width - (i*pFont->Width)) & 0xFFFF) >= pFont->Width)){
		lcd_DisplayChar(pDrawProp, ref_column, y0, *Text);
		ref_column += pFont->Width;		/* Υ˰ư */
		Text++;							/* ʸ˰ư */
		i++;
	}
}

/*
 *  ʸ
 *  param1  pDrawProp: Pointer to Draw Prop
 *  param2  x0: ʸ Xɸ
 *  param3  y0: ʸ Yɸ
 *  param4  Text: ʸؤΥݥ
 */
void
lcd_DisplayStringAtLine(LCD_DrawProp_t *pDrawProp, uint16_t Line, uint8_t *ptr)
{
	sFONT    *pFont = pDrawProp->pFont;

	lcd_DisplayStringAt(pDrawProp, 0, Line * pFont->Height, ptr, LEFT_MODE);
}

